/*
 * Copyright (C) 2008 ViXS Systems Inc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** include files **/
#include <config.h>

#include <dfb_types.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <fbdev/fb.h>

#include <fusion/arena.h>
#include <fusion/fusion.h>
#include <fusion/reactor.h>
#include <fusion/shmalloc.h>

#include <directfb.h>

#include <direct/messages.h>
#include <direct/util.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/screens.h>
#include <core/state.h>
#include <core/gfxcard.h>
#include <core/layers.h>
#include <core/surface.h>
//PORT TO 1.4.3\
//Added include <core/core.h> to get support for functions used in the XC3 driver 
#include <core/core.h>

//PORT TO 1.4.3: Removed (looks like its not needed)
//#include <core/surfacemanager.h>
#include <core/system.h>

#include <fbdev/fbdev.h>

#include <gfx/convert.h>
#include <gfx/util.h>

#include <misc/conf.h>

#include <core/graphics_driver.h>
#include <string.h>
#include <directfb_strings.h>

#include "xcdfb.h"
#include "xcioctl.h"
#include "xcmmio.h"

//Changed to CoreGraphicsDevice
static int driver_probe( CoreGraphicsDevice *device );

static void xcodeEngineReset( void *drv, void *dev );

static DFBResult xcodeEngineSync( void *drv, void *dev );

static void xcodeCheckState( void *drv, void *dev,
                              CardState *state, DFBAccelerationMask accel );

static void xcodeSetState( void *drv, void *dev,
                            GraphicsDeviceFuncs *funcs,
                            CardState *state, DFBAccelerationMask accel );
static bool xcodeFillRectangle( void *drv, void *dev, DFBRectangle *rect );

static bool xcodeBlit( void *drv, void *dev, DFBRectangle *rect, int dx, int dy );




static DFBResult
xcodeWaitVSync( CoreScreen *screen,
                void       *driver_data,
                void       *screen_data );

static DFBResult
xcodeInitScreen( CoreScreen           *screen,
                 CoreGraphicsDevice       *device,
                 void                 *driver_data,
                 void                 *screen_data,
                 DFBScreenDescription *description );

/** private functions **/
int adjust_phys_offset(int offset);

/** public data **/


/** private data **/

static const DirectFBSurfaceDrawingFlagsNames( draw_flags );
static const DirectFBSurfaceBlittingFlagsNames( blit_flags );
static const DirectFBAccelerationMaskNames( accel_functions);

/** external functions **/

/** external data **/
extern DisplayLayerFuncs xcOverlayFuncs;
extern DisplayLayerFuncs  xcOldPrimaryFuncs;
extern DisplayLayerFuncs xcPrimaryFuncs;

extern void    *xcOldPrimaryDriverData;


/** local definitions **/
DFB_GRAPHICS_DRIVER( xcode )

/** default settings  **/

/* driver capability flags */


#define XC3_SUPPORTED_DRAWINGFLAGS \
               (DSDRAW_NOFX )

#define XC3_SUPPORTED_BLITTINGFLAGS \
               (DSBLIT_NOFX)

#define XC3_SUPPORTED_DRAWINGFUNCTIONS \
               (DFXL_FILLRECTANGLE )


#define XC3_SUPPORTED_BLITTINGFUNCTIONS \
               (DFXL_BLIT)

/**************************************************************************************************/

static const char *draw_name( DFBSurfaceDrawingFlags flag )
{
     int i;


     for (i=0; i<D_ARRAY_SIZE(draw_flags); i++) 
     {
          if (draw_flags[i].flag == flag)
          {
            return draw_flags[i].name;
          }
     }
     return "<invalid draw flag>";
}

static const char *blit_name( DFBSurfaceBlittingFlags flag )
{
     int i;


     for (i=0; i<D_ARRAY_SIZE(blit_flags); i++) 
     {
          if (blit_flags[i].flag == flag)
          {
            return blit_flags[i].name;
          }
     }
     return "<invalid blit flag>";
}

static const char *accel_name( DFBAccelerationMask mask )
{
     int i;

     for (i=0; i<D_ARRAY_SIZE(accel_functions); i++) 
     {
          if (accel_functions[i].mask == mask)
          {
            return accel_functions[i].name;
          }
     }
     return "<invalid accel flag>";
}



void print_flags(int type, int flags)
{
     int i,flag_mask;

     //
     //DEBUG only: Print out requested draw flags
     // 
     flag_mask=1;

     while (flags)
     {
          if (flags & 1)
          {
               if (type ==1)
               {
                    D_DEBUG( "%s....\n", draw_name(flag_mask));
               }
               else if (type == 2)
               {
                    D_DEBUG( "%s....\n", blit_name(flag_mask));

               }
               else
               {
                    D_DEBUG( "%s....\n", accel_name(flag_mask));

               }
          }
          flags>>=1;
          flag_mask<<=1;
     }
}



/*
 *
 *Function Name:adjust_phys_offset
 *
 *Parameters: offset -> Offset from DFB framebuffer
 *
 *Description:
 *            Adjusts the offset to correctly point to the actual physical address within the XC frmaebuffer.
 *            The DFB frambuffer is mapped from linear address 0, but in reality
 *            the frmabuffer can be located anywhere within the XC 192MB frmabuffer.
 *
 *Returns:
 *        Adjusted Offset
 *
*/
int adjust_phys_offset(int offset)
{
     return ( (dfb_system_video_memory_physical(offset) - XC3_SOC_MEM_PHYS) );
}

/*
 *
 *Function Name:check_src_pixel_format
 *
 *Parameters: format -> Describes the src surface pixel format
 *
 *Description:
 *            Determines whether or not the current ASIC supports format as SRC
 *
 *Returns:
 *        TRUE if supported, otherwise FALSE
 *
*/
static __inline int check_src_pixel_format(int format)
{
     if ( (format == DSPF_RGB32) ||(format  == DSPF_ARGB) )
     {    
          return DFB_TRUE;
     }

     return DFB_FALSE;
}


/*
 *
 *Function Name:check_dst_pixel_format
 *
 *Parameters: format -> Describes the dst surface pixel format
 *
 *Description:
 *            Determines whether or not the current ASIC supports format as DST
 *
 *Returns:
 *        TRUE if supported, otherwise FALSE
 *
*/
static __inline DFBResult check_dst_pixel_format(int format)
{
     if ( (format == DSPF_RGB32) ||(format  == DSPF_ARGB) )
     {    
          return DFB_TRUE;
     }

     return DFB_FALSE;

}


/*
 *
 *Function Name:xcode_validate_DESTINATION
 *
 *Parameters: 
 *             mdev -> Pointer to device data
 *             state-> Pointer to cardstate decribing destination properties
 *
 *Description:
 *           Called by xcodesetstate() to ensure that the destination properties are properly set
 *           for execution of rendering functions.
 *
 *Returns:
 *         NA
*/
static inline void
xcode_validate_DESTINATION( pxc_dfb_device_data mdev,
                              CardState         *state )
{
     CoreSurface *destination	   =	state->destination;
     
     D_DEBUG( "xcode_validate_DESTINATION \n" );

     //DFB 1.4.3: Changed this to access the correct information (to get size)
     D_DEBUG( "width %x \n", destination->config.size.w );
     D_DEBUG( "height %x \n", destination->config.size.h );

     //
     // Remember destination parameters for usage in rendering functions. //
     // 
     
     mdev->dst_phys   = state->dst.phys;  
     mdev->dst_pitch  = state->dst.pitch;
     mdev->dst_width  = destination->config.size.w;
     mdev->dst_height = destination->config.size.h;

     mdev->dst_format = destination->config.format;
     mdev->dst_bpp    = DFB_BYTES_PER_PIXEL( mdev->dst_format );

     D_DEBUG( "xcode validate DESTINATION PHYS = %08x:\n", mdev->dst_phys );
     D_DEBUG( "xcode validate DESTINATION PITCH = %08x:\n", mdev->dst_pitch );
     D_DEBUG( "xcode validate DESTINATION FORMAT = %08x:\n", mdev->dst_format );


     /* Set the flag. */
     XCODE_VALIDATE_STATE( DESTINATION );
}

/*
 *
 *Function Name:xcode_validate_COLOR
 *
 *Parameters: 
 *             mdev -> Pointer to device data
 *             state-> Pointer to cardstate decribing color properties
 *
 *Description:
 *           Called by xcodesetstate to ensure that the color variables are properly set
 *           for execution of rendering functions.
 *
 *Returns:
 *         NA
*/

static inline void
xcode_validate_COLOR( pxc_dfb_device_data mdev,
                              CardState         *state )
{

     DFBColor       color = state->color;
     u32            clr;

     D_DEBUG( "xcode_validate_COLOR \n" );

     switch (state->destination->config.format) 
     {
          
          case DSPF_RGB32:
                   clr = PIXEL_RGB32( color.r,
                                      color.g,
                                      color.b );
                   break;
          case DSPF_ARGB:
                   clr = PIXEL_ARGB( color.a,
                                     color.r,
                                     color.g,
                                     color.b );
                   break;
          default:
                   D_BUG( "unexpected pixelformat! %08x \n", state->destination->config.format );
                   return;
     }

     mdev->fill_color = clr;

     D_DEBUG( "fill color =%08x \n", clr );


     /* Set the flag. */
     XCODE_VALIDATE_STATE( COLOR );

}

/*
 *
 *Function Name:xcode_validate_SOURCE
 *
 *Parameters: 
 *             mdev -> Pointer to device data
 *             state-> Pointer to cardstate decribing destination properties
 *
 *Description:
 *           Called by xcodesetstate() to ensure that the source properties are properly set
 *           for execution of rendering functions.
 *
 *Returns:
 *         NA
*/

static inline void
xcode_validate_SOURCE( pxc_dfb_device_data mdev,
                              CardState         *state )
{
     CoreSurface *source =    state->source;
     D_DEBUG( "xcode_validate_SOURCE \n" );

     D_DEBUG( "width %x \n", source->config.size.w );
     D_DEBUG( "height %x \n", source->config.size.h );

     /* Remember destination parameters for usage in rendering functions. */
     mdev->src_phys   = state->src.phys;

     mdev->src_pitch  = state->src.pitch;
     mdev->src_width  = source->config.size.w;
     mdev->src_height = source->config.size.h;
     mdev->src_format = source->config.format;
     mdev->src_bpp    = DFB_BYTES_PER_PIXEL( mdev->src_format );

     D_DEBUG( "xcode validate SOURCE PHYS = %08x:\n", mdev->src_phys );
     D_DEBUG( "xcode validate SOURCE PITCH = %08x:\n", mdev->src_pitch );
     D_DEBUG( "xcode validate SOURCE FORMAT = %08x:\n", mdev->src_format );

     /* Set the flag. */
     XCODE_VALIDATE_STATE( SOURCE );

}





/*
 *
 *Function Name:xcode_validate_dfb_drawflags
 *
 *Parameters: 
 *             flags -> Drawing flags requested
 *
 *Description:
 *           Called by xcodeCheckState() to verify that we do indeed support 
 *           the requested drawing operation
 *
 *Returns:
 *         DFB_OK if supported, otherwise DFB_UNSUPPORTED
*/

static DFBResult xcode_validate_dfb_drawflags( u32 flags)
{
     DFBResult result=DFB_OK;
     int  i,flag_mask;

     if (flags & ~XC3_SUPPORTED_DRAWINGFLAGS)
     {
          D_BUG( "asked for unsupported drawing function ....\n");
          result = DFB_UNSUPPORTED;
     }
     else
     {
          D_DEBUG( "asked for supported drawing functions ....\n");
     }


     if (flags)
     {
          print_flags(1,flags);     
     }
     else
     {
          //
          //Must bo NOFX
          // 
          D_DEBUG( "%s....\n", draw_name(DSDRAW_NOFX));

     }
     return result;

}

/*
 *
 *Function Name:xcode_validate_dfb_blitflags
 *
 *Parameters: 
 *             flags -> Blitting flags requested
 *
 *Description:
 *           Called by xcodeCheckState() to verify that we do indeed support 
 *           the requested blitting operation
 *
 *Returns:
 *         DFB_OK if supported, otherwise DFB_UNSUPPORTED
*/
static DFBResult xcode_validate_dfb_blitflags( u32 flags)
{
     DFBResult result=DFB_OK;
     int  i,flag_mask;

     if (flags & ~XC3_SUPPORTED_BLITTINGFLAGS)
     {
          D_BUG( "asked for unsupported blitting flags....\n");
          result = DFB_UNSUPPORTED;
     }
     else
     {
          D_DEBUG( "asked for supported blitting functions ....\n");
     }


     if (flags)
     {
          print_flags(2,flags);     
     }
     else
     {
          //
          //Must bo NOFX
          // 
          D_DEBUG( "%s....\n", draw_name(DSBLIT_NOFX));

     }
     return result;



}

/*
 *
 *Function Name:xcode_validate_dfb_accelfunction
 *
 *Parameters: 
 *             flags -> Type of operation requested(StretchBLT etc)
 *
 *Description:
 *           Called by xcodeCheckState() to verify that we do indeed support 
 *           the requested operation
 *
 *Returns:
 *         DFB_OK if supported, otherwise DFB_UNSUPPORTED
*/
static DFBResult xcode_validate_dfb_accelfunction( u32 accel)
{
     DFBResult result=DFB_OK;
     int  i;

     if (accel & ~(XC3_SUPPORTED_DRAWINGFUNCTIONS | XC3_SUPPORTED_BLITTINGFUNCTIONS) )
     {
          D_BUG( "asked for unsupported accel functions ....\n");
          result = DFB_UNSUPPORTED;
     }
     else
     {
          D_DEBUG( "asked for supported accel functions ....\n");

     }

     if (accel)
     {
          //
          //print out the drawing functions first
          //
          if (DFB_DRAWING_FUNCTION(accel))
          {
               accel = DFB_DRAWING_FUNCTION(accel);
          
               print_flags(2,accel);     
          }
          else
          {
               //
               //print out BLIT functions
               //
               accel >>= 16;

               print_flags(1,accel);     
          
          }
     }
     else
     {
          //
          //DFXL_NONE
          // 
          D_DEBUG( "%s....\n", accel_name(DFXL_NONE));
     }

     return result;
}


/**************************************************************************************************/




/*
 *
 *Function Name:xcodeEngineReset
 *
 *Parameters: 
 *             drv-> Pointer to driver data
 *             dev-> Pointer to device data
 *
 *Description:
 *             Reset engine
 *Returns:
 *             NA
*/
static void xcodeEngineReset( void *drv, void *dev )
{
     D_DEBUG(" %s enter:\n",__FUNCTION__);
     D_DEBUG(" %s exit:\n",__FUNCTION__);

}



/*
 *
 *Function Name:xcodeEngineSync
 *
 *Parameters: 
 *             drv-> Pointer to driver data
 *             dev-> Pointer to device data
 *
 *Description:
 *             SYnc engine, for XC3, this menas the DMA engine
 *Returns:
 *             DFB_OK when engine is IDLE
*/

static DFBResult xcodeEngineSync( void *drv, void *dev )
{

     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;

     u32  dma_bits_to_check = (XC_DMA_STATUS_INDIRECT0_MASK);

     D_DEBUG(" %s enter:\n",__FUNCTION__);

     //
     //Waitfor IDLE
     // 
     while ( (xc_in32(mdrv->mmio_base,XC_DMA_STATUS)) & (dma_bits_to_check) )
     {
          D_DEBUG( "DMA Engine still busy!!\n" );
     }

     D_DEBUG(" %s exit:\n",__FUNCTION__);



     return DFB_OK;
}



/*
 *
 *Function Name:xcodeCheckState
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *state-> Indicates current state of card(clr keys etc)
 *           accel-> Indicates what type of operation is being performed
 *
 *Description:
 *           Parses the accel parameters to ensure current ASIC can carry out intended  
 *           operation, if it can't, it doesnt set any bits within the state->accel variable
 *
 *Returns:
 *           N/A
*/
static void xcodeCheckState( void *drv, void *dev,
                              CardState *state, DFBAccelerationMask accel )
{

     int supported_accel=0;   

     D_DEBUG(" %s enter:\n",__FUNCTION__);

     D_DEBUG( "driver accel= %08x \n", accel );

     /*
      * Check desired function against support functions.
     */
     if (xcode_validate_dfb_accelfunction(accel) == DFB_OK)
     {
          /*
           * Check destination pixel format
          */
          if ( check_dst_pixel_format(state->destination->config.format) )
          {
               /*
                * Check if we drawing or blitting  
               */
               if (DFB_DRAWING_FUNCTION( accel ) ) 
               {
                    /* Return if unsupported drawing flags are set. */
                    if (xcode_validate_dfb_drawflags(state->drawingflags)== DFB_OK)
                    {
                         supported_accel = accel;
                    }
               }
               else 
               {
                    /*
                     * We are doing some sor of BLT
                    */
                    if (xcode_validate_dfb_blitflags(state->blittingflags)== DFB_OK)
                    {
                         //
                         // No format conversion supported. //
                         // 
                         if (state->source->config.format != state->destination->config.format)
                         {
                              D_BUG( "blitting clr conversion requested \n");

                         }          
                         else if (!( check_src_pixel_format(state->source->config.format)) )
                         {
                              D_BUG( "src pixel format not supported %08x \n",
                                                       state->source->config.format );
                         }
                         else
                         {
                              //
                              //Setup supported functions/settings 
                              // 
                              supported_accel = accel;
                         }
               
                    }
               }
          }
          else
          {
               D_BUG( "dst pixel format not supported %08x \n",
                                        state->destination->config.format );

          }
     }

     //
     //The supported_accel variable is setup whenever the driver has determined that the 
     //request settings can be accomodated, otherwise, the value will be 0!
     // 
     state->accel |= supported_accel;

     D_DEBUG( "xcode DFB driver state->accel %08x \n", state->accel );

     D_DEBUG(" %s exit:\n",__FUNCTION__);

}


/*
 *
 *Function Name:xcodeSetState
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *funcs-> Table of supported functions which can be changed dynamically
 *           *state-> Indicates current state of card(clr keys etc)
 *           accel-> Indicates what type of operation is being performed
 *
 *Description:
 *
 *Returns:
 *
*/
static void xcodeSetState( void *drv, void *dev,
                            GraphicsDeviceFuncs *funcs,
                            CardState *state, DFBAccelerationMask accel )
{

     pxc_dfb_device_data mdev = (pxc_dfb_device_data) dev;
     StateModificationFlags modified = state->modified;

     D_DEBUG(" %s enter:\n",__FUNCTION__);

     /*
      * 1) Invalidate hardware states
      *
      * Each modification to the hw independent state invalidates one or more hardware states.
     */

     //
     // Simply invalidate all? 
     // 
     
     if (modified == SMF_ALL) 
     {
          XCODE_INVALIDATE_STATE( ALL );
     }
     else if (modified)
     {
          //
          // Invalidate destination registers. 
          // 
          if (modified & SMF_DESTINATION)
          {
               XCODE_INVALIDATE_STATE( DESTINATION | COLOR );
          }
          else if (modified & SMF_COLOR)
          {     
               XCODE_INVALIDATE_STATE( COLOR );
          }
          if (modified & SMF_SOURCE)
          {
               XCODE_INVALIDATE_STATE( SOURCE );
          }
     }

     /*
      * 2) Validate hardware states
      *
      * Each function has its own set of states that need to be validated.
      */

     //
     // Always validate destination... 
     // 
     XCODE_CHECK_VALID_STATE( DESTINATION );

     switch (accel)
     {
          case DFXL_FILLRECTANGLE:
                    //
                    // ...require valid drawing color. 
                    // 
                    XCODE_CHECK_VALID_STATE( COLOR );

                    /*
                     * 3) Tell which functions can be called without further validation, i.e. SetState()
                     *
                     * When the hw independent state is changed, this collection is reset.
                     */
                    state->set = XC3_SUPPORTED_DRAWINGFUNCTIONS;
                    break;
          case DFXL_BLIT:
                    //
                    // ...require valid source. 
                    // 
                    XCODE_CHECK_VALID_STATE( SOURCE );
                    state->set = XC3_SUPPORTED_BLITTINGFUNCTIONS;
                    break;
          default:
                    D_BUG( "unexpected drawing/blitting function" );
                    break;
     }

     /*
      * 4) Clear modification flags
      *
      * All flags have been evaluated in 1) and remembered for further validation.
      * If the hw independent state is not modified, this function won't get called
      * for subsequent rendering functions, unless they aren't defined by 3).
      */
     state->modified = 0;

     D_DEBUG(" %s exit:\n",__FUNCTION__);

}


/*
 *
 *Function Name:xcodeFillRectangle
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *rect-> Rectangle to be filled
 *
 *Description:
 *           FIlls a rectangle on the DST surface.
 *
 *Returns:
 *           true or false
*/
static bool xcodeFillRectangle( void *drv, void *dev, DFBRectangle *rect )
{
     pxc_dfb_driver_data pmdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;

     u32 width,height,i,dma_dq_ptr;
     volatile u8 *mmio     = pmdrv->mmio_base;
     u8  *dst_addr;
     u8  *pDMADesc;

     D_DEBUG(" %s enter:\n",__FUNCTION__);

     D_DEBUG( "xcode DFB driver xcodeFillRectangle = %x %x %x %x\n",rect->x,rect->y,rect->w,rect->h );


     width = rect->w  * pmdev->dst_bpp;

     height = rect->h;

     //
     //Adjust starting co-ordinates
     // 
     dst_addr = (u8*) ((pmdev->dst_phys));

     dst_addr+=(rect->y * pmdev->dst_pitch);
     dst_addr+=(rect->x * pmdev->dst_bpp);


     //
     //Select DMA-LIST aperture
     // 


     pDMADesc=(u8*)((u32)pmdrv->shared->uDMAListOffset+
          (DMA_DESC_APER_SIZE * pmdev->dma_list_index)) ; //Either 0 or 1

     dma_dq_ptr = (u32)pDMADesc;

     
     pDMADesc += (u32) pmdrv->mmaper_base;

     D_DEBUG( "height = %x, width = %x, dst = %x, phys=%x\n",height,width,dst_addr, pmdev->dst_phys );

     /*
      * Configure DMA, we basically fill one scan line at a time.
      * 
     */
     for(i=0; i< height; i++)
     {
          ((pxc_dfb_driver_dma_desc) pDMADesc)->src_addr_0_31 = pmdev->fill_color;
          ((pxc_dfb_driver_dma_desc) pDMADesc)->src_addr_32_63 = pmdev->fill_color;
          ((pxc_dfb_driver_dma_desc) pDMADesc)->dst_addr_0_31 = (u32)dst_addr;
          ((pxc_dfb_driver_dma_desc) pDMADesc)->dst_addr_32_63 = 0;
          
          ((pxc_dfb_driver_dma_desc) pDMADesc)->dma_cmd = (width | 
               ( (1 << XC_DMA_DIRECT_DMACMD_A_DSTSPACE_SHIFT)   |
                 (1 << XC_DMA_DIRECT_DMACMD_A_FILL_MODE_SHIFT ) 
               ));
          ((pxc_dfb_driver_dma_desc) pDMADesc)->crypto_cw = 0;
     
          dst_addr+=pmdev->dst_pitch;
     
          pDMADesc = (pxc_dfb_driver_dma_desc) pDMADesc+1;
     
     }
     
     //
     //Set EOL
     // 
     pDMADesc = (pxc_dfb_driver_dma_desc) pDMADesc - 1;
     
     ((pxc_dfb_driver_dma_desc) pDMADesc)->dma_cmd |=
                    (1 <<  XC_DMA_ACTIVE_CMD_EOL_SHIFT); 


     /*
      * SYNC DMA engine
     */
     xcodeEngineSync(drv,dev);

     /*
      * Kick off DMA(make sure to add the application offset to FB offset)
     */
     D_DEBUG( "kicking off dma DQ_PTR = %x \n",(adjust_phys_offset(dma_dq_ptr))  );
     xc_out32(mmio,XC_DMA_DQ_PTR0,(adjust_phys_offset(dma_dq_ptr)) );

     pmdev->dma_list_index ^=1;

     D_DEBUG(" %s exit:\n",__FUNCTION__);

     return true;

}


/*
 *
 *Function Name:xcodeBlit
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *rect-> Rectangle describing source size
 *           dx-> Destination x-coordinate of BLT   
 *           dy-> Destination y-coordinate of BLT   
 *
 *Description:
 *           Performs a SRC copy.
 *
 *Returns:
 *           true or false
*/
static bool xcodeBlit( void *drv, void *dev, DFBRectangle *rect, int dx, int dy )
{

     pxc_dfb_driver_data pmdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;

     volatile u8 *mmio     = pmdrv->mmio_base;
     
     volatile void *pDMADesc   = pmdrv->shared->pDMAListAddress;
     volatile u8  *dst_addr, *src_addr;
     u32  blt_height,blt_width,segment_width,segment_height,max_segment_height,blt_size=0;

     D_DEBUG(" %s enter:\n",__FUNCTION__);

     /*
      * Adjust DST address
     */
     dst_addr = (u8*) pmdev->dst_phys;
     dst_addr+=(dy * pmdev->dst_pitch);
     dst_addr+=(dx * pmdev->dst_bpp);

     /*
      * Adjust SRC address
     */
     src_addr = (u8*) pmdev->src_phys;
     src_addr+=(rect->y * pmdev->src_pitch);
     src_addr+=(rect->x * pmdev->src_bpp);

     blt_width = rect->w * pmdev->src_bpp;
     blt_height = rect->h;


     /*
      * At this point we should determine whether or not the source is still
      * valid and if it is, not setup the SRC parameters, but since on jaguar/panther
      * we dont have a real 2D engine, we will skip that step
     */
     D_DEBUG( "xcode DFB driver xcodeBlit dx = %x dy = %x\n",dx,dy );

     D_DEBUG( "xcode DFB driver xcodeBlit source rect = %x %x %x %x\n",rect->x,rect->y,rect->w,rect->h );

     D_DEBUG( "pDMADesc = %x\n",pDMADesc );

     /*
      * Sync engine
     */
     xcodeEngineSync(drv,dev);

     /*
      * Due to the XC3 hardware limitations, we can oly do blits upto 4kb in size,
      * 
     */
     while(blt_width)
     {
          if (blt_width > DMA_BLT_MAX_WIDTH)
          {
               segment_width = ((DMA_BLT_MAX_WIDTH) & ~3);
          }
          else
          {
               segment_width = blt_width;
          }
          /*
           * Configure DMA
          */
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->src_addr_0_31 = (u32) src_addr;
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->src_addr_32_63 = 0;
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->dst_addr_0_31 = (u32) dst_addr;
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->dst_addr_32_63 = 0;
          
          
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->dma_cmd = (
                                                               (1 << XC_DMA_DIRECT_DMACMD_A_SRCSPACE_SHIFT) |
                                                               (1 << XC_DMA_DIRECT_DMACMD_A_DSTSPACE_SHIFT) |
                                                               (1 << XC_DMA_ACTIVE_CMD_EXTENDED_SHIFT) 
                                                               );
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->crypto_cw = 0;
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->ubase_addr = 0;
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->vbase_addr = 0;
          
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->dma_ext_cmd_0_31 = (
                                                                        ( pmdev->src_pitch) |
                                                                        ( pmdev->dst_pitch << XC_DMA_EXT_CMD_SPECIFIER_DSTPITCH_SHIFT )
                                                                        );
          ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->dma_ext_cmd_32_63 = (
                                                                         (segment_width) |
                                                                         (blt_height << XC_DMA_EXT_CMD_SPECIFIER_BLTHEIGHT_SHIFT)
                                                                         );

          pDMADesc = (pxc_dfb_driver_ext_dma_desc) pDMADesc+1;

          blt_width-=segment_width;
          src_addr+=segment_width;
          dst_addr+=segment_width;

     }

     pDMADesc= (pxc_dfb_driver_ext_dma_desc) pDMADesc -1;

     ((pxc_dfb_driver_ext_dma_desc) pDMADesc)->dma_cmd |= 
                                   (1 << XC_DMA_ACTIVE_CMD_EOL_SHIFT);


     /*
      * Kick off DMA(make sure to add the application offset to FB offset)
     */
     D_DEBUG( "kicking off ext dma DQ_PTR = %x \n",adjust_phys_offset(pmdrv->shared->uDMAListOffset) );
     xc_out32(mmio,XC_DMA_DQ_PTR0, adjust_phys_offset(pmdrv->shared->uDMAListOffset) );

     D_DEBUG(" %s exit:\n",__FUNCTION__);

     return true;

}


/*
 *
 *Function Name:driver_probe
 *
 *Parameters: 
 *             GraphicsDevice->Framebuffer device
 *
 *Description:
 *             Check the framebuffer device to determine if its ours
 *             
 *Returns:
 *             Returns TRUE if our device, otherwise FALSE
 *
*/
static int
driver_probe( CoreGraphicsDevice *device )
{
     int  iRetVal=0;

     D_DEBUG(" %s enter:\n",__FUNCTION__);

     switch (dfb_gfxcard_get_accelerator( device ))
     {
          case FB_ACCEL_VIXS_XC3:
          {
               iRetVal=1;
          }
     }

     D_DEBUG(" %s exit with %x:\n",__FUNCTION__,iRetVal);

     return iRetVal;
}


/*
 *
 *Function Name:driver_get_info
 *
 *Parameters: 
 *             GraphicsDevice -> Framebuffer device
 *             GraphicsDriverInfo -> DirectFB driver info structure
 *Description:
 *             Gets the driver specific information(name, vendor etc)
 *Returns:
 *             Driver must fill out the relevant fields of the info structure
 *
*/

static void
driver_get_info( CoreGraphicsDevice     *device,
                 GraphicsDriverInfo *info )
{
     D_DEBUG(" %s enter:\n",__FUNCTION__);

     /* fill driver info structure */
     snprintf( info->name,
               DFB_GRAPHICS_DRIVER_INFO_NAME_LENGTH,
               XC_DFB_DRIVER_NAME );

     snprintf( info->vendor,
               DFB_GRAPHICS_DRIVER_INFO_VENDOR_LENGTH,
               XC_DFB_DRIVER_VENDOR);

     info->version.major = XC_DFB_DRV_VERSION_MAJ;
     info->version.minor = XC_DFB_DRV_VERSION_MIN;

     info->driver_data_size = sizeof (xc_driver_data);
     info->device_data_size = sizeof (xc_device_data);

     D_DEBUG(" %s exit:\n",__FUNCTION__);

}


/*
 *
 *Function Name:driver_init_driver
 *
 *Parameters: 
 *             GraphicsDevice -> Framebuffer device
 *             GraphicsDeviceFuncs -> DirectFB driver functions structure
 *             driver_data -> driver specific structure
 *             device_data -> device specific structure
 *             core -> DirectFB core object
 *Description:
 *             Fills out the driver specific functions and maps in the MMIO
 *
 *Returns:
 *             DFB_OK
*/

static DFBResult
driver_init_driver( CoreGraphicsDevice      *device,
                    GraphicsDeviceFuncs *funcs,
                    void                *driver_data,
                    void                *device_data,
                    CoreDFB             *core )
{

     DFBResult      result = DFB_OK;
	 XcdrvShared    *shared = NULL;
     FusionSHMPoolShared *pool;

     D_DEBUG(" %s enter:\n",__FUNCTION__);

     

     if (dfb_core_is_master( core))
	 {
	      D_DEBUG ("dfb core is master \n");
          pool      = dfb_core_shmpool( core );

          shared = (XcdrvShared*) SHCALLOC( pool, 1, sizeof(XcdrvShared) );
          if (!shared) {
               result = D_OOSHM();
               return result;
          }

          fusion_arena_add_shared_field( dfb_core_arena( core ), "xcdrv", shared );
	 }
	 else
	 {
	      D_DEBUG ("dfb core is not master \n");
          fusion_arena_get_shared_field( dfb_core_arena( core ),
                                    "xcdrv", &shared );
	      
	 }

     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) driver_data;
     pxc_dfb_device_data mdev = (pxc_dfb_device_data) device_data;

     mdrv->shared = shared;


     mdrv->mmio_base = dfb_gfxcard_map_mmio( device, 0, -1 ); 

     //
     // Map the aperture
     //
     mdrv->mmaper_base = dfb_gfxcard_memory_virtual(device,0);

     D_DEBUG( "aperture = %x", mdrv->mmio_base );
     D_DEBUG( "xcode DFB driver mmio = %08x \n", mdrv->mmio_base );

     if (mdrv->mmio_base)
     {

          D_DEBUG( "Board ID = %x", xc_in32(mdrv->mmio_base,XC_RBM_BOARD_ID)  );

          //
          //Set chip type
          // 
          if ( (xc_in32(mdrv->mmio_base,XC_RBM_BOARD_ID) & XC_RBM_BOARD_ID_TECH_ID_MASK) == XC_RBM_TECH_ID_JAGUAR )
          {
               mdev->chip_type = CHIP_XCIII_JAGUAR;
          }
          else
          {
               mdev->chip_type = CHIP_XCIII_PANTHER;

          }
          mdrv->pdevice_data = (pxc_dfb_device_data) device_data;
     
          mdrv->accelerator = dfb_gfxcard_get_accelerator( device );
     
          funcs->EngineReset   = xcodeEngineReset;
          funcs->EngineSync    = xcodeEngineSync;
          funcs->CheckState    = xcodeCheckState;
          funcs->SetState      = xcodeSetState;
          funcs->FillRectangle = xcodeFillRectangle;
          funcs->Blit =          xcodeBlit;
     
          //
          //If we running on Panther, Init overlay layer
          // 
          if (mdev->chip_type == CHIP_XCIII_PANTHER)
          {
               D_DEBUG( "Init overlay surface \n" );

               dfb_layers_register( dfb_screens_at(DSCID_PRIMARY),
                                    driver_data, &xcOverlayFuncs );

          }
     }
     else 
     {
          result=DFB_IO;

          D_BUG( "no mmio found\n");

     }

     if (!dfb_core_is_master( core ))
	 {
	 }

     D_DEBUG(" %s exit:\n",__FUNCTION__);

     return result;
}


/*
 *
 *Function Name:driver_init_device
 *
 *Parameters: 
 *             GraphicsDevice -> Framebuffer device
 *             GraphicsDeviceInfo -> DirectFB device capabilities
 *             driver_data -> driver specific structure
 *             device_data -> device specific structure
 *Description:
 *             Fills out the device specific functions(BLT etc) and capabilities(colorkey etc)
 *
 *Returns:
 *             DFB_OK
*/

static DFBResult
driver_init_device( CoreGraphicsDevice     *device,
                    GraphicsDeviceInfo *device_info,
                    void               *driver_data,
                    void               *device_data )
{
     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) driver_data;
     pxc_dfb_device_data mdev = (pxc_dfb_device_data) device_data;
     FBDev *dfb_fbdev = dfb_system_data();
     app_ioctl_command dfb_ioctl_cmd;
     int offset;



     D_DEBUG( "%s enter:\n",__FUNCTION__);

     /* fill out device info*/
     if (mdev->chip_type == CHIP_XCIII_PANTHER)
     {
          snprintf( device_info->name,
                    DFB_GRAPHICS_DRIVER_INFO_NAME_LENGTH,
                    XC_REV_A_DFB_DEVICE_NAME );
     }
     else
     {
          snprintf( device_info->name,
                    DFB_GRAPHICS_DRIVER_INFO_NAME_LENGTH,
                    XC_REV_B_DFB_DEVICE_NAME );

     }

     snprintf( device_info->vendor,
               DFB_GRAPHICS_DRIVER_INFO_VENDOR_LENGTH,
               XC_DFB_DRIVER_VENDOR);

     /* fill device info */
     device_info->caps.flags = CCF_CLIPPING | CCF_READSYSMEM;

     device_info->caps.drawing  = XC3_SUPPORTED_DRAWINGFLAGS;

     device_info->caps.blitting = XC3_SUPPORTED_BLITTINGFLAGS;

     device_info->caps.accel    = XC3_SUPPORTED_DRAWINGFUNCTIONS |
                                  XC3_SUPPORTED_BLITTINGFUNCTIONS;

     device_info->limits.surface_byteoffset_alignment = 0x2000;
     device_info->limits.surface_bytepitch_alignment  = 1;
     device_info->limits.surface_pixelpitch_alignment = 1;


     //
     //Allocate memory for the shred structures between the FB and DFB
     //drivers 
     offset = dfb_gfxcard_reserve_memory( device, sizeof(sw_dve_fifo) );

     if (offset)
     {
          mdrv->shared->uSWFifoOffset = offset;
          mdrv->shared->pSWFifoAddress = dfb_system_video_memory_virtual(offset);
          // Debug: clear buffer
          memset(mdrv->shared->pSWFifoAddress,0x0, sizeof(sw_dve_fifo));
          D_DEBUG( " offset = % x \n", offset);
          D_DEBUG( " pFIFOAddress = % x \n", mdrv->shared->pSWFifoAddress);

          //
          //Inform FB driver of shared structure
          // 
          dfb_ioctl_cmd.uSize = sizeof(app_ioctl_command);
          dfb_ioctl_cmd.uIOCTLCommand = CMD_DFB_IOCTL_SET_SHARED_STRUCT_OFFSET;
          dfb_ioctl_cmd.dfb_shared_struct_offset.uOffset = offset;


          if (ioctl( dfb_fbdev->fd,CMD_APPLICATION_IOCTL , &dfb_ioctl_cmd ) )
          {
               D_ERROR("failed to pass shared structure offset to FB driver");
          }
     }
     else
     {
          D_DEBUG( "failed to allocate shared stucture:\n",__FUNCTION__);

     }

     //
     //Allocate space for DMA LIST (64K)
     // 
     offset = dfb_gfxcard_reserve_memory( device, DMA_DESC_APER_SIZE*2 );

     if (offset)
     {
          mdrv->shared->uDMAListOffset = offset;
          mdrv->shared->pDMAListAddress = dfb_system_video_memory_virtual(offset);

          // Debug: clear buffer
          memset(mdrv->shared->pDMAListAddress,0x0, sizeof(DMA_DESC_APER_SIZE *2));
          D_DEBUG( " DMA List offset = % x \n", offset);
          D_DEBUG( " DMA List pFIFOAddress = % x \n", mdrv->shared->pSWFifoAddress);

          dfb_ioctl_cmd.uSize = sizeof(app_ioctl_command);
          dfb_ioctl_cmd.uIOCTLCommand = CMD_APP_IOCTL_SET_DMA_DESC_OFFSET;
          dfb_ioctl_cmd.dfb_shared_struct_offset.uOffset = offset;


          if (ioctl( dfb_fbdev->fd,CMD_APPLICATION_IOCTL , &dfb_ioctl_cmd ) )
          {
               D_ERROR("failed to pass DMA shared structure offset to FB driver");
          }
     }
     else
     {
          D_DEBUG( "failed to allocate DMA stucture:\n",__FUNCTION__);

     }

     D_DEBUG( "%s exit:\n",__FUNCTION__);

     return DFB_OK;
}


/*
 *
 *Function Name:driver_close_device
 *
 *Parameters: 
 *             GraphicsDevice -> Framebuffer device
 *             driver_data -> driver specific structure
 *             device_data -> device specific structure
 *Description:
 *             Perform any cleanup or statistic loggging before device is shutdown.
 *
 *Returns:
 *             NA
*/

static void
driver_close_device( CoreGraphicsDevice *device,
                     void           *driver_data,
                     void           *device_data )
{

     D_DEBUG(" %s enter:\n",__FUNCTION__);
     D_DEBUG(" %s exit:\n",__FUNCTION__);


}
/*
 *
 *Function Name:driver_close_driver
 *
 *Parameters: 
 *             GraphicsDevice -> Framebuffer device
 *             driver_data -> driver specific structure
 *Description:
 *             Perform any cleanup or statistic loggging before driver is unloaded.
 *
 *Returns:
 *             NA
*/

static void
driver_close_driver( CoreGraphicsDevice *device,
                     void           *driver_data )
{

     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) driver_data;

     D_DEBUG(" %s enter:\n",__FUNCTION__);

     /*unmap MMIO*/
     dfb_gfxcard_unmap_mmio( device, mdrv->mmio_base, -1 );

     D_DEBUG(" %s exit:\n",__FUNCTION__);

}


/************************** Primary Screen functions **************************/

