#include <config.h>

#include <dfb_types.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>


#include <directfb.h>

#include <direct/messages.h>
#include <direct/util.h>

#include <asm/cachectl.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/screens.h>
#include <core/state.h>
#include <core/system.h>
#include <core/gfxcard.h>
#include <core/layers.h>
#include <core/surface.h>
#include <core/palette.h>

#include <display/idirectfbsurface.h>

#include <gfx/convert.h>
#include <gfx/util.h>
#include <gfx/clip.h>

#include <misc/conf.h>

#include <math.h>

#include "viper.h"
#include "xcmmio.h"
#include "xc4fb_if.h"
#include "xcode4_2d.h"
#include "xcode4_osd.h"
#include "xcioctl.h"

D_DEBUG_DOMAIN( Xcode2D, "Xcode2D", "Xcode2D" );

D_DEBUG_DOMAIN( XcodeBlit, "XcodeBlit", "XcodeBlit" );

#define DPRINTF(...) D_DEBUG_AT( Xcode2D, __VA_ARGS__)

#define BLITPRINTF(...) D_DEBUG_AT( XcodeBlit, __VA_ARGS__)

#define DPRINTF_VALUE(x) DPRINTF( #x"=0x%x\n", x)

#define G2D_NEGATIVE_SIGN_MASK 0x4000
#define G2D_VALUE_MASK 0x3fff

//GIC hardware flaws.
#define GIC_HEIGHT_LIMIT 0x1000
#define GIC_WIDTH_LIMIT 0xf00

#define TRANSFER_G2D_POSITION_NUMBER(output,input) \
     if (input >= 0) \
     { \
         output = input;\
     }\
     else \
     { \
         output = G2D_NEGATIVE_SIGN_MASK | ((u32)input & G2D_VALUE_MASK ); \
     }


/*
 *
 *Function Name:Set_Source_Color_Key
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *g2d_op_ctl_value-> operation contorl value 
 *
 *Description:
 *           set the color compare register for source color key, the color key of g2d_op_ctl_value
             will be turned on
 *
 *Returns:
 *           none
*/
static __inline void Set_Source_Color_Key(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev, u32 * g2d_op_ctl_value) 
{ 
    DPRINTF( "source color key  %08x:\n", pmdev->src_colorkey );

    u32 fifo_entry;

    //Ensure there is enough space in the FIFO
    WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 3);

    //color compare function register
    REG_WRITE( G2D_CLR_CMP_FNCT, COMPARE_FNCT_EQUAL);
 

    //color compare color register
    REG_WRITE(G2D_CLR_CMP_CLR, pmdev->src_colorkey); 

    //Mask for color compare function
    REG_WRITE(G2D_CLR_CMP_MASK, COMPARE_MASK_NO_ALPHA); 

    //turn the the color key
    *g2d_op_ctl_value = *g2d_op_ctl_value | (COLOR_KEY_ON << COLOR_KEY_SHIFT);

}

/*
 *
 *Function Name:Set_Destination_Color_Key
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *g2d_op_ctl_value-> operation contorl value 
 *
 *Description:
 *           set the color compare register for destination color key, the color key of g2d_op_ctl_value
             will be turned on
 *
 *Returns:
 *           none
*/
static __inline void Set_Destination_Color_Key(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev, u32 * g2d_op_ctl_value) 
{ 
    u32 fifo_entry;
    DPRINTF( "destination color key  %08x:\n", pmdev->dst_colorkey );

    //Ensure there is enough space in the FIFO
    WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 3);

    //color compare function register
    REG_WRITE( G2D_CLR_CMP_FNCT, ((COMPARE_INPUT_DST_DATA << COMPARE_INPUT_SHIFT) | COMPARE_FNCT_NOT_EQUAL));
 

    //color compare color register
    REG_WRITE(G2D_CLR_CMP_CLR, pmdev->dst_colorkey); 

    //Mask for color compare function
    REG_WRITE(G2D_CLR_CMP_MASK, COMPARE_MASK_NO_ALPHA); 

    //turn the color key
    *g2d_op_ctl_value = *g2d_op_ctl_value | (COLOR_KEY_ON << COLOR_KEY_SHIFT);

}

/*
 *
 *Function Name:G2d_Operation_Start_For_Blit
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *g2d_op_ctl_value-> operation contorl value
 *           base_size : fifo space required
 *
 *Description:
 *           require enough fifo space for Blit operation, set the registers for
 *           source or destination color key if necessary
 *
 *Returns:
 *           none
*/
static __inline void G2d_Operation_Start_For_Blit(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev, u32 * g2d_op_ctl_value, u32 base_size) 
{
     if (pmdev->blittingflags & DSBLIT_SRC_COLORKEY)
     {
         //Set the registers for source color key
         Set_Source_Color_Key(mdrv, pmdev, g2d_op_ctl_value);
     }
     else if (pmdev->blittingflags & DSBLIT_DST_COLORKEY)
     {
         //Set the registers for source color key
         Set_Destination_Color_Key(mdrv, pmdev, g2d_op_ctl_value);
     }
}

/*
 *
 *Function Name:G2d_Operation_Start_For_FillRectangel
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *g2d_op_ctl_value-> operation contorl value
 *           base_size : fifo space required
 *
 *Description:
 *           require enough fifo space for Filling Rectangle operation, set foreground color
 *           register, set the registers for source or destination color key if necessary
 *
 *Returns:
 *           none
*/
static __inline void G2d_Operation_Start_For_FillRectangel(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev, u32 * g2d_op_ctl_value, u32 base_size) 
{

     u32 size_of_operation = base_size;
     u32 fifo_entry;
     
     if (pmdev->drawingflags & DSDRAW_BLEND)
     {
         if ((pmdev->dst_blend == DSBF_ZERO) && (pmdev->src_blend == DSBF_ZERO))
         {
              pmdev->fill_color = 0;
         }
         else
         {
              *g2d_op_ctl_value = *g2d_op_ctl_value | (GFX_ALPHA_MODE_FGND_COLOR_ALPHA << GFX_ALPHA_MODE_SHIFT);
         }
     }

     if (pmdev->drawingflags & DSDRAW_DST_COLORKEY)
     {
         Set_Destination_Color_Key(mdrv, pmdev, g2d_op_ctl_value) ;

         WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 1);

         //Fill Color
         REG_WRITE(G2D_FRGND_COLOR, pmdev->fill_color);
     }
     else
     {
         WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 1);

         //Fill Color
         REG_WRITE(G2D_FRGND_COLOR,pmdev->fill_color);
     }

}

/*
 *
 *Function Name:Set_Clip_Region
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *
 *Description:
 *           set the top,left, bottom, right of region
 *
 *Returns:
 *           none
*/
static __inline void Set_Clip_Region(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev) 
{
     u32 tmp;
     u32 fifo_entry;

     //Ensure there is enough space in the FIFO
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 2);

     //
     //top left of the region
     // 
     REG_WRITE(G2D_SC_TOP_LEFT, 
              ((pmdev->clip.y1 << SC_TOP_SHIFT) | pmdev->clip.x1));

     //
     //bottom right of the region
     // 
     REG_WRITE(G2D_SC_BOTTOM_RIGHT, 
              ((pmdev->clip.y2 << SC_BOTTOM_SHIFT) | pmdev->clip.x2));

}

static __inline void xcodeLockG2d(pxc_dfb_driver_data mdrv)
{
     pthread_mutex_lock(&mdrv->shared->g2d_mutex);
}

static __inline void xcodeUnlockG2d(pxc_dfb_driver_data mdrv)
{
     pthread_mutex_unlock(&mdrv->shared->g2d_mutex);
}

static void check_gic_palette(pxc_dfb_driver_data mdrv, pxc_dfb_device_data mdev)
{
    u32 fifo_entries;

    // update palette table
    if(mdev->text_entries_num)
    {
        int i;

        //Enable CLUT programe
        WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 1);
        GIC0_REG_WRITE(GIC0_CONFIG_CTRL,
          GIC_PROGRAM_ALL_CLUT << GIC0_SRC_CLUT_PROG_EN_SHIFT
      );
      
      for (i = 0 ; i < mdev->text_entries_num ; i ++)
      {
          WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 1);
          GIC0_REG_WRITE(GIC0_LUT_PROGM, mdrv->pdevice_data->text_palette[i]);
      }

      mdev->text_entries_num = 0;
    }
}

static void check_g2d_palette(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev)
{
    if(pmdev->text_entries_num_g2d)
    {
        int i;
        
        xcode4_waitfifo(mdrv, pmdev, 1);
        
        //G2D 8LIT INDEX
        // 
        REG_WRITE(G2D_8LUT_INDEX, 0);
        
        for (i = 0; i < pmdev->text_entries_num_g2d; i ++)
        {
            xcode4_waitfifo(mdrv, pmdev, 1);
            
            REG_WRITE(G2D_LUT_DATA, pmdev->text_palette[i]);    /* Load up the G2D palette ... */
        }

        pmdev->text_entries_num_g2d = 0;
    }
}

/*
 *
 *Function Name:Set_Destination_Registers
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *
          dst_pixel_format_color : destination color type
 *
          dx : the left position 
 *
          dy : the top position
 *
 *Description:
 *           set G2D_DST_OFFSET, G2D_DST_PITCH_FORMAT, G2D_DST_Y_X registers for rectangle 
 *           operations(fill, draw, blit)
 *
 *Returns:
 *           none
*/
static __inline void Set_Destination_Registers(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev, int dst_pixel_format_color, int dx, int dy ) 
{

     u32 u32_dx;
     u32 u32_dy;
     u32 fifo_entry;

     TRANSFER_G2D_POSITION_NUMBER(u32_dx, dx);
     TRANSFER_G2D_POSITION_NUMBER(u32_dy, dy);

     //Ensure there is enough space in the FIFO
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 3);

     //
     //DST offset
     // 
     REG_WRITE(G2D_DST_OFFSET,pmdev->dst_phys);

     //
     //DST pitch/bpp
     // 
     REG_WRITE(G2D_DST_PITCH_FORMAT, 
              (pmdev->dst_pitch | ( dst_pixel_format_color << DST_PIX_FORMAT_SHIFT)));

     //
     //DST x,y
     // 
     REG_WRITE(G2D_DST_Y_X, 
              (u32_dx | (u32_dy << DST_Y_SHIFT)) );


}

/*
 *
 *Function Name:Set_Source_Registers
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *
          src_pixel_format_color : source color type
 *
          *rect : the source rectangle 
 *
 *Description:
 *           set G2D_SRC_PITCH_FORMAT, G2D_SRC_H_W, G2D_SRC_Y_X registers for rectangle 
 *           operations(fill, draw, blit)
 *
 *Returns:
 *           none
*/
static __inline void Set_Source_Registers(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev, int src_pixel_format_color, DFBRectangle *rect ) 
{
     u32 value_of_src_pitch_format;
     u32 u32_dx;
     u32 u32_dy;
     u32 fifo_entry;

     TRANSFER_G2D_POSITION_NUMBER(u32_dx, rect->x);
     TRANSFER_G2D_POSITION_NUMBER(u32_dy, rect->y);

     //if ((pmdev->blittingflags & DSBLIT_SRC_PREMULTIPLY ) ||
     //    (pmdev->blittingflags & DSBLIT_SRC_PREMULTCOLOR))
     //{
     //    value_of_src_pitch_format = SRC_SCALE_BLT_ALPHA_PREMULT_MASK | pmdev->src_pitch | (src_pixel_format_color << SRC_PIX_FORMAT_SHIFT);
     //}
     //else
     //{
     //    value_of_src_pitch_format = pmdev->src_pitch | (src_pixel_format_color << SRC_PIX_FORMAT_SHIFT);
     //}

     value_of_src_pitch_format = pmdev->src_pitch | (src_pixel_format_color << SRC_PIX_FORMAT_SHIFT);


     //Ensure there is enough space in the FIFO
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 4);

     //
     //SRC offset
     // 
     REG_WRITE(G2D_SRC_OFFSET,pmdev->src_phys);

     //
     //SRC pitch/bpp
     // 
     REG_WRITE(G2D_SRC_PITCH_FORMAT, 
              value_of_src_pitch_format);

     //
     //SRC x,y
     // 
     REG_WRITE(G2D_SRC_Y_X, 
              (u32_dx | (u32_dy << SRC_Y_SHIFT)) );

     //
     //SRC (width/height)
     // 
     REG_WRITE(G2D_SRC_H_W,
              ((rect->w) | (rect->h << SRC_H_SHIFT)));

     //Check palette
     check_g2d_palette(mdrv, pmdev);
}

/*
 *
 *Function Name:Set_Source_Registers_For_Stretch_Blit
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *
          src_pixel_format_color : source color type
 *
          *rect : the source rectangle 
 *
 *Description:
 *           set G2D_SRC_PITCH_FORMAT, G2D_SRC_H_W, G2D_SRC_Y_X registers for rectangle 
 *           stretch blit operations
 *
 *Returns:
 *           none
*/
static __inline void Set_Source_Registers_For_Stretch_Blit(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev, int src_pixel_format_color, DFBRectangle *rect ) 
{
     u32 value_of_src_pitch_format;
     u32 u32_dx;
     u32 u32_dy;
     u32 fifo_entry;

     TRANSFER_G2D_POSITION_NUMBER(u32_dx, rect->x);
     TRANSFER_G2D_POSITION_NUMBER(u32_dy, rect->y);

     if (pmdev->blittingflags & DSBLIT_INDEX_TRANSLATION)
     {
         value_of_src_pitch_format = (pmdev->src_pitch * 2) | (src_pixel_format_color << SRC_PIX_FORMAT_SHIFT);
     }
     else
     {
         value_of_src_pitch_format = pmdev->src_pitch | (src_pixel_format_color << SRC_PIX_FORMAT_SHIFT);
     }

     if(dfb_config->g2d_linear_stretch_blit == 0)
     {
         if((pmdev->blittingflags & DSBLIT_LINEAR_STRETCH) ||
            ((pmdev->src_width <= pmdev->dst_width) || (pmdev->src_height <= pmdev->dst_height)))
         {
              value_of_src_pitch_format = value_of_src_pitch_format & (~SCALING_TYPE_MASK);
              value_of_src_pitch_format = value_of_src_pitch_format | (SCALING_TYPE_LERP << SCALING_TYPE_SHIFT);
         }
         else
         {
              value_of_src_pitch_format = value_of_src_pitch_format | (SCALING_TYPE_NEIGHBOR << SCALING_TYPE_SHIFT);
         }
     }
     else if(dfb_config->g2d_linear_stretch_blit == 1)
     {
          value_of_src_pitch_format = value_of_src_pitch_format & (~SCALING_TYPE_MASK);
          value_of_src_pitch_format = value_of_src_pitch_format | (SCALING_TYPE_LERP << SCALING_TYPE_SHIFT);
     }
     else
     {
          value_of_src_pitch_format = value_of_src_pitch_format | (SCALING_TYPE_NEIGHBOR << SCALING_TYPE_SHIFT);
     }

     //Ensure there is enough space in the FIFO
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 4);

     //
     //SRC offset
     // 
     REG_WRITE(G2D_SRC_OFFSET,pmdev->src_phys);

     //
     //SRC pitch/bpp
     // 
     REG_WRITE(G2D_SRC_PITCH_FORMAT, 
              value_of_src_pitch_format);

     //
     //SRC x,y
     // 
     REG_WRITE(G2D_SRC_Y_X, 
              (u32_dx | (u32_dy << SRC_Y_SHIFT)) );

     //
     //SRC (width/height)
     // 
     REG_WRITE(G2D_SRC_H_W,
              ((rect->w) | (rect->h << SRC_H_SHIFT)));

     //Check palette
     check_g2d_palette(mdrv, pmdev);
}

/*
 *
 *Function Name:G2d_Operation_Finish
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *
          *g2d_op_ctl_value : operation control value
 *
          *rect : the destination rectangle 
 *
 *Description:
 *           set G2D_OP_CTL, G2D_DST_H_W_INIT, registers for rectangle 
 *           operations(fill, draw, blit)
 *
 *Returns:
 *           none
*/
static __inline void G2d_Operation_Finish(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev, u32 * g2d_op_ctl_value, DFBRectangle *rect)
{
     u32 gfx_op ;
     u32 fifo_entry;

     gfx_op = (*g2d_op_ctl_value & GFX_OP_MASK ) >> GFX_OP_SHIFT;
     
     if  ((pmdev->drawingflags & DSDRAW_XOR) && 
           (gfx_op == XC_G2D_OP_FILL_RECT))
     {
         *g2d_op_ctl_value = *g2d_op_ctl_value | (GFX_ROP_XORPEN << GFX_ROP_SHIFT) | (GFX_ROP_MODE_MIXED_ENABLE << GFX_ROP_MODE_SHIFT); 
     }
     else if (pmdev->blittingflags & DSBLIT_XOR)
     {
         if ((gfx_op == XC_G2D_OP_COPY_RECT) ||
             (gfx_op == XC_G2D_OP_SCALE_RECT))

         {
             *g2d_op_ctl_value = *g2d_op_ctl_value | (GFX_ROP_XORPEN << GFX_ROP_SHIFT) | (GFX_ROP_MODE_MIXED_ENABLE << GFX_ROP_MODE_SHIFT); 
         }
     }

     //Ensure there is enough space in the FIFO
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 3);

     //G2D OP
     REG_WRITE(G2D_OP_CTL, *g2d_op_ctl_value);

     REG_WRITE(G2D_GUI_CONTROL, 0);
     
     /* Insert this since the GIC/G2D need to be synced up */
     Wait_For_Gic0_Idle;
       
     //Initiator(width/height)
     REG_WRITE(G2D_DST_H_W_INIT, ((rect->w) | (rect->h << DST_H_SHIFT)));
}

/*
 *
 *Function Name:G2d_Operation_Finish_Stretch_Blit
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *
          *g2d_op_ctl_value : operation control value
 *
          *src_rect : the source rectangle 
 *
          *dst_rect : the destination rectangle 
 *
 *Description:
 *           set G2D_SCALERYX, G2D_OP_CTL, G2D_DST_H_W_INIT, registers for rectangle 
 *           stretch blit operations
 *
 *Returns:
 *           none
*/
static __inline void G2d_Operation_Finish_Stretch_Blit( pxc_dfb_driver_data mdrv, 
                                                        pxc_dfb_device_data pmdev, 
                                                        u32 * g2d_op_ctl_value, 
                                                        DFBRectangle *src_rect,
                                                        DFBRectangle *dst_rect )
{
     u32 scaler_x;
     u32 scaler_y;
     u32 fifo_entry;

     if ((pmdev->drawingflags & DSDRAW_XOR) || 
         (pmdev->blittingflags & DSBLIT_XOR))
     {
        *g2d_op_ctl_value = *g2d_op_ctl_value | (GFX_ROP_XORPEN << GFX_ROP_SHIFT) | (GFX_ROP_MODE_MIXED_ENABLE << GFX_ROP_MODE_SHIFT); 
     }

     scaler_x = (src_rect->w << SCALER_MULTIPLY_SHIFT) / dst_rect->w;
     scaler_y = (src_rect->h << SCALER_MULTIPLY_SHIFT) / dst_rect->h;
     
     //Ensure there is enough space in the FIFO
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 4);

     //G2D SCALER FRACYX
     REG_WRITE(G2D_SCALERYX, (scaler_x | (scaler_y << YSCALER_SHIFT)));

     //G2D OP
     REG_WRITE(G2D_OP_CTL, *g2d_op_ctl_value);

     REG_WRITE(G2D_GUI_CONTROL, 0);
    
     Wait_For_Gic0_Idle;
     //Initiator(width/height)
     REG_WRITE(G2D_DST_H_W_INIT, ((dst_rect->w) | (dst_rect->h << DST_H_SHIFT)));
}


/**********************************************************************************************************************/
void Xcode4_Trigger_Gic0_Action(pxc_dfb_driver_data mdrv, 
                                pxc_dfb_device_data mdev,
                                gic_opt_struct * gic_ctl_ptr)
{
    u32 fifo_entries;

    DPRINTF("Xcode4_Trigger_Gic0_Action enter:\n");

    check_gic_palette(mdrv,mdev);

    //make sure GIC0 command fifo is enough
    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 8);

    GIC0_REG_WRITE(GIC0_SRC1_ADR0, gic_ctl_ptr->gic_src1_adr0);
    GIC0_REG_WRITE(GIC0_SRC1_ADR1, gic_ctl_ptr->gic_src1_adr1);
    GIC0_REG_WRITE(GIC0_SRC1_ADR2, gic_ctl_ptr->gic_src1_adr2);
    GIC0_REG_WRITE(GIC0_SRC1_H_W, gic_ctl_ptr->gic_src1_h_w);  
    GIC0_REG_WRITE(GIC0_SRC1_PITCH, gic_ctl_ptr->gic_src1_pitch);
    GIC0_REG_WRITE(GIC0_SRC1_Y_X, gic_ctl_ptr->gic_src1_y_x);
    GIC0_REG_WRITE(GIC0_SRC1_MODE, gic_ctl_ptr->gic_src1_mode); 
    GIC0_REG_WRITE(GIC0_ALPHA_MODE_SRC1, gic_ctl_ptr->gic_alpha_mode_src1);

    //make sure GIC0 command fifo is enough
    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 9);

    //configure src2 as graphic background, src1 will be overwritten by src1
    GIC0_REG_WRITE(GIC0_SRC2_ADR0, gic_ctl_ptr->gic_src2_adr0);
    GIC0_REG_WRITE(GIC0_SRC2_ADR1, gic_ctl_ptr->gic_src2_adr1);
    GIC0_REG_WRITE(GIC0_SRC2_ADR2, gic_ctl_ptr->gic_src2_adr2);

    GIC0_REG_WRITE(GIC0_SRC2_H_W, gic_ctl_ptr->gic_src2_h_w);
    GIC0_REG_WRITE(GIC0_SRC2_Y_X, gic_ctl_ptr->gic_src2_y_x);
    GIC0_REG_WRITE(GIC0_SRC2_PITCH, gic_ctl_ptr->gic_src2_pitch);

    GIC0_REG_WRITE(GIC0_SRC2_MODE, gic_ctl_ptr->gic_src2_mode); 
    GIC0_REG_WRITE(GIC0_ALPHA_MODE_SRC2, gic_ctl_ptr->gic_alpha_mode_src2);
    GIC0_REG_WRITE(GIC0_BACKGRD_COLOR, gic_ctl_ptr->gic_backgrd_color);

    //make sure GIC0 command fifo is enough
    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 6);

    GIC0_REG_WRITE(GIC0_DST_ADR0, gic_ctl_ptr->gic_dst_adr0);
    GIC0_REG_WRITE(GIC0_DST_ADR1, gic_ctl_ptr->gic_dst_adr1);
    GIC0_REG_WRITE(GIC0_DST_ADR2, gic_ctl_ptr->gic_dst_adr2);

    GIC0_REG_WRITE(GIC0_DST_H_W, gic_ctl_ptr->gic_dst_h_w);
    GIC0_REG_WRITE(GIC0_DST_PITCH, gic_ctl_ptr->gic_dst_pitch);
    GIC0_REG_WRITE(GIC0_DST_MODE, gic_ctl_ptr->gic_dst_mode);

    if(gic_ctl_ptr->gic_trigger_command & GIC0_CSC_SRC_SEL_MASK)
    {
        //make sure GIC0 command fifo is enough
        WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 8);
        GIC0_REG_WRITE(GIC0_SRC_CSC_PARM0, gic_ctl_ptr->gic1_src_csc_parm0);
        GIC0_REG_WRITE(GIC0_SRC_CSC_PARM1, gic_ctl_ptr->gic1_src_csc_parm1);
        GIC0_REG_WRITE(GIC0_SRC_CSC_PARM2, gic_ctl_ptr->gic1_src_csc_parm2);
        GIC0_REG_WRITE(GIC0_SRC_CSC_PARM3, gic_ctl_ptr->gic1_src_csc_parm3);
        GIC0_REG_WRITE(GIC0_SRC_CSC_PARM4, gic_ctl_ptr->gic1_src_csc_parm4);
        GIC0_REG_WRITE(GIC0_SRC_CSC_PARM5, gic_ctl_ptr->gic1_src_csc_parm5);
        GIC0_REG_WRITE(GIC0_SRC_CSC_PARM6, gic_ctl_ptr->gic1_src_csc_parm6);
        GIC0_REG_WRITE(GIC0_SRC_CSC_PARM7, gic_ctl_ptr->gic1_src_csc_parm7);
    }

    if(gic_ctl_ptr->gic_trigger_command & GIC0_DST_CLIP_EN_MASK)
    {
        WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 2);
        GIC0_REG_WRITE( GIC0_CLIP_TOP_LEFT, gic_ctl_ptr->gic_clip_top_left);
        GIC0_REG_WRITE( GIC0_CLIP_BOT_RIGHT, gic_ctl_ptr->gic_clip_bot_right);
    }

    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 1);
    GIC0_REG_WRITE(GIC0_CONFIG_CTRL, gic_ctl_ptr->gic_trigger_command);

    DPRINTF("Xcode4_Trigger_Gic0_Action exit\n");
}


/*
 *
 *Function Name:check_src_pixel_format
 *
 *Parameters: format -> Describes the src surface pixel format
 *
 *Description:
 *            Determines whether or not the current ASIC supports format as SRC
 *
 *Returns:
 *        TRUE if supported, otherwise FALSE
 *
*/
static __inline int check_src_pixel_format(int format)
{
     if ( (format == DSPF_RGB32) ||
          (format == DSPF_ARGB) ||
          (format == DSPF_RGB16) ||
          (format == DSPF_ARGB1555) ||
          (format == DSPF_ARGB4444) ||
          (format == DSPF_LUT8) ||
          (format == DSPF_ALUT44) ||
          (format == DSPF_A1) ||
          (format == DSPF_A8) ||
          (format == DSPF_I420) ||
          (format == DSPF_YV16) ||
          (format == DSPF_YUV444P) ||
          (format == DSPF_LUT2))
     {    
          return DFB_TRUE;
     }

     return DFB_FALSE;
}


/*
 *
 *Function Name:check_dst_pixel_format
 *
 *Parameters: format -> Describes the dst surface pixel format
 *
 *Description:
 *            Determines whether or not the current ASIC supports format as DST
 *
 *Returns:
 *        TRUE if supported, otherwise FALSE
 *
*/
static __inline DFBResult check_dst_pixel_format(int format)
{
     if ( (format == DSPF_RGB32) ||
          (format == DSPF_ARGB) ||
          (format == DSPF_RGB16) ||
          (format == DSPF_ARGB1555) ||
          (format == DSPF_ARGB4444) ||
          (format == DSPF_LUT8) ||
          (format == DSPF_ALUT44) ||
          (format == DSPF_A1) ||
          (format == DSPF_I420) ||
          (format == DSPF_LUT2))
     {    
          return DFB_TRUE;
     }

     return DFB_FALSE;
}

static __inline DFBResult check_src_premultiply(DFBSurfaceDrawingFlags blittingflags, DFBSurfaceBlendFunction blend)
{
    return ((blittingflags & DSBLIT_SRC_PREMULTIPLY) || 
           ((blittingflags & (DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_BLEND_COLORALPHA)) && (blend == DSBF_SRCALPHA)));
}

static __inline DFBResult check_color_premultiply(DFBSurfaceDrawingFlags drawingflags, DFBSurfaceBlendFunction blend)
{
    return (drawingflags & DSDRAW_SRC_PREMULTIPLY) || 
           ((drawingflags & DSDRAW_BLEND) && blend == DSBF_SRCALPHA);
}

/*
 *
 *Function Name:xcode_validate_DESTINATION
 *
 *Parameters: 
 *             mdev -> Pointer to device data
 *             state-> Pointer to cardstate decribing destination properties
 *
 *Description:
 *           Called by xcodesetstate() to ensure that the destination properties are properly set
 *           for execution of rendering functions.
 *
 *Returns:
 *         NA
*/
static inline void
xcode_validate_DESTINATION( pxc_dfb_device_data mdev,
                              CardState         *state)
{
     CoreSurface   *destination = state->destination;

     DPRINTF( "xcode_validate_DESTINATION \n" );

     DPRINTF( "width %x \n", destination->config.size.w );
     DPRINTF( "height %x \n", destination->config.size.h );

     //
     // Remember destination parameters for usage in rendering functions. //
     // 

     mdev->dst_phys   = state->dst.phys;  

     mdev->dst_width  = destination->config.size.w;
     mdev->dst_height = destination->config.size.h;

     mdev->dst_format = destination->config.format;
     mdev->dst_caps   = destination->config.caps;
     mdev->dst_bpp    = DFB_BYTES_PER_PIXEL( mdev->dst_format );
     mdev->dst_pitch  = state->dst.pitch;
     mdev->dst_color  = state->color;
     mdev->dst_colorkey = state->dst_colorkey;
     mdev->clip = state->clip;
     mdev->modified = state->modified;
     mdev->dst_addr = (u32)state->dst.addr;
     mdev->dst_blend = state->dst_blend;
     mdev->src_blend = state->src_blend;

     DPRINTF( "xcode validate DESTINATION PHYS = %08x:\n", mdev->dst_phys );
     DPRINTF( "xcode validate DESTINATION PITCH = %08x:\n", mdev->dst_pitch );
     DPRINTF( "xcode validate DESTINATION FORMAT = %08x:\n", mdev->dst_format );

     /* Set the flag. */
     XCODE_VALIDATE_STATE( DESTINATION );
}

/*
 *
 *Function Name:xcode_validate_COLOR
 *
 *Parameters: 
 *             mdev -> Pointer to device data
 *             state-> Pointer to cardstate decribing color properties
 *
 *Description:
 *           Called by xcodesetstate to ensure that the color variables are properly set
 *           for execution of rendering functions.
 *
 *Returns:
 *         NA
*/

static inline void
xcode_validate_COLOR( pxc_dfb_device_data mdev,
                      CardState         *state )
{

     DFBColor       color = state->color;
     u32            clr;

     DPRINTF( "xcode_validate_COLOR \n" );
     
     if (check_color_premultiply(state->drawingflags, state->src_blend))
     {
          u16 ca = color.a + 1;

          color.r = (color.r * ca) >> 8;
          color.g = (color.g * ca) >> 8;
          color.b = (color.b * ca) >> 8;
     }
     
     switch (state->destination->config.format) 
     {
          
          case DSPF_RGB16:
          case DSPF_ARGB1555:
          case DSPF_ARGB4444:
          case DSPF_RGB32:
          case DSPF_ARGB:
          case DSPF_A1:
                   clr = PIXEL_ARGB( color.a,
                                     color.r,
                                     color.g,
                                     color.b );
                   break;

          case DSPF_LUT8:
          case DSPF_ALUT44:
          case DSPF_LUT2:
                   clr = state->color_index ;
                   break;
          default:
                   DPRINTF( "unexpected pixelformat! %08x \n", state->destination->config.format );
                   return;
     }

     mdev->fill_color = clr;

     DPRINTF( "fill color =%08x \n", clr );

     /* Set the flag. */
     XCODE_VALIDATE_STATE( COLOR );

}

/*
 *
 *Function Name:xcode_validate_SOURCE
 *
 *Parameters: 
 *             mdev -> Pointer to device data
 *             state-> Pointer to cardstate decribing destination properties
 *
 *Description:
 *           Called by xcodesetstate() to ensure that the source properties are properly set
 *           for execution of rendering functions.
 *
 *Returns:
 *         NA
*/

static inline void
xcode_validate_SOURCE( pxc_dfb_device_data mdev,
                       CardState         *state)
{
     CoreSurface   *source = state->source;

     DPRINTF( "xcode_validate_SOURCE \n" );
     DPRINTF( "width %x \n", source->config.size.w );

     DPRINTF( "height %x \n", source->config.size.h);

     /* Remember destination parameters for usage in rendering functions. */
     mdev->src_phys   = state->src.phys;
     mdev->src_width  = source->config.size.w;
     mdev->src_height = source->config.size.h;
     mdev->src_format = source->config.format;
     mdev->src_caps   = source->config.caps;
     mdev->src_bpp    = DFB_BYTES_PER_PIXEL( mdev->src_format );
     mdev->src_pitch  = state->src.pitch;
     mdev->src_addr = (u32)state->src.addr;
     mdev->src_blend = state->src_blend;
     mdev->src_colorkey = state->src_colorkey;

     DPRINTF( "xcode validate SOURCE PHYS = %08x:\n", mdev->src_phys );
     DPRINTF( "xcode validate SOURCE PITCH = %08x:\n", mdev->src_pitch );
     DPRINTF( "xcode validate SOURCE FORMAT = %08x:\n", mdev->src_format );

     /* Set the flag. */
     XCODE_VALIDATE_STATE( SOURCE );
}

static bool
xcode_check_g2d_capability(CardState *state)
{
    if((state->blittingflags & DSBLIT_BLEND_ALPHACHANNEL ||
        state->blittingflags & DSBLIT_BLEND_COLORALPHA))
    {
        if (((state->src_blend == DSBF_ONE) && (state->dst_blend == DSBF_ZERO)) ||
            ((state->src_blend == DSBF_ZERO) && (state->dst_blend == DSBF_ZERO)))
        {
             return true;
        }
        else
        {
             return false;
        }
    }
    return true;
}

static bool 
xcode_check_palette(void *drv, void *dev, CardState *state)
{
    
    unsigned int num_entries;
    int i;
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
    CoreSurface      *destination = state->destination;
    CoreSurface      *source      = state->source;
    bool ret;
    DFBColor *src_entries;
    u32 a, r, g, b;
    u32 pal_value;
    u32 update_palette;
    u32 fifo_entries;
         
    DPRINTF("xcode_check_palette enter \n");

    ret = true;

    pmdev->text_entries_num = 0;

    pmdev->text_entries_num_g2d = 0;

    if ( ((destination != NULL) && (source != NULL)) )
    {
        if ((destination->config.format == source->config.format) &&
             (DFB_PIXELFORMAT_IS_INDEXED(destination->config.format) ||
             !DFB_COLOR_BITS_PER_PIXEL(destination->config.format)))
        {
            if (!dfb_palette_equal( source->palette, destination->palette ))
            {
                printf("the source is not equal to destination \n");
                ret = false;
            }
        }
    }

    if (source != NULL)
    {

        BLITPRINTF("Palette blittingflags = 0x%08x\n", state->blittingflags);

        /* If font color is transparent, we can't handle it */
        if(source->config.format == DSPF_A8 && state->color.a == 0 && (state->blittingflags & DSBLIT_COLORIZE))
        {
            ret = false;
        }
        
        /* If it is A8 -> LUT surface, then ignore it */
        if ((DFB_PIXELFORMAT_IS_INDEXED(source->config.format) || 
            (source->config.format == DSPF_A8 && (!DFB_PIXELFORMAT_IS_INDEXED(destination->config.format) ))
             ) &&
            (ret == true))
        {            
            /* We generate a pseudo palette for A8, so this code is not valid for this ... */
            if (source->config.format != DSPF_A8)
            {
                src_entries = source->palette->entries;
        
                if (source->palette->num_entries > 256)
                {
                    num_entries = 256;
                }
                else
                {
                    num_entries = source->palette->num_entries;
                }
            }
            else
            {
                num_entries =   256;
            }

            pmdev->text_entries_num = num_entries;

            pmdev->text_entries_num_g2d = num_entries;
            
            /* 
            *   Check the following conditions:
            *   1. RGBcolor = previousRGBcolor
            *   2. blittingflags = previousBlittingflags
            *   3. destSurfCaps = previousdestSurfCaps
            *   If any of these tests fail, we must regenerate the palette... 
            */
            
            update_palette  =   
                (pmdev->last_text_r != state->color.r)   ||
                (pmdev->last_text_g != state->color.g)   ||
                (pmdev->last_text_b != state->color.b)   ||
                (pmdev->last_text_flags != state->blittingflags) ||
                (pmdev->last_text_format != source->config.format) ||
                (pmdev->last_dest_caps != (destination->config.caps & DSCAPS_PREMULTIPLIED) );

            //Decide on whether we load up the GIC or the G2D 
            //with the palette for the set color ... 
            if (source->config.format == DSPF_A8)
            {
                BLITPRINTF("COLOR = 0x%08X\n", 
                    (state->color.a << 24) | 
                    (state->color.r << 16) |
                    (state->color.g << 8) |
                    (state->color.b));
                
                for (i = 0 ; i < 256 ; i++)
                {
                    a   =    i;

                    if (state->blittingflags & DSBLIT_BLEND_ALPHACHANNEL)
                    {
                        if (state->blittingflags & DSBLIT_BLEND_COLORALPHA)
                        {
                            a = (a * (state->color.a+1)) >> 8;
                        }
                        else if (state->blittingflags & DSBLIT_COLORIZE)
                        {
                            a = (a * (state->color.a+1)) >> 8;
                        }
                    }
                    else
                    {
                        if (state->blittingflags & DSBLIT_BLEND_COLORALPHA)
                        {
                            a = state->color.a;
                        }
                        else if (state->blittingflags & DSBLIT_COLORIZE)
                        {
                            a = (a * (state->color.a+1)) >> 8;
                        }
                    }
                    
                    if (state->blittingflags & DSBLIT_COLORIZE)
                    {
                        //r   =   pmdev->src_color_r;
                        //g   =   pmdev->src_color_g;
                        //b   =   pmdev->src_color_b;
                        //The above cannot be done since it is not guaranteed that this information
                        //will be set in the structure ... 
                        r       =   state->color.r;
                        g       =   state->color.g;
                        b       =   state->color.b;
                    }
                    else
                    {
                        r   =   0xFF;
                        g   =   0xFF;
                        b   =   0xFF;
                    }

                    if (state->blittingflags & DSBLIT_SRC_PREMULTIPLY)
                    {
                        r   =   ((r + 1)   * i) >> 8;
                        g   =   ((g + 1)   * i) >> 8;
                        b   =   ((b + 1)   * i) >> 8;
                    }

                    if ((state->blittingflags & (DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_BLEND_COLORALPHA)) && 
                        state->src_blend == DSBF_SRCALPHA)
                    {
                        r   =   ((r + 1)   * a) >> 8;
                        g   =   ((g + 1)   * a) >> 8;
                        b   =   ((b + 1)   * a) >> 8;
                    }
                    
                    pal_value   =   (a << 24) |
                                    (r << 16) |
                                    (g << 8)  |
                                    (b << 0)  ;
                                    
                    pmdev->text_palette[i]  =   pal_value;
                }

                BLITPRINTF("->PALETTE[128] = 0x%08X\n", pmdev->text_palette[128]);
            }
            else
            {
                BLITPRINTF("PALETTE = 0x%08X\n", 
                    (src_entries[0].a << 24) | 
                    (src_entries[0].r << 16) |
                    (src_entries[0].g << 8) |
                    (src_entries[0].b));

                BLITPRINTF("COLOR = 0x%08X\n", 
                    (state->color.a << 24) | 
                    (state->color.r << 16) |
                    (state->color.g << 8) |
                    (state->color.b));
                
                for (i = 0 ; i < num_entries ; i++)
                {
                    a = src_entries[i].a;
                    r = src_entries[i].r;
                    g = src_entries[i].g;
                    b = src_entries[i].b;

                    if(state->blittingflags & DSBLIT_BLEND_ALPHACHANNEL)
                    {
                        if(state->blittingflags & DSBLIT_BLEND_COLORALPHA)
                        {
                            a = (a * (state->color.a + 1)) >> 8;
                        }
                    }
                    else
                    {
                        if(state->blittingflags & DSBLIT_BLEND_COLORALPHA)
                        {
                            a = state->color.a;
                        }
                    }
                    
                    if (state->blittingflags & DSBLIT_COLORIZE)
                    {
                        //r   =   pmdev->src_color_r;
                        //g   =   pmdev->src_color_g;
                        //b   =   pmdev->src_color_b;
                        //The above cannot be done since it is not guaranteed that this information
                        //will be set in the structure ... 
                        r       =   r * (state->color.r + 1) >> 8;
                        g       =   g * (state->color.g + 1) >> 8;
                        b       =   b * (state->color.b + 1) >> 8;
                    }

                    if (state->blittingflags & DSBLIT_SRC_PREMULTIPLY)
                    {
                        r   =   (r   * (a+1)) >> 8;
                        g   =   (g   * (a+1)) >> 8;
                        b   =   (b   * (a+1)) >> 8;
                    }

                    if ((state->blittingflags & (DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_BLEND_COLORALPHA)) && 
                        state->src_blend == DSBF_SRCALPHA)
                    {
                        r   =   (r   * (a+1)) >> 8;
                        g   =   (g   * (a+1)) >> 8;
                        b   =   (b   * (a+1)) >> 8;
                    }
                    
                    pal_value   =   (a << 24) |
                                    (r << 16) |
                                    (g << 8)  |
                                    (b << 0)  ;
                                    
                    pmdev->text_palette[i]  =   pal_value;
                }

                BLITPRINTF("->PALETTE = 0x%08X\n", pmdev->text_palette[0]);
            }
        }
    }

    return ret;

}



/*
 *
 *Function Name:xcode_validate_dfb_drawflags
 *
 *Parameters: 
 *             flags -> Drawing flags requested
 *
 *Description:
 *           Called by xcodeCheckState() to verify that we do indeed support 
 *           the requested drawing operation
 *
 *Returns:
 *         DFB_OK if supported, otherwise DFB_UNSUPPORTED
*/

static DFBResult xcode_validate_dfb_drawflags( u32 flags)
{
     DFBResult result=DFB_OK;
     int  i;

     if (flags & ~XC4_SUPPORTED_DRAWINGFLAGS)
     {
          DPRINTF( "asked for unsupported drawing function ....\n");
          result = DFB_UNSUPPORTED;
     }
     else
     {
          DPRINTF( "asked for supported drawing functions ....\n");
     }

     return result;
}

/*
 *
 *Function Name:xcode_validate_dfb_blitflags
 *
 *Parameters: 
 *             flags -> Blitting flags requested
 *
 *Description:
 *           Called by xcodeCheckState() to verify that we do indeed support 
 *           the requested blitting operation
 *
 *Returns:
 *         DFB_OK if supported, otherwise DFB_UNSUPPORTED
*/
static DFBResult xcode_validate_dfb_blitflags( u32 flags)
{
     DFBResult result=DFB_OK;
     int  i;

     if (flags & ~XC4_SUPPORTED_BLITTINGFLAGS)
     {
          DPRINTF( "asked for unsupported blitting flags....\n");
          result = DFB_UNSUPPORTED;
     }
     else
     {
          DPRINTF( "asked for supported blitting functions ....\n");
     }

     return result;
}

/*
 *
 *Function Name:xcode_validate_dfb_accelfunction
 *
 *Parameters: 
 *             flags -> Type of operation requested(StretchBLT etc)
 *
 *Description:
 *           Called by xcodeCheckState() to verify that we do indeed support 
 *           the requested operation
 *
 *Returns:
 *         DFB_OK if supported, otherwise DFB_UNSUPPORTED
*/
static DFBResult xcode_validate_dfb_accelfunction( u32 accel)
{
     DFBResult result=DFB_OK;
     int  i;

     if (accel & ~(XC4_SUPPORTED_DRAWINGFUNCTIONS | XC4_SUPPORTED_BLITTINGFUNCTIONS) )
     {
          DPRINTF( "asked for unsupported accel functions ....\n");
          result = DFB_UNSUPPORTED;
     }
     else
     {
          DPRINTF( "asked for supported accel functions ....\n");

     }

     return result;
}

extern void set_hd_comp_enable(pxc_dfb_driver_data mdrv, int enable);

/*
 *
 *Function Name:xcodeEngineReset
 *
 *Parameters: 
 *             drv-> Pointer to driver data
 *             dev-> Pointer to device data
 *
 *Description:
 *             Reset engine
 *Returns:
 *             NA
*/
void xcodeEngineReset( void *drv, void *dev )
{
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data mdev = (pxc_dfb_device_data) dev;
    app_ioctl_command dfb_ioctl_cmd;

    u32 tmp;

    DPRINTF( "xcode DFB driver xcodeEngineReset enter:\n" );

    //
    //For reseting engine, we use toggle the G2D and FES GC_RESET register
    // 

    RESET_TOGGLE(mdrv, CG_RESET_REG1, G2D_RESET_MASK);

    RESET_TOGGLE(mdrv, CG_RESET_REG1, GIC_RESET_MASK);

    REG_WRITE(GIC_CFIFO_CTRL, GIC_BOTH_CMD_FIF0);

    REG_WRITE(DISP_HD_COMP_ALPHA, 0);

    REG_WRITE(DISP_HD_COMP_ALPHA_MODE, 0x80801003);

    tmp = REG_READ(CG_RESET_REG) & DISP_RESET_MASK;

    if (tmp == DISP_RESET_MASK)
    {
        RESET_TOGGLE(mdrv, CG_RESET_REG, DISP_RESET_MASK);
    }

    //wait screen cleared
    usleep(20000);
    
    //reset compositor and memory controller
    set_hd_comp_enable(mdrv, 0);

    REG_WRITE( DISP_SFT_RESET, 0x0000);

    usleep(1);

    REG_WRITE( DISP_SFT_RESET, 0xffff);

    //wait reset stable
    usleep(20000);

    REG_WRITE(DISP_HD_COMP_ALPHA, 0);

    REG_WRITE(G2D_LOWER_VALID_MEM, 0x00000000);

    REG_WRITE(G2D_TOP_VALID_MEM, HIGHVALID_MASK);

    REG_WRITE(G2D_BUFFER_CTRL, SRCBUFFER_EMERGENCY_FULL_MASK);

    //change G2D_MARB_CTRL\MEM_DENSITY to 0 according to David Yeh's suggestion
    tmp = REG_READ(G2D_MARB_CTRL);
    tmp &= ~G2D_MARB_CTRL_MEM_DENSITY_MASK;
    REG_WRITE(G2D_MARB_CTRL, tmp);

    DPRINTF( "xcode DFB driver xcodeEngineReset exit:\n" );

}


/*************************************************************************************************** 
 
  return 1 means not support port duff type
 
***************************************************************************************************/
bool xcode_check_porter_duff(CardState *state)
{
    bool ret_val = 0;

    if ((state->src_blend == DSBF_SRCALPHA) && (state->dst_blend == DSBF_INVSRCALPHA))
    {
        //DSPD_NONE
        ;
    }
    else if ((state->src_blend == DSBF_ZERO) && (state->dst_blend == DSBF_ZERO))
    {
        //DSPD_CLEAR
        ;
    }
    else if ((state->src_blend == DSBF_ONE) && (state->dst_blend == DSBF_ZERO))
    {
        //DSPD_SRC
        ;
    }
    else if ((state->src_blend == DSBF_ONE) && (state->dst_blend == DSBF_INVSRCALPHA))
    {
        //DSPD_SRC_OVER
        ;
    }
    else
    {
        ret_val = 1;
    }

    return ret_val;
}

/*
 *
 *Function Name:xcodeCheckState
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *state-> Indicates current state of card(clr keys etc)
 *           accel-> Indicates what type of operation is being performed
 *
 *Description:
 *           Parses the accel parameters to ensure current ASIC can carry out intended  
 *           operation, if it can't, it doesnt set any bits within the state->accel variable
 *
 *Returns:
 *           N/A
*/
void xcodeCheckState( void *drv, void *dev,
                      CardState *state, DFBAccelerationMask accel )
{

     int supported_accel=0;   

     DPRINTF( "xcodeCheckState enter:\n" );

     DPRINTF( "driver accel= %08x \n", accel );

     /*
      * Check desired function against support functions.
     */
     if (xcode_check_porter_duff(state))
     {
         DPRINTF( "unsupported Porter Duff, src_blend 0x%x dst_blend 0x%x \n", state->src_blend, state->dst_blend);
     }
     else if (xcode_validate_dfb_accelfunction(accel) == DFB_OK)
     {
          /*
          * Check destination pixel format
          */
          if (check_dst_pixel_format(state->destination->config.format))
          {
               /*
               * Check if we drawing or blitting  
               */
               if (DFB_DRAWING_FUNCTION( accel ) ) 
               {
                    /* Return if unsupported drawing flags are set. */
                    if (xcode_validate_dfb_drawflags(state->drawingflags)== DFB_OK)
                    {
                         supported_accel = accel;
                    }
               }
               else 
               {
                    /*
                    * We are doing some sor of BLT
                    */
                    if (xcode_validate_dfb_blitflags(state->blittingflags)== DFB_OK)
                    {
                         //
                         // No format conversion supported. //
                         // 
                         if ((!check_src_pixel_format(state->source->config.format))  ||
                             (!xcode_check_palette(drv, dev, state)))
                         {
                              DPRINTF( "src pixel format not supported, %08x \n",
                                                       state->source->config.format );
                         }
                         else
                         {
                              //
                              //Setup supported functions/settings 
                              // 
                              supported_accel = accel;
                         }
                    }
               }
          }
          else
          {
               DPRINTF( "dst pixel format not supported %08x \n",
                                        state->destination->config.format );
          }
     }

     //
     //The supported_accel variable is setup whenever the driver has determined that the 
     //request settings can be accomodated, otherwise, the value will be 0!
     // 
     state->accel |= supported_accel;

     DPRINTF( "xcode DFB driver state->accel %08x \n", state->accel );

     DPRINTF( "xcodeCheckState exit:\n" );

}

void xcodeDetermineWorkingRegion(void *drv, void *dev, DFBRectangle *rect, int dx, int dy)
{
    /* This function will simply determine the working region... */
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
    
    /* 
    *   This function can be greatly simplified by doing the following:
    *   a. Forming a rectangle with xywh = [dx dy rect->w rect->h] (Save a copy as well)
    *   b. Invoke dfb_clip_rectangle with the rectangle and the current clip rectangle
    *   c. Compare output of (b) to the original rectangle (dx and dy)
    *   d. Adjust source surface x and y by the difference between original and new
    */
    
    
    DFBRectangle rect_dst_adjust, rect_src_adjust;
    int dbx, dby;
    u32 offset_w = 0, offset_h = 0;
    
    /* Copy over source rectangle information... */
    //memcpy(&rect_dst_adjust, rect, sizeof(DFBRectangle));
    rect_dst_adjust.x  =   dx; rect_dst_adjust.y = dy;
    rect_dst_adjust.w  =   rect->w; rect_dst_adjust.h = rect->h;
    
    /* Comparison rectangle */
    rect_src_adjust.x  =   rect->x; rect_src_adjust.y = rect->y;
    rect_src_adjust.w  =   rect->w; rect_src_adjust.h = rect->h;
    
    DPRINTF("Original dest rectangle [XYWH]: %i %i %i %i\n",
                rect_dst_adjust.x, rect_dst_adjust.y, rect_dst_adjust.w, rect_dst_adjust.h);
    DPRINTF("Original source rectangle [XYWH]: %i %i %i %i\n",
                rect_src_adjust.x, rect_src_adjust.y, rect_src_adjust.w, rect_src_adjust.h);
    
    DPRINTF("dx: %i, dy: %i\n", dx, dy);
    /* Find the intersection of the clipping rectangle (0 0 -> nnnn) */
    dfb_clip_rectangle(&pmdev->clip, &rect_dst_adjust);
   
    DPRINTF("After dfb_clip_rectangle [XYWH]: %u %u %u %u\n",
                rect_dst_adjust.x, rect_dst_adjust.y, rect_dst_adjust.w, rect_dst_adjust.h);
    /* Compare rect_dst_adjust.x and rect_dst_adjust.y with dx and dy, determine direction of change
    *  and adjust rect_src_adjust.x and rect_src_adjust.y accordingly...
    *  I.e rect_dst_adjust.n = dn ; Do nothing (no change)
    *      rect_dst_adjust.n > dn ; Adjusted to the right, or below (only probable case, add to src.x and src.y the delta )
    *      rect_dst_adjust.n < dn ; Adjusted to the left, or above (do not handle this for now)
    *  Assume that the source width and height have already been accounted for... 
    */
    if (rect_dst_adjust.x > dx)
    {
        offset_w = ABS(rect_dst_adjust.x - dx);
    }
    else if (rect_dst_adjust.x < dx)
    {
        offset_w = 0;
        DPRINTF("WARN: rect_dst_adjust.x is < dx !!\n");
    }
    else
    {
        offset_w = 0;
    }
    DPRINTF("Calculated offset_w: %u\n", offset_w);
    
    /* Now process Y portion of this... */
    if (rect_dst_adjust.y > dy)
    {
        offset_h = ABS(rect_dst_adjust.y - dy);
    }
    else if (rect_dst_adjust.y < dy)
    {
        offset_h    =   0;
        DPRINTF("WARN: rect_dst_adjust.y is < dy !!\n");
    }
    else
    {
        offset_h = 0;
    }
    DPRINTF("Calculated offset_h: %u\n", offset_h);
    
    rect_src_adjust.x+=offset_w;
    rect_src_adjust.y+=offset_h;
    
    /* 
    *  rect_src_adjust has the source rectangle (the working region, xywh).. 
    *  While rect_dst_adjust.x and rect_dst_adjust.y have the adjusted X and Y on dest
    *  Copy over the adjusted widths and heights..
    */
    rect_src_adjust.w = rect_dst_adjust.w;
    rect_src_adjust.h = rect_dst_adjust.h;
    
    pmdev->target_dx    =   rect_dst_adjust.x;
    pmdev->target_dy    =   rect_dst_adjust.y;
    
    pmdev->rectsrc_x    =   rect_src_adjust.x;
    pmdev->rectsrc_y    =   rect_src_adjust.y;
    pmdev->rectsrc_w    =   rect_src_adjust.w;
    pmdev->rectsrc_h    =   rect_src_adjust.h;
    if (pmdev->target_dx < 0 || pmdev->target_dy < 0)
    {
        DPRINTF("target_dx or target_dy is < 0!!\n");
        DPRINTF("target_dx: %i, target_dy: %i\n", 
                    pmdev->target_dx,
                    pmdev->target_dy);
    }
    DPRINTF("Adjusted source rectangle...\n");
    DPRINTF("XYWH = %i %i %i %i\n", rect_src_adjust.x, rect_src_adjust.y,
                rect_src_adjust.w, rect_src_adjust.h);
    DPRINTF("Adjusted destination rectangle...\n");
    DPRINTF("XYWH = %i %i %i %i\n", rect_dst_adjust.x, rect_dst_adjust.y,
                rect_dst_adjust.w, rect_dst_adjust.h);
}


void xcodeDetermineStretchWorkingRegion(void *drv, void *dev, DFBRectangle *srect, DFBRectangle *drect)
{
     /* This function will simply determine the working region when using a stretch operation.... */
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
    
    DFBRectangle    source_rect, dest_rect;
    
    /* Copy over source/dest regions into a temporary buffer ... */
    
    source_rect.x   =   srect->x;   source_rect.y   =   srect->y;
    source_rect.w   =   srect->w;   source_rect.h   =   srect->h;
    
    dest_rect.x     =   drect->x;   dest_rect.y     =   drect->y;
    dest_rect.w     =   drect->w;   dest_rect.h     =   drect->h;
    
    DPRINTF("SRC[XYWH]: %i %i %i %i\n",
            source_rect.x, source_rect.y, 
            source_rect.w, source_rect.h 
            );
    
    DPRINTF("DST[XYWH]: %i %i %i %i\n",
            dest_rect.x, dest_rect.y,
            dest_rect.w, dest_rect.h
            );

    /* Invoke dfb_clip_stretchblit to clip both rectangles... */
    dfb_clip_stretchblit(&pmdev->clip, &source_rect, &dest_rect);
    
    pmdev->rect_scalesrc_w  =   source_rect.w;
    pmdev->rect_scalesrc_h  =   source_rect.h;
    
    pmdev->rectsrc_x        =   source_rect.x;
    pmdev->rectsrc_y        =   source_rect.y;
    
    pmdev->target_dx        =   dest_rect.x;
    pmdev->target_dy        =   dest_rect.y;
    
    pmdev->rectsrc_w        =   dest_rect.w;
    pmdev->rectsrc_h        =   dest_rect.h;
    
    /* 
    *   Parameters have been calculated and are as follows: 
    *   rect_scalesrc_w/rect_scalesrc_h : Source rectangle (width and height)
    *   rectsrc_x/rectsrc_y : Starting position on source
    *   target_dx, target_dy: Target position on the destination surface
    *   rectsrc_w/rectsrc_h: Target width and height on the destination surface...
    */
    return;
}

static u32 xcodeGicGetFormatCode(DFBSurfacePixelFormat dpf)
{
    u32 gic_code;
    
    switch (dpf)
    {
        case DSPF_LUT8:
        case DSPF_A8:
            gic_code    =    GIC_LUT8;
            break;
            
        case DSPF_ARGB:
            gic_code    =    GIC_AYUV444;
            break;
            
        case DSPF_RGB16:
            gic_code    =    GIC_RGB565;
            break;
            
        default:
            gic_code     =    GIC_AYUV444;
            break;
    }
    return gic_code;
}

static int get_pixel_format_color(DFBSurfacePixelFormat dsf_color)
{
    int pixel_format_color;

    switch (dsf_color)
    {
        case DSPF_RGB32:
        case DSPF_ARGB:
            pixel_format_color = PIX_FORMAT_RGB32;
            break;

        case DSPF_RGB16:
            pixel_format_color = PIX_FORMAT_RGB565;
            break;

        case DSPF_ARGB1555:
            pixel_format_color = PIX_FORMAT_RGB1555;
            break;

        case DSPF_ARGB4444:
            pixel_format_color = PIX_FORMAT_RGB4444;
            break;

        case DSPF_LUT8:
        case DSPF_A8:
            pixel_format_color = PIX_FORMAT_LUT8;
            break;

        case DSPF_ALUT44:
            pixel_format_color = PIX_FORMAT_LUT4;
            break;

        case DSPF_LUT2:
            pixel_format_color = PIX_FORMAT_LUT2;
            break;

        case DSPF_A1:
            pixel_format_color = PIX_FORMAT_MONOCHROME;
            break;

        default:
            pixel_format_color = PIX_FORMAT_UNKNOWN;
            break;
    }

    return pixel_format_color;
}

static int get_gic_pixel_bytes( int format)
{
     int bytes = 0;
     switch(format)
     {
          case GIC_LUT8:
               bytes = 1;
               break;
          case GIC_RGB565:
          case GIC_ARGB1555:
               bytes = 2;
               break;
          case GIC_AYUV444:
          case GIC_ARGB8888:
               bytes = 4;
               break;
     }
     return bytes;
}

static void xcodeGicAdjustOffset( gic_opt_struct *gic_opt )
{
     int src0_addr, src0_x, src0_y, src0_pitch, src0_bytes, src0_h, src0_w;
     int src1_addr, src1_x, src1_y, src1_pitch, src1_bytes, src1_h, src1_w;
     
     int src0_x_old, src1_x_old, dst_x_old;
     int dst_addr, dst_x, dst_y, dst_pitch, dst_bytes;
     int dst_pixels, dst_w, dst_h;
     int gic_offset;
     int clipx1, clipy1, clipx2, clipy2;
     
     src0_addr = gic_opt->gic_src1_adr0;
     src0_x = gic_opt->gic_src1_y_x & GIC0_SRC1_X_MASK;
     src0_y = (gic_opt->gic_src1_y_x & GIC0_SRC1_Y_MASK) >> GIC0_SRC1_Y_SHIFT;
     src0_bytes = get_gic_pixel_bytes(gic_opt->gic_src1_mode & GIC0_SRC1_FMT_MASK);
     src0_pitch = gic_opt->gic_src1_pitch;
     src0_w = gic_opt->gic_src1_h_w & GIC0_SRC1_W_MASK;
     src0_h = (gic_opt->gic_src1_h_w & GIC0_SRC1_H_MASK) >> GIC0_SRC1_H_SHIFT;

     src1_addr = gic_opt->gic_src2_adr0;
     src1_x = gic_opt->gic_src2_y_x & GIC0_SRC2_X_MASK;
     src1_y = (gic_opt->gic_src2_y_x & GIC0_SRC2_Y_MASK) >> GIC0_SRC2_Y_SHIFT;
     src1_bytes = get_gic_pixel_bytes(gic_opt->gic_src2_mode & GIC0_SRC2_FMT_MASK);
     src1_pitch = gic_opt->gic_src2_pitch;
     src1_w = gic_opt->gic_src2_h_w & GIC0_SRC2_W_MASK;
     src1_h = (gic_opt->gic_src2_h_w & GIC0_SRC2_H_MASK) >> GIC0_SRC2_H_SHIFT;

     dst_addr = gic_opt->gic_dst_adr0;
     dst_x = gic_opt->gic_dst_y_x & GIC0_DST_X_MASK;
     dst_y = (gic_opt->gic_dst_y_x & GIC0_DST_Y_MASK) >> GIC0_DST_Y_SHIFT;
     dst_bytes = get_gic_pixel_bytes(gic_opt->gic_dst_mode & GIC0_DST_FMT_MASK);
     dst_pitch = gic_opt->gic_dst_pitch;
     dst_w = gic_opt->gic_dst_h_w & GIC0_DST_W_MASK;
     dst_h = (gic_opt->gic_dst_h_w & GIC0_DST_H_MASK) >> GIC0_DST_H_SHIFT;

     if(!src0_bytes || !dst_bytes)
     {
          return;
     }
     /* 
     *  The data comes in from the upper layers as follows : 
     *  SRC1_PITCH = src pitch
     *  SRC1Y_X = (rectsrc_x, rectsrc_y) -> XY index into source image
     *  SRC1_ADR0 = source surface base address
     *  SRC1_H_W = (rectsrc_w, rectsrc_h) -> WH of part from source image
     *  SRC2_PITCH = dst pitch
     *  SRC2_ADR0 = destination address
     *  SRC2Y_X   = (target_dx, target_dy)
     *  SRC2H_W   = (rectsrc_w, rectsrc_h)
     *  DST_PITCH = dst pitch
     *  DST_H_W   = (rectsrc_w, rectsrc_h)
     *  DST_ADR0  = Destination Address
     *  DST_YX    = (target_dx, target_dy)
     */
     
     //Address is 32byte aligned
     src0_addr = src0_addr + src0_y * src0_pitch + src0_x * src0_bytes >> 5 << 5;
     src0_x = src0_x % (32 / src0_bytes);
     src0_y = 0;

     if(src1_bytes)
     {
         src1_addr = src1_addr + src1_y * src1_pitch + src1_x * src1_bytes >> 5 << 5;
         src1_x = src1_x % (32 / src1_bytes);
         src1_y = 0;
     }

     dst_pixels = (32 / dst_bytes);
     dst_addr = dst_addr + dst_y * dst_pitch + dst_x * dst_bytes >> 5 << 5;
     dst_x = dst_x % dst_pixels;
     dst_y = 0;
     
     clipx1 =   dst_x;
     clipx2 =   (dst_x + dst_w) - 1;
     clipy1 =   0;
     clipy2 =   dst_h-1;
     
     /* We need to extend the width as well, if we are off */
     /* In addition, the clipping rectangle may need to be adjusted */
     /* dst and src2 does not need to be positioned exactly ... */
     
     //dst_w+=    src0_x;
     //calculate relative offset
     
     src0_x_old = src0_x;
     src1_x_old = src1_x;
     
     dst_x_old  = dst_x;
     
     
     src0_x = dst_x - src0_x;
     src1_x = dst_x - src1_x;
     
     //make sure offset is always non-negative
     
     
     //This will break in the case of A8 -> ARGB since we will need to do multiple adjustments
     //to make the offset positive 
     while (src0_x < 0)
     {
          dst_addr -= 32;
          src0_x += dst_pixels;
          src1_x += dst_pixels;
          dst_x  += dst_pixels;
          dst_w  += dst_pixels;
          
          clipx1 += dst_pixels;
          clipx2 += dst_pixels;
     }
     
     /* 
     *  When doing the adjustments, we take the largest width and use that
     *  for the operation
     */
     if ( (src0_w + src0_x) > (dst_w + dst_x) )
     {
        dst_w   = (src0_w + src0_x);
     
     }
     else if ( (src0_w + src0_x) <= (dst_w + dst_x) )
     {
        dst_w   =   (dst_w + dst_x);   
     }
     else
     {
        dst_w   =   (dst_w + dst_x);
     }
     
     
     /* Need to also take into account the case in which dstx != 0, but src0x == 0*/
     
     //set clip window (this may need to be different)
     gic_opt->gic_clip_top_left = ((clipx1 << GIC0_CLIP_LEFT_SHIFT) & GIC0_CLIP_LEFT_MASK)  | ((clipy1 << GIC0_CLIP_TOP_SHIFT) & GIC0_CLIP_TOP_MASK);
     //gic_opt->gic_clip_bot_right = (dst_x + dst_w) | ( (dst_h-1) << GIC0_CLIP_BOT_SHIFT);
     gic_opt->gic_clip_bot_right = ((clipx2 << GIC0_CLIP_RIGHT_SHIFT) & GIC0_CLIP_RIGHT_MASK)  | ((clipy2 << GIC0_CLIP_BOT_SHIFT) & GIC0_CLIP_BOT_MASK);
     gic_opt->gic_trigger_command |= GIC0_DST_CLIP_EN_MASK;

     //adjust settings
     gic_opt->gic_src1_adr0 = src0_addr;
     gic_opt->gic_src1_y_x = src0_x;
     gic_opt->gic_src2_adr0 = dst_addr;
     gic_opt->gic_src2_y_x = 0;
     gic_opt->gic_dst_adr0 = dst_addr;
     gic_opt->gic_dst_y_x = 0;
     gic_opt->gic_dst_h_w = dst_w | (dst_h << GIC0_DST_H_SHIFT);
     gic_opt->gic_src1_h_w  = dst_w | (dst_h << GIC0_SRC1_H_SHIFT);
     gic_opt->gic_src2_h_w  = dst_w | (dst_h << GIC0_SRC2_H_SHIFT);     
    
}

static bool xcodeIsIndexedColor( u32 format)
{
    return DFB_PIXELFORMAT_IS_INDEXED(format) || !DFB_COLOR_BITS_PER_PIXEL(format);
}


static bool xcodeGicLimitationCheck(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev)
{
     if (pmdev->rectsrc_h < GIC_MIN_Y_SIZE || pmdev->rectsrc_w < GIC_MIN_X_SIZE)
     {
        /* Fail blit if we cannot handle properly... */
        return false;
     }
     
     return true;
}


static bool xcodeNeedPremultiply( void *drv, void *dev )
{
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;

    if( xcodeIsIndexedColor(pmdev->src_format) )
    {
        return false;
    }

    if( pmdev->blittingflags & DSBLIT_SRC_PREMULTIPLY && pmdev->src_blend == DSBF_SRCALPHA)
    {
        return true;
    }

    if( !(pmdev->blittingflags & (DSBLIT_BLEND_COLORALPHA | DSBLIT_COLORIZE | DSBLIT_SRC_PREMULTCOLOR)) )
    {
        return false;
    }
    
    return true;
}

static bool xcodeGicPremultiply( void *drv, void *dev)
{
     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
     gic_opt_struct gic_ctl = {0};
     u8 a_factor = 255, r_factor = 255, g_factor = 255, b_factor = 255;
     int fixed_alpha = 0;


     //check limitation
     if(!xcodeGicLimitationCheck(mdrv, pmdev))
     {
          return false;
     }

     if(pmdev->src_format != DSPF_ARGB && pmdev->src_format != DSPF_RGB16)
     {
        /* Only ARGB format can be handled */
        return false;
     }

     BLITPRINTF("PREMULTIPLY blittingflags = %x, color = %x\n", pmdev->blittingflags, 
        (pmdev->src_color_alpha << 24) | (pmdev->src_color_r << 16) | (pmdev->src_color_g << 8) | (pmdev->src_color_b));

     /* Calculate CSC coeffient */
     if(pmdev->blittingflags & DSBLIT_BLEND_ALPHACHANNEL)
     {
        if(pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA)
        {
            a_factor = pmdev->src_color_alpha;
            if(pmdev->src_caps & DSCAPS_PREMULTIPLIED)
            {
                r_factor = pmdev->src_color_alpha;
                g_factor = pmdev->src_color_alpha;
                b_factor = pmdev->src_color_alpha;
            }
        }
     }
     else
     {
        if(pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA)
        {
            a_factor = pmdev->src_color_alpha;
            fixed_alpha = 1;
        }
     }

     /* Check COLORIZE flag */
     if(pmdev->blittingflags & DSBLIT_COLORIZE)
     {
        r_factor = (r_factor + 1) * pmdev->src_color_r >> 8;
        g_factor = (g_factor + 1) * pmdev->src_color_g >> 8;
        b_factor = (b_factor + 1) * pmdev->src_color_b >> 8;
     }

     BLITPRINTF("FACTOR = %d,%d,%d,%d FIXED = %d\n", 
        a_factor, r_factor, g_factor, b_factor, fixed_alpha);

     BLITPRINTF("TEMPADDR = 0x%x\n", pmdev->temp_render_physical);

     /* If RGB16, we need to convert to ARGB first */
     if(pmdev->src_format == DSPF_RGB16)
     {
          gic_ctl.gic_src1_mode = GIC_DEFAULT_PACK_ORDER | xcodeGicGetFormatCode(pmdev->src_format); 

          gic_ctl.gic_alpha_mode_src1 = GIC_ALPHA_MODE_CONSTANT | (a_factor << GIC0_SRC1_ALPHA_C_SHIFT);
 
          /* Check if both PREMULTIPLY and SRCALPHA are set */
          if(!(pmdev->blittingflags & DSBLIT_SRC_PREMULTIPLY && pmdev->src_blend == DSBF_SRCALPHA))
          {
              gic_ctl.gic_alpha_mode_src1 |= GIC_ALPHA_MODE_NOPREMULTIPLY;
          }

          gic_ctl.gic_src1_y_x = ((pmdev->rectsrc_x << GIC0_SRC1_X_SHIFT) & GIC0_SRC1_X_MASK) |                             
                                      ((pmdev->rectsrc_y << GIC0_SRC1_Y_SHIFT) & GIC0_SRC1_Y_MASK);

          gic_ctl.gic_src1_adr0 = pmdev->src_phys;

          gic_ctl.gic_src1_pitch = pmdev->src_pitch;

          /* Positioning SRC1 on the destination surface */
          gic_ctl.gic_src1_h_w = ((pmdev->rectsrc_h << GIC0_SRC1_H_SHIFT) & GIC0_SRC1_H_MASK) | 
                                      ((pmdev->rectsrc_w << GIC0_SRC1_W_SHIFT) & GIC0_SRC1_W_MASK);

          /* Setup the DST as the destination as well */     
          gic_ctl.gic_dst_mode = xcodeGicGetFormatCode(DSPF_ARGB); 

          /* Use temperary buffer */
          gic_ctl.gic_dst_adr0 = pmdev->temp_render_physical;

          gic_ctl.gic_dst_y_x = 0;

          gic_ctl.gic_dst_pitch = (DFB_BYTES_PER_PIXEL(DSPF_ARGB) * pmdev->rectsrc_w + 31) >> 5 << 5;

          gic_ctl.gic_dst_h_w = ((pmdev->rectsrc_h << GIC0_DST_H_SHIFT) & GIC0_DST_H_MASK) |
                                        ((pmdev->rectsrc_w << GIC0_DST_W_SHIFT) & GIC0_DST_W_MASK);

          /* Set flag to indicate that all sources and the destination are ready */
          gic_ctl.gic_opt_status |= GIC_DST_READY | GIC_SRC1_READY | GIC_SRC2_READY ;

          gic_ctl.gic1_src_csc_parm0  = ((GIC_CSC_ALPHASCALE * r_factor) << GIC0_C00_SHIFT);
          gic_ctl.gic1_src_csc_parm1  = 0x0;
          gic_ctl.gic1_src_csc_parm2  = ((GIC_CSC_ALPHASCALE * g_factor) << GIC0_C11_SHIFT);
          gic_ctl.gic1_src_csc_parm3  = 0x0;
          gic_ctl.gic1_src_csc_parm4  = ((GIC_CSC_ALPHASCALE * b_factor) << GIC0_C22_SHIFT);
          gic_ctl.gic1_src_csc_parm5  = 0x0;
          gic_ctl.gic1_src_csc_parm6  = 0xFF00FF00;
          gic_ctl.gic1_src_csc_parm7  = 0x0000FF00;

          gic_ctl.gic_src2_adr0 = pmdev->src_phys;
          gic_ctl.gic_src2_h_w = 1;
          gic_ctl.gic_src2_pitch = pmdev->src_pitch;
          gic_ctl.gic_src2_y_x = 0;
          gic_ctl.gic_src2_mode = GIC_DEFAULT_PACK_ORDER | xcodeGicGetFormatCode(pmdev->src_format);
          gic_ctl.gic_alpha_mode_src2 = GIC_ALPHA_MODE_CONSTANT | GIC0_SRC2_ALPHA_PREMUL_MASK;

          gic_ctl.gic_trigger_command = 
              GIC0_OP_KICKOFF_MASK |
             (GIC0_CSC_SEL_SRC1 << GIC0_CSC_SRC_SEL_SHIFT);

          /* Adjust gic setting according to h/w limitation */
          xcodeGicAdjustOffset(&gic_ctl);

          /* Make certain G2D is finished working on whatever it is doing */
          xcode4_waitidle(mdrv,pmdev);

          /* Start Gic Blit*/     
          Xcode4_Trigger_Gic0_Action(mdrv, pmdev, &gic_ctl);

          /* Modify source information */     
          pmdev->src_phys = pmdev->temp_render_physical;
          pmdev->rectsrc_x = 0;
          pmdev->rectsrc_y = 0;
          pmdev->src_format = DSPF_ARGB;
          pmdev->src_pitch = gic_ctl.gic_dst_pitch;

          return true;
     }

     /* First pass, adjust A, G, B */
     
     gic_ctl.gic_src1_mode = GIC_ALPHA_R_SWAP_PACK_ORDER | xcodeGicGetFormatCode(pmdev->src_format); //ARGB

     gic_ctl.gic_alpha_mode_src1 = GIC_ALPHA_MODE_PERPIXEL | GIC_ALPHA_MODE_NOPREMULTIPLY;

     gic_ctl.gic_src1_y_x = ((pmdev->rectsrc_x << GIC0_SRC1_X_SHIFT) & GIC0_SRC1_X_MASK) |                             
                                 ((pmdev->rectsrc_y << GIC0_SRC1_Y_SHIFT) & GIC0_SRC1_Y_MASK);


     gic_ctl.gic_src1_adr0 = pmdev->src_phys;


     gic_ctl.gic_src1_pitch = pmdev->src_pitch;

     /* Positioning SRC1 on the destination surface */
     gic_ctl.gic_src1_h_w = ((pmdev->rectsrc_h << GIC0_SRC1_H_SHIFT) & GIC0_SRC1_H_MASK) | 
                                 ((pmdev->rectsrc_w << GIC0_SRC1_W_SHIFT) & GIC0_SRC1_W_MASK);

     /* Setup the DST as the destination as well */     
     gic_ctl.gic_dst_mode = xcodeGicGetFormatCode(pmdev->src_format); //ARGB

     /* Use temperary buffer */
     gic_ctl.gic_dst_adr0 = pmdev->temp_render_physical;

     gic_ctl.gic_dst_y_x = 0;

     gic_ctl.gic_dst_pitch = pmdev->src_pitch;

     gic_ctl.gic_dst_h_w = ((pmdev->rectsrc_h << GIC0_DST_H_SHIFT) & GIC0_DST_H_MASK) |
                                   ((pmdev->rectsrc_w << GIC0_DST_W_SHIFT) & GIC0_DST_W_MASK);

     gic_ctl.gic_backgrd_color = 0;
          
     /* Set flag to indicate that all sources and the destination are ready */
     gic_ctl.gic_opt_status |= GIC_DST_READY | GIC_SRC1_READY | GIC_SRC2_READY ;

     gic_ctl.gic1_src_csc_parm0  = ((GIC_CSC_ALPHASCALE * a_factor) << GIC0_C00_SHIFT);
     gic_ctl.gic1_src_csc_parm1  = 0x0;
     gic_ctl.gic1_src_csc_parm2  = ((GIC_CSC_ALPHASCALE * g_factor) << GIC0_C11_SHIFT);
     gic_ctl.gic1_src_csc_parm3  = 0x0;
     gic_ctl.gic1_src_csc_parm4  = ((GIC_CSC_ALPHASCALE * b_factor) << GIC0_C22_SHIFT);
     gic_ctl.gic1_src_csc_parm5  = 0x0;

     if(fixed_alpha)
     {
          gic_ctl.gic1_src_csc_parm6  = (0xFF00 << 16) | (a_factor << 8) | (a_factor);
     }
     else
     {
          gic_ctl.gic1_src_csc_parm6  = 0xFF00FF00;
     }
     
     gic_ctl.gic1_src_csc_parm7  = 0x0000FF00;

     /* Enable the CSC on SRC1... */
     gic_ctl.gic_trigger_command = 
        (SRC1_OVER_BACKGROUND << GIC0_BLEND_LEVELS_SHIFT) |
         GIC0_OP_KICKOFF_MASK |
        (GIC0_CSC_SEL_SRC1 << GIC0_CSC_SRC_SEL_SHIFT);

     /* Adjust gic setting according to h/w limitation */
     xcodeGicAdjustOffset(&gic_ctl);

     /* Make certain G2D is finished working on whatever it is doing */
     xcode4_waitidle(mdrv,pmdev);

     /* Start Gic Blit*/     
     Xcode4_Trigger_Gic0_Action(mdrv, pmdev, &gic_ctl);

     /* Second pass, adjust R */
     gic_ctl.gic_src1_adr0       = pmdev->temp_render_physical;

     gic_ctl.gic_src1_y_x = gic_ctl.gic_dst_y_x;
   
     gic_ctl.gic_dst_adr0        = pmdev->temp_render_physical;
    
     gic_ctl.gic_src1_mode       = GIC_ALPHA_R_SWAP_PACK_ORDER | GIC_ARGB8888;
    
     gic_ctl.gic_backgrd_color   =   0;
    
     gic_ctl.gic_src1_h_w = (pmdev->rectsrc_w << GIC0_SRC1_W_SHIFT) | (pmdev->rectsrc_h << GIC0_SRC1_H_SHIFT);
     gic_ctl.gic_dst_h_w  = (pmdev->rectsrc_w << GIC0_SRC1_W_SHIFT) | (pmdev->rectsrc_h << GIC0_SRC1_H_SHIFT);
            
    
     /* Setup the pitch register(s) */
     gic_ctl.gic_src1_pitch = pmdev->src_pitch;
     gic_ctl.gic_dst_pitch  =  pmdev->src_pitch;
        
            
     /* Setup destination mode */
     gic_ctl.gic_dst_mode    =   GIC_ARGB8888;

     
     /* Check if both PREMULTIPLY and SRCALPHA are set */
     if(pmdev->blittingflags & DSBLIT_SRC_PREMULTIPLY && pmdev->src_blend == DSBF_SRCALPHA)
     {
         /* pre-multiply is needed */
         gic_ctl.gic_alpha_mode_src1 = GIC_ALPHA_MODE_PERPIXEL;
     }
     else
     {
         /* This will also tell the GIC to pre-multiply */
         gic_ctl.gic_alpha_mode_src1 = GIC_ALPHA_MODE_PERPIXEL | GIC_ALPHA_MODE_NOPREMULTIPLY;
     }

     /* 
     *   Define the CSC matrix for the alpha scale... 
     *   Remember that the correct alpha is in the R component...
     *   Also remember that the G and B components must remain the same...
     */
     gic_ctl.gic1_src_csc_parm0  = ((GIC_CSC_ALPHASCALE * r_factor) << GIC0_C00_SHIFT);
     gic_ctl.gic1_src_csc_parm1  = 0x0;
     gic_ctl.gic1_src_csc_parm2  = ((GIC_CSC_ALPHASCALE * 255) << GIC0_C11_SHIFT);
     gic_ctl.gic1_src_csc_parm3  = 0x0;
     gic_ctl.gic1_src_csc_parm4  = ((GIC_CSC_ALPHASCALE * 255) << GIC0_C22_SHIFT);
     gic_ctl.gic1_src_csc_parm5  = 0x0;
     gic_ctl.gic1_src_csc_parm6  = 0xFF00FF00;
     gic_ctl.gic1_src_csc_parm7  = 0x0000FF00;

     /* Enable the CSC on SRC1... */
     gic_ctl.gic_trigger_command = 
        (SRC1_OVER_BACKGROUND << GIC0_BLEND_LEVELS_SHIFT) |
         GIC0_OP_KICKOFF_MASK |
        (GIC0_CSC_SEL_SRC1 << GIC0_CSC_SRC_SEL_SHIFT);

     gic_ctl.gic_opt_status |= GIC_DST_READY | GIC_SRC1_READY | GIC_SRC2_READY | GIC_BACKGRD_COLOR;

     /* Trigger modulate */
     Xcode4_Trigger_Gic0_Action(mdrv, pmdev, &gic_ctl);

     /* Modify source information */     
     pmdev->src_phys = pmdev->temp_render_physical;
     pmdev->rectsrc_x = 0;
     pmdev->rectsrc_y = 0;

     return true;
}


static bool xcodeGicBasic(pxc_dfb_driver_data mdrv, pxc_dfb_device_data pmdev)
{
     gic_opt_struct gic_ctl = {0};
     u32 blittingflags = pmdev->blittingflags;

     //check limitation
     if(!xcodeGicLimitationCheck(mdrv, pmdev))
     {
          return false;
     }

     if(pmdev->dst_format != DSPF_ARGB && pmdev->dst_format != DSPF_RGB16)
     {
        /* Fail blit if dest is not RGB format */
        return false;
     }

     /* Check Porter Duff Rule */
     if(!(pmdev->src_blend == DSBF_ONE && pmdev->dst_blend == DSBF_INVSRCALPHA) &&
        !(pmdev->src_blend == DSBF_SRCALPHA && pmdev->dst_blend == DSBF_INVSRCALPHA))
     {
        /* Now our GIC can only handle the following case:
         * Ao = A1 + (1-A1) * A2
         * Co = [A1] * C1 + (1-A1)*[A2]*C2
         * From DFB's description:
         * pixel color = sc * cf[sf] + dc * cf[df]
         * pixel alpha = sa * af[sf] + da * af[df]
         * We can only handle DSPD_CLEAR, DSPD_SRC, DSPD_SRC_OVER correctly
         * However, since DSPD_NONE is used wildly, we still use hardware to do it
         * Other rules are not supported.
         */
        return false;
     }
     
     /* Check if alpha is zero */
     if ((pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA) && !pmdev->src_color_alpha)
     {
        return true;
     }

     /* Check if premultiply is needed */
     if(xcodeNeedPremultiply(mdrv, pmdev))
     {
        if(!xcodeGicPremultiply(mdrv, pmdev))
        {
            return false;
        }
     }

     /* We will take care of this ourselves in the palette generation for A8 fonts ... */
     if((check_src_premultiply(pmdev->blittingflags, pmdev->src_blend) && 
         !xcodeIsIndexedColor(pmdev->src_format)))
     {
          gic_ctl.gic_alpha_mode_src1 = GIC_ALPHA_MODE_PERPIXEL ;
     }
     else
     {
          gic_ctl.gic_alpha_mode_src1 = 
            GIC_ALPHA_MODE_PERPIXEL | GIC_ALPHA_MODE_NOPREMULTIPLY;
     }

     if(blittingflags & DSBLIT_DST_PREMULTIPLY)
     {
          gic_ctl.gic_alpha_mode_src2 = GIC_ALPHA_MODE_PERPIXEL;
     }
     else
     {
          gic_ctl.gic_alpha_mode_src2 = 
            GIC_ALPHA_MODE_PERPIXEL | GIC_ALPHA_MODE_NOPREMULTIPLY;
     }

     gic_ctl.gic_src1_mode = GIC_DEFAULT_PACK_ORDER | xcodeGicGetFormatCode(pmdev->src_format);

     gic_ctl.gic_src1_pitch = pmdev->src_pitch;
    
     gic_ctl.gic_src1_y_x = ((pmdev->rectsrc_x << GIC0_SRC1_X_SHIFT) & GIC0_SRC1_X_MASK) |                             
                                 ((pmdev->rectsrc_y << GIC0_SRC1_Y_SHIFT) & GIC0_SRC1_Y_MASK);

     gic_ctl.gic_src1_adr0 = pmdev->src_phys;

     /* Positioning SRC1 on the destination surface */
     gic_ctl.gic_src1_h_w = ((pmdev->rectsrc_h << GIC0_SRC1_H_SHIFT) & GIC0_SRC1_H_MASK) | 
                                 ((pmdev->rectsrc_w << GIC0_SRC1_W_SHIFT) & GIC0_SRC1_W_MASK);

     gic_ctl.gic_src2_mode = GIC_DEFAULT_PACK_ORDER | xcodeGicGetFormatCode(pmdev->dst_format);

     gic_ctl.gic_src2_pitch = pmdev->dst_pitch;

     gic_ctl.gic_src2_y_x = ((pmdev->target_dx<< GIC0_SRC2_X_SHIFT) & GIC0_SRC2_X_MASK) |                             
                                 ((pmdev->target_dy << GIC0_SRC2_Y_SHIFT) & GIC0_SRC2_Y_MASK);

     gic_ctl.gic_src2_adr0 = pmdev->dst_phys;

     gic_ctl.gic_src2_h_w = ((pmdev->rectsrc_h << GIC0_SRC2_H_SHIFT) & GIC0_SRC2_H_MASK) | 
                                      ((pmdev->rectsrc_w << GIC0_SRC2_W_SHIFT) & GIC0_SRC2_W_MASK);

     /* Setup the DST as the destination as well */     
     gic_ctl.gic_dst_mode = xcodeGicGetFormatCode(pmdev->dst_format); //ARGB

     gic_ctl.gic_dst_pitch = pmdev->dst_pitch;
     
     gic_ctl.gic_dst_subfld = 0x32;

     gic_ctl.gic_dst_h_w = ((pmdev->rectsrc_h << GIC0_DST_H_SHIFT) & GIC0_DST_H_MASK) |
                                   ((pmdev->rectsrc_w << GIC0_DST_W_SHIFT) & GIC0_DST_W_MASK);
     
     gic_ctl.gic_dst_adr0 = pmdev->dst_phys;

     gic_ctl.gic_dst_y_x = ((pmdev->target_dx<< GIC0_DST_X_SHIFT) & GIC0_DST_X_MASK) |                             
                                 ((pmdev->target_dy << GIC0_DST_Y_SHIFT) & GIC0_DST_Y_MASK);
     gic_ctl.gic_trigger_command = GIC0_OP_KICKOFF_MASK | GIC0_BLEND_EN_MASK;
     
     /* Set flag to indicate that all sources and the destination are ready */
     gic_ctl.gic_opt_status |= GIC_DST_READY | GIC_SRC1_READY | GIC_SRC2_READY ;

     /* Adjust gic setting according to h/w limitation */
     xcodeGicAdjustOffset(&gic_ctl);

     /* Make certain G2D is finished working on whatever it is doing */
     xcode4_waitidle(mdrv,pmdev);

     /* Start Gic Blit*/     
     Xcode4_Trigger_Gic0_Action(mdrv, pmdev, &gic_ctl);

     return true;
}


/* This is a tricky part for BTD100 only */
static bool PipelineGicStretchBlit(void* drv, void* dev, bool draw)
{
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
    int need_sync = 0;

    if(mdrv->master)
    {
        //Only apply to slave
        return false;
    }

    if(pmdev->last_temp_size)
    {
        //Wait for last G2D done
        xcode4_waitidle(mdrv,pmdev);
        
        //Save environment
        int target_dx = pmdev->target_dx;
        int target_dy = pmdev->target_dy;
        int rectsrc_x = pmdev->rectsrc_x;
        int rectsrc_y = pmdev->rectsrc_y;
        int rectsrc_w = pmdev->rectsrc_w;
        int rectsrc_h = pmdev->rectsrc_h;
        int src_phys = pmdev->src_phys;
        int src_pitch = pmdev->src_pitch;
        int src_width = pmdev->src_width;
        int src_height = pmdev->src_height;
        int src_format = pmdev->src_format;
        int dst_format = pmdev->dst_format;
        int dst_phys = pmdev->dst_phys;
        int dst_width = pmdev->dst_width;
        int dst_height = pmdev->dst_height;
        int dst_pitch =pmdev->dst_pitch; 
        int flag = pmdev->blittingflags;
        DFBSurfaceBlendFunction src_blend = pmdev->src_blend;
        DFBSurfaceBlendFunction dst_blend = pmdev->dst_blend;
        DFBRegion clip = pmdev->clip;

        //Load previous environment
        pmdev->rectsrc_x = 0;
        pmdev->rectsrc_y = 0;
        pmdev->rectsrc_w = pmdev->last_temp_size & 0xFFFF;
        pmdev->rectsrc_h = pmdev->last_temp_size >> 16;
        pmdev->target_dx = pmdev->last_temp_position & 0xFFFF;
        pmdev->target_dy = pmdev->last_temp_position >> 16;
        pmdev->dst_format = DSPF_ARGB;
        pmdev->src_format = DSPF_ARGB;
        int dst_bytes_per_line;
        int dst_bytes_pixel = DFB_BYTES_PER_PIXEL(pmdev->dst_format);
        dst_bytes_per_line  =  dst_bytes_pixel * pmdev->rectsrc_w;
        if (dst_bytes_per_line & (32 - 1) )
        {
            dst_bytes_per_line &= (~(32-1));
            dst_bytes_per_line+=    32;            
        }
        pmdev->src_phys     =   pmdev->last_temp_physical;
        pmdev->src_pitch    =   dst_bytes_per_line;
        pmdev->src_width    =   pmdev->rectsrc_w;
        pmdev->src_height   =   pmdev->rectsrc_h;
        pmdev->src_format   =   pmdev->dst_format;
        pmdev->dst_phys     =   pmdev->last_temp_target;
        pmdev->dst_width    =   pmdev->last_temp_rect & 0xFFFF;
        pmdev->dst_height   =   pmdev->last_temp_rect >> 16;
        pmdev->dst_pitch    =   pmdev->last_temp_pitch;
        pmdev->clip         =   pmdev->last_temp_clip;
        pmdev->blittingflags =  DSBLIT_BLEND_ALPHACHANNEL;
        pmdev->src_blend    =   DSBF_ONE;
        pmdev->dst_blend    =   DSBF_INVSRCALPHA;

        //Gic operation
        xcodeLockG2d(mdrv);
        xcodeGicBasic(mdrv, pmdev);
        xcodeUnlockG2d(mdrv);

        //Load current environment
        pmdev->rectsrc_y = rectsrc_y;
        pmdev->rectsrc_w = rectsrc_w;
        pmdev->rectsrc_w = rectsrc_w;
        pmdev->rectsrc_h = rectsrc_h;
        pmdev->target_dx = target_dx;
        pmdev->target_dy = target_dy;
        pmdev->src_phys     =   src_phys;
        pmdev->src_pitch    =   src_pitch;
        pmdev->src_width    =   src_width;
        pmdev->src_height   =   src_height;
        pmdev->src_format   =   src_format;
        pmdev->dst_format   =   dst_format;
        pmdev->dst_phys     =   dst_phys;
        pmdev->dst_width    =   dst_width;
        pmdev->dst_height   =   dst_height;
        pmdev->dst_pitch    =   dst_pitch;
        pmdev->blittingflags =  flag;
        pmdev->src_blend    =   src_blend;
        pmdev->dst_blend    =   dst_blend;
        pmdev->clip         =   clip;
            
        //Clear status
        pmdev->last_temp_size = 0;
        pmdev->last_temp_position = 0;
        pmdev->last_temp_target = 0;
        pmdev->last_temp_pitch = 0;
    }
    else
    {
        need_sync = 1;
    }
        
    if( draw && 
        pmdev->src_width  <= 150 && 
        pmdev->src_height <= 150 && 
        pmdev->dst_format == DSPF_ARGB &&
        pmdev->src_format == DSPF_ARGB &&
        pmdev->src_blend  == DSBF_ONE &&
        pmdev->dst_blend  == DSBF_INVSRCALPHA &&
        pmdev->blittingflags == DSBLIT_BLEND_ALPHACHANNEL)
    {
        //Calculate parameters
        int dst_bytes_per_line;
        int dst_bytes_pixel = DFB_BYTES_PER_PIXEL(pmdev->dst_format);
        dst_bytes_per_line  =  dst_bytes_pixel * pmdev->rectsrc_w;
        if (dst_bytes_per_line & (32 - 1) )
        {
            dst_bytes_per_line &= (~(32-1));
            dst_bytes_per_line+=    32;            
        }
        int scaler_x = (pmdev->rect_scalesrc_w << SCALER_MULTIPLY_SHIFT) / pmdev->rectsrc_w;
        int scaler_y = (pmdev->rect_scalesrc_h << SCALER_MULTIPLY_SHIFT) / pmdev->rectsrc_h;
        
        //Save status
        pmdev->last_temp_size = pmdev->rectsrc_w | (pmdev->rectsrc_h << 16);
        pmdev->last_temp_position = pmdev->target_dx | (pmdev->target_dy << 16);
        pmdev->last_temp_target = pmdev->dst_phys;
        pmdev->last_temp_pitch = pmdev->dst_pitch;
        pmdev->last_temp_clip = pmdev->clip;
        pmdev->last_temp_rect = pmdev->dst_width | (pmdev->dst_height << 16);
    
        //Get next buffer
        if(pmdev->last_temp_physical == pmdev->temp_render_extra_physical)
        {
            pmdev->last_temp_physical = pmdev->temp_render_extra_physical + TEMP_RENDER_EXTRA_SIZE / 2;
        }
        else
        {
            pmdev->last_temp_physical = pmdev->temp_render_physical;
        }
    
        //G2D OP
        xcodeLockG2d(mdrv);
        
        xcode4_waitfifo(mdrv, pmdev, 10);
        
        REG_WRITE(G2D_DST_OFFSET, pmdev->last_temp_physical);
        REG_WRITE(G2D_DST_Y_X, 0x00000000);
        REG_WRITE(G2D_DST_PITCH_FORMAT, 
            ((dst_bytes_per_line << G2D_DST_PITCH_FORMAT_DST_PITCH_SHIFT) & G2D_DST_PITCH_FORMAT_DST_PITCH_MASK) |
            ((get_pixel_format_color(pmdev->dst_format) << DST_PIX_FORMAT_SHIFT) & DST_PIX_FORMAT_MASK) |
            ((GFX_ROP_ALPHA_FORMAT_SRC_ALPHA << ROP_ALPHA_FORMAT_SHIFT) & ROP_ALPHA_FORMAT_MASK)  
        );
        
        REG_WRITE(G2D_SRC_OFFSET, pmdev->src_phys);
        REG_WRITE(G2D_SRC_Y_X,  
                    ((pmdev->rectsrc_x << SRC_X_SHIFT) & SRC_X_MASK) |
                    ((pmdev->rectsrc_y << SRC_Y_SHIFT) & SRC_Y_MASK)
                  );
        REG_WRITE(G2D_SRC_H_W,
                    ((pmdev->rect_scalesrc_w << SRC_W_SHIFT) & SRC_W_MASK) |
                    ((pmdev->rect_scalesrc_h << SRC_H_SHIFT) & SRC_H_MASK)
                  );

        u32 scale_type;
        if ((pmdev->rect_scalesrc_w <= pmdev->rectsrc_w) ||
            (pmdev->rect_scalesrc_h <= pmdev->rectsrc_h)
                  )
        {
            scale_type = SCALING_TYPE_LERP;
        }
        else
        {
            scale_type = SCALING_TYPE_NEIGHBOR;
        }

        REG_WRITE(G2D_SRC_PITCH_FORMAT,
            ((pmdev->src_pitch << G2D_SRC_PITCH_FORMAT_SRC_PITCH_SHIFT) & G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK) |
            ((get_pixel_format_color(pmdev->src_format) << SRC_PIX_FORMAT_SHIFT) & SRC_PIX_FORMAT_MASK)  |
            ((scale_type << SCALING_TYPE_SHIFT) & SCALING_TYPE_MASK)
            );
        
        REG_WRITE(G2D_SCALERYX, 
                    ((scaler_x << XSCALER_SHIFT) & XSCALER_MASK)  |
                    ((scaler_y << YSCALER_SHIFT) & YSCALER_MASK)
                  );
        
        int source_direction    =   0;
        
        if (pmdev->blittingflags & DSBLIT_FLIP_VERTICAL)
        {
            /* Change Source direction is now right to left ... */
            source_direction |= 0x1;    /* Flip vertical flag (internal to this function only)*/
        }
        
        if (pmdev->blittingflags & DSBLIT_FLIP_HORIZONTAL)
        {
            /* Change source direction to include bottom to top ...*/
            source_direction |= 0x2;    /* Flip horizontal flag */
            
        }
        
        if (pmdev->blittingflags & DSBLIT_ROTATE180)
        {
            source_direction = 0x3; /* Flip both vertical and horizontal ... */
        }
        
        int src_dir_mask    =   0;
        if (source_direction == 0x1)
        {
            /* Left to right, bottom to top */
            src_dir_mask    =  ((LEFT_TO_RIGHT_BOT_TO_TOP << BLTDIR_SHIFT) & BLTDIR_MASK);
        }
        else if (source_direction == 0x2)
        {
            /* Right to left, top to bottom */
            src_dir_mask    =   ((RIGHT_TO_LEFT_TOP_TO_BOT << BLTDIR_SHIFT) & BLTDIR_MASK);
        }
        else if (source_direction == 0x3)
        {
            /* Right to left, bottom to top */
            src_dir_mask    =   ((RIGHT_TO_LEFT_BOT_TO_TOP << BLTDIR_SHIFT) & BLTDIR_MASK);
        }
        
        REG_WRITE(G2D_OP_CTL,
                    ((XC_G2D_OP_SRC_SEL_BLT     << SRC_SELECT_SHIFT) & SRC_SELECT_MASK)     |
                    ((XC_G2D_OP_SCALE_RECT      << GFX_OP_SHIFT) & GFX_OP_MASK)             |
                    ((LEFT_TO_RIGHT_TOP_TO_BOT  << DEST_BLT_DIR_SHIFT) & DEST_BLT_DIR_MASK) |
                    ((GFX_SCISSOR_DISABLE       << SCISSOR_DIS_SHIFT) & SCISSOR_DIS_MASK)   |
                    ((SRC_ALPHA_WRITE_OUT       << SRC_ALPHA_SHIFT) & SRC_ALPHA_MASK)       |
                    ((GFX_ALPHA_MODE_NO_ALPHA   << GFX_ALPHA_MODE_SHIFT) & GFX_ALPHA_MODE_MASK) |
                    ((src_dir_mask))
                  );

        //Check if need sync
        if(need_sync)
        {
            Wait_For_Gic0_Idle;
        }
        
        REG_WRITE(G2D_DST_H_W_INIT,
                    ((pmdev->rectsrc_w << DST_W_SHIFT) & DST_W_MASK)    |
                    ((pmdev->rectsrc_h << DST_H_SHIFT) & DST_H_MASK)
                  );
        
        xcodeUnlockG2d(mdrv);
        
        return true;
    }

    return false;
}


/*
 *
 *Function Name:xcodeEngineSync
 *
 *Parameters: 
 *             drv-> Pointer to driver data
 *             dev-> Pointer to device data
 *
 *Description:
 *             Sync the G2D engine
 *Returns:
 *             DFB_OK when engine is IDLE
*/

DFBResult xcodeEngineSync( void *drv, void *dev )
{
     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data mdev = (pxc_dfb_device_data) dev;

     DPRINTF( "xcode DFB driver xcodeEngineSync enter:\n" );

     //complete GIC stretch pipeline
     PipelineGicStretchBlit(drv, dev, false);

     xcode4_waitidle(mdrv,mdev);
     
     Wait_For_Gic0_Idle;    

     DPRINTF( "xcode DFB driver xcodeEngineSync exit:\n" );

     return DFB_OK;
}

static bool xcodeGicFillRectangle( void *drv, void *dev, DFBRectangle *rect )
{
     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
     
     DFBRectangle rect_dst;
     
     u32     fifo_entries;
     u32     tmp;
     DFBBoolean returnValue;
     u32    buffer_size;
     u32    offset;
     
     gic_opt_struct gic_ctl;
     
     DPRINTF( "xcode DFB driver xcodeGicFillRectangleFunc enter:\n" );

     DPRINTF( "xcode DFB driver xcodeGicFillRectangleFunc rect = %x %x %x %x\n",rect->x,rect->y,rect->w,rect->h );
    
     DPRINTF( "copy rect to rect_dst.\n");   

     rect_dst.x = rect->x;
     rect_dst.y = rect->y;
     rect_dst.w = rect->w;       
     rect_dst.h = rect->h;

     DPRINTF("Before call dfb_clip_rectangle(), rect_dst = %x %x %x %x\n", rect_dst.x, rect_dst.y, rect_dst.w, rect_dst.h );  

     DPRINTF( "Call dfb_clip_rectangle(). \n");
     
     // Find the intersection of the clipping rectangle
     returnValue = dfb_clip_rectangle(&pmdev->clip, &rect_dst);

     DPRINTF( " dfb_clip_rectangle() call is done.  returnValue = %d \n", returnValue);
     DPRINTF("After call dfb_clip_rectangle(), rect_dst = %x %x %x %x\n", rect_dst.x, rect_dst.y, rect_dst.w, rect_dst.h ); 

     if (rect_dst.w < 4 || rect_dst.h < 4)
     {
          return false;
     }

     buffer_size =   rect_dst.h * pmdev->dst_pitch;

     if (buffer_size > TEMP_RENDER_SIZE )
     {
        rect_dst.h /= 2;
        xcodeGicFillRectangle(drv, dev, &rect_dst);
        rect_dst.y += rect_dst.h;
        xcodeGicFillRectangle(drv, dev, &rect_dst);
        return true;
     }

     u32 src_surface_address = pmdev->temp_render_line_physical;
     u32 dst_surface_address = pmdev->dst_phys + rect_dst.y * pmdev->dst_pitch + rect_dst.x * DFB_BYTES_PER_PIXEL(pmdev->dst_format);
     u32 dst_surface_width = rect_dst.w;
     u32 dst_surface_height = rect_dst.h;
     u32 dst_surface_pitch = pmdev->dst_pitch;

     u32 top = 0;
     u32 bottom = rect_dst.h - 1;
     u32 left = 0;
     u32 right = rect_dst.w - 1;

     u32 r = pmdev->src_color_r;
     u32 g = pmdev->src_color_g;
     u32 b = pmdev->src_color_b;

     // check if premultiply is needed
     if(check_color_premultiply(pmdev->drawingflags, pmdev->src_blend))
     {
          r = r * pmdev->src_color_alpha >> 8;
          g = g * pmdev->src_color_alpha >> 8;
          b = b * pmdev->src_color_alpha >> 8;
     }

     gic_ctl.gic_opt_status = GIC_DST_READY | GIC_SRC1_READY | GIC_SRC2_READY | GIC_BACKGRD_COLOR;

     if(dst_surface_address & 0x1F)
     {
         // check DST format
         if(!DFB_BYTES_PER_PIXEL(pmdev->dst_format))
         {
              return true;
         }
         
         offset = (dst_surface_address & 0x1F) / DFB_BYTES_PER_PIXEL(pmdev->dst_format);
         dst_surface_address &= ~0x1F;
         dst_surface_width += offset;
         left += offset;
         right += offset;
     }

     xcodeLockG2d(mdrv);

     //Pass 1, Generate a single-line color buffer
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entries, 7);
     REG_WRITE(G2D_FRGND_COLOR, PIXEL_ARGB(pmdev->src_color_alpha, r, g, b));
     REG_WRITE(G2D_DST_OFFSET, src_surface_address);
     REG_WRITE(G2D_DST_PITCH_FORMAT, ( PIX_FORMAT_RGB32 << DST_PIX_FORMAT_SHIFT));
     REG_WRITE(G2D_DST_Y_X, 0);
     REG_WRITE(G2D_OP_CTL, (XC_G2D_OP_FILL_RECT << GFX_OP_SHIFT)  | (1 <<SCISSOR_DIS_SHIFT ) | (SRC_ALPHA_WRITE_OUT << SRC_ALPHA_SHIFT) | XC_G2D_OP_SRC_SEL_FRGND_CLR);
     REG_WRITE(G2D_GUI_CONTROL, 0);
     REG_WRITE(G2D_DST_H_W_INIT,((dst_surface_width) | (1 << DST_H_SHIFT)));

     //Pass 2, GIC blit
     gic_ctl.gic_opt_status = GIC_DST_READY | GIC_SRC1_READY | GIC_SRC2_READY ;
     gic_ctl.gic_src1_adr0  = src_surface_address;
     gic_ctl.gic_src1_h_w   = (dst_surface_width << GIC0_SRC1_W_SHIFT) | (dst_surface_height << GIC0_SRC1_H_SHIFT);
     gic_ctl.gic_src1_pitch = 0; //single line
     gic_ctl.gic_src1_y_x  = 0;
     gic_ctl.gic_src1_mode = GIC_DEFAULT_PACK_ORDER | xcodeGicGetFormatCode(DSPF_ARGB);
     gic_ctl.gic_alpha_mode_src1 = 0x12 ;

     gic_ctl.gic_src2_adr0 =  dst_surface_address;
     gic_ctl.gic_src2_h_w = (dst_surface_width << GIC0_SRC2_W_SHIFT) | (dst_surface_height << GIC0_SRC2_H_SHIFT) ;
     gic_ctl.gic_src2_pitch = dst_surface_pitch; 
     gic_ctl.gic_src2_y_x = 0;
     gic_ctl.gic_src2_mode = GIC_DEFAULT_PACK_ORDER | xcodeGicGetFormatCode(pmdev->dst_format);

     if (pmdev->drawingflags & DSDRAW_DST_PREMULTIPLY)
     {
          gic_ctl.gic_alpha_mode_src2 = 0x02 ;
     }
     else
     {
          gic_ctl.gic_alpha_mode_src2 = 0x12 ;
     }

     gic_ctl.gic_dst_adr0 = dst_surface_address ;
     gic_ctl.gic_dst_h_w = (dst_surface_width << GIC0_DST_W_SHIFT) | (dst_surface_height << GIC0_DST_H_SHIFT) ;
     gic_ctl.gic_dst_pitch = dst_surface_pitch ;
     gic_ctl.gic_dst_y_x = 0; 
     gic_ctl.gic_dst_mode = xcodeGicGetFormatCode(pmdev->dst_format) ;

     gic_ctl.gic_clip_top_left = (top << GIC0_CLIP_TOP_SHIFT) | (left << GIC0_CLIP_LEFT_SHIFT) ;
     gic_ctl.gic_clip_bot_right = (bottom << GIC0_CLIP_BOT_SHIFT) | (right << GIC0_CLIP_RIGHT_SHIFT) ;

     gic_ctl.gic_trigger_command =  GIC0_OP_KICKOFF_MASK |
                                                     GIC0_BLEND_EN_MASK |
                                                     GIC0_DST_CLIP_EN_MASK;

     xcode4_waitidle(mdrv,pmdev);

     Xcode4_Trigger_Gic0_Action(mdrv, pmdev, &gic_ctl);

     xcodeUnlockG2d(mdrv);

     DPRINTF( "xcode DFB driver xcodeGicFillRectangleFunc exit:\n" );

     return true;

}

bool xcodeFillRectangle( void *drv, void *dev, DFBRectangle *rect )
{
     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
     int dst_pixel_format_color;
     u32 g2d_op_ctl_value;
     u32 size_of_operation = BASE_SIZE_OF_FILLRECTANGLE;

     DPRINTF( "xcode DFB driver xcodeFillRectangle = %x %x %x %x\n",rect->x,rect->y,rect->w,rect->h );

     BLITPRINTF("Fill drawingflags = 0x%x, fill_color = %x, dst = %x\n", pmdev->drawingflags, pmdev->fill_color, pmdev->dst_phys);
     
     //convert the directfb color format to xcode4 g2d color format
     //return false;

     dst_pixel_format_color = get_pixel_format_color(pmdev->dst_format);
     if (dst_pixel_format_color == PIX_FORMAT_UNKNOWN)
     {
         DPRINTF( "xcode DFB driver xcodeFillRectangle, PIX_FORMAT_UNKNOWN\n" );
         return false;
     }

     //check if GIC fill is needed
     if ((pmdev->drawingflags & DSDRAW_BLEND) && ((pmdev->fill_color & 0xFF000000) != 0xFF000000))
     {      
        /* NOTE: Previous code was like ...
             *   xcodeGicFillRectangleFunc(mdrv, pmdev, rect);
             *   return true; 
             *  However, there are cases in which xcodeGicFillRectangle cannot handle 
             *  the fillRectangle request. In these cases, it would return true even
             *  if we didn't actually perform the operation.
             */
         return xcodeGicFillRectangle(mdrv, pmdev, rect);
     }

     //set the filling rectangle operation
     if (pmdev->modified & SMF_CLIP)
     {
         //turn on the scissor feature
         g2d_op_ctl_value = (XC_G2D_OP_FILL_RECT << GFX_OP_SHIFT) | (SRC_ALPHA_WRITE_OUT << SRC_ALPHA_SHIFT);
         size_of_operation = size_of_operation + SIZE_OF_SET_CLIP_REGION;
     }
     else
     {
         //disable scissor feature
         g2d_op_ctl_value = (XC_G2D_OP_FILL_RECT << GFX_OP_SHIFT)  | (1 <<SCISSOR_DIS_SHIFT ) | (SRC_ALPHA_WRITE_OUT << SRC_ALPHA_SHIFT);
     }

     //turn on the source foreground color
     g2d_op_ctl_value = g2d_op_ctl_value | (XC_G2D_OP_SRC_SEL_FRGND_CLR);

     xcodeLockG2d(mdrv);
     
     //start filling operation
     G2d_Operation_Start_For_FillRectangel(mdrv, pmdev, &g2d_op_ctl_value, size_of_operation);

     //check if need to set clip region
     if (pmdev->modified & SMF_CLIP)
     {
         Set_Clip_Region(mdrv, pmdev);
     }

     //set destination registers
     Set_Destination_Registers(mdrv, pmdev, dst_pixel_format_color, rect->x, rect->y);

     //finish fill operation 
     G2d_Operation_Finish(mdrv, pmdev, &g2d_op_ctl_value, rect);

     xcodeUnlockG2d(mdrv);

     DPRINTF( "xcode DFB driver xcodeFillRectangle exit:\n" );

     return true;
}

static bool Xcode4_BlitOnly_ARGB_To_YUV(void *drv, void *dev)
{
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
    gic_opt_struct gic_ctl;
    bool blit_success;

    DPRINTF("Xcode4_BlitOnly_ARGB_To_YUV enter \n");
    DPRINTF("pmdev->src_phys 0x%x \n", pmdev->src_phys);
    DPRINTF("pmdev->src_pitch 0x%x \n", pmdev->src_pitch);
    DPRINTF("pmdev->dst_phys 0x%x \n", pmdev->dst_phys);
    DPRINTF("pmdev->dst_pitch 0x%x \n", pmdev->dst_pitch);

    if (pmdev->blittingflags & (~XC4_GIC_SUPPORTED_BLITTINGFLAGS))
    {
        blit_success = false;
    }
    else
    {
        xcodeLockG2d(mdrv);
        
        /* Make certain G2D is finished working on whatever it is doing */
        xcode4_waitidle(mdrv,pmdev);

        blit_success = true;
        memset (&gic_ctl, 0, sizeof(gic_opt_struct));


        gic_ctl.gic_src1_pitch = pmdev->src_pitch;        
        gic_ctl.gic_src1_mode = GIC_ARGB8888 |
				                (ALPHA_24_TO_31 << GIC0_SRC2_A_PACK_ORDER_SHIFT) |
                                (RED_16_TO_23 << GIC0_SRC2_YR_PACK_ORDER_SHIFT) |
                                (GREEN_8_TO_15 << GIC0_SRC2_UG_PACK_ORDER_SHIFT) |
                                (BLUE_0_TO_7 << GIC0_SRC2_VB_PACK_ORDER_SHIFT);


        unsigned int dst_uv_pitch;
        unsigned int y_size, uv_size;
        int times_w, times_h;

        if (pmdev->dst_format == DSPF_YUV444P) {
            times_w = 1;
            times_h = 1;

            gic_ctl.gic_dst_mode = GIC_8_BIT_YUV_SURFACE | (2 << GIC0_DST_PLANAR_MODE_SHIFT);//TODO: GIC has problem with YUV444
        } 
        else if (pmdev->dst_format == DSPF_YV16) {
            times_w = 2;
            times_h = 1;
            
            gic_ctl.gic_dst_mode = GIC_8_BIT_YUV_SURFACE | (1 << GIC0_DST_PLANAR_MODE_SHIFT);
        }
        else if (pmdev->dst_format == DSPF_I420) {
            times_w = 2;
            times_h = 2;
            
            gic_ctl.gic_dst_mode = GIC_8_BIT_YUV_SURFACE;
        }

        dst_uv_pitch = pmdev->dst_pitch / times_w;
        uv_size = (pmdev->dst_height / times_h) * dst_uv_pitch;
        y_size = pmdev->dst_height * pmdev->dst_pitch;

        if (pmdev->dst_pitch % times_w) {
            DPRINTF("dst UV pitch is not integer!\n");
        }

        if (pmdev->dst_height % times_h) {
            DPRINTF("dst UV height is not integer!\n");
        }

        if (dst_uv_pitch & 0x1f) {
             DPRINTF("dst UV pitch is not 32 aligned!\n");
        }
        
        gic_ctl.gic_dst_pitch = pmdev->dst_pitch | ((dst_uv_pitch/GIC_ALIGN_DIV) << GIC0_DST_UV_PITCH_SHIFT);
        

  
        //set the alpha mode mask
        if (pmdev->blittingflags & DSBLIT_BLEND_ALPHACHANNEL)
        {
            gic_ctl.gic_alpha_mode_src1 = GIC_PER_PIXEL_ALPHA | (0 << GIC0_SRC1_ALPHA_C_SHIFT);
        }
        else if (pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA)
        {
            gic_ctl.gic_alpha_mode_src1 = GIC_CONSTANT_ALPHA | 
                            (pmdev->dst_color.a << GIC0_SRC1_ALPHA_C_SHIFT);
        }
        else
        {
            gic_ctl.gic_alpha_mode_src1 = GIC_CONSTANT_ALPHA | (0xff << GIC0_SRC1_ALPHA_C_SHIFT);
        }

        if ((pmdev->blittingflags & DSBLIT_SRC_PREMULTCOLOR ) ||
            (pmdev->blittingflags & DSBLIT_SRC_PREMULTIPLY ))
        {
            gic_ctl.gic_alpha_mode_src1 |= GIC0_SRC1_ALPHA_PREMUL_MASK;
        }

        //Setup RGB to YCbCr conversion
        gic_ctl.gic1_src_csc_parm0 = 0x10200839;
        gic_ctl.gic1_src_csc_parm1 = 0xFB440322;
        gic_ctl.gic1_src_csc_parm2 = 0x0E0CF6B1;
        gic_ctl.gic1_src_csc_parm3 = 0xF43A0E0C;
        gic_ctl.gic1_src_csc_parm4 = 0x0010FDBB;
        gic_ctl.gic1_src_csc_parm5 = 0x00800080;
        gic_ctl.gic1_src_csc_parm6 = 0xFF00FF00;
        gic_ctl.gic1_src_csc_parm7 = 0x0000FF00;

        //preset the trigger command
        gic_ctl.gic_trigger_command = GIC0_OP_KICKOFF_MASK |
                                      (SRC1_OVER_BACKGROUND << GIC0_BLEND_LEVELS_SHIFT)| 
                                      GIC0_BLEND_EN_MASK |
                                      (GIC0_BLEND_DOMAIN_YCBCR << GIC0_BLEND_DOMAIN_SHIFT) |
                                      (GIC0_CSC_SEL_SRC1 << GIC0_CSC_SRC_SEL_SHIFT); 

        gic_ctl.gic_dst_y_x = 0 | (0 << GIC0_DST_Y_SHIFT); //always from (0, 0) to bypass the GIC flaw.

        gic_ctl.gic_opt_status |= GIC_DST_READY | GIC_SRC1_READY | GIC_SRC2_READY;
        

        int dst_curr_width;
        
        int curr_height;

        int src_gap;

        int org_src_adr0;
        int org_dst_adr0, org_dst_adr1, org_dst_adr2;

        int actual_advance_pixels;

        int src_y = pmdev->rectsrc_y;
        int dst_y = pmdev->target_dy;
        int remaining_height = pmdev->rectsrc_h;

        /*change the values that are different for each loop*/
        while (remaining_height > 0)
        {
            int src_x = pmdev->rectsrc_x;
            int dst_x = pmdev->target_dx;
            int dst_remaining_width = pmdev->rectsrc_w;
            
            while (dst_remaining_width > 0) 
            {
                unsigned int src_offset = src_y * pmdev->src_pitch + src_x * 4;
                org_src_adr0 = pmdev->src_phys + src_offset;
                
                DPRINTF("before src_adr0 = 0x%x\n", org_src_adr0);

                //find the nearest 32 aligned start address.
                int src_gap_bytes = org_src_adr0 - (org_src_adr0 & ~0x1f);
                D_ASSERT((src_gap_bytes & 0x3) == 0); 
                src_gap = src_gap_bytes >> 2; 

                gic_ctl.gic_src1_adr0 = org_src_adr0 - src_gap_bytes;

                DPRINTF("after src_adr0 = 0x%x\n", gic_ctl.gic_src1_adr0);

                
                int dst_uv_gap, dst_y_gap;
                
                unsigned int dst_y_offset = dst_y * pmdev->dst_pitch + dst_x;
                unsigned int dst_uv_offset;

                int residue_uv;  //used to represent if the starting x is at the odd position.
                if (pmdev->dst_format == DSPF_YUV444P) {
                    dst_uv_offset = dst_y * dst_uv_pitch + dst_x;
                    residue_uv = 0;
                } 
                else if (pmdev->dst_format == DSPF_YV16) {
                    //this warning will not cause problem.
                    if (dst_x & 0x1) {
                        DPRINTF("dst x is not even for YV16!\n");
                    }
                        
                    dst_uv_offset = dst_y * dst_uv_pitch + dst_x / 2;
                    residue_uv = dst_x & 1;
                }
                else if (pmdev->dst_format == DSPF_I420) {
                    //those warnings will not cause problem.
                    if (dst_x & 0x1) {
                        DPRINTF("dst x is not even for I420!\n");
                    }
                    if (dst_y & 0x1) {
                        DPRINTF("dst y is not even for I420!\n");
                    }

                    dst_uv_offset = dst_y / 2 * dst_uv_pitch + dst_x / 2;
                    residue_uv = dst_x & 1;
                }

                org_dst_adr0 = pmdev->dst_phys + dst_y_offset;
                org_dst_adr1 = pmdev->dst_phys + y_size + dst_uv_offset;
                org_dst_adr2 = org_dst_adr1 + uv_size;

                DPRINTF("before alignment dst addr0 = 0x%x, addr1= 0x%x, add2=0x%x\n", 
                        org_dst_adr0, org_dst_adr1, org_dst_adr2);

                //find the nearest 32 aligned start address.
                dst_uv_gap = org_dst_adr1 - (org_dst_adr1 & ~0x1f);
                dst_y_gap = dst_uv_gap * times_w + residue_uv;
                DPRINTF("residue_uv = %d, dst_uv_gap = %d, dst_y_gap = %d\n", 
                            residue_uv, dst_uv_gap, dst_y_gap);

                gic_ctl.gic_dst_adr0 = org_dst_adr0 - dst_y_gap;
                gic_ctl.gic_dst_adr1 = org_dst_adr1 - dst_uv_gap;
                gic_ctl.gic_dst_adr2 = org_dst_adr2 - dst_uv_gap;
                
                DPRINTF("after alignment dst addr0 = 0x%x, addr1= 0x%x, add2=0x%x\n", 
                              gic_ctl.gic_dst_adr0, gic_ctl.gic_dst_adr1, gic_ctl.gic_dst_adr2);                


                DPRINTF("src_gap = %d, dst_y_gap =%d\n", src_gap, dst_y_gap);

                //make sure the dst_gap is larger than src_y_gap, 
                if (src_gap > dst_y_gap) {
                    dst_y_gap += ((src_gap - dst_y_gap) / 8 + 1) * 8;
                    DPRINTF("adjusting gap, dst_y_gap = %d\n", dst_y_gap);
                }

		
                gic_ctl.gic_dst_adr0 = org_dst_adr0 - dst_y_gap;
		        gic_ctl.gic_dst_adr1 = org_dst_adr1 - dst_y_gap / times_w;
		        gic_ctl.gic_dst_adr2 = org_dst_adr2 - dst_y_gap / times_w;

                D_ASSERT((gic_ctl.gic_dst_adr0 & 0x1f) == 0);     
                D_ASSERT((gic_ctl.gic_dst_adr1 & 0x1f) == 0);  
                D_ASSERT((gic_ctl.gic_dst_adr2 & 0x1f) == 0);
                
                DPRINTF("after adjusting dst_adr0 = 0x%x\n", gic_ctl.gic_dst_adr0); 


                if (pmdev->dst_format == DSPF_YV16) {
                    u32 tmp = gic_ctl.gic_dst_adr1;
                    gic_ctl.gic_dst_adr1 = gic_ctl.gic_dst_adr2;
                    gic_ctl.gic_dst_adr2 = tmp;
                }

                dst_remaining_width += dst_y_gap;
                dst_curr_width = dst_remaining_width > GIC_WIDTH_LIMIT ? GIC_WIDTH_LIMIT : dst_remaining_width;

                DPRINTF("dst_remaining_width = %d, dst_curr_width = %d\n", dst_remaining_width, dst_curr_width);
                
                int src_x_in_dst = dst_y_gap - src_gap;
                int src_curr_width = dst_curr_width - src_x_in_dst;
                DPRINTF("src_x_in_dst = %d, src_curr_width = %d\n", src_x_in_dst, src_curr_width);

                curr_height = remaining_height > GIC_HEIGHT_LIMIT ? GIC_HEIGHT_LIMIT : remaining_height;

                gic_ctl.gic_src1_h_w = src_curr_width | (curr_height << GIC0_SRC1_H_SHIFT);
                gic_ctl.gic_dst_h_w = dst_curr_width | (curr_height << GIC0_DST_H_SHIFT);

                if (dst_y_gap) {
                    gic_ctl.gic_trigger_command |= GIC0_DST_CLIP_EN_MASK;
                    gic_ctl.gic_clip_top_left = dst_y_gap | (0 << GIC0_CLIP_TOP_SHIFT);
                    gic_ctl.gic_clip_bot_right = (dst_curr_width - 1) | ((curr_height - 1) << GIC0_CLIP_BOT_SHIFT);
                    DPRINTF("clip left=%d, right = %d\n", dst_y_gap, dst_curr_width - 1);
                }

                //GIC0_SRC1_Y_X denotes the coordinate of the top-left point of source surface in the destination
                //surface. Both surface need to start at 32-aligned address. GIC0_CLIP_TOP_LEFT comes into
                //effect if starting address need to be aligned.
                gic_ctl.gic_src1_y_x = src_x_in_dst | (0 << GIC0_SRC1_Y_SHIFT);
                
                Xcode4_Trigger_Gic0_Action(mdrv, pmdev, &gic_ctl);

                actual_advance_pixels = dst_curr_width - dst_y_gap;
                
                dst_remaining_width -= dst_curr_width;
                src_x += actual_advance_pixels;
                dst_x += actual_advance_pixels;
                                
                DPRINTF("src_x advances %d\n", actual_advance_pixels);     
            }

            remaining_height -= curr_height;
            src_y += curr_height;
            dst_y += curr_height;
        }

        xcodeUnlockG2d(mdrv);
    }
    
    return blit_success;
}

static bool Xcode4_BlitOnly_YUV_To_ARGB(void *drv, void *dev)
{
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
    gic_opt_struct gic_ctl;
    bool blit_success;

    DPRINTF("Xcode4_BlitOnly_YUV_To_ARGB enter \n");
    DPRINTF("pmdev->src_phys 0x%x \n", pmdev->src_phys);
    DPRINTF("pmdev->src_pitch 0x%x \n", pmdev->src_pitch);
    DPRINTF("pmdev->dst_phys 0x%x \n", pmdev->dst_phys);
    DPRINTF("pmdev->dst_pitch 0x%x \n", pmdev->dst_pitch);

    if (pmdev->blittingflags & (~XC4_GIC_SUPPORTED_BLITTINGFLAGS))
    {
        blit_success = false;
    }
    else
    {
        xcodeLockG2d(mdrv);
        
        /* Make certain G2D is finished working on whatever it is doing */
        xcode4_waitidle(mdrv,pmdev);

        blit_success = true;

        /******* the values below never change for each loop, so place them outside of the loop******/
        memset (&gic_ctl, 0, sizeof(gic_opt_struct));

        unsigned int src_uv_pitch;
        unsigned int uv_size;
        int times_w, times_h;

        if (pmdev->src_format == DSPF_YUV444P) {
            times_w = 1;
            times_h = 1;

            gic_ctl.gic_src1_mode = GIC_8_BIT_YUV_SURFACE | (2 << GIC0_SRC1_PLANAR_MODE_SHIFT);//TODO: GIC has problem with YUV444
        } 
        else if (pmdev->src_format == DSPF_YV16) {
            times_w = 2;
            times_h = 1;
            
            gic_ctl.gic_src1_mode = GIC_8_BIT_YUV_SURFACE | (1 << GIC0_SRC1_PLANAR_MODE_SHIFT);
        }
        else if (pmdev->src_format == DSPF_I420) {
            times_w = 2;
            times_h = 2;
            
            gic_ctl.gic_src1_mode = GIC_8_BIT_YUV_SURFACE;
        }

        src_uv_pitch = pmdev->src_pitch / times_w;
        uv_size = (pmdev->src_height / times_h) * src_uv_pitch;

        if (pmdev->src_pitch % times_w) {
            DPRINTF("src UV pitch is not integer!\n");
        }

        if (pmdev->src_height % times_h) {
            DPRINTF("src UV height is not integer!\n");
        }

        if (src_uv_pitch & 0x1f) {
             DPRINTF("src UV pitch is not 32 aligned!\n");
        }

        unsigned int y_size = pmdev->src_height * pmdev->src_pitch;
        
        gic_ctl.gic_src1_pitch = pmdev->src_pitch  | ((src_uv_pitch/GIC_ALIGN_DIV) << GIC0_SRC1_UV_PITCH_SHIFT);
        

        gic_ctl.gic_alpha_mode_src1 = GIC_ALPHA_MODE_CONSTANT | (0xff << GIC0_SRC1_ALPHA_C_SHIFT);
        
        //Setup RGB to YCbCr conversion
        gic_ctl.gic1_src_csc_parm0 = 0x0000253F; 
        gic_ctl.gic1_src_csc_parm1 = 0x253F3312;
        gic_ctl.gic1_src_csc_parm2 = 0xE5FCF37D;
        gic_ctl.gic1_src_csc_parm3 = 0x4093253F;
        gic_ctl.gic1_src_csc_parm4 = 0x00100000;
        gic_ctl.gic1_src_csc_parm5 = 0x00800080;
        gic_ctl.gic1_src_csc_parm6  = 0xFF00FF00;
        gic_ctl.gic1_src_csc_parm7  = 0x0000FF00;
        
        //preset the trigger command
        gic_ctl.gic_trigger_command = GIC0_OP_KICKOFF_MASK |
                                      (SRC1_OVER_BACKGROUND << GIC0_BLEND_LEVELS_SHIFT) |     
                                      GIC0_BLEND_EN_MASK | 
                                      //(0 << GIC0_BLEND_DOMAIN_SHIFT) |
                                      (GIC0_CSC_SEL_SRC1 << GIC0_CSC_SRC_SEL_SHIFT); 

        gic_ctl.gic_dst_pitch = pmdev->dst_pitch;
        gic_ctl.gic_dst_y_x = 0 | (0 << GIC0_DST_Y_SHIFT); //always from (0, 0) to bypass the GIC flaw.
        gic_ctl.gic_dst_mode = GIC_ARGB8888;
        gic_ctl.gic_opt_status |= GIC_DST_READY | GIC_SRC1_READY | GIC_SRC2_READY;

        int dst_curr_width;
        
        int curr_height;

        int dst_gap;

        int org_src1_adr0, org_src1_adr1, org_src1_adr2;
        int org_dst_adr0;

        int actual_advance_pixels;

        int src_y = pmdev->rectsrc_y;
        int dst_y = pmdev->target_dy;
        int remaining_height = pmdev->rectsrc_h;


        /*change the values that are different for each loop*/
        while (remaining_height > 0)
        {
            int src_x = pmdev->rectsrc_x;
            int dst_x = pmdev->target_dx;
            int dst_remaining_width = pmdev->rectsrc_w;
            
            while (dst_remaining_width > 0) 
            {
                int src_uv_gap, src_y_gap;
                
                unsigned int src_offset = src_y * pmdev->src_pitch + src_x;
                unsigned int src_uv_offset;

                int residue_uv;  //used to represent if the starting x is at the odd position.
                if (pmdev->src_format == DSPF_YUV444P) {
                    src_uv_offset = src_y * src_uv_pitch + src_x;
                    residue_uv = 0;
                } 
                else if (pmdev->src_format == DSPF_YV16) {
                    if (src_x & 0x1) {
                        DPRINTF("src x is not even for YV16!\n");
                    }
                        
                    src_uv_offset = src_y * src_uv_pitch + src_x / 2;
                    residue_uv = src_x & 1;
                }
                else if (pmdev->src_format == DSPF_I420) {
                    if (src_x & 0x1) {
                        DPRINTF("src x is not even for I420!\n");
                    }
                    if (src_y & 0x1) {
                        DPRINTF("src y is not even for I420!\n");
                    }

                    src_uv_offset = src_y / 2 * src_uv_pitch + src_x / 2;
                    residue_uv = src_x & 1;
                }

                org_src1_adr0 = pmdev->src_phys + src_offset;
                org_src1_adr1 = pmdev->src_phys + y_size + src_uv_offset;
                org_src1_adr2 = org_src1_adr1 + uv_size;

                //find the nearest 32 aligned start address.
                src_uv_gap = org_src1_adr1 - (org_src1_adr1 & ~0x1f);
                src_y_gap = src_uv_gap * times_w + residue_uv;
                DPRINTF("residue_uv = %d, src_uv_gap = %d, src_y_gap = %d\n", residue_uv, src_uv_gap, src_y_gap);

                gic_ctl.gic_src1_adr0 = org_src1_adr0 - src_y_gap;
                gic_ctl.gic_src1_adr1 = org_src1_adr1 - src_uv_gap;
                gic_ctl.gic_src1_adr2 = org_src1_adr2 - src_uv_gap;
                
                DPRINTF("after addr0 = 0x%x, addr1= 0x%x, add2=0x%x\n", gic_ctl.gic_src1_adr0, 
                              gic_ctl.gic_src1_adr1, gic_ctl.gic_src1_adr2);                
                D_ASSERT((gic_ctl.gic_src1_adr0 & 0x1f) == 0);     
                D_ASSERT((gic_ctl.gic_src1_adr1 & 0x1f) == 0);  
                D_ASSERT((gic_ctl.gic_src1_adr2 & 0x1f) == 0);

                if (pmdev->src_format == DSPF_YV16) {
                    u32 tmp = gic_ctl.gic_src1_adr1;
                    gic_ctl.gic_src1_adr1 = gic_ctl.gic_src1_adr2;
                    gic_ctl.gic_src1_adr2 = tmp;
                }

                unsigned int dst_offset = dst_y * pmdev->dst_pitch + dst_x * 4;
                org_dst_adr0 = pmdev->dst_phys + dst_offset;
                
                DPRINTF("before dst_adr0 = 0x%x\n", org_dst_adr0);

                //find the nearest 32 aligned start address.
                int dst_gap_bytes = org_dst_adr0 - (org_dst_adr0 & ~0x1f);
                dst_gap = dst_gap_bytes >> 2; 

                DPRINTF("src_y_gap =%d, dst_gap = %d\n", src_y_gap, dst_gap);

                //make sure the dst_gap is larger than src_y_gap, 
                if(src_y_gap > dst_gap) {
                    dst_gap += ((src_y_gap - dst_gap) / 8 + 1) * 8;
                    DPRINTF("adjusting gap, dst_gap = %d\n", dst_gap);
                }
                dst_gap_bytes = dst_gap << 2;
                gic_ctl.gic_dst_adr0 = org_dst_adr0 - dst_gap_bytes;
                
                DPRINTF("after dst_adr0 = 0x%x\n", gic_ctl.gic_dst_adr0);    
                D_ASSERT((gic_ctl.gic_dst_adr0 & 0x1f) == 0); 

                dst_remaining_width += dst_gap;
                dst_curr_width = dst_remaining_width > GIC_WIDTH_LIMIT ? GIC_WIDTH_LIMIT : dst_remaining_width;

                DPRINTF("dst_remaining_width = %d, dst_curr_width = %d\n", dst_remaining_width, dst_curr_width);
                
                int src_x_in_dst = dst_gap - src_y_gap;
                int src_curr_width = dst_curr_width - src_x_in_dst;
                DPRINTF("src_x_in_dst = %d, src_curr_width = %d\n", src_x_in_dst, src_curr_width);

                curr_height = remaining_height > GIC_HEIGHT_LIMIT ? GIC_HEIGHT_LIMIT : remaining_height;

                gic_ctl.gic_src1_h_w = src_curr_width | (curr_height << GIC0_SRC1_H_SHIFT);
                gic_ctl.gic_dst_h_w = dst_curr_width | (curr_height << GIC0_DST_H_SHIFT);

                if (dst_gap) {
                    gic_ctl.gic_trigger_command |= GIC0_DST_CLIP_EN_MASK;
                    gic_ctl.gic_clip_top_left = dst_gap | (0 << GIC0_CLIP_TOP_SHIFT);
                    gic_ctl.gic_clip_bot_right = (dst_curr_width - 1) | ((curr_height - 1) << GIC0_CLIP_BOT_SHIFT);
                    DPRINTF("clip left=%d, right = %d\n", dst_gap, dst_curr_width - 1);
                }

                //GIC0_SRC1_Y_X denotes the coordinate of the top-left point of source surface in the destination
                //surface. Both surface need to start at 32-aligned address. GIC0_CLIP_TOP_LEFT comes into
                //effect if starting address need to be aligned.
                gic_ctl.gic_src1_y_x = src_x_in_dst | (0 << GIC0_SRC1_Y_SHIFT);
                
                Xcode4_Trigger_Gic0_Action(mdrv, pmdev, &gic_ctl);

                actual_advance_pixels = dst_curr_width - dst_gap;
                
                dst_remaining_width -= dst_curr_width;
                src_x += actual_advance_pixels;
                dst_x += actual_advance_pixels;
                                
                DPRINTF("src_x advances %d\n", actual_advance_pixels);     
            }

            remaining_height -= curr_height;
            src_y += curr_height;
            dst_y += curr_height;
        }

        xcodeUnlockG2d(mdrv);
    }
    
    return blit_success;
}

bool xcodeBlit( void *drv, void *dev, DFBRectangle *rect, int dx, int dy)
{
     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
     u32 dst_pixel_format_color;
     u32 src_pixel_format_color;
     u32 g2d_op_ctl_value;
     int base_size_ofoperation = BASE_SIZE_OF_BLIT;
     u32 *surface_mem;
     u32 pixvalue;
     u32 value;
     u32 blend = 0;
     u32 fifo_entry;
     
     DPRINTF( "xcode DFB driver xcodeBlit enter:\n" );
     DPRINTF( "xcode DFB driver xcodeBlit = %x %x %x %x %x %x\n",rect->x,rect->y,rect->w,rect->h, dx, dy );

     BLITPRINTF("G2D blittingflags = 0x%x, blend = %d,%d, src = %x, dst = %x\n", pmdev->blittingflags, pmdev->src_blend, pmdev->dst_blend, pmdev->src_phys, pmdev->dst_phys);

     if(pmdev->src_pitch > G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK - 3)
     {
          DPRINTF("Limitation: source pitch 0x%x greater than G2D hardware maximum 0x%x!\n", 
                         pmdev->src_pitch, G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK - 3);
          return false;
     }
     
     if(pmdev->dst_pitch > G2D_DST_PITCH_FORMAT_DST_PITCH_MASK - 3)
     {
          DPRINTF("Limitation: dst pitch 0x%x greater than G2D hardware maximum 0x%x!\n", 
                         pmdev->dst_pitch, G2D_DST_PITCH_FORMAT_DST_PITCH_MASK - 3);
          return false;
     }

     
     xcodeDetermineWorkingRegion(drv, dev, rect, dx, dy);

     //set copy rectangle feature
     if (pmdev->modified & SMF_CLIP)
     {
        //turn on the clip
         g2d_op_ctl_value = (XC_G2D_OP_COPY_RECT << GFX_OP_SHIFT) |(SRC_ALPHA_WRITE_OUT << SRC_ALPHA_SHIFT);
         base_size_ofoperation = base_size_ofoperation + SIZE_OF_SET_CLIP_REGION;
     }
     else
     {
         g2d_op_ctl_value = (SRC_ALPHA_WRITE_OUT << SRC_ALPHA_SHIFT) | (XC_G2D_OP_COPY_RECT << GFX_OP_SHIFT)  | (1 <<SCISSOR_DIS_SHIFT );
     }

     /* xcodeBlit can only handle
      * DSPD_CLEAR
      * DSPD_SRC
      * DSPD_NONE for DST w/o alpha
      */
    if ((pmdev->blittingflags & DSBLIT_BLEND_ALPHACHANNEL) ||
        (pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA))
    {
         if ((pmdev->src_blend == DSBF_ZERO) && 
             (pmdev->dst_blend == DSBF_ZERO))
         {
              DFBRectangle temp_rect;
              temp_rect.x = dx;
              temp_rect.y = dy;
              temp_rect.w = rect->w;
              temp_rect.h = rect->h;

              BLITPRINTF("Clear rectangle\n");

              //porter duff clear, just set 0
              pmdev->fill_color = 0;
              pmdev->drawingflags = 0;
              return xcodeFillRectangle(drv, dev, &temp_rect);
         }
         else if(pmdev->dst_blend != DSBF_ZERO)
         {
             if ((pmdev->blittingflags & DSBLIT_BLEND_ALPHACHANNEL) ||
                  (pmdev->blittingflags & DSBLIT_SRC_PREMULTIPLY))
             {
                 //turn on the source color alpha
                 g2d_op_ctl_value = g2d_op_ctl_value | (GFX_ALPHA_MODE_SRC_ALPHA << GFX_ALPHA_MODE_SHIFT);
                 pmdev->blittingflags &= ~DSBLIT_SRC_PREMULTIPLY;
             }
             else if ((pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA) ||
                      (pmdev->blittingflags & DSBLIT_SRC_PREMULTCOLOR))
             {
                 //turn on the foreground color alpha
                 g2d_op_ctl_value = g2d_op_ctl_value | (GFX_ALPHA_MODE_FGND_COLOR_ALPHA << GFX_ALPHA_MODE_SHIFT);
                 pmdev->blittingflags &= ~DSBLIT_BLEND_COLORALPHA;
                 base_size_ofoperation ++;
             }
             blend = 1;
         }
     }

     if((pmdev->blittingflags & DSBLIT_SRC_PREMULTIPLY) && !blend)
     {
         /* Fake to enable premultiply */
         pmdev->src_blend = DSBF_SRCALPHA;
     }

     if ((pmdev->dst_format == DSPF_ARGB) &&
         (is_ycbcr[DFB_PIXELFORMAT_INDEX(pmdev->src_format)]) &&
         (!(pmdev->blittingflags & 
            (DSBLIT_BLEND_ALPHACHANNEL 
             | DSBLIT_BLEND_COLORALPHA 
             | DSBLIT_SRC_COLORKEY 
             | DSBLIT_DST_COLORKEY
             | DSBLIT_SRC_PREMULTIPLY
             | DSBLIT_DST_PREMULTIPLY
             | DSBLIT_DEMULTIPLY)
             )))
     {
#ifndef GIC_YUV444_NO_FLAW          
          if(pmdev->src_format == DSPF_YUV444P) {//to circumvent GIC YUV444 flaw
               DPRINTF("Limitation: use software blit to circumvent GIC YUV444 flaw.\n");
               return false;               
          }
#endif          
          return Xcode4_BlitOnly_YUV_To_ARGB(drv, dev);
     }
     

     if ((pmdev->dst_format == DSPF_I420) &&
         (pmdev->src_format == DSPF_ARGB))
     {
         return Xcode4_BlitOnly_ARGB_To_YUV(drv, dev);
     }

     //convert the directfb color format to xcode4 g2d color format
     dst_pixel_format_color = get_pixel_format_color(pmdev->dst_format);
     if (dst_pixel_format_color == PIX_FORMAT_UNKNOWN)
     {
         return false;
     }

     //convert the directfb color format to xcode4 g2d color format
     src_pixel_format_color = get_pixel_format_color(pmdev->src_format);
     if (src_pixel_format_color == PIX_FORMAT_UNKNOWN)
     {
         return false;
     }

     xcodeLockG2d(mdrv);

     /* Check if premultiply is needed */
     if(xcodeNeedPremultiply(drv, dev))
     {
        if(!xcodeGicPremultiply(drv, dev))
        {
            /* Fail blit if premultiply failed */
            xcodeUnlockG2d(mdrv);
            return false;
        }
        else
        {
            /* Modify src rect */
            rect->x = 0;
            rect->y = 0;
            src_pixel_format_color = get_pixel_format_color(pmdev->src_format);
        }
     }
     
      /* Flips/rotations are usually handled by this register ... */
      /* 
      *   By default, we program the G2D for Left to right, top to bottom on destination ...
      *   By controlling the direction on the source, we can create a "flip" effect (hopefully ... )
      *   source direction is left to right, top to bottom ... 
      */
      int source_direction    =   0;

      if (pmdev->blittingflags & DSBLIT_FLIP_VERTICAL)
      {
        /* Change Source direction is now right to left ... */
        source_direction |= 0x1;    /* Flip vertical flag (internal to this function only)*/
      }

      if (pmdev->blittingflags & DSBLIT_FLIP_HORIZONTAL)
      {
        /* Change source direction to include bottom to top ...*/
        source_direction |= 0x2;    /* Flip horizontal flag */
      }

      if (pmdev->blittingflags & DSBLIT_ROTATE180)
      {
        source_direction = 0x3; /* Flip both vertical and horizontal ... */
      }

      int src_dir_mask    =   0;
      
      if (source_direction == 0x1)
      {
        /* Left to right, bottom to top */
        src_dir_mask    =  ((LEFT_TO_RIGHT_BOT_TO_TOP << BLTDIR_SHIFT) & BLTDIR_MASK)
                                 | ((LEFT_TO_RIGHT_TOP_TO_BOT  << DEST_BLT_DIR_SHIFT) & DEST_BLT_DIR_MASK);
      }
      else if (source_direction == 0x2)
      {
        /* Right to left, top to bottom */
        src_dir_mask    =   ((RIGHT_TO_LEFT_TOP_TO_BOT << BLTDIR_SHIFT) & BLTDIR_MASK)
                                | ((LEFT_TO_RIGHT_TOP_TO_BOT  << DEST_BLT_DIR_SHIFT) & DEST_BLT_DIR_MASK);
      }
      else if (source_direction == 0x3)
      {
        /* Right to left, bottom to top */
        src_dir_mask    =   ((RIGHT_TO_LEFT_BOT_TO_TOP << BLTDIR_SHIFT) & BLTDIR_MASK)
                                | ((LEFT_TO_RIGHT_TOP_TO_BOT  << DEST_BLT_DIR_SHIFT) & DEST_BLT_DIR_MASK);
      }

      g2d_op_ctl_value |= src_dir_mask;
           
     //     start blit operation
     G2d_Operation_Start_For_Blit(mdrv, pmdev, &g2d_op_ctl_value, base_size_ofoperation);

     //Ensure there is enough space in the FIFO
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 1);
     
     //Fill Color
     REG_WRITE(G2D_FRGND_COLOR,(pmdev->dst_color.a << 24));

     //check if setting clip region
     if (pmdev->modified & SMF_CLIP)
     {
         Set_Clip_Region(mdrv, pmdev);
     }

     //set destination registers
     Set_Destination_Registers(mdrv, pmdev, dst_pixel_format_color, dx, dy);

     //set source registers
     Set_Source_Registers(mdrv, pmdev, src_pixel_format_color, rect);

     //finish the operation
     G2d_Operation_Finish(mdrv, pmdev, &g2d_op_ctl_value, rect);

     xcodeUnlockG2d(mdrv);;

     DPRINTF( "xcode DFB driver xcodeBlit exit:\n" );

     return true;
}

static bool xcodeGicStretchBlit(void* drv, void* dev, DFBRectangle* src_rec, DFBRectangle* dst_rec);

static bool xcodeGicBlit( void *drv, void *dev, DFBRectangle *rect, int dx, int dy )
{
     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
     u32 blittingflags = pmdev->blittingflags;
     bool ret;

     DPRINTF( "xcode DFB driver xcodeGicBlit enter:\n" );

     BLITPRINTF("GIC blittingflags = 0x%x, blend = %d,%d, src = %x, dst = %x\n", pmdev->blittingflags, pmdev->src_blend, pmdev->dst_blend, pmdev->src_phys, pmdev->dst_phys);

     /* Get region */
     xcodeDetermineWorkingRegion(drv, dev, rect, dx, dy);

     /* Check if alpha blending is needed */
     if(!(blittingflags & (DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_BLEND_COLORALPHA)))
     {
        /* Clear premultiply flag */
        pmdev->src_blend = DSBF_ONE;
        pmdev->dst_blend = DSBF_ZERO;
        return xcodeBlit(drv, dev, rect, dx, dy);
     }

     /* Check if rotate is needed */
     if (blittingflags & (DSBLIT_FLIP_VERTICAL |DSBLIT_FLIP_HORIZONTAL |DSBLIT_ROTATE180))
     {
         DFBRectangle dst_rect;
         dst_rect.x = dx;
         dst_rect.y = dy;
         dst_rect.w = rect->w;
         dst_rect.h = rect->h;
         return xcodeGicStretchBlit(drv, dev, rect, &dst_rect);
     }

     xcodeLockG2d(mdrv);

     ret = xcodeGicBasic(mdrv, pmdev);

     xcodeUnlockG2d(mdrv);
     
     DPRINTF( "xcode DFB driver xcodeGicBlit exit:\n" );
     
     return ret;
}


/*
 *
 *Function Name:xcodeG2dStretchBlit
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *src_rec-> Rectangle describing source size
 *           *dst_rec-> Rectangle describing destination size   
 *
 *Description:
 *           Performs a stretch SRC copy.
 *
 *Returns:
 *           true or false
*/
bool xcodeStretchBlit(void* drv, void* dev, DFBRectangle* src_rec, DFBRectangle* dst_rec)
{
     pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
     pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
     int dst_pixel_format_color;
     int src_pixel_format_color;
     u32 g2d_op_ctl_value;
     int base_size_ofoperation = BASE_SIZE_OF_STRETCH_BLIT;
     bool ret_value;
     u32 *src_addr;
     u32 *dst_addr;
     int i;
     DFBRectangle temp_rect;
     u32 temp;
     bool temp_ret;
     u32 fifo_entry;

     DPRINTF( "xcode DFB driver xcodeStretchBlit enter:\n" );
     DPRINTF( "xcode DFB driver xcodeStretchBlit = %x %x %x %x %x %x %x %x\n",
              src_rec->x, src_rec->y, src_rec->w, src_rec->h, dst_rec->x, dst_rec->y, dst_rec->w, dst_rec->h );

     if(pmdev->src_pitch > G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK - 3)
     {
          DPRINTF("Limitation: source pitch 0x%x greater than G2D hardware maximum 0x%x!\n", 
                    pmdev->src_pitch, G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK - 3);
          return false;
     }

     if(pmdev->dst_pitch > G2D_DST_PITCH_FORMAT_DST_PITCH_MASK - 3)
     {
          DPRINTF("Limitation: dst pitch 0x%x greater than G2D hardware maximum 0x%x!\n", 
                         pmdev->dst_pitch, G2D_DST_PITCH_FORMAT_DST_PITCH_MASK - 3);
          return false;
     }

     if(dst_rec->w <= src_rec->w / 8 || dst_rec->h <= src_rec->h / 8) {
          DPRINTF("Limitation: shrink ratio exceeds hardware limitation 1/8\n");
          return false;
     }

     xcodeDetermineStretchWorkingRegion(drv, dev, src_rec, dst_rec);

     //convert the directfb color format to xcode4 g2d color format
     dst_pixel_format_color = get_pixel_format_color(pmdev->dst_format);
     if (dst_pixel_format_color == PIX_FORMAT_UNKNOWN)
     {
        return false;
     }

     //convert the directfb color format to xcode4 g2d color format
     src_pixel_format_color = get_pixel_format_color(pmdev->src_format);
     if (src_pixel_format_color == PIX_FORMAT_UNKNOWN)
     {
        return false;
     }

     //set scale rectangle feature
     if (pmdev->modified & SMF_CLIP)
     {
       //turn on the scissor feature
        g2d_op_ctl_value = (XC_G2D_OP_SCALE_RECT << GFX_OP_SHIFT) | (SRC_ALPHA_WRITE_OUT << SRC_ALPHA_SHIFT);
        base_size_ofoperation = base_size_ofoperation + SIZE_OF_SET_CLIP_REGION;
     }
     else
     {
       //disable the scissor feature
        g2d_op_ctl_value = (XC_G2D_OP_SCALE_RECT << GFX_OP_SHIFT)  | (1 <<SCISSOR_DIS_SHIFT ) | (SRC_ALPHA_WRITE_OUT << SRC_ALPHA_SHIFT);
     }

     if ((pmdev->blittingflags & DSBLIT_BLEND_ALPHACHANNEL) ||
             (pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA))
     {
          if ((pmdev->src_blend == DSBF_ZERO) && 
             (pmdev->dst_blend == DSBF_ZERO))
          {
             temp_rect.x = dst_rec->x;
             temp_rect.y = dst_rec->y;
             temp_rect.w = dst_rec->w;
             temp_rect.h = dst_rec->h;

             //porter duff clear, just set 0
             temp = pmdev->fill_color;
             pmdev->fill_color = 0;
             temp_ret = xcodeFillRectangle(drv, dev, &temp_rect);
             pmdev->fill_color = temp;
             return temp_ret;
          }
          else if (pmdev->dst_blend != DSBF_ZERO)
          {
              if ((pmdev->blittingflags & DSBLIT_BLEND_ALPHACHANNEL) ||
                  (pmdev->blittingflags & DSBLIT_SRC_PREMULTIPLY))
              {
                 //turn on the source color alpha
                 g2d_op_ctl_value = g2d_op_ctl_value | (GFX_ALPHA_MODE_SRC_ALPHA << GFX_ALPHA_MODE_SHIFT);
              }
              else if ((pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA) ||
                      (pmdev->blittingflags & DSBLIT_SRC_PREMULTCOLOR))
              {
                 //turn on the foreground color alpha
                 g2d_op_ctl_value = g2d_op_ctl_value | (GFX_ALPHA_MODE_FGND_COLOR_ALPHA << GFX_ALPHA_MODE_SHIFT);
                 base_size_ofoperation ++;
              }
              pmdev->blittingflags &= ~(DSBLIT_SRC_PREMULTCOLOR | DSBLIT_BLEND_COLORALPHA);
          }
     }

     BLITPRINTF("Stretch blittingflags = 0x%x, blend = %d,%d, src = %x, dst = %x\n", pmdev->blittingflags, pmdev->src_blend, pmdev->dst_blend, pmdev->src_phys, pmdev->dst_phys);

     xcodeLockG2d(mdrv);

     /* Check if premultiply is needed */
     if(xcodeNeedPremultiply(drv, dev))
     {
        pmdev->rectsrc_w = pmdev->rect_scalesrc_w;
        pmdev->rectsrc_h = pmdev->rect_scalesrc_h;

        if(!xcodeGicPremultiply(drv, dev))
        {
            /* Fail blit if premultiply failed */
            xcodeUnlockG2d(mdrv);
            return false;
        }

        src_rec->x = 0;
        src_rec->y = 0;
        src_pixel_format_color = get_pixel_format_color(pmdev->src_format);
     }

	 /* Flips/rotations are usually handled by this register ... */
	 /* 
	   *   By default, we program the G2D for Left to right, top to bottom on destination ...
	   *   By controlling the direction on the source, we can create a "flip" effect (hopefully ... )
	   *   source direction is left to right, top to bottom ... 
	   */
     int source_direction    =   0;

     if (pmdev->blittingflags & DSBLIT_FLIP_VERTICAL)
     {
       /* Change Source direction is now right to left ... */
       source_direction |= 0x1;    /* Flip vertical flag (internal to this function only)*/
     }

     if (pmdev->blittingflags & DSBLIT_FLIP_HORIZONTAL)
     {
       /* Change source direction to include bottom to top ...*/
       source_direction |= 0x2;    /* Flip horizontal flag */
     }

     if (pmdev->blittingflags & DSBLIT_ROTATE180)
     {
       source_direction = 0x3; /* Flip both vertical and horizontal ... */
     }

     int src_dir_mask    =   0;
      
     if (source_direction == 0x1)
     {
       /* Left to right, bottom to top */
       src_dir_mask    =  ((LEFT_TO_RIGHT_BOT_TO_TOP << BLTDIR_SHIFT) & BLTDIR_MASK)
                                | ((LEFT_TO_RIGHT_TOP_TO_BOT  << DEST_BLT_DIR_SHIFT) & DEST_BLT_DIR_MASK);
     }
     else if (source_direction == 0x2)
     {
       /* Right to left, top to bottom */
       src_dir_mask    =   ((RIGHT_TO_LEFT_TOP_TO_BOT << BLTDIR_SHIFT) & BLTDIR_MASK)
                               | ((LEFT_TO_RIGHT_TOP_TO_BOT  << DEST_BLT_DIR_SHIFT) & DEST_BLT_DIR_MASK);
     }
     else if (source_direction == 0x3)
     {
       /* Right to left, bottom to top */
       src_dir_mask    =   ((RIGHT_TO_LEFT_BOT_TO_TOP << BLTDIR_SHIFT) & BLTDIR_MASK)
                               | ((LEFT_TO_RIGHT_TOP_TO_BOT  << DEST_BLT_DIR_SHIFT) & DEST_BLT_DIR_MASK);
     }

     g2d_op_ctl_value |= src_dir_mask;

     //start blit operation
     G2d_Operation_Start_For_Blit(mdrv, pmdev, &g2d_op_ctl_value, base_size_ofoperation);

     //Ensure there is enough space in the FIFO
     WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 1);

     REG_WRITE(G2D_FRGND_COLOR,(pmdev->dst_color.a << 24));

     //check if setting clip region
     if (pmdev->modified & SMF_CLIP)
     {
        Set_Clip_Region(mdrv, pmdev);
     }

     //set destination registers
     Set_Destination_Registers(mdrv, pmdev, dst_pixel_format_color, dst_rec->x, dst_rec->y);

     //set source registers
     Set_Source_Registers_For_Stretch_Blit(mdrv, pmdev, src_pixel_format_color, src_rec);

     //finish the stretch blit operation
     G2d_Operation_Finish_Stretch_Blit(mdrv, pmdev, &g2d_op_ctl_value, src_rec, dst_rec);

     xcodeUnlockG2d(mdrv);

     DPRINTF( "xcode DFB driver xcodeStretchBlit exit:\n" );

     return true;

}

/*
 *
 *Function Name:xcodeG2dStretchBlit
 *
 *Parameters: 
 *           drv-> Xcode driver context
 *           dev-> Current device context(only 1 supported for now)
 *           *src_rec-> Rectangle describing source size
 *           *dst_rec-> Rectangle describing destination size   
 *
 *Description:
 *           Performs a stretch SRC copy.
 *
 *Returns:
 *           true or false
*/
static bool xcodeGicStretchBlit(void* drv, void* dev, DFBRectangle* src_rec, DFBRectangle* dst_rec)
{
    pxc_dfb_driver_data mdrv = (pxc_dfb_driver_data) drv;
    pxc_dfb_device_data pmdev = (pxc_dfb_device_data) dev;
    
    u32 temp_pitch;
    u32 temp_width;
    u32 temp_height;
    /* Destination will always be ARGB... */
    u32 src_bytes_pixel;
    u32 dst_bytes_pixel;
    u32 dst_bytes_per_line;
    u32 required_buffer_size;
    u32 temp_physical; 
    DFBRectangle strRect;
     u32 scaler_x, scaler_y;
    u32 source_direction, src_dir_mask;
    bool ret;

    if(pmdev->src_pitch > G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK - 3)
    {
         DPRINTF("Limitation: source pitch 0x%x greater than G2D hardware maximum 0x%x!\n", 
                    pmdev->src_pitch, G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK - 3);
         return false;
    }

    if(pmdev->dst_pitch > G2D_DST_PITCH_FORMAT_DST_PITCH_MASK - 3)
    {
        DPRINTF("Limitation: dst pitch 0x%x greater than G2D hardware maximum 0x%x!\n", 
                     pmdev->dst_pitch, G2D_DST_PITCH_FORMAT_DST_PITCH_MASK - 3);
        return false;
    }

    if ((dst_rec->w <= GIC_MIN_X_SIZE) ||(dst_rec->h <= GIC_MIN_X_SIZE))
    {
        return false;
    }
    else if ((src_rec->w / dst_rec->w) >= G2D_DOWNSCALE_LIMIT)
    {
        return false;
    }
    else if ((src_rec->h / dst_rec->h) >= G2D_DOWNSCALE_LIMIT)
    {
        return false;
    }

    if(pmdev->dst_format != DSPF_ARGB && pmdev->dst_format != DSPF_RGB16)
    {
        /* Fail blit if dest is not RGB format */
        return false;
    }

    /* Check Porter Duff Rule */
    if(!(pmdev->src_blend == DSBF_ONE && pmdev->dst_blend == DSBF_INVSRCALPHA) &&
        !(pmdev->src_blend == DSBF_SRCALPHA && pmdev->dst_blend == DSBF_INVSRCALPHA))
    {
        /* Now our GIC can only handle the following case:
         * Ao = A1 + (1-A1) * A2
         * Co = [A1] * C1 + (1-A1)*[A2]*C2
         * From DFB's description:
         * pixel color = sc * cf[sf] + dc * cf[df]
         * pixel alpha = sa * af[sf] + da * af[df]
         * We can only handle DSPD_CLEAR, DSPD_SRC, DSPD_SRC_OVER correctly
         * However, since DSPD_NONE is used wildly, we still use hardware to do it
         * Other rules are not supported.
         */
        return false;
    }

    /* Check if alpha blending is needed */
    if(!(pmdev->blittingflags & (DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_BLEND_COLORALPHA)))
    {
        /* Clear premultiply flag */
        pmdev->src_blend = DSBF_ONE;
        pmdev->dst_blend = DSBF_ZERO;
        return xcodeStretchBlit(drv, dev, src_rec, dst_rec);
    }

    /* Check if alpha is zero */
    if ((pmdev->blittingflags & DSBLIT_BLEND_COLORALPHA) && !pmdev->src_color_alpha)
    {
       return true;
    }

    /* If the temporary buffer allocation has failed, do not do anything, fail the call */
    
    /*
    *   Steps to perform this:
    *   1.  Use G2D to take source region and stretch into the temporary buffer
    *   For this one, we may need to adjust the pitch of the destination (for the temp buffer),
    *   to a more reasonable value to make it more memory friendly...
    *   In addition, the alpha component must be left as is...
    *   Rotations can be handled by us here in the G2D function by manipulating the Blit directions
    *   on the source and the destination. 
    *
    *   2. Pass on the region in the temporary buffer to xcodeGicBlit and perform as is...
    *
    */
    
    
    /*
    *   Aug 04/2011:
    *   NOTE: It seems like there is some contention going on here. Basically during
    *   BTD100 run with the Scaling-Size=150, it looks like the next G2D scale is being
    *   done before the GIC can finish the blend. If we can direct the G2D scale to an unused portion
    *   of the temporary buffer, we can speed this part even further.
    *
    */

    BLITPRINTF("GIC Stretch blittingflags = 0x%x, blend = %d,%d, src = %x, dst = %x\n", pmdev->blittingflags, pmdev->src_blend, pmdev->dst_blend, pmdev->src_phys, pmdev->dst_phys);

    xcodeDetermineStretchWorkingRegion(drv, dev, src_rec, dst_rec);
    
    src_bytes_pixel	=	DFB_BYTES_PER_PIXEL(pmdev->src_format);
	dst_bytes_pixel	=	DFB_BYTES_PER_PIXEL(pmdev->dst_format);
    
    dst_bytes_per_line  =   dst_bytes_pixel * pmdev->rectsrc_w;
    
    /* Let us align this value to GIC align (32 bytes) ... */
    if (dst_bytes_per_line & (32 - 1) )
    {
        /* Add what is necessary to align to the next 32-byte boundary .. */
        dst_bytes_per_line &= (~(32-1));
        dst_bytes_per_line+=    32;
    
    }
    D_DEBUG("W: %u, P: 0x%X\n", 
            pmdev->rectsrc_w, dst_bytes_per_line);

    required_buffer_size    =   
        (pmdev->rectsrc_h * dst_bytes_per_line);
    
    if (required_buffer_size > TEMP_RENDER_SIZE)
    {
        D_DEBUG("Required size for temp buffer [%u] is greater than actual size [%u]\n",
            required_buffer_size, TEMP_RENDER_SIZE);
        D_DEBUG("This blit cannot be performed by hardware...\n");
        return false;
    }

    if (pmdev->src_pitch > G2D_MAX_PITCH_SIZE)
    {
        return false;
    }
    
    /* Calculate scaling co-efficient for the G2D ... */
    scaler_x = (pmdev->rect_scalesrc_w << SCALER_MULTIPLY_SHIFT) / pmdev->rectsrc_w;
    scaler_y = (pmdev->rect_scalesrc_h << SCALER_MULTIPLY_SHIFT) / pmdev->rectsrc_h;
    D_DEBUG("Scaler Coeff [X Y] in U3.12: 0x%X, 0x%X\n",
             scaler_x, scaler_y
           );

    //Try pipeline Gic Stretch Blit
    if(PipelineGicStretchBlit(drv, dev, true))
    {
        return true;
    }

    xcodeLockG2d(mdrv);

    //added by xwu: wait for posible GIC0 operations
    Wait_For_Gic0_Idle;

   /* Setup G2D for stretching operation ... */
   xcode4_waitfifo(mdrv, pmdev, 10);
   
   REG_WRITE(G2D_DST_OFFSET, pmdev->temp_render_physical);
   REG_WRITE(G2D_DST_Y_X, 0x00000000);
   REG_WRITE(G2D_DST_PITCH_FORMAT, 
       ((dst_bytes_per_line << G2D_DST_PITCH_FORMAT_DST_PITCH_SHIFT) & G2D_DST_PITCH_FORMAT_DST_PITCH_MASK) |
       ((get_pixel_format_color(pmdev->dst_format) << DST_PIX_FORMAT_SHIFT) & DST_PIX_FORMAT_MASK) |
       ((GFX_ROP_ALPHA_FORMAT_SRC_ALPHA << ROP_ALPHA_FORMAT_SHIFT) & ROP_ALPHA_FORMAT_MASK)  
   );
   
   REG_WRITE(G2D_SRC_OFFSET, pmdev->src_phys);
   REG_WRITE(G2D_SRC_Y_X,  
               ((pmdev->rectsrc_x << SRC_X_SHIFT) & SRC_X_MASK) |
               ((pmdev->rectsrc_y << SRC_Y_SHIFT) & SRC_Y_MASK)
             );
   REG_WRITE(G2D_SRC_H_W,
               ((pmdev->rect_scalesrc_w << SRC_W_SHIFT) & SRC_W_MASK) |
               ((pmdev->rect_scalesrc_h << SRC_H_SHIFT) & SRC_H_MASK)
             );

   //Check palette
   check_g2d_palette(mdrv, pmdev);

   u32 scale_type;
        
   if ((pmdev->rect_scalesrc_w <= pmdev->rectsrc_w) ||
       (pmdev->rect_scalesrc_h <= pmdev->rectsrc_h))
   {
       scale_type = SCALING_TYPE_LERP;
   }
   else
   {
       scale_type = SCALING_TYPE_NEIGHBOR;
   }

   REG_WRITE(G2D_SRC_PITCH_FORMAT,
       ((pmdev->src_pitch << G2D_SRC_PITCH_FORMAT_SRC_PITCH_SHIFT) & G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK) |
       ((get_pixel_format_color(pmdev->src_format) << SRC_PIX_FORMAT_SHIFT) & SRC_PIX_FORMAT_MASK)  |
       ((SCALING_TYPE_LERP << SCALING_TYPE_SHIFT) & SCALING_TYPE_MASK)
       );
   
   REG_WRITE(G2D_SCALERYX, 
               ((scaler_x << XSCALER_SHIFT) & XSCALER_MASK)  |
               ((scaler_y << YSCALER_SHIFT) & YSCALER_MASK)
             );
   
   /* Flips/rotations are usually handled by this register ... */
   /* 
   *   By default, we program the G2D for Left to right, top to bottom on destination ...
   *   By controlling the direction on the source, we can create a "flip" effect (hopefully ... )
   *   source direction is left to right, top to bottom ... 
   */
   source_direction    =   0;
   
   if (pmdev->blittingflags & DSBLIT_FLIP_VERTICAL)
   {
       /* Change Source direction is now right to left ... */
       source_direction |= 0x1;    /* Flip vertical flag (internal to this function only)*/
   }
   
   if (pmdev->blittingflags & DSBLIT_FLIP_HORIZONTAL)
   {
       /* Change source direction to include bottom to top ...*/
       source_direction |= 0x2;    /* Flip horizontal flag */
       
   }
   
   if (pmdev->blittingflags & DSBLIT_ROTATE180)
   {
       source_direction = 0x3; /* Flip both vertical and horizontal ... */
   }
   
   src_dir_mask    =   0;
   if (source_direction == 0x1)
   {
       /* Left to right, bottom to top */
       src_dir_mask    =  ((LEFT_TO_RIGHT_BOT_TO_TOP << BLTDIR_SHIFT) & BLTDIR_MASK);
   }
   else if (source_direction == 0x2)
   {
       /* Right to left, top to bottom */
       src_dir_mask    =   ((RIGHT_TO_LEFT_TOP_TO_BOT << BLTDIR_SHIFT) & BLTDIR_MASK);
   }
   else if (source_direction == 0x3)
   {
       /* Right to left, bottom to top */
       src_dir_mask    =   ((RIGHT_TO_LEFT_BOT_TO_TOP << BLTDIR_SHIFT) & BLTDIR_MASK);
   }
   

   REG_WRITE(G2D_OP_CTL,
               ((XC_G2D_OP_SRC_SEL_BLT     << SRC_SELECT_SHIFT) & SRC_SELECT_MASK)     |
               ((XC_G2D_OP_SCALE_RECT      << GFX_OP_SHIFT) & GFX_OP_MASK)             |
               ((LEFT_TO_RIGHT_TOP_TO_BOT  << DEST_BLT_DIR_SHIFT) & DEST_BLT_DIR_MASK) |
               ((GFX_SCISSOR_DISABLE       << SCISSOR_DIS_SHIFT) & SCISSOR_DIS_MASK)   |
               ((SRC_ALPHA_WRITE_OUT       << SRC_ALPHA_SHIFT) & SRC_ALPHA_MASK)       |
               ((GFX_ALPHA_MODE_NO_ALPHA   << GFX_ALPHA_MODE_SHIFT) & GFX_ALPHA_MODE_MASK) |
               ((src_dir_mask))
             );
   
   
   REG_WRITE(G2D_DST_H_W_INIT,
               ((pmdev->rectsrc_w << DST_W_SHIFT) & DST_W_MASK)    |
               ((pmdev->rectsrc_h << DST_H_SHIFT) & DST_H_MASK)
             );
   
   
   /* Wait for G2D to finish the stretch blit... */
   xcode4_waitidle(mdrv, pmdev);
    
    /* 
    *   With this complete, let us pass on the source onto xcodeGicBlit (with all the regular information), 
    *   xcodeGicBlit will be able to know if it is being invoked directly (or indirectly through xcodeGicStretchBlit),
    *   by observing "current_op" in the device structure. If it is set to DFXL_STRETCHBLIT, it is being invoked
    *   indirectly...
    */
    
    
    /* Pass on the dimensions of the image in the temporary buffer ... */
    pmdev->rectsrc_x = 0;
    pmdev->rectsrc_y = 0;
    
    /* Point to temporary buffer ... */
    pmdev->src_phys     =   pmdev->temp_render_physical;
    pmdev->src_pitch    =   dst_bytes_per_line;
    pmdev->src_width    =   pmdev->rectsrc_w;
    pmdev->src_height   =   pmdev->rectsrc_h;

    /* Our source format will now be our destination format ... */
    pmdev->src_format   =   pmdev->dst_format;

    /* Remove flip flags */
    pmdev->blittingflags &= ~(DSBLIT_FLIP_VERTICAL |DSBLIT_FLIP_HORIZONTAL |DSBLIT_ROTATE180);

    ret = xcodeGicBasic(drv, dev);

    xcodeUnlockG2d(mdrv);

    return ret;
}

void xcodeSetState( void *drv, void *dev,
                    GraphicsDeviceFuncs *funcs,
                    CardState *state, DFBAccelerationMask accel)
{

     pxc_dfb_device_data mdev = (pxc_dfb_device_data) dev;
     StateModificationFlags modified = state->modified;

     DPRINTF( "xcode DFB driver xcodeSetState enter:\n" );

     mdev->current_op   =   accel;
     /*
      * 1) Invalidate hardware states
      *
      * Each modification to the hw independent state invalidates one or more hardware states.
     */

     //
     // Simply invalidate all? 
     // 
     
     if (modified == SMF_ALL) 
     {
        XCODE_INVALIDATE_STATE( ALL );
    
        /* Load our color palette */
        if (mdev->src_color_r == state->color.r &&
            mdev->src_color_g == state->color.g &&
            mdev->src_color_b == state->color.b &&
            mdev->src_color_alpha == state->color.a
            )
        {
            mdev->src_color_update = 0;
            /* Clear the update flag */
        }
        else
        {
            mdev->src_color_update = 1;
            /* Set the palette update flag and 
            *  transfer color
            */
            mdev->src_color_r = state->color.r;
            mdev->src_color_g = state->color.g;
            mdev->src_color_b = state->color.b;
            mdev->src_color_alpha =
                state->color.a;
        
        }
        //xcodeGicSetColor(drv, dev, state->color);
        //mdev->src_color_alpha = state->color.a;
    
     }
     else if (modified)
     {
          //
          // Invalidate destination registers. 
          // 
          if (modified & SMF_DESTINATION)
          {
               XCODE_INVALIDATE_STATE( DESTINATION | COLOR );
          }
          else if (modified & SMF_COLOR)
          {     
               XCODE_INVALIDATE_STATE( COLOR );
               /* Setup the GIC color */
               if (mdev->src_color_r == state->color.r &&
                    mdev->src_color_g == state->color.g &&
                    mdev->src_color_b == state->color.b &&
                    mdev->src_color_alpha == state->color.a
                )
                {
                    mdev->src_color_update = 0;
                    /* Clear the update flag */
                }
                else
                {
                    mdev->src_color_update = 1;
                    /* Set the palette update flag and 
                    *  transfer color
                    */
                    mdev->src_color_r = state->color.r;
                    mdev->src_color_g = state->color.g;
                    mdev->src_color_b = state->color.b;
                    mdev->src_color_alpha =
                        state->color.a;
        
                } 
          }
          if (modified & SMF_SOURCE)
          {
               XCODE_INVALIDATE_STATE( SOURCE );
          }
     }

     DPRINTF( "xcodeCheckState here!!!!\n" );

     /*
      * 2) Validate hardware states
      *
      * Each function has its own set of states that need to be validated.
      */

     //
     // Always validate destination... 
     // 
     XCODE_CHECK_VALID_STATE( DESTINATION );

     switch (accel)
     {
          case DFXL_FILLRECTANGLE:
              //
              // ...require valid drawing color. 
              // 
              XCODE_CHECK_VALID_STATE_NO_LOCK( COLOR );
     
              /*
               * 3) Tell which functions can be called without further validation, i.e. SetState()
               *
               * When the hw independent state is changed, this collection is reset.
               */
              state->set = XC4_SUPPORTED_DRAWINGFUNCTIONS;
              mdev->drawingflags = state->drawingflags;
     
              break;
          case DFXL_BLIT:
              //
              // ...require valid drawing color. 
              // 
              XCODE_CHECK_VALID_STATE_NO_LOCK( COLOR );
              //
              // ...require valid source. 
              // 
              XCODE_CHECK_VALID_STATE( SOURCE );
              state->set = XC4_SUPPORTED_BLITTINGFUNCTIONS;
     
              mdev->blittingflags = state->blittingflags;

              /* 
              * Use if (SRC_PREMULTCOLOR only, or DSPD_SRC_OVER with blend enable,
              * or with DSPF_A8 src_format ... 
              */
              if(xcode_check_g2d_capability(state))
              {
                    DPRINTF("Using xcodeBlit...\n");
                    funcs->Blit = xcodeBlit;
              }
              else
              {
                    DPRINTF("Using xcodeGicBlit...\n");
                    funcs->Blit = xcodeGicBlit;
              }
              break;
          case DFXL_STRETCHBLIT:
              //
              // ...require valid drawing color. 
              // 
              XCODE_CHECK_VALID_STATE_NO_LOCK( COLOR );
              //
              // ...require valid source. 
              // 
              XCODE_CHECK_VALID_STATE( SOURCE );
              state->set = XC4_SUPPORTED_BLITTINGFUNCTIONS;

              mdev->blittingflags = state->blittingflags;

              if(xcode_check_g2d_capability(state))
              {
                   DPRINTF("Using xcodeStretchBlit...\n");
                   funcs->StretchBlit  = xcodeStretchBlit;  
              }
              else
              {
                   DPRINTF("Using xcodeGicStretchBlit...\n");
                   funcs->StretchBlit  = xcodeGicStretchBlit;
              }
              break;

          default:
              DPRINTF( "unexpected drawing/blitting function" );
              break;
     }


     /*
      * 4) Clear modification flags
      *
      * All flags have been evaluated in 1) and remembered for further validation.
      * If the hw independent state is not modified, this function won't get called
      * for subsequent rendering functions, unless they aren't defined by 3).
      */
     state->modified = 0;

     //make sure that every the checkstate function will be called
     state->checked = 0;

     //complete GIC stretch pipeline
     PipelineGicStretchBlit(drv, dev, false);

     DPRINTF( "xcode DFB driver xcodeSetState exit:\n" );

}

/******************************************************************************************************************/

void Wait_For_Fifo_Entries(pxc_dfb_driver_data mdrv,
                           u32 status_reg, 
                           u32 free_ent_mask, 
                           u32 free_ent_shift,
                           u32 requested_fifo_space)
{
    u32 cur_entries;
    int timeout = 1000000;

    cur_entries = (REG_READ(status_reg) & free_ent_mask) >> free_ent_shift;

    while (cur_entries < requested_fifo_space)
    {
        timeout --;
        if (timeout == 0)
            break;

        usleep(1);

        cur_entries = (REG_READ(status_reg) & free_ent_mask) >> free_ent_shift;
    }

    return;
}


/******************************************************************************************************************/

/* xcode 4 G2D blit and scale blit with rotation ( 0, 90, 180 or 270 degrees ) */
/* This function does not use scissors.                                        */

int xcode4_g2d_scale_and_rotate ( xc_driver_data *mdrv,
                                  int srcPhys, int dfbSrcFmt, int srcX, int srcY, int srcW, int srcH, int srcPitch,
                                  int dstPhys, int dfbDstFmt, int dstX, int dstY, int dstW, int dstH, int dstPitch, 
                                  int rotation )
{
    u32 g2dReg;
    u32 bltW = dstW, bltH = dstH;
    u32 dummy, fifo_entries = 9;
    bool scale = ( ( srcW != dstW ) || ( srcH != dstH ) ); 
    int g2dSrcFmt = get_pixel_format_color( dfbSrcFmt );
    int g2dDstFmt = get_pixel_format_color( dfbDstFmt );

    if( g2dSrcFmt == PIX_FORMAT_UNKNOWN || g2dDstFmt == PIX_FORMAT_UNKNOWN )
    {
        return -1;
    }
    if( !( srcPhys && dstPhys ) )
    {
        return -2;
    }

    /* if we scaling we need an extra FIFO entry for the G2D_SCALERYX register */
    if ( scale )
    {
        fifo_entries++;
    }
    WAIT_FOR_G2D_FIFO_ENTRIES( dummy, fifo_entries );

    /* set up operation */
    g2dReg = 0;
    MEM_SET_FLD( g2dReg, G2D_OP_CTL, SCISSOR_DIS, 1 ); /* scissors not needed */
    MEM_SET_FLD( g2dReg, G2D_OP_CTL, SRC_ALPHA,   1 ); /* write out src alpha */
    if ( scale )
    {
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, GFX_OP, 6 ); /* scale blit */
    }
    else
    {
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, GFX_OP, 0 ); /* blit */
    }

    if ( rotation == 90 )
    {
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, BLTDIR,       1 ); /* 90 degree rotation */
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, DEST_BLT_DIR, 2 );
    }
    else if ( rotation == 180 )
    {
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, BLTDIR,       1 ); /* 180 degree rotation */
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, DEST_BLT_DIR, 4 );
    }
    else if ( rotation == 270 )
    {
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, BLTDIR,       1 ); /* 270 degree rotation */
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, DEST_BLT_DIR, 3 );
    }
    else
    {
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, BLTDIR,       0 ); /* 0 degree rotation */
        MEM_SET_FLD( g2dReg, G2D_OP_CTL, DEST_BLT_DIR, 0 );
    }
    REG_WRITE( G2D_OP_CTL,  g2dReg );

    /* set up scaling factor register */
    if ( scale )
    {
        g2dReg = 0;
        MEM_SET_FLD( g2dReg, G2D_SCALERYX, XSCALER, ( srcW << SCALER_MULTIPLY_SHIFT ) / dstW );
        MEM_SET_FLD( g2dReg, G2D_SCALERYX, YSCALER, ( srcH << SCALER_MULTIPLY_SHIFT ) / dstH );
        REG_WRITE  ( G2D_SCALERYX, g2dReg );
    }

    /* set up src registers */
    REG_WRITE( G2D_SRC_OFFSET, srcPhys );

    g2dReg = 0;
    if ( !dfb_config->g2d_linear_stretch_blit && srcW > dstW && srcH > dstH )
    {
        MEM_SET_FLD( g2dReg, G2D_SRC_PITCH_FORMAT, SCALING_TYPE, 1 ); /* pick nearest */
    }
    MEM_SET_FLD( g2dReg, G2D_SRC_PITCH_FORMAT, SRC_PIX_FORMAT, g2dSrcFmt );
    MEM_SET_FLD( g2dReg, G2D_SRC_PITCH_FORMAT, G2D_SRC_PITCH_FORMAT_SRC_PITCH, srcPitch );
    REG_WRITE  ( G2D_SRC_PITCH_FORMAT, g2dReg );

    g2dReg = 0;
    MEM_SET_FLD( g2dReg, G2D_SRC_Y_X, SRC_X, srcX );
    MEM_SET_FLD( g2dReg, G2D_SRC_Y_X, SRC_Y, srcY );
    REG_WRITE  ( G2D_SRC_Y_X, g2dReg );

    g2dReg = 0;
    MEM_SET_FLD( g2dReg, G2D_SRC_H_W, SRC_W, srcW );
    MEM_SET_FLD( g2dReg, G2D_SRC_H_W, SRC_H, srcH );
    REG_WRITE  ( G2D_SRC_H_W, g2dReg );

    /* set up dst registers */
    REG_WRITE( G2D_DST_OFFSET, dstPhys );

    g2dReg = 0;
    if ( rotation == 90 || rotation == 270 )
    {
        bltW = dstH;
        bltH = dstW;
        MEM_SET_FLD( g2dReg, G2D_DST_PITCH_FORMAT, DST_TRANSLATE, 1 ); 
    }
    MEM_SET_FLD( g2dReg, G2D_DST_PITCH_FORMAT, DST_PIX_FORMAT, g2dDstFmt );
    MEM_SET_FLD( g2dReg, G2D_DST_PITCH_FORMAT, G2D_DST_PITCH_FORMAT_DST_PITCH, dstPitch );
    REG_WRITE  ( G2D_DST_PITCH_FORMAT, g2dReg );

    g2dReg = 0;
    MEM_SET_FLD( g2dReg, G2D_DST_Y_X, DST_X, dstX );
    MEM_SET_FLD( g2dReg, G2D_DST_Y_X, DST_Y, dstY );
    REG_WRITE  ( G2D_DST_Y_X, g2dReg );

    g2dReg = 0;
    MEM_SET_FLD( g2dReg, G2D_DST_H_W_INIT, DST_W, bltW );
    MEM_SET_FLD( g2dReg, G2D_DST_H_W_INIT, DST_H, bltH );
    REG_WRITE  ( G2D_DST_H_W_INIT, g2dReg );

    return 0;
}

/******************************************************************************************************************/

