/*  
*  Copyright (c) 2006-2010 VIXS Systems Inc. All rights reserved
*
*  FILENAME: xcode4_adv7511.c
*
*  DESCRIPTION: ADV7511 DirectFB driver extension for XCode4
*
*  AUTHOR:
*
*  HISTORY:
*
*/
#include <config.h>

#include <dfb_types.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>


#include <directfb.h>

#include <direct/messages.h>
#include <direct/util.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/screens.h>
#include <core/state.h>
#include <core/system.h>
#include <core/gfxcard.h>
#include <core/layers.h>
#include <core/surface.h>
#include <core/palette.h>

#include <display/idirectfbsurface.h>

#include <gfx/convert.h>
#include <gfx/util.h>

#include <misc/conf.h>

#include "xcode4_adv7511.h"

/* Video mode settings */
#include "xcode4_adv7511_modes.h"
#include "drv_if.h"

#include "xcodeapi.h"

xcode_handle_t g_xcode_handle;


int ADV7511_Init()
{
	
	g_xcode_handle	=	open(XCODE_DEVICE_NAME, O_RDWR);

	if (g_xcode_handle <= 0)
	{
		D_DEBUG("Failed to open Xcode AV driver @ %s, external HDMI will not be available for use!\n", XCODE_DEVICE_NAME);
		return -1;
	}

	/* Successful initialization */
	return 0;
}

int ADV7511_UpdatePacket(int packet_id, unsigned char *packet_data, unsigned int len)
{
	int r_val = 0;
	unsigned char packet_checksum = 0;
	unsigned char packet_base;
	char *packet_name;	
	unsigned char packet_accum;
	unsigned char reg_value;
	int i;
		
	if (g_xcode_handle <= 0)
	{
		/* Silently fail */
		return -1;
	}	
	
	switch (packet_id)
	{
		case ADV7511_SPD_PACKET:
			packet_base	=	ADV7511_SPD_PACKET_ADDR;
			packet_name =   "Source Product Description";
			break;
		
		case ADV7511_MPEG_PACKET:
			packet_base =	ADV7511_MPEG_PACKET_ADDR;
			packet_name =	"MPEG Source";
			break;
		
		case ADV7511_ACP_PACKET:
			packet_base =	ADV7511_ACP_PACKET_ADDR;
			packet_name =	"Audio Content Protection";
			break;
		
		case ADV7511_ISRC1_PACKET:
			packet_base =	ADV7511_ISRC1_PACKET_ADDR;
			packet_name	=	"ISRC1";
			break;

		case ADV7511_ISRC2_PACKET:
			packet_base	=	ADV7511_ISRC2_PACKET_ADDR;
			packet_name =	"ISRC2";
			break;
		
		case ADV7511_GM_PACKET:
			packet_base	=	ADV7511_GM_PACKET_ADDR;
			packet_name =	"Gamut Metadata";
			break;

		case ADV7511_VSI_PACKET:
			packet_base	=	ADV7511_VSI_PACKET_ADDR;
			packet_name	=	"Vendor Specific";
			break;
		
		case ADV7511_SPARE1_PACKET:
			packet_base	=	ADV7511_SPAREPACKET1_ADDR;
			packet_name	=	"ADV7511 Spare Packet1";
		
		default:
			/* Fail */
			D_DEBUG("Invalid HDMI packet specifier.\n");
			return -1;
			break;
				
	}
	D_DEBUG("Updating %s InfoFrame..\n", packet_name);
	/*	Compute the checksum of the packet passed in (it is composted of 3 bytes and the packet body) */	
	/* 
		The checksum is computed such that the following condition is true : 
		3 bytes of packet header + checksum + 28 bytes of body = 0
		checksum = 0 - sum(28 bytes of body) - sum(3 bytes packet header) or
		checksum = 0 - ( sum(28 bytes of body) + sum(3 bytes packet header) )
		NOTE: Using byte-wide arithmetic
	*/
	packet_accum = 0;
	D_DEBUG("Packet size (including header) is %u bytes.\n", len + 3);
	
	if (len > 28)
	{
		D_DEBUG("WARNING: Packet Length is greater than 28, clamping to 28 bytes...\n");
		len = 28;
		packet_data[2] = len;
	}
		
	for (i = 0 ; i < len + 3 ; i++)
	{
		D_DEBUG("Byte %u: 0x%X\n", i, packet_data[i]);
		if (i != 3) 
		{
			packet_accum += packet_data[i]; /* Skip checksum byte specified to us */
		}
				
	}
	D_DEBUG("Accumulator byte 0x%X\n", packet_accum);
	packet_checksum = 0 - packet_accum;
	D_DEBUG("Checksum byte 0x%X\n", packet_checksum);
	
	/* Set the packet update flag for our specific packet */
	i2c_write(ADV7511_PACKET_I2C_ADDR, ADV7511_PACKET_UPDATE, packet_base + 0x1F);
	usleep(I2C_SLEEP_US);	
	/* Update our packet */ 	
	
	for (i = 0 ; i < len + 3 ; i++)
	{
		if (i != 3)
		{
			i2c_write(ADV7511_PACKET_I2C_ADDR, packet_data[i], packet_base+i); 
			/* Regular write to the I2C packet data */
		}
		else
		{
			i2c_write(ADV7511_PACKET_I2C_ADDR, packet_checksum, packet_base+i);
			/* Write the checksum value to the register */
		}	
		usleep(I2C_SLEEP_US);	
	}
	/*
		Fill the rest of the packet with 0's
	*/		
	for ( ; i < 0x1F ; i++)
	{
		i2c_write(ADV7511_PACKET_I2C_ADDR, 0x00, packet_base+i);
		usleep(I2C_SLEEP_US);
		/* Write 0 byte to the rest of the packet so checksum is not broken */
		
	}	
	/* Clear the packet update flag for our specific packet */
	i2c_write(ADV7511_PACKET_I2C_ADDR, ADV7511_PACKET_NOUPDATE, packet_base + 0x1F);
	usleep(I2C_SLEEP_US);
	/* Done updating */	
	
	return 0;	
}


int ADV7511_RunScript(unsigned char *script)
{
    /* Execute the script for the ADV7511 */
    int i;
    unsigned char value;
    while (1) 
    {
	    if (script[i] != 0 ) 
	    {
	        i2c_write(ADV7511_I2C_ADDR, script[i+1], script[i]);
	        D_DEBUG("ADV7511 write to 0x%X: 0x%X\n", script[i], script[i+1]); 
	        usleep(I2C_SLEEP_US);
	        i2c_read(ADV7511_I2C_ADDR, &value, script[i]);
	    
	        D_DEBUG("ADV7511 value [0x%x]= 0x%X\n", script[i], value);
	        i+=2;
	    }
	    else 
	    {
	        break;
	    }
    }

    return 0;
}

int ADV7511_SetMode(struct display_mode_struct *mode)
{
    int return_value;
    /* Analyze the display_mode_struct passed to us and change the display mode */
    unsigned char value;
	
	/* Used to set the vendor specific InfoFrame for certain modes */
	unsigned char packet[6];
	
      
        
    if (g_xcode_handle < 0) 
    {
		return_value =  -1;	/* Indicate failure*/
		D_DEBUG("Cannot set mode, because the AV driver could not be opened!\n");
    }
    else
    {	
		return_value	=   0;
		i2c_read(FPGA_I2C_ADDR, &value, FPGA_I2C_MUX_REG);
    
		value |= FPGA_I2C_SYNC_TO_ADV7511;	/* Switch the sync mux FPGA input to drive the ADV7511 */        
    
		D_DEBUG("FPGA @ 0x%X reg 0x%X value: 0x%X\n", FPGA_I2C_ADDR, FPGA_I2C_MUX_REG, value);
		i2c_write(FPGA_I2C_ADDR, value, FPGA_I2C_MUX_REG);

	
	
		switch (mode->resolution)
		{
			case DSOR_720_480:
				/* 480i or 480p60/59.94/50 */
				ADV7511_DisablePacket(ADV7511_VSI_PACKET);
				if (mode->scanmode == DSESM_PROGRESSIVE)
				{
				/* Progressive  */
					ADV7511_RunScript((unsigned char *)&porta_480p60);
				}
				else
				{
					/* Interlaced. currently not working */
					ADV7511_RunScript((unsigned char *)&porta_480i60);
				}
				break;
	
			case DSOR_720_576:
				/* 576i or 576p, this mode is not supported as we have no way of testing this */
				ADV7511_DisablePacket(ADV7511_VSI_PACKET);
				return_value = -1;
				break;
	
			case DSOR_1280_720:
				/* 720p60/59.94/50 (not sure if 50 is supported) */
				ADV7511_DisablePacket(ADV7511_VSI_PACKET);
				ADV7511_RunScript((unsigned char *)&porta_720p60);
				break;
	    
			case DSOR_1920_1080:
				/* 1080i or 1080p */
				ADV7511_DisablePacket(ADV7511_VSI_PACKET);
				if (mode->scanmode == DSESM_PROGRESSIVE)
				{
					ADV7511_RunScript((unsigned char *)&porta_1080p60);
				}
				else
				{
					ADV7511_RunScript((unsigned char *)&porta_1080i60);
				}
				break;


			/* SPECIAL CASES (3D, etc...) */
			case DSOR_1280_1470:
			case DSOR_1920_1080_3D_SIDE_BY_SIDE:
			case DSOR_1920_2205:
			
				packet[0] = HDMI_VSD_PACKET_ID;
				packet[1] = HDMI_VSD_VERSION_ID;
				packet[2] = 0x06; /* Length */
				
				packet[3] = 0x00; /* Subroutine will calculate checksum */	
				packet[4] = 0x03;
				packet[5] = 0x0C;
				packet[6] = 0x00;
				packet[7] = STRUCTURE_3D;
				if (mode->resolution == DSOR_1280_1470)
				{
					packet[8] = (FRAME_PACKING_3D << MODE_SHIFT_3D);
				
					ADV7511_RunScript((unsigned char *)&porta_720p60fp);
					/* Program packet */
				}
				else if (mode->resolution == DSOR_1920_1080_3D_SIDE_BY_SIDE)
				{
					packet[8] = (SIDE_BY_SIDE_3D << MODE_SHIFT_3D);
				
					ADV7511_RunScript((unsigned char *)&porta_1080p60);
				}
				else if (mode->resolution == DSOR_1920_2205)
				{
					packet[8] = (FRAME_PACKING_3D << MODE_SHIFT_3D);
				
					ADV7511_RunScript((unsigned char *)&porta_1080p24fp);
				}		
			
				/* Update the VSI packet and enable sending it over HDMI */
				ADV7511_UpdatePacket(ADV7511_VSI_PACKET, (unsigned char *)&packet, 6);
				ADV7511_EnablePacket(ADV7511_VSI_PACKET); 

			default:
				/* Assume a bad mode and return a failure  */
				return_value = -1;
				break;
		}      
    }
    return return_value;
}
int ADV7511_EnablePacket(int packet_id)
{
	int return_value = 0;
	unsigned char packet_mask = 0;
	unsigned char packet_ctrl;	
	if (g_xcode_handle < 0)
    {
		return_value =  -1; 
		/* Silently fail */
	}
	else
	{
		D_DEBUG("Enabling HDMI packet.\n");
		switch (packet_id)
		{
			case ADV7511_SPD_PACKET:
				packet_mask = ADV7511_SPD_PACKET_MASK;
				break;
			
			case ADV7511_MPEG_PACKET:
				packet_mask = ADV7511_MPEG_PACKET_MASK;
				break;
			
			case ADV7511_ACP_PACKET:
				packet_mask = ADV7511_ACP_PACKET_MASK;
				break;
			
			case ADV7511_ISRC1_PACKET:
			case ADV7511_ISRC2_PACKET:
				/* Both packets will be sent if ISRC packet is enabled */
				packet_mask	= ADV7511_ISRC_PACKET_MASK;
				break;
			
			case ADV7511_GM_PACKET:
				packet_mask	= ADV7511_GM_PACKET_MASK;
				break;
			
			case ADV7511_SPARE1_PACKET:
				packet_mask = ADV7511_SPARE1_PACKET_MASK;
				break;
			
			case ADV7511_VSI_PACKET:
				packet_mask = ADV7511_VSI_PACKET_MASK;
				break;
			
			default:
				packet_mask = 0x00;
				D_DEBUG("Invalid packet indentifier.\n");
				break;
			

		}
		
		/* Read the current register value */
		i2c_read(ADV7511_I2C_ADDR, &packet_ctrl, ADV7511_PACKET_CTRL_REG);
		packet_ctrl |= packet_mask;	/* Enable the specific packet */
		
		i2c_write(ADV7511_I2C_ADDR, packet_ctrl, ADV7511_PACKET_CTRL_REG);
		usleep(I2C_SLEEP_US); 
		return_value = 0;		
	}
	return return_value;
}

int ADV7511_DisablePacket(int packet_id)
{
	int return_value = 0;
	unsigned char packet_mask = 0;
	unsigned char packet_ctrl = 0;
	if (g_xcode_handle < 0)
	{
		return_value = -1;
		/* Silently fail */
	}
	else
	{
		D_DEBUG("Disabling HDMI packet.\n");
		switch (packet_id)
		{
			case ADV7511_SPD_PACKET:
				packet_mask = ADV7511_SPD_PACKET_MASK;
				break;
			
			case ADV7511_MPEG_PACKET:
				packet_mask = ADV7511_MPEG_PACKET_MASK;
				break;
			
			case ADV7511_ACP_PACKET:
				packet_mask = ADV7511_ACP_PACKET_MASK;
				break;
			
			case ADV7511_ISRC1_PACKET:
			case ADV7511_ISRC2_PACKET:
				/* Both packets will be sent if ISRC packet is enabled */
				packet_mask	= ADV7511_ISRC_PACKET_MASK;
				break;
			
			case ADV7511_GM_PACKET:
				packet_mask	= ADV7511_GM_PACKET_MASK;
				break;
			
			case ADV7511_SPARE1_PACKET:
				packet_mask = ADV7511_SPARE1_PACKET_MASK;
				break;
			
			case ADV7511_VSI_PACKET:
				packet_mask = ADV7511_VSI_PACKET_MASK;
				break;
			
			default:
				packet_mask = 0x00; /* To be non-destructive */
				D_DEBUG("Invalid packet indentifier.\n");
				break;
			

		}
		
		/* Read the current register value */
		i2c_read(ADV7511_I2C_ADDR, &packet_ctrl, ADV7511_PACKET_CTRL_REG);
		packet_ctrl &= (~packet_mask);	/* Disable the specific packet */
		
		i2c_write(ADV7511_I2C_ADDR, packet_ctrl, ADV7511_PACKET_CTRL_REG);
		usleep(I2C_SLEEP_US); 
		return_value = 0;		

	
	}	

	return return_value;
}


static int i2c_read(unsigned char dev_addr, unsigned char *byte, unsigned int addr)
{
    unsigned char wvalue[80], rvalue[80];
    
    wvalue[0]	=   addr;
    
    xc_i2c_write_then_read(XC4_I2C_CHANNEL, dev_addr, wvalue, rvalue, 1, 1);
    *byte   =	rvalue[0];
    
    return 0;
}
static int i2c_write(unsigned char dev_addr, unsigned char byte, unsigned int addr)
{
    unsigned char wvalue[80];
    wvalue[0]	=   addr;
    wvalue[1]	=   byte;
    
    xc_i2c_write(XC4_I2C_CHANNEL, dev_addr, wvalue, 2, 1, 1);
    
    return 0;
}



static int xc_i2c_read(int I2CChannel, int Address, void *data, size_t n, int start, int stop)
{
    int ret;

    if (ret = xcode_i2c_read(g_xcode_handle, I2CChannel, Address, data, n, MASTER_I2C_ATOMIC_PERIOD, MASTER_I2C_TIMEOUT, start, stop))
    {
	D_DEBUG("I2C Read failed @ Address: 0x%X\n", Address);
    }
    else
    {
    }
    return ret;
}

static int xc_i2c_write(int I2CChannel, int Address, const void *data, size_t n, int start,int stop)
{
    int ret;
    
    if( g_xcode_handle == -1)
        return 1;

    if (ret = xcode_i2c_write(g_xcode_handle, I2CChannel, Address, data, n, MASTER_I2C_ATOMIC_PERIOD, MASTER_I2C_TIMEOUT, start, stop))
    {
	D_DEBUG("I2C write failed @ address: 0x%X\n", Address);
    }
    else
    {
    }
    return ret;
}

static int xc_i2c_write_then_read(int I2CChannel, int Address, const void *wdata, void *rdata, size_t wn, size_t rn)
{
    int error;
    int start = 1;
    int stop = 1;

    if (error = xc_i2c_write(I2CChannel, Address, wdata, wn, start, 0))
    {
        return error;
    }
    return xc_i2c_read(I2CChannel, Address, rdata, rn, 1, stop);
}


static int xcode_i2c_read(xcode_handle_t XCodeHandle, int I2CChannel, int Address, void *data, size_t n, int atomic_period, int timeout, int start, int stop)
{
    IOCTLI2CCOMMAND I2CCommand;

    memset(&I2CCommand, 0, sizeof(IOCTLI2CCOMMAND));
    I2CCommand.ioctl_header.command = CMD_I2C_COMMAND_STREAM;
    I2CCommand.ioctl_header.size_of_buf = sizeof(IOCTLI2CCOMMAND);
    I2CCommand.num_bytes = n;
    I2CCommand.rop = I2CREAD;
    I2CCommand.slave_address = Address;
    I2CCommand.atomic_period = atomic_period;
    I2CCommand.address_mode = I2CCONFIGADDRESSMODE7BIT;
    I2CCommand.timeout = timeout;
    I2CCommand.i2c_block_number = I2CChannel;
    I2CCommand.startbit = start;
    I2CCommand.stopbit = stop;
	I2CCommand.ioctl_header.return_value = IOCTL_RET_CMD_FAILED;
	I2CCommand.ioctl_header.time_out_value = IOCTL_TIMEOUTVALUE;

    ioctl(XCodeHandle, IOCTL_CMD, &I2CCommand);

    if (I2CCommand.ioctl_header.return_value == IOCTL_RET_CMD_COMPLETE)
    {
        memcpy(data, I2CCommand.data_array, n);
        return 0;
    }
    else
    {
        //xcode_print_error_code(I2CCommand.ioctl_header.return_value, "I2C_READ");
        return I2CCommand.ioctl_header.return_value;
    }
}


static int xcode_i2c_write(xcode_handle_t XCodeHandle, int I2CChannel, int Address, const void *data, size_t n, int atomic_period, int timeout, int start, int stop)
{
    IOCTLI2CCOMMAND I2CCommand;

    memset(&I2CCommand, 0, sizeof(IOCTLI2CCOMMAND));
    memcpy(I2CCommand.data_array, data, n);
    I2CCommand.ioctl_header.command = CMD_I2C_COMMAND_STREAM;
    I2CCommand.ioctl_header.size_of_buf = sizeof(IOCTLI2CCOMMAND);
    I2CCommand.num_bytes = n;
    I2CCommand.rop = I2CWRITE;
    I2CCommand.slave_address = (unsigned int)Address;
    I2CCommand.atomic_period = atomic_period;
    I2CCommand.address_mode = I2CCONFIGADDRESSMODE7BIT;
    I2CCommand.timeout = timeout;
    I2CCommand.i2c_block_number = (unsigned int)I2CChannel;
    I2CCommand.startbit = start;
    I2CCommand.stopbit = stop;
	I2CCommand.ioctl_header.return_value = IOCTL_RET_CMD_FAILED;
	I2CCommand.ioctl_header.time_out_value = IOCTL_TIMEOUTVALUE;

    ioctl(XCodeHandle, IOCTL_CMD, &I2CCommand);

    if (I2CCommand.ioctl_header.return_value == IOCTL_RET_CMD_COMPLETE) 
    {
        return 0;
    }
    else
    {
        //xcode_print_error_code(I2CCommand.ioctl_header.return_value, "I2C_WRITE");
        return I2CCommand.ioctl_header.return_value;
    }
}





