#include <config.h>
#include <dfb_types.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>


#include <directfb.h>
#include <direct/messages.h>
#include <direct/util.h>

#include <core/coredefs.h>
#include <core/coretypes.h>
#include <core/screens.h>
#include <core/state.h>
#include <core/system.h>
#include <core/gfxcard.h>
#include <core/layers.h>
#include <core/layers_internal.h>
#include <core/surface.h>
#include <core/palette.h>

#include <gfx/convert.h>
#include <gfx/util.h>
#include <gfx/clip.h>

#include <misc/conf.h>

#include "viper.h"
#include "xcmmio.h"
#include "xc4fb_if.h"
#include "xcode4_2d.h"
#include "xcode4_osd.h"
#include "xcioctl.h"
#include "xcdfb.h"


#define max(a,b) ((a) > (b) ? a : b)
#define min(a,b) ((a) < (b) ? a : b)

extern DFBResult XcodeFlipSDFrame(CoreLayer *layer, 
                                         const DFBRegion *update,
                                         CoreSurfaceBufferLock *lock);

D_DEBUG_DOMAIN( Merge, "osd_merge", "xcode4_osd_merge" );

#define DPRINTF(...) D_DEBUG_AT( Merge, __VA_ARGS__)

#define ERAISE(x) {int err = x; if(err) return err;}

//////////////////////////////////////////////////////////////////

typedef enum
{
    AP_PLANE = TOP_GRA1_LAYER_ID,
    IG_PLANE = TOP_GRA0_LAYER_ID,
    BG_PLANE = BACKGROUND_GRA_LAYER_ID,
    G0_PLANE = XCODE4_GRA0_LAYER_ID,
    V0_PLANE = XCODE4_VID0_LAYER_ID,
    V1_PLANE = XCODE4_VID1_LAYER_ID,
}xc4_plane_id_t;

#define FORCE_3D_SIDE_BY_SIDE   1
#define FORCE_3D_TOP_AND_BOT  2

typedef struct
{
    int id;
    int address;
    int format;
    int x;
    int y;
    int width;
    int height;
    int pitch;
    int is_3d;
    int force_3d;
    int stereo_depth;
    int alpha;
    int dest_x;
    int dest_y;
    int dest_w;
    int dest_h;
    int premultiplied;
}xc4_plane_info_t;

typedef struct
{
    int id;
    int src_x;
    int src_y;
    int src_w;
    int src_h;
    int depth;
}xc4_stereo_info_t;

//
// get temp buffer
//
static int xcode4_get_temp_addr(pxc_dfb_driver_data mdrv)
{
    return mdrv->pdevice_data->temp_render_physical;
}

static int get_pixel_format_color(DFBSurfacePixelFormat dsf_color)
{
    int pixel_format_color;

    switch (dsf_color)
    {
        case DSPF_RGB32:
        case DSPF_ARGB:
            pixel_format_color = PIX_FORMAT_RGB32;
            break;

        case DSPF_RGB16:
            pixel_format_color = PIX_FORMAT_RGB565;
            break;

        case DSPF_ARGB1555:
            pixel_format_color = PIX_FORMAT_RGB1555;
            break;

        case DSPF_ARGB4444:
            pixel_format_color = PIX_FORMAT_RGB4444;
            break;

        case DSPF_LUT8:
        case DSPF_A8:
            pixel_format_color = PIX_FORMAT_LUT8;
            break;

        case DSPF_ALUT44:
            pixel_format_color = PIX_FORMAT_LUT4;
            break;

        case DSPF_LUT2:
            pixel_format_color = PIX_FORMAT_LUT2;
            break;

        case DSPF_A1:
            pixel_format_color = PIX_FORMAT_MONOCHROME;
            break;

        default:
            pixel_format_color = PIX_FORMAT_UNKNOWN;
            break;
    }

    return pixel_format_color;
}


static u32 xcodeGicGetFormatCode(DFBSurfacePixelFormat dpf)
{
    u32 gic_code;
    
    switch (dpf)
    {
        case DSPF_LUT8:
        case DSPF_A8:
            gic_code    =    GIC_LUT8;
            break;
            
        case DSPF_ARGB:
            gic_code    =    GIC_AYUV444;
            break;
            
        case DSPF_RGB16:
            gic_code    =    GIC_RGB565;
            break;
            
        default:
            gic_code     =    GIC_AYUV444;
            break;
    }
    return gic_code;
}


//
// xcode 4 G2D scale
//
static int xcode4_g2d_scale_depth(   
                    pxc_dfb_driver_data mdrv,
                    int src_addr,
                    int src_fmt,
                    int src_x,
                    int src_y,
                    int src_w,
                    int src_h,
                    int src_pitch,
                    int dst_addr,
                    int dst_fmt,
                    int dst_x,
                    int dst_y,
                    int dst_w,
                    int dst_h,
                    int dst_pitch,
                    int depth
                    )
{
    u32 fifo_entry;
    u32 scaler_x, scaler_y;
    u32 cmd;

    DPRINTF("%s\n", __FUNCTION__);

    DPRINTF("src_addr = %x, src_fmt = %x, (%d, %d, %d, %d), src_pitch = %d\n",
        src_addr, src_fmt, src_x, src_y, src_w, src_h, src_pitch);
    
    DPRINTF("dst_addr = %x, dst_fmt = %x, (%d, %d, %d, %d), dst_pitch = %d\n",
        dst_addr, dst_fmt, dst_x + depth, dst_y, dst_w, dst_h, dst_pitch);

    int sf = get_pixel_format_color(src_fmt);
    
    int df = get_pixel_format_color(dst_fmt);

    if(sf == PIX_FORMAT_UNKNOWN || df == PIX_FORMAT_UNKNOWN)
        return -1;
    
    if(!(src_addr && dst_addr))
        return -2;

    //Ensure there is enough space in the FIFO
    WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 4);
    
    //
    //top left of the region
    // 
    REG_WRITE(G2D_SC_TOP_LEFT, dst_x | (dst_y << SC_TOP_SHIFT));

    //
    //bottom right of the region
    //
    REG_WRITE(G2D_SC_BOTTOM_RIGHT, 
             (((dst_y + dst_h - 1) << SC_BOTTOM_SHIFT) | (dst_x + dst_w - 1)));

    //
    //DST offset
    // 
    REG_WRITE(G2D_DST_OFFSET, dst_addr);

    //
    //DST pitch/bpp
    // 
    REG_WRITE(G2D_DST_PITCH_FORMAT, 
              (dst_pitch | ( df << DST_PIX_FORMAT_SHIFT)));

    //Ensure there is enough space in the FIFO
    WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 4);
    
    //
    //DST x,y
    // 
    REG_WRITE(G2D_DST_Y_X, ((dst_y << DST_Y_SHIFT) | (dst_x + depth)));

    //
    //SRC offset
    // 
    REG_WRITE(G2D_SRC_OFFSET, src_addr);

    //
    //SRC pitch/bpp
    // 
    if (!dfb_config->g2d_linear_stretch_blit && src_w > dst_h && src_h > dst_h)
    {
        REG_WRITE(G2D_SRC_PITCH_FORMAT, 
            (src_pitch | ( sf << SRC_PIX_FORMAT_SHIFT) | (1 << SCALING_TYPE_SHIFT)));
     }
     else
     {
        REG_WRITE(G2D_SRC_PITCH_FORMAT, 
            (src_pitch | ( sf << DST_PIX_FORMAT_SHIFT)));
     }

    //
    //SRC x,y
    // 
    REG_WRITE(G2D_SRC_Y_X, ((src_y << SRC_Y_SHIFT) | (src_x)));

    //Ensure there is enough space in the FIFO
    WAIT_FOR_G2D_FIFO_ENTRIES(fifo_entry, 5);

    scaler_x = (src_w << SCALER_MULTIPLY_SHIFT) / dst_w;
    scaler_y = (src_h << SCALER_MULTIPLY_SHIFT) / dst_h;
    
    //
    //SRC (width/height)
    // 
    REG_WRITE(G2D_SRC_H_W,
             ((src_w) | (src_h << SRC_H_SHIFT)));

    //G2D SCALER FRACYX
    REG_WRITE(G2D_SCALERYX,(scaler_x | (scaler_y << YSCALER_SHIFT)));

    cmd = (SRC_ALPHA_WRITE_OUT << SRC_ALPHA_SHIFT) | (XC_G2D_OP_COPY_RECT << GFX_OP_SHIFT)  | (1 <<SCISSOR_DIS_SHIFT );

    //G2D OP
    if(src_w != dst_w || src_h != dst_h)
    {
        cmd |= (XC_G2D_OP_SCALE_RECT << GFX_OP_SHIFT);
    }

    //Depth
    if(depth)
    {
        cmd &= ~(1 <<SCISSOR_DIS_SHIFT );
    }

    REG_WRITE(G2D_OP_CTL, cmd);

    REG_WRITE(G2D_GUI_CONTROL,0);

    //Initiator(width/height)    
    REG_WRITE(G2D_DST_H_W_INIT,((dst_w) | (dst_h << DST_H_SHIFT)));

    return 0;
}

//
// xcode 4 GIC copy
//
static int xcode4_gic_src1_over_src2(
                    pxc_dfb_driver_data mdrv,
                    int src0_addr,
                    int src0_fmt,
                    int src0_x,
                    int src0_y,
                    int src0_w,
                    int src0_h,
                    int src0_pitch,
                    int src1_addr,
                    int src1_fmt,
                    int src1_x,
                    int src1_y,
                    int src1_w,
                    int src1_h,
                    int src1_pitch,
                    int dst_addr,
                    int dst_fmt,
                    int dst_w,
                    int dst_h,
                    int dst_pitch,
                    int premultiplied
                    )
{
    gic_opt_struct current_gic_ctl;
    u32 fifo_entries;
    int offset;
    int top = 0, left = 0, bottom = dst_h - 1, right = dst_w - 1;
    int need_clip = 0;

    DPRINTF("%s\n", __FUNCTION__);

    DPRINTF("src0_addr = %x, src0_fmt = %x, (%d, %d, %d, %d), src0_pitch = %d\n",
        src0_addr, src0_fmt, src0_x, src0_y, src0_w, src0_h, src0_pitch);

    DPRINTF("src1_addr = %x, src1_fmt = %x, (%d, %d, %d, %d), src1_pitch = %d\n",
        src1_addr, src1_fmt, src1_x, src1_y, src1_w, src1_h, src1_pitch);

    DPRINTF("dst_addr = %x, dst_fmt = %x, (%d, %d, %d, %d), dst_pitch = %d\n",
        dst_addr, dst_fmt, 0, 0, dst_w, dst_h, dst_pitch);

    if(dst_fmt != DSPF_ARGB)
        return -1;

    if(!(src0_addr && src1_addr && dst_addr))
        return -2;

    if(!dst_w || !dst_h)
        return -3;

    memset(&current_gic_ctl, 0, sizeof(current_gic_ctl));

    //DST must be 32 bytes alignment
    if(dst_addr & 0x1F)
    {
        offset = (dst_addr & 0x1F) / 4;
        dst_addr &= ~0x1F;
        dst_w += offset;
        src0_x += offset;
        src1_x += offset;
        left += offset;
        right += offset;
        need_clip = 1;
    }

    if(src0_x < 0 || src1_x < 0)
    {
        offset = -MIN(src0_x, src1_x);
        offset = (offset + 7) >> 3 << 3;
        dst_w += offset;
        dst_addr -= offset * 4;
        src0_x += offset;
        src1_x += offset;
        left += offset;
        right += offset;
        need_clip = 1;
    }

    if(src0_y < 0 || src1_y < 0)
    {
        offset = -MIN(src0_y, src1_y);
        dst_h += offset;
        dst_addr -= offset * dst_pitch;
        src0_y += offset;
        src1_y += offset;
        top += offset;
        bottom += offset;
        need_clip = 1;
    }

    if(need_clip)
    {
        DPRINTF("Adjust to:\n");
        
        DPRINTF("src0_addr = %x, src0_fmt = %x, (%d, %d, %d, %d), src0_pitch = %d\n",
            src0_addr, src0_fmt, src0_x, src0_y, src0_w, src0_h, src0_pitch);

        DPRINTF("src1_addr = %x, src1_fmt = %x, (%d, %d, %d, %d), src1_pitch = %d\n",
            src1_addr, src1_fmt, src1_x, src1_y, src1_w, src1_h, src1_pitch);

        DPRINTF("dst_addr = %x, dst_fmt = %x, (%d, %d, %d, %d), dst_pitch = %d\n",
            dst_addr, dst_fmt, 0, 0, dst_w, dst_h, dst_pitch);
    }

    current_gic_ctl.gic_src1_adr0 = src0_addr;
    current_gic_ctl.gic_src1_h_w = src0_w | (src0_h << GIC0_SRC1_H_SHIFT);
    current_gic_ctl.gic_src1_pitch = src0_pitch;
    current_gic_ctl.gic_src1_y_x = src0_x | (src0_y << GIC0_SRC1_Y_SHIFT);

    current_gic_ctl.gic_src1_mode = xcodeGicGetFormatCode(src0_fmt) | 
                             (ALPHA_24_TO_31 << GIC0_SRC1_A_PACK_ORDER_SHIFT) |
                             (RED_16_TO_23 << GIC0_SRC1_YR_PACK_ORDER_SHIFT) |
                             (GREEN_8_TO_15 << GIC0_SRC1_UG_PACK_ORDER_SHIFT) |
                             (BLUE_0_TO_7 << GIC0_SRC1_VB_PACK_ORDER_SHIFT);

    if(current_gic_ctl.gic_src1_h_w)
    {
        current_gic_ctl.gic_alpha_mode_src1 = premultiplied ? (GIC_ALPHA_MODE_PERPIXEL | GIC0_SRC2_ALPHA_PREMUL_MASK) : GIC_ALPHA_MODE_PERPIXEL;
    }
    else
    {
        current_gic_ctl.gic_src1_h_w = 0x10001;
        current_gic_ctl.gic_alpha_mode_src1 = GIC_ALPHA_MODE_CONSTANT | GIC0_SRC2_ALPHA_PREMUL_MASK;
    }

    current_gic_ctl.gic_src2_adr0 = src1_addr;
    current_gic_ctl.gic_src2_h_w = src1_w | (src1_h << GIC0_SRC2_H_SHIFT);
    current_gic_ctl.gic_src2_pitch = src1_pitch;
    current_gic_ctl.gic_src2_y_x = src1_x | (src1_y << GIC0_SRC2_Y_SHIFT);
    
    current_gic_ctl.gic_src2_mode = xcodeGicGetFormatCode(src1_fmt) | 
                             (ALPHA_24_TO_31 << GIC0_SRC2_A_PACK_ORDER_SHIFT) |
                             (RED_16_TO_23 << GIC0_SRC2_YR_PACK_ORDER_SHIFT) |
                             (GREEN_8_TO_15 << GIC0_SRC2_UG_PACK_ORDER_SHIFT) |
                             (BLUE_0_TO_7 << GIC0_SRC2_VB_PACK_ORDER_SHIFT);

    if(current_gic_ctl.gic_src2_h_w)
    {
        current_gic_ctl.gic_alpha_mode_src2 = GIC_ALPHA_MODE_PERPIXEL | GIC0_SRC2_ALPHA_PREMUL_MASK;
    }
    else
    {
        current_gic_ctl.gic_src2_h_w = 0x10001;
        current_gic_ctl.gic_alpha_mode_src2 = GIC_ALPHA_MODE_CONSTANT | GIC0_SRC2_ALPHA_PREMUL_MASK;
    }

    current_gic_ctl.gic_clip_top_left = (left) | ((top << GIC0_CLIP_TOP_SHIFT) & GIC0_CLIP_TOP_MASK);
    current_gic_ctl.gic_clip_bot_right = (right) |((bottom << GIC0_CLIP_BOT_SHIFT) & GIC0_CLIP_BOT_MASK);

    current_gic_ctl.gic_dst_adr0 = dst_addr;
    current_gic_ctl.gic_dst_h_w = dst_w | (dst_h << GIC0_DST_H_SHIFT);
    current_gic_ctl.gic_dst_pitch = dst_pitch;
    current_gic_ctl.gic_dst_y_x = 0;
    
    current_gic_ctl.gic_dst_mode = GIC_ARGB8888;
    
    current_gic_ctl.gic_trigger_command = GIC0_OP_KICKOFF_MASK | 
                                             GIC0_BLEND_EN_MASK;

    if(need_clip)
    {
        current_gic_ctl.gic_trigger_command |= GIC0_DST_CLIP_EN_MASK;
    }

    gic_opt_struct * gic_ctl_ptr = &current_gic_ctl;

    //make sure GIC0 command fifo is enough
    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 8);
    
    GIC0_REG_WRITE(GIC0_SRC1_ADR0, gic_ctl_ptr->gic_src1_adr0);
    GIC0_REG_WRITE(GIC0_SRC1_ADR1, gic_ctl_ptr->gic_src1_adr1);
    GIC0_REG_WRITE(GIC0_SRC1_ADR2, gic_ctl_ptr->gic_src1_adr2);
    GIC0_REG_WRITE(GIC0_SRC1_H_W, gic_ctl_ptr->gic_src1_h_w);  
    GIC0_REG_WRITE(GIC0_SRC1_PITCH, gic_ctl_ptr->gic_src1_pitch);
    GIC0_REG_WRITE(GIC0_SRC1_Y_X, gic_ctl_ptr->gic_src1_y_x);
    GIC0_REG_WRITE(GIC0_SRC1_MODE, gic_ctl_ptr->gic_src1_mode); 
    GIC0_REG_WRITE(GIC0_ALPHA_MODE_SRC1, gic_ctl_ptr->gic_alpha_mode_src1);

    //make sure GIC0 command fifo is enough
    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 8);

    //configure src2 as graphic background, src2 will be overwritten by src1
    GIC0_REG_WRITE(GIC0_SRC2_ADR0, gic_ctl_ptr->gic_src2_adr0);
    GIC0_REG_WRITE(GIC0_SRC2_ADR1, gic_ctl_ptr->gic_src2_adr1);
    GIC0_REG_WRITE(GIC0_SRC2_ADR2, gic_ctl_ptr->gic_src2_adr2);
    GIC0_REG_WRITE(GIC0_SRC2_H_W, gic_ctl_ptr->gic_src2_h_w);
    GIC0_REG_WRITE(GIC0_SRC2_Y_X, gic_ctl_ptr->gic_src2_y_x);
    GIC0_REG_WRITE(GIC0_SRC2_PITCH, gic_ctl_ptr->gic_src2_pitch);
    GIC0_REG_WRITE(GIC0_SRC2_MODE, gic_ctl_ptr->gic_src2_mode); 
    GIC0_REG_WRITE(GIC0_ALPHA_MODE_SRC2, gic_ctl_ptr->gic_alpha_mode_src2);

    //make sure GIC0 command fifo is enough
    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 7);

    GIC0_REG_WRITE(GIC0_DST_ADR0, gic_ctl_ptr->gic_dst_adr0);
    GIC0_REG_WRITE(GIC0_DST_ADR1, gic_ctl_ptr->gic_dst_adr1);
    GIC0_REG_WRITE(GIC0_DST_ADR2, gic_ctl_ptr->gic_dst_adr2);
    GIC0_REG_WRITE(GIC0_DST_H_W, gic_ctl_ptr->gic_dst_h_w);
    GIC0_REG_WRITE(GIC0_DST_PITCH, gic_ctl_ptr->gic_dst_pitch);
    GIC0_REG_WRITE(GIC0_DST_Y_X, gic_ctl_ptr->gic_dst_y_x);
    GIC0_REG_WRITE(GIC0_DST_MODE, gic_ctl_ptr->gic_dst_mode);

    //make sure GIC0 command fifo is enough
    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 2);
    
    GIC0_REG_WRITE( GIC0_CLIP_TOP_LEFT, gic_ctl_ptr->gic_clip_top_left);
    GIC0_REG_WRITE( GIC0_CLIP_BOT_RIGHT, gic_ctl_ptr->gic_clip_bot_right);

    //make sure GIC0 command fifo is enough
    WAIT_FOR_GIC0_FIFO_ENTRIES(fifo_entries, 1);

    //start to kick off gic0 to do the combine
    GIC0_REG_WRITE(GIC0_CONFIG_CTRL, gic_ctl_ptr->gic_trigger_command);

    return 0;
}

static int xcode4_gic_clear(
                    pxc_dfb_driver_data mdrv,
                    int dst_addr,
                    int dst_fmt,
                    int dst_x,
                    int dst_y,
                    int dst_w,
                    int dst_h,
                    int dst_pitch
                    )
{
    return xcode4_gic_src1_over_src2(mdrv,
        xcode4_get_temp_addr(mdrv),
        dst_fmt,
        0,
        0,
        0,
        0,
        dst_pitch,
        xcode4_get_temp_addr(mdrv),
        dst_fmt,
        0,
        0,
        0,
        0,
        dst_pitch,
        dst_addr + dst_y * dst_pitch + dst_x * DFB_BYTES_PER_PIXEL(dst_fmt),
        dst_fmt,
        dst_w,
        dst_h,
        dst_pitch,
        1
        );
}

static int xcode4_gic_copy(   
                    pxc_dfb_driver_data mdrv,
                    int src_addr,
                    int src_fmt,
                    int src_x,
                    int src_y,
                    int src_w,
                    int src_h,
                    int src_pitch,
                    int dst_addr,
                    int dst_fmt,
                    int dst_x,
                    int dst_y,
                    int dst_w,
                    int dst_h,
                    int dst_pitch,
                    int depth,
                    int premultiplied
                    )
{
    return xcode4_gic_src1_over_src2(mdrv,
        src_addr + src_x * DFB_BYTES_PER_PIXEL(src_fmt) + src_y * src_pitch,
        src_fmt,
        depth,
        0,
        src_w,
        src_h,
        src_pitch,
        src_addr,
        src_fmt,
        0,
        0,
        0,
        0,
        src_pitch,
        dst_addr + dst_x * DFB_BYTES_PER_PIXEL(dst_fmt) + dst_y * dst_pitch,
        dst_fmt,
        dst_w,
        dst_h,
        dst_pitch,
        premultiplied
        );
}

//
// scale virtual layer coordinates to match physical source size of source viewport used for GRA0
//
void scale_virtual_layer_coords( pxc_dfb_layer_data g0_layer_data, pxc_dfb_layer_data virtual_layer_data, DFBRectangle *outRect )
{
    if(g0_layer_data->dest_width && g0_layer_data->dest_height)
    {
        outRect->w = virtual_layer_data->dest_width * g0_layer_data->source_width / g0_layer_data->dest_width;
        outRect->h = virtual_layer_data->dest_height * g0_layer_data->source_height / g0_layer_data->dest_height;
        outRect->x = virtual_layer_data->x * g0_layer_data->source_width / g0_layer_data->dest_width;
        outRect->y = virtual_layer_data->y * g0_layer_data->source_height / g0_layer_data->dest_height;

        if( outRect->w > g0_layer_data->source_width || outRect->h > g0_layer_data->source_height )
        {
            outRect->x = 0;
            outRect->y = 0;
            outRect->w = g0_layer_data->source_width;
            outRect->h = g0_layer_data->source_height;
        }
    }
    else
    {
        outRect->x = 0;
        outRect->y = 0;
        outRect->w = g0_layer_data->source_width;
        outRect->h = g0_layer_data->source_height;
    }

    DPRINTF( "virtual layer scaling x=%d, y=%d, w=%d, h=%d to x=%d, y=%d, w=%d, h=%d\n",
             virtual_layer_data->x, virtual_layer_data->y, virtual_layer_data->dest_width, virtual_layer_data->dest_height,
             outRect->x, outRect->y, outRect->w, outRect->h );
}

//
// get plane info
//
static xc4_plane_info_t xcode4_get_plane_info(pxc_dfb_driver_data mdrv, int id)
{
    xc4_plane_info_t info;
    info.id = id;
    CoreLayer * layer = dfb_layer_at (id);
    pxc_dfb_layer_data m_layer_data = (pxc_dfb_layer_data) layer->layer_data;
    CoreLayer *g0_layer = dfb_layer_at( G0_PLANE );
    pxc_dfb_layer_data g0_layer_data = (pxc_dfb_layer_data)g0_layer->layer_data;
    DFBRectangle scale_rect;

    info.address = m_layer_data->y_start_addr;
    info.format = m_layer_data->format;
    info.width = m_layer_data->source_width;
    info.height = m_layer_data->source_height;
    info.pitch = m_layer_data->y_pitch;

    info.is_3d = (info.width < info.height);

    if((m_layer_data->z_depth & 0xFF000000) == 0x80000000)
    {
        info.stereo_depth = (short)(m_layer_data->z_depth & 0xFFFF);
        info.force_3d = (m_layer_data->z_depth >> 16) & 0xFF;
    }
    else
    {
        info.stereo_depth = m_layer_data->z_depth;
        info.force_3d = 0;
    }
    
    info.alpha = m_layer_data->alpha;
    info.x = m_layer_data->source_x;
    info.y = m_layer_data->source_y;
    info.dest_x = m_layer_data->x;
    info.dest_y = m_layer_data->y;
    info.dest_w = m_layer_data->dest_width;
    info.dest_h = m_layer_data->dest_height;
    info.premultiplied = (m_layer_data->surface_caps & DSCAPS_PREMULTIPLIED) != 0;

    switch(id)
    {
        case AP_PLANE:
            info.address = mdrv->shared->vir_gra_gic_ctl.gic_src1_adr0;
            scale_virtual_layer_coords( g0_layer_data, m_layer_data, &scale_rect );            
            info.dest_x = scale_rect.x;
            info.dest_y = scale_rect.y;
            info.dest_w = scale_rect.w;
            info.dest_h = scale_rect.h;
            break;
        case IG_PLANE:
            info.address = mdrv->shared->vir_gra_gic_ctl.gic_src2_adr0;
            scale_virtual_layer_coords( g0_layer_data, m_layer_data, &scale_rect );            
            info.dest_x = scale_rect.x;
            info.dest_y = scale_rect.y;
            info.dest_w = scale_rect.w;
            info.dest_h = scale_rect.h;
            break;
        case BG_PLANE:
            scale_virtual_layer_coords( g0_layer_data, m_layer_data, &scale_rect );            
            info.dest_x = scale_rect.x;
            info.dest_y = scale_rect.y;
            info.dest_w = scale_rect.w;
            info.dest_h = scale_rect.h;
            break;
        case G0_PLANE:
            info.premultiplied = (REG_READ(DISP_HD_COMP_ALPHA_MODE) & PREMULT_DISABLE_G0_MASK) != 0;
            info.address = mdrv->shared->vir_gra_gic_ctl.gic_dst_adr0;
            break;
        default:
            break;
    }

    DPRINTF("layer %d: addr=%x, fmt=%x, (%d,%d)->(%d,%d), pitch=%d, 3d=%d, depth=%d, alpha=%d, premul=%d\n",
        id, info.address, info.format, info.width, info.height, info.dest_w, info.dest_h, info.pitch, info.is_3d, info.stereo_depth, info.alpha, info.premultiplied);

    return info;
}

static void xcode4_check_force_3d(xc4_plane_info_t * src, xc4_plane_info_t * dst)
{
    if(src->force_3d)
    {
        dst->is_3d = 1;
        dst->force_3d = src->force_3d;
    }
}

static xc4_stereo_info_t xcode4_get_stereo_info(pxc_dfb_driver_data mdrv, int stereo_eye, xc4_plane_info_t * plane_info)
{
    xc4_stereo_info_t info;

    memset(&info, 0, sizeof(info));

    int delta = (1 - stereo_eye * 2) * plane_info->stereo_depth;

    info.id = plane_info->id;
    
    switch(mdrv->shared->current_hd_resolution)
    {
        case DSOR_1920_2205:
            if(plane_info->is_3d)
            {
                info.src_x = 0;
                info.src_y = stereo_eye * (1080 + 45);
                info.src_w = 1920;
                info.src_h = 1080;
                info.depth = 0;
            }
            else
            {
                if(plane_info->id == G0_PLANE)
                {
                    info.src_x = 0;
                    info.src_y = 0;
                    info.src_w = 1920;
                    info.src_h = 1080;
                }
                else
                {
                    info.src_x = 0;
                    info.src_y = 0;
                    info.src_w = plane_info->width;
                    info.src_h = plane_info->height;
                }
                info.depth = delta;
            }
            break;
        case DSOR_1280_1470:
            if(plane_info->is_3d)
            {
                info.src_x = 0;
                info.src_y = stereo_eye * (720 + 30);
                info.src_w = 1280;
                info.src_h = 720;
                info.depth = 0;
            }
            else
            {
                if(plane_info->id == G0_PLANE)
                {
                    info.src_x = 0;
                    info.src_y = 0;
                    info.src_w = 1280;
                    info.src_h = 720;
                }
                else
                {
                    info.src_x = 0;
                    info.src_y = 0;
                    info.src_w = plane_info->width;
                    info.src_h = plane_info->height;
                }
                info.depth = delta;
            }
            break;
        case DSOR_1920_1080_3D_SIDE_BY_SIDE:
            if(plane_info->is_3d)
            {
                if(plane_info->id == G0_PLANE)
                {
                    info.src_x = stereo_eye * 960;
                    info.src_y = 0;
                    info.src_w = 960;
                    info.src_h = 1080;
                }
                else
                {
                    info.src_x = stereo_eye * 1920;
                    info.src_y = 0;
                    info.src_w = 1920;
                    info.src_h = 1080;
                }
                info.depth = 0;
            }
            else
            {
                if(plane_info->id == G0_PLANE)
                {
                    info.src_x = 0;
                    info.src_y = 0;
                    info.src_w = 1920;
                    info.src_h = 1080;
                }
                else
                {
                    info.src_x = 0;
                    info.src_y = 0;
                    info.src_w = plane_info->width;
                    info.src_h = plane_info->height;
                }
                info.depth = delta / 2;
            }
            break;
        default:
            if(plane_info->is_3d)
            {
                switch(plane_info->force_3d)
                {
                    case FORCE_3D_SIDE_BY_SIDE:
                        info.src_x = stereo_eye * plane_info->width / 2;
                        info.src_y = 0;
                        info.src_w = plane_info->width / 2;
                        info.src_h = plane_info->height;
                        break;
                    case FORCE_3D_TOP_AND_BOT:
                        info.src_x = 0;
                        info.src_y = stereo_eye * plane_info->height / 2;
                        info.src_w = plane_info->width;
                        info.src_h = plane_info->height / 2;
                        break;
                    default:
                        info.src_x = 0;
                        info.src_y = 0;
                        info.src_w = plane_info->width;
                        info.src_h = plane_info->height;
                        break;
                }
                info.depth = 0;
            }
            else
            {
                info.src_x = 0;
                info.src_y = 0;
                info.src_w = plane_info->width;
                info.src_h = plane_info->height;
                info.depth = delta;
            }
            break;
    }

    return info;
}

static int xcode4_get_stereo_window(pxc_dfb_driver_data mdrv, int stereo_eye, int *pX, int *pY, int *pW, int *pH)
{
    DPRINTF("Original window (%d,%d,%d,%d)\n",
        *pX, *pY, *pW, *pH);
    
    switch(mdrv->shared->current_hd_resolution)
    {
        case DSOR_1920_2205:
            *pY += stereo_eye * (1080 + 45);
            break;
        case DSOR_1280_1470:
            *pY += stereo_eye * (720 + 30);
            break;
        case DSOR_1920_1080_3D_SIDE_BY_SIDE:
            *pX = *pX + (stereo_eye * 960);
            break;
        default:
            return -1;
    }

    DPRINTF("Modified window (%d,%d,%d,%d)\n",
        *pX, *pY, *pW, *pH);
    
    return 0;
}

//
// G2D plane clear
//
static int xcode4_plane_clear(pxc_dfb_driver_data mdrv, xc4_plane_id_t src0_id, xc4_plane_id_t dummy_id)
{
    xc4_plane_info_t src_info;

    DPRINTF("ID=%d Clear\n", src0_id);
    
    src_info = xcode4_get_plane_info(mdrv, src0_id);

    ERAISE(xcode4_gic_clear(mdrv,
        src_info.address,
        src_info.format,
        0,
        0,
        src_info.width,
        src_info.height,
        src_info.pitch));

    return 0;
}

//
// G2D plane copy
//
static int xcode4_plane_copy(pxc_dfb_driver_data mdrv, xc4_plane_id_t src_id, xc4_plane_id_t dst_id)
{
    xc4_plane_info_t src_info, dst_info;
    int no_need_premultiply;

    DPRINTF("ID=%d->%d\n", src_id, dst_id);

    src_info = xcode4_get_plane_info(mdrv, src_id);
    dst_info = xcode4_get_plane_info(mdrv, dst_id);

    xcode4_check_force_3d(&src_info, &dst_info);

    no_need_premultiply = src_info.premultiplied || !dst_info.premultiplied;

    if (dst_info.is_3d)
    {
        xc4_stereo_info_t stereo_src, stereo_dst;

        stereo_src = xcode4_get_stereo_info(mdrv, 0, &src_info);

        stereo_dst = xcode4_get_stereo_info(mdrv, 0, &dst_info);

        if ((stereo_src.src_w == stereo_dst.src_w) && (stereo_src.src_h == stereo_dst.src_h))
        {
            ERAISE(xcode4_gic_copy(mdrv,
                src_info.address,
                src_info.format,
                stereo_src.src_x,
                stereo_src.src_y,
                stereo_src.src_w,
                stereo_src.src_h,
                src_info.pitch,
                dst_info.address ,
                dst_info.format,
                stereo_dst.src_x,
                stereo_dst.src_y,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                stereo_src.depth,
                no_need_premultiply));
        }
        else
        {
            Wait_For_Gic0_Idle;
            ERAISE(xcode4_g2d_scale_depth(mdrv,
                src_info.address,
                src_info.format,
                stereo_src.src_x,
                stereo_src.src_y,
                stereo_src.src_w,
                stereo_src.src_h,
                src_info.pitch,
                dst_info.address ,
                dst_info.format,
                stereo_dst.src_x,
                stereo_dst.src_y,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                stereo_src.depth));
            Wait_For_G2d_Idle;
            mdrv->shared->last_frame_update = 1;
        }

        stereo_src = xcode4_get_stereo_info(mdrv, 1, &src_info);

        stereo_dst = xcode4_get_stereo_info(mdrv, 1, &dst_info);

        if ((stereo_src.src_w == stereo_dst.src_w) && (stereo_src.src_h == stereo_dst.src_h))
        {
            ERAISE(xcode4_gic_copy(mdrv,
                src_info.address,
                src_info.format,
                stereo_src.src_x,
                stereo_src.src_y,
                stereo_src.src_w,
                stereo_src.src_h,
                src_info.pitch,
                dst_info.address ,
                dst_info.format,
                stereo_dst.src_x,
                stereo_dst.src_y,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                stereo_src.depth,
                no_need_premultiply));
        }
        else
        {
            Wait_For_Gic0_Idle;
            ERAISE(xcode4_g2d_scale_depth(mdrv,
                src_info.address,
                src_info.format,
                stereo_src.src_x,
                stereo_src.src_y,
                stereo_src.src_w,
                stereo_src.src_h,
                src_info.pitch,
                dst_info.address ,
                dst_info.format,
                stereo_dst.src_x,
                stereo_dst.src_y,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                stereo_src.depth));
            Wait_For_G2d_Idle;
            mdrv->shared->last_frame_update = 1;
        }
    }
    else
    {
        if(src_info.is_3d)
        {
            if(src_info.height == 2205)
            {
                src_info.dest_w = src_info.width = 1920;
                src_info.dest_h = src_info.height = 1080;
            }
            else if(src_info.height == 1470)
            {
                src_info.dest_w = src_info.width = 1280;
                src_info.dest_h = src_info.height = 720;
            }
            else
            {
                src_info.dest_w = src_info.width = 1920;
                src_info.dest_h = src_info.height = 1080;
            }
        }

        if(src_info.width == src_info.dest_w && src_info.height == src_info.dest_h)
        {
            ERAISE(xcode4_gic_copy(mdrv,
                src_info.address,
                src_info.format,
                src_info.x,
                src_info.y,
                src_info.width,
                src_info.height,
                src_info.pitch,
                dst_info.address,
                dst_info.format,
                src_info.dest_x,
                src_info.dest_y,
                src_info.dest_w,
                src_info.dest_h,
                dst_info.pitch,
                0,
                no_need_premultiply));
        }
        else
        {
            if ((src_info.dest_w < dst_info.width) || (src_info.dest_h < dst_info.height) || (src_info.dest_x != dst_info.x) || (src_info.dest_y != dst_info.y))
            {
                xcode4_gic_clear(mdrv,
                    dst_info.address,
                    dst_info.format,
                    0,
                    0,
                    dst_info.width,
                    dst_info.height,
                    dst_info.pitch);
            }

            Wait_For_Gic0_Idle;
            ERAISE(xcode4_g2d_scale_depth(mdrv,
                src_info.address,
                src_info.format,
                src_info.x,
                src_info.y,
                src_info.width,
                src_info.height,
                src_info.pitch,
                dst_info.address,
                dst_info.format,
                src_info.dest_x,
                src_info.dest_y,
                src_info.dest_w,
                src_info.dest_h,
                dst_info.pitch,
                0));
            Wait_For_G2d_Idle;
            mdrv->shared->last_frame_update = 1;
        }
    }

    return 0;
}

//
// GIC plane copy
//
static int xcode4_plane_merge(pxc_dfb_driver_data mdrv, xc4_plane_id_t src_id, xc4_plane_id_t dst_id)
{
    xc4_plane_info_t src_info, dst_info;

    DPRINTF("ID=%d+%d->%d\n", src_id, dst_id, dst_id);

    src_info = xcode4_get_plane_info(mdrv, src_id);
    
    dst_info = xcode4_get_plane_info(mdrv, dst_id);

    xcode4_check_force_3d(&src_info, &dst_info);

    if(dst_info.is_3d)
    {
        xc4_stereo_info_t stereo_src, stereo_dst;

        stereo_src = xcode4_get_stereo_info(mdrv, 0, &src_info);

        stereo_dst = xcode4_get_stereo_info(mdrv, 0, &dst_info);

        if ((stereo_src.src_w == stereo_dst.src_w) && (stereo_src.src_h == stereo_dst.src_h))
        {
            xcode4_gic_src1_over_src2(mdrv,
                src_info.address +  stereo_src.src_x * DFB_BYTES_PER_PIXEL(src_info.format) + stereo_src.src_y * src_info.pitch,
                src_info.format,
                stereo_src.depth,
                0,
                stereo_src.src_w,
                stereo_src.src_h,
                src_info.pitch,
                dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
                dst_info.format,
                0,
                0,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
                dst_info.format,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                src_info.premultiplied
                );
        }
        else
        {
            Wait_For_Gic0_Idle;
            xcode4_g2d_scale_depth(mdrv,
                src_info.address,
                src_info.format,
                stereo_src.src_x,
                stereo_src.src_y,
                stereo_src.src_w,
                stereo_src.src_h,
                src_info.pitch,
                xcode4_get_temp_addr(mdrv),
                dst_info.format,
                0,
                0,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                0);
            Wait_For_G2d_Idle;
            mdrv->shared->last_frame_update = 1;

            xcode4_gic_src1_over_src2(mdrv,
                xcode4_get_temp_addr(mdrv),
                dst_info.format,
                stereo_src.depth,
                0,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
                dst_info.format,
                0,
                0,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
                dst_info.format,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                src_info.premultiplied
                );
        }

        stereo_src = xcode4_get_stereo_info(mdrv, 1, &src_info);

        stereo_dst = xcode4_get_stereo_info(mdrv, 1, &dst_info);

        if ((stereo_src.src_w == stereo_dst.src_w) && (stereo_src.src_h == stereo_dst.src_h))
        {
            xcode4_gic_src1_over_src2(mdrv,
                src_info.address +  stereo_src.src_x * DFB_BYTES_PER_PIXEL(src_info.format) + stereo_src.src_y * src_info.pitch,
                src_info.format,
                stereo_src.depth,
                0,
                stereo_src.src_w,
                stereo_src.src_h,
                src_info.pitch,
                dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
                dst_info.format,
                0,
                0,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
                dst_info.format,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                src_info.premultiplied
                );
        }
        else
        {
            Wait_For_Gic0_Idle;
            xcode4_g2d_scale_depth(mdrv,
                src_info.address,
                src_info.format,
                stereo_src.src_x,
                stereo_src.src_y,
                stereo_src.src_w,
                stereo_src.src_h,
                src_info.pitch,
                xcode4_get_temp_addr(mdrv),
                dst_info.format,
                0,
                0,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                0);
            Wait_For_G2d_Idle;
            mdrv->shared->last_frame_update = 1;

            xcode4_gic_src1_over_src2(mdrv,
                xcode4_get_temp_addr(mdrv),
                dst_info.format,
                stereo_src.depth,
                0,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
                dst_info.format,
                0,
                0,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
                dst_info.format,
                stereo_dst.src_w,
                stereo_dst.src_h,
                dst_info.pitch,
                src_info.premultiplied
                );
        }
    }
    else
    {
        if(src_info.width == src_info.dest_w && src_info.height == src_info.dest_h)
        {
            xcode4_gic_src1_over_src2(mdrv,
                src_info.address +  src_info.x * DFB_BYTES_PER_PIXEL(src_info.format) + src_info.y * src_info.pitch,
                src_info.format,
                src_info.dest_x,
                src_info.dest_y,
                src_info.dest_w,
                src_info.dest_h,
                src_info.pitch,
                dst_info.address + dst_info.x * DFB_BYTES_PER_PIXEL(dst_info.format) + dst_info.y * dst_info.pitch,
                dst_info.format,
                0,
                0,
                src_info.dest_w,
                src_info.dest_h,
                dst_info.pitch,
                dst_info.address,
                dst_info.format,
                src_info.dest_w,
                src_info.dest_h,
                dst_info.pitch,
                src_info.premultiplied
                );
        }
        else
        {
            Wait_For_Gic0_Idle;
            xcode4_g2d_scale_depth(mdrv,
                src_info.address,
                src_info.format,
                src_info.x,
                src_info.y,
                src_info.width,
                src_info.height,
                src_info.pitch,
                xcode4_get_temp_addr(mdrv),
                dst_info.format,
                0,
                0,
                src_info.dest_w,
                src_info.dest_h,
                dst_info.pitch,
                0);
            Wait_For_G2d_Idle;
            mdrv->shared->last_frame_update = 1;

            xcode4_gic_src1_over_src2(mdrv,
                xcode4_get_temp_addr(mdrv),
                dst_info.format,
                src_info.dest_x,
                src_info.dest_y,
                src_info.dest_w,
                src_info.dest_h,
                dst_info.pitch,
                dst_info.address + dst_info.x * DFB_BYTES_PER_PIXEL(dst_info.format) + dst_info.y * dst_info.pitch,
                dst_info.format,
                0,
                0,
                src_info.dest_w,
                src_info.dest_h,
                dst_info.pitch,
                dst_info.address + dst_info.x * DFB_BYTES_PER_PIXEL(dst_info.format) + dst_info.y * dst_info.pitch,
                dst_info.format,
                src_info.dest_w,
                src_info.dest_h,
                dst_info.pitch,
                src_info.premultiplied
                );
        }
    }

    return 0;
}

//
// GIC plane copy
//
static int xcode4_plane_merge2(pxc_dfb_driver_data mdrv, xc4_plane_id_t src_id, xc4_plane_id_t src2_id, xc4_plane_id_t dst_id)
{
    xc4_plane_info_t src_info, src2_info, dst_info;

    DPRINTF("ID=%d+%d->%d\n", src_id, src2_id, dst_id);

    src_info = xcode4_get_plane_info(mdrv, src_id);

    src2_info = xcode4_get_plane_info(mdrv, src2_id);
    
    dst_info = xcode4_get_plane_info(mdrv, dst_id);

    xcode4_check_force_3d(&src_info, &dst_info);

    xcode4_check_force_3d(&src2_info, &dst_info);

    if(dst_info.is_3d)
    {
        xc4_stereo_info_t stereo_src, stereo_src2, stereo_dst;

        stereo_src = xcode4_get_stereo_info(mdrv, 0, &src_info);

        stereo_src2 = xcode4_get_stereo_info(mdrv, 0, &src2_info);

        stereo_dst = xcode4_get_stereo_info(mdrv, 0, &dst_info);

        if ((stereo_src.src_w != stereo_dst.src_w) || (stereo_src.src_h != stereo_dst.src_h) ||
            (stereo_src2.src_w != stereo_dst.src_w) || (stereo_src2.src_h != stereo_dst.src_h))
        {
            ERAISE(xcode4_plane_copy(mdrv, src2_id, dst_id));

            ERAISE(xcode4_plane_merge(mdrv, src_id, dst_id));

            return 0;
        }

        xcode4_gic_src1_over_src2(mdrv,
            src_info.address +  stereo_src.src_x * DFB_BYTES_PER_PIXEL(src_info.format) + stereo_src.src_y * src_info.pitch,
            src_info.format,
            stereo_src.depth,
            0,
            stereo_src.src_w,
            stereo_src.src_h,
            src_info.pitch,
            src2_info.address +  stereo_src2.src_x * DFB_BYTES_PER_PIXEL(src2_info.format) + stereo_src2.src_y * src2_info.pitch,
            src2_info.format,
            stereo_src2.depth,
            0,
            stereo_src2.src_w,
            stereo_src2.src_h,
            src2_info.pitch,
            dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * src_info.pitch,
            dst_info.format,
            stereo_dst.src_w,
            stereo_dst.src_h,
            dst_info.pitch,
            src_info.premultiplied
            );

        stereo_src = xcode4_get_stereo_info(mdrv, 1, &src_info);

        stereo_src2 = xcode4_get_stereo_info(mdrv, 1, &src2_info);

        stereo_dst = xcode4_get_stereo_info(mdrv, 1, &dst_info);

        xcode4_gic_src1_over_src2(mdrv,
            src_info.address +  stereo_src.src_x * DFB_BYTES_PER_PIXEL(src_info.format) + stereo_src.src_y * src_info.pitch,
            src_info.format,
            stereo_src.depth,
            0,
            stereo_src.src_w,
            stereo_src.src_h,
            src_info.pitch,
            src2_info.address +  stereo_src2.src_x * DFB_BYTES_PER_PIXEL(src2_info.format) + stereo_src2.src_y * src2_info.pitch,
            src2_info.format,
            stereo_src2.depth,
            0,
            stereo_src2.src_w,
            stereo_src2.src_h,
            src2_info.pitch,
            dst_info.address + stereo_dst.src_x * DFB_BYTES_PER_PIXEL(dst_info.format) + stereo_dst.src_y * dst_info.pitch,
            dst_info.format,
            stereo_dst.src_w,
            stereo_dst.src_h,
            dst_info.pitch,
            src_info.premultiplied
            );
    }
    else
    {
        if(src2_info.width != src2_info.dest_w || src2_info.height != src2_info.dest_h ||
           src_info.width != src_info.dest_w || src_info.height != src_info.dest_h)
        {
            ERAISE(xcode4_plane_copy(mdrv, src2_id, dst_id));

            return xcode4_plane_merge(mdrv, src_id, dst_id);
        }
        
        xcode4_gic_src1_over_src2(mdrv,
            src_info.address +  src_info.x * DFB_BYTES_PER_PIXEL(src_info.format) + src_info.y * src_info.pitch,
            src_info.format,
            src_info.dest_x,
            src_info.dest_y,
            src_info.dest_w,
            src_info.dest_h,
            src_info.pitch,
            src2_info.address +  src2_info.x * DFB_BYTES_PER_PIXEL(src2_info.format) + src2_info.y * src2_info.pitch,
            src2_info.format,
            src2_info.dest_x,
            src2_info.dest_y,
            src2_info.dest_w,
            src2_info.dest_h,
            src2_info.pitch,
            dst_info.address,
            dst_info.format,
            dst_info.width,
            dst_info.height,
            dst_info.pitch,
            src_info.premultiplied
            );
    }

    return 0;
}

//
// G2D plane fill
//
static int xcode4_plane_fill(pxc_dfb_driver_data mdrv, xc4_plane_id_t dst_id, xc4_plane_id_t dummy_id)
{
    xc4_plane_info_t dst_info;

    pxc_dfb_layer_data layer;

    int sx, sy, sw, sh;

    DPRINTF("ID=%d Fill\n", dst_id);

    dst_info = xcode4_get_plane_info(mdrv, dst_id);
    
    if(!dst_info.is_3d)
    {
        layer = (pxc_dfb_layer_data) dfb_layer_at (V0_PLANE)->layer_data;

        sx = REG_READ(DISP_HD_COMP_VID0_POS) & DISP_HD_COMP_VID0_POS_POS_X_MASK;
        sy = (REG_READ(DISP_HD_COMP_VID0_POS) & DISP_HD_COMP_VID0_POS_POS_Y_MASK) >> DISP_HD_COMP_VID0_POS_POS_Y_SHIFT;
        sw = REG_READ(DISP_HD_VID0_SCALED_SIZE ) & DISP_HD_VID0_SCALED_SIZE_H_SCALED_SIZE_MASK,
        sh = (REG_READ(DISP_HD_VID0_SCALED_SIZE ) & DISP_HD_VID0_SCALED_SIZE_V_SCALED_SIZE_MASK) >> DISP_HD_VID0_SCALED_SIZE_V_SCALED_SIZE_SHIFT;

        if(layer->alpha)
        {
            ERAISE(xcode4_gic_clear(mdrv,
                dst_info.address,
                dst_info.format,
                sx,
                sy,
                sw,
                sh,
                dst_info.pitch));
        }

        layer = (pxc_dfb_layer_data) dfb_layer_at (V1_PLANE)->layer_data;

        sx = REG_READ(DISP_HD_COMP_VID1_POS) & DISP_HD_COMP_VID1_POS_POS_X_MASK;
        sy = (REG_READ(DISP_HD_COMP_VID1_POS) & DISP_HD_COMP_VID1_POS_POS_Y_MASK) >> DISP_HD_COMP_VID1_POS_POS_Y_SHIFT;
        sw = REG_READ(DISP_HD_VID1_SCALED_SIZE ) & DISP_HD_VID1_SCALED_SIZE_H_SCALED_SIZE_MASK,
        sh = (REG_READ(DISP_HD_VID1_SCALED_SIZE ) & DISP_HD_VID1_SCALED_SIZE_V_SCALED_SIZE_MASK) >> DISP_HD_VID1_SCALED_SIZE_V_SCALED_SIZE_SHIFT;
        
        if(layer->alpha)
        {
            ERAISE(xcode4_gic_clear(mdrv,
                dst_info.address,
                dst_info.format,
                sx,
                sy,
                sw,
                sh,
                dst_info.pitch));
        }
    }
    else
    {
        int dx, dy, dw, dh;

        layer = (pxc_dfb_layer_data) dfb_layer_at (V0_PLANE)->layer_data;

        sx = REG_READ(DISP_HD_COMP_VID0_POS) & DISP_HD_COMP_VID0_POS_POS_X_MASK;
        sy = (REG_READ(DISP_HD_COMP_VID0_POS) & DISP_HD_COMP_VID0_POS_POS_Y_MASK) >> DISP_HD_COMP_VID0_POS_POS_Y_SHIFT;
        sw = REG_READ(DISP_HD_VID0_SCALED_SIZE ) & DISP_HD_VID0_SCALED_SIZE_H_SCALED_SIZE_MASK,
        sh = (REG_READ(DISP_HD_VID0_SCALED_SIZE ) & DISP_HD_VID0_SCALED_SIZE_V_SCALED_SIZE_MASK) >> DISP_HD_VID0_SCALED_SIZE_V_SCALED_SIZE_SHIFT;
        
        if(layer->alpha)
        {
            dx = sx;
            dy = sy;
            dw = sw;
            dh = sh;

            xcode4_get_stereo_window(mdrv, 0, &dx, &dy, &dw, &dh);

            ERAISE(xcode4_gic_clear(mdrv,
                dst_info.address,
                dst_info.format,
                dx,
                dy,
                dw,
                dh,
                dst_info.pitch));

            dx = sx;
            dy = sy;
            dw = sw;
            dh = sh;

            xcode4_get_stereo_window(mdrv, 1, &dx, &dy, &dw, &dh);
            
            ERAISE(xcode4_gic_clear(mdrv,
                            dst_info.address,
                            dst_info.format,
                            dx,
                            dy,
                            dw,
                            dh,
                            dst_info.pitch));
        }
    }

    return 0;
}

static int xcode4_plane_replace(pxc_dfb_driver_data mdrv, xc4_plane_id_t src_id, xc4_plane_id_t dst_id)
{
    xc4_plane_info_t src_info, dst_info;
    int no_need_premultiply;

    DPRINTF("ID=%d->%d\n", src_id, dst_id);

    src_info = xcode4_get_plane_info(mdrv, src_id);
    dst_info = xcode4_get_plane_info(mdrv, dst_id);

    xcode4_check_force_3d(&src_info, &dst_info);

    no_need_premultiply = src_info.premultiplied || !dst_info.premultiplied;

    if( no_need_premultiply && 
        !src_info.is_3d && 
        !dst_info.is_3d && 
        src_info.width == src_info.dest_w && 
        src_info.height == src_info.dest_h &&
        src_info.width == dst_info.width &&
        src_info.height == dst_info.height)
    {
        mdrv->shared->vir_gra_gic_ctl.gic_dst_adr0 = src_info.address;
        mdrv->shared->vir_gra_gic_ctl.gic_dst_pitch = src_info.pitch;
        DPRINTF("Show %d directly\n", src_id);
        return 0;
    }

    return 1;
}

//
// Merge planes according to mode
//
static int xcode4_merge_mode(pxc_dfb_driver_data mdrv, int mode)
{
    //bits of mode
    //bit0: AP_PLANE exist
    //bit1: IG_PLANE exist
    //bit2: BG_PLANE exist
    //others: Reserved

    int err = 0;

    pthread_mutex_lock(&mdrv->shared->g2d_mutex);

    switch(mode)
    {
        case 0x00: //empty
            err = (xcode4_plane_clear(mdrv, G0_PLANE, 0));
            break;
        case 0x01: //AP
            if((err = xcode4_plane_replace(mdrv, AP_PLANE, G0_PLANE)))
            {
                err = (xcode4_plane_copy(mdrv, AP_PLANE, G0_PLANE));
            }
            break;
        case 0x02: //IG
            if((err = xcode4_plane_replace(mdrv, IG_PLANE, G0_PLANE)))
            {
                err = (xcode4_plane_copy(mdrv, IG_PLANE, G0_PLANE));
            }
            break;
        case 0x03: //AP+IG
            err = (xcode4_plane_merge2(mdrv, AP_PLANE, IG_PLANE, G0_PLANE));
            break;
        case 0x04: //BG
            err = (xcode4_plane_copy(mdrv, BG_PLANE, G0_PLANE));
            err = (xcode4_plane_fill(mdrv, G0_PLANE, 0));
            break;
        case 0x05: //BG+AP
            err = (xcode4_plane_copy(mdrv, BG_PLANE, G0_PLANE));
            err = (xcode4_plane_fill(mdrv, G0_PLANE, 0));
            err = (xcode4_plane_merge(mdrv, AP_PLANE, G0_PLANE));
            break;
        case 0x06: //BG+IG
            err = (xcode4_plane_copy(mdrv, BG_PLANE, G0_PLANE));
            err = (xcode4_plane_fill(mdrv, G0_PLANE, 0));
            err = (xcode4_plane_merge(mdrv, IG_PLANE, G0_PLANE));
            break;
        case 0x07: //BG+IG+AP
            err = (xcode4_plane_copy(mdrv, BG_PLANE, G0_PLANE));
            err = (xcode4_plane_fill(mdrv, G0_PLANE, 0));
            err = (xcode4_plane_merge(mdrv, IG_PLANE, G0_PLANE));
            err = (xcode4_plane_merge(mdrv, AP_PLANE, G0_PLANE));
            break;
        default:
            break;
    }

    Wait_For_Gic0_Idle;

    pthread_mutex_unlock(&mdrv->shared->g2d_mutex);;

    return 0;
}

static int xcode4_build_mode(int ap_exist, int ig_exist, int bg_exist)
{
    return (bg_exist << 2) | (ig_exist << 1) | ap_exist;
}

static int xcode4_get_ig_exist(pxc_dfb_driver_data mdrv)
{
    return (mdrv->shared->vir_gra_gic_ctl.gic_alpha_mode_src2 != GIC_CONSTANT_ALPHA) &&
        (mdrv->shared->vir_gra_gic_ctl.gic_opt_status & GIC_SRC2_READY);
}

static int xcode4_get_ap_exist(pxc_dfb_driver_data mdrv)
{
    return (mdrv->shared->vir_gra_gic_ctl.gic_alpha_mode_src1 != GIC_CONSTANT_ALPHA) &&
        (mdrv->shared->vir_gra_gic_ctl.gic_opt_status & GIC_SRC1_READY);
}

int Background_Layer_Alpha_On(pxc_dfb_driver_data mdrv)
{
    CoreLayer * bg_layer = dfb_layer_at (BACKGROUND_GRA_LAYER_ID);
    if(bg_layer && (mdrv->shared->vir_gra_gic_ctl.gic_opt_status & GIC_BG_READY))
    {
        pxc_dfb_layer_data m_layer_data = (pxc_dfb_layer_data) bg_layer->layer_data;
        if(m_layer_data && m_layer_data->alpha)
        {
            return 1;
        }
    }
    return 0;
}

static int xcode4_get_bg_exist(pxc_dfb_driver_data mdrv)
{
    if(Background_Layer_Alpha_On(mdrv))
    {
        CoreLayer * layer;
        pxc_dfb_layer_data vid0_layer, vid1_layer, gra0_layer;
        int full_video = 0;
        int is_3d = 0;
        
        layer = dfb_layer_at (V0_PLANE);
        vid0_layer = (pxc_dfb_layer_data) layer->layer_data;

        layer = dfb_layer_at (G0_PLANE);
        gra0_layer = (pxc_dfb_layer_data) layer->layer_data;

        if(vid0_layer->layer_is_set && vid0_layer->alpha)
        {
            int target_width, target_height;
            switch(mdrv->shared->current_hd_resolution)
            {
                case DSOR_1920_2205:
                    target_width = 1920;
                    target_height = 1080;
                    is_3d = 1;
                    break;
                case DSOR_1280_1470:
                    target_width = 1280;
                    target_height = 720;
                    is_3d = 1;
                    break;
                case DSOR_1920_1080_3D_SIDE_BY_SIDE:
                    target_width = 960;
                    target_height = 1080;
                    is_3d = 1;
                    break;
                default:
                    target_width = gra0_layer->dest_width;
                    target_height = gra0_layer->dest_height;
                    break;
            }
            if(vid0_layer->dest_width >= target_width && vid0_layer->dest_height >= target_height)
            {
                DPRINTF("Video 0 is full screen\n");
                full_video = 1;
            }
        }

        if(!full_video && !is_3d)
        {
            layer = dfb_layer_at (V1_PLANE);
            vid1_layer = (pxc_dfb_layer_data) layer->layer_data;

            if(vid1_layer->layer_is_set && 
                vid1_layer->alpha &&
                vid1_layer->dest_width >= gra0_layer->dest_width &&
                vid1_layer->dest_height >= gra0_layer->dest_height)
            {
                DPRINTF("Video 1 is full screen\n");
                full_video = 1;
            }
        }

        return !full_video;
    }
    
    return 0;
}

static int xcode4_osd_merge(pxc_dfb_driver_data mdrv, int layerId)
{
    if(mdrv->shared->vir_gra_gic_ctl.gic_opt_status & GIC_DST_READY)
    {
        int ap_exist = xcode4_get_ap_exist(mdrv);
        int ig_exist = xcode4_get_ig_exist(mdrv);
        int bg_exist = xcode4_get_bg_exist(mdrv);
        int mode = xcode4_build_mode(ap_exist, ig_exist, bg_exist);

        DPRINTF("FLIP %d, MODE 0x%x\n", layerId, mode);
        ERAISE(xcode4_merge_mode(mdrv, mode));

        //Wait for sync if merge is needed
        if(mode > 3)
        {
            mdrv->shared->last_frame_update = 1;
        }
    }
    return 0;
}

static void Toggle_Virtual_Graphic_Dst_Buffer(pxc_dfb_driver_data mdrv)
{
    u32 tmp;
    u32 base_frm_cnt;

    //Write frame buffer
    REG_WRITE(DISP_HD_GRA0_ADR, mdrv->shared->vir_gra_gic_ctl.gic_dst_adr0);
    REG_WRITE(DISP_HD_GRA0_BUF, mdrv->shared->vir_gra_gic_ctl.gic_dst_pitch);

    //Update next buffer pointer and pitch
    if(mdrv->shared->vir_gra_gic_ctl.gic_dst_adr0 == mdrv->shared->vir_gra_gic_ctl.dst_buffer1)
    {
        mdrv->shared->vir_gra_gic_ctl.gic_dst_adr0 = mdrv->shared->vir_gra_gic_ctl.dst_buffer2;
    }
    else
    {
        mdrv->shared->vir_gra_gic_ctl.gic_dst_adr0 = mdrv->shared->vir_gra_gic_ctl.dst_buffer1;
    }
    mdrv->shared->vir_gra_gic_ctl.gic_dst_pitch = mdrv->shared->vir_gra_gic_ctl.dst_pitch;

    //WAIT_FOP_NEXT_HD_FRAME;
    mdrv->shared->last_frame_counter = GET_HD_FRM_COUNT;

    return;

}

void Virtual_Layer_Combine_To_Physical_Hd_Gra0(pxc_dfb_driver_data    mdrv,
                                               void                  *layer_data,
                                               CoreSurfaceBufferLock *lock,
                                               int layer_id,
                                               DFBSurfaceFlipFlags flags)
{
    CoreLayer * bg_layer;
    pxc_dfb_layer_data bg_layer_data;
    u32 video0_screen_width;
    u32 video0_screen_height;
    u32 video1_screen_width;
    u32 video1_screen_height;
    int active_layer_id = -1;
    int current_frame_count = 0;

    pthread_mutex_lock(&mdrv->shared->virtual_layer_mutex);

    if(lock)
    {
        if (layer_id == TOP_GRA0_LAYER_ID)
        {
            mdrv->shared->vir_gra_gic_ctl.gic_src2_adr0 = lock->phys;
        }
        else if (layer_id == TOP_GRA1_LAYER_ID)
        {
            mdrv->shared->vir_gra_gic_ctl.gic_src1_adr0 = lock->phys;
        }
    }

    if(mdrv->shared->last_frame_update)
    {
        //wait for last refresh complete
        while ((mdrv->shared->last_frame_counter == GET_HD_FRM_COUNT) &&
               (REG_READ(DISP_HD_COMP_CTRL) & DISP_HD_COMP_CTRL_COMP_EN_MASK))
        {
            usleep(1);
        }
        mdrv->shared->last_frame_update = 0;
    }
    else if(mdrv->shared->last_frame_counter == GET_HD_FRM_COUNT)
    {
        //cancel last update
        mdrv->shared->vir_gra_gic_ctl.gic_dst_adr0 = mdrv->shared->last_frame_address;
        mdrv->shared->vir_gra_gic_ctl.gic_dst_pitch = mdrv->shared->last_frame_pitch;
        Toggle_Virtual_Graphic_Dst_Buffer( mdrv);
    }

    //save frame count before merge
    current_frame_count = GET_HD_FRM_COUNT;

    if(!xcode4_osd_merge(mdrv, layer_id))
    {
        mdrv->shared->last_frame_address = REG_READ(DISP_HD_GRA0_ADR);
        mdrv->shared->last_frame_pitch = REG_READ(DISP_HD_GRA0_BUF);
        Toggle_Virtual_Graphic_Dst_Buffer( mdrv);
    }

    //if flip take place during merging, then we show it next time
    if(current_frame_count != GET_HD_FRM_COUNT)
    {
        mdrv->shared->last_frame_update = 1;
    }

    if(dfb_config->soft_mirror_sd)
    {
        XcodeFlipSDFrame(dfb_layer_at(XCODE4_GRA0_LAYER_ID), NULL, NULL);
    }

    pthread_mutex_unlock(&mdrv->shared->virtual_layer_mutex);

}

