/*
   (c) Copyright 2001-2009  The world wide DirectFB Open Source Community (directfb.org)
   (c) Copyright 2000-2004  Convergence (integrated media) GmbH

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>,
              Andreas Hundt <andi@fischlustig.de>,
              Sven Neumann <neo@directfb.org>,
              Ville Syrjälä <syrjala@sci.fi> and
              Claudio Ciccani <klan@users.sf.net>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>

#include <directfb.h>

#include <core/fonts.h>
#include <core/gfxcard.h>
#include <core/surface.h>
#include <core/surface_buffer.h>

#include <gfx/convert.h>

#include <media/idirectfbfont.h>

#include <direct/mem.h>
#include <direct/memcpy.h>
#include <direct/messages.h>
#include <direct/utf8.h>
#include <direct/util.h>

#include <misc/conf.h>
#include <misc/util.h>

//#define FT_BITMAP_DEBUG

#undef SIZEOF_LONG
#include <ft2build.h>
#include <freetype2/freetype/ftsynth.h>
#include FT_OUTLINE_H
#include FT_TRIGONOMETRY_H
#include FT_OUTLINE_H
#include FT_BITMAP_H


#include FT_GLYPH_H
/* ViXS specific modification ...   */
#include FT_STROKER_H       /* Stroker to generate outlines ... */
/* ViXS specific modifcations end ..*/
#ifndef FT_LOAD_TARGET_MONO
    /* FT_LOAD_TARGET_MONO was added in FreeType-2.1.3, we have to use (less good)
       FT_LOAD_MONOCHROME with older versions. Make it an alias for code simplicity. */
    #define FT_LOAD_TARGET_MONO FT_LOAD_MONOCHROME
#endif

#define MY_FT_PIX_FLOOR( x )     ( (x) & ~63 )

static void New_GlyphSlot_Embolden( FT_GlyphSlot  slot);

static void print_bitmap(FT_BitmapGlyph pglyph);


static DFBResult
Probe( IDirectFBFont_ProbeContext *ctx );

static DFBResult
Construct( IDirectFBFont      *thiz,
	   ... );

#include <direct/interface_implementation.h>

DIRECT_INTERFACE_IMPLEMENTATION( IDirectFBFont, FT2 )

static FT_Library      library           = NULL;
static int             library_ref_count = 0;
static pthread_mutex_t library_mutex     = PTHREAD_MUTEX_INITIALIZER;

#define KERNING_CACHE_MIN    0
#define KERNING_CACHE_MAX  127
#define KERNING_CACHE_SIZE (KERNING_CACHE_MAX - KERNING_CACHE_MIN + 1)


/* Default outline parameters ... */
#define FT2_DEFAULT_OUTLINE_WIDTH       0x10000
#define FT2_DEFAULT_OUTLINE_OPACITY     0xFFFF0000

#define KERNING_DO_CACHE(a,b)      ((a) >= KERNING_CACHE_MIN && \
                                    (a) <= KERNING_CACHE_MAX && \
                                    (b) >= KERNING_CACHE_MIN && \
                                    (b) <= KERNING_CACHE_MAX)

#define KERNING_CACHE_ENTRY(a,b)   \
     (data->kerning[(a)-KERNING_CACHE_MIN][(b)-KERNING_CACHE_MIN])

#define CHAR_INDEX(c)    (((c) < 256) ? data->indices[c] : FT_Get_Char_Index( data->face, c ))

#define GB18030_TWO_BYTES_ENCODING \
((pos + 1) < length) && \
((bytes[pos] >= 0x81) && (bytes[pos] <= 0xfe)) && \
(((bytes[pos+1] >= 0x40) && (bytes[pos+1] <= 0x7e)) || ((bytes[pos+1] >= 0x80) && (bytes[pos+1] <= 0xfe)))

#define GB18030_FOUR_BYTES_ENCODING \
((pos + 3) < length) && \
((bytes[pos] >= 0x81) && (bytes[pos] <= 0xfe)) && \
((bytes[pos+1] >= 0x30) && (bytes[pos+1] <= 0x39)) && \
((bytes[pos+2] >= 0x81) && (bytes[pos+2] <= 0xfe)) && \
((bytes[pos+3] >= 0x30) && (bytes[pos+3] <= 0x39))



/* 
*  If data->outline_render is true AND layer = 0 -> give outline ONLY 
*  If data->outline_render is true AND layer = 1 -> give base font
*  If data->outline_render is false, simply ignore layer and give base only... 
*
*  By making it this way, it is much easier to hook up the DFB API layer to the 
*  outline rendering...   
*/

typedef struct {
     FT_Face      face;
     FT_Stroker   stroker;          /* Stroker object for defining the outline ... */
     int          disable_charmap;
     int          fixed_advance;
     /* ViXS modification start: */
     int          outline_width;    /* Width of the outline in 16.16 fixed point */
     int          outline_opacity;  /* Opacity of the surrounding outline in 16.16 fixed point ... */
     bool         outline_render;   /* Flag used to indicate if the application has requested outlines for this instance ..*/
     /* ViXS modification end: */
     bool         fixed_clip;
     unsigned int indices[256];
} FT2ImplData;

typedef struct {
     signed char x;
     signed char y;
} KerningCacheEntry;

typedef struct {
     FT2ImplData base;

     KerningCacheEntry kerning[KERNING_CACHE_SIZE][KERNING_CACHE_SIZE];
} FT2ImplKerningData;

/**********************************************************************************************************************/

static DFBResult
ft2UTF8GetCharacterIndex( CoreFont     *thiz,
                          unsigned int  character,
                          unsigned int *ret_index )
{
     FT2ImplData *data = thiz->impl_data;

     D_MAGIC_ASSERT( thiz, CoreFont );

     if (data->disable_charmap)
          *ret_index = character;
     else {
          pthread_mutex_lock ( &library_mutex );

          *ret_index = CHAR_INDEX( character );

          pthread_mutex_unlock ( &library_mutex );
     }

     return DFB_OK;
}

static DFBResult
ft2UTF8DecodeText( CoreFont       *thiz,
                   const void     *text,
                   int             length,
                   unsigned int   *ret_indices,
                   int            *ret_num )
{
     int pos = 0, num = 0;
     const u8 *bytes   = text;
     FT2ImplData *data = thiz->impl_data;
     DFBTextEncodingID current_encoding = thiz->current_encoding; 

     D_MAGIC_ASSERT( thiz, CoreFont );
     D_ASSERT( text != NULL );
     D_ASSERT( length >= 0 );
     D_ASSERT( ret_indices != NULL );
     D_ASSERT( ret_num != NULL );

     pthread_mutex_lock ( &library_mutex );

     while (pos < length) {
          unsigned int c;

          if ((current_encoding == DTEID_UTF16BE) ||
              (current_encoding == DTEID_GB2312) ||
              (current_encoding == DTEID_BIG5) ||
              (current_encoding == DTEID_KSC5601_1987))
          {
               if ((pos + 1) < length) 
               {
                    c = (bytes[pos] << 8) | bytes[pos+1];
                    pos = pos + 2;
               }
               else
               {
                    break;
               }
          }
          else if (current_encoding == DTEID_GB18030_2000) 
          {
               if (bytes[pos] < 128) 
               {
                    c = bytes[pos++];
               }
               else if (GB18030_TWO_BYTES_ENCODING) 
               {
                    c = (bytes[pos] << 8) | bytes[pos+1];
                    pos = pos + 2;
               }
               else if (GB18030_FOUR_BYTES_ENCODING) 
               {
                    c = (bytes[pos] << 24) | (bytes[pos+1] << 16) | (bytes[pos+2] << 8) | bytes[pos+3];
                    pos = pos + 4;
               }
               else
               {
                    break;
               }

          }
          else if (current_encoding == DTEID_SHIFT_JIS) 
          {
               if ((bytes[pos] >= 0xa1) && (bytes[pos] <= 0xdf)) 
               {
                    c = bytes[pos++];
               }
               else if ((pos + 1) < length)   
               {
                    c = (bytes[pos] << 8) | bytes[pos+1];
                    pos = pos + 2;
               }
               else
               {
                    break;
               }
          }
          else if (bytes[pos] < 128)
          {
               c = bytes[pos++];
          }
          else {
               c = DIRECT_UTF8_GET_CHAR( &bytes[pos] );
               pos += DIRECT_UTF8_SKIP(bytes[pos]);
          }

          if (data->disable_charmap)
               ret_indices[num++] = c;
          else
               ret_indices[num++] = CHAR_INDEX( c );
     }

     pthread_mutex_unlock ( &library_mutex );

     *ret_num = num;

     return DFB_OK;
}

static const CoreFontEncodingFuncs ft2UTF8Funcs = {
     .GetCharacterIndex = ft2UTF8GetCharacterIndex,
     .DecodeText        = ft2UTF8DecodeText,
};

/**********************************************************************************************************************/

static DFBResult
ft2Latin1GetCharacterIndex( CoreFont     *thiz,
                            unsigned int  character,
                            unsigned int *ret_index )
{
     FT2ImplData *data = thiz->impl_data;

     D_MAGIC_ASSERT( thiz, CoreFont );

     if (data->disable_charmap)
          *ret_index = character;
     else
          *ret_index = data->indices[character];

     return DFB_OK;
}

static DFBResult
ft2Latin1DecodeText( CoreFont       *thiz,
                     const void     *text,
                     int             length,
                     unsigned int   *ret_indices,
                     int            *ret_num )
{
     int i;
     const u8 *bytes   = text;
     FT2ImplData *data = thiz->impl_data;

     D_MAGIC_ASSERT( thiz, CoreFont );
     D_ASSERT( text != NULL );
     D_ASSERT( length >= 0 );
     D_ASSERT( ret_indices != NULL );
     D_ASSERT( ret_num != NULL );

     if (data->disable_charmap) {
          for (i=0; i<length; i++)
               ret_indices[i] = bytes[i];
     }
     else {
          for (i=0; i<length; i++)
               ret_indices[i] = data->indices[bytes[i]];
     }

     *ret_num = length;

     return DFB_OK;
}

static const CoreFontEncodingFuncs ft2Latin1Funcs = {
     .GetCharacterIndex = ft2Latin1GetCharacterIndex,
     .DecodeText        = ft2Latin1DecodeText,
};

/**********************************************************************************************************************/

static DFBResult
render_glyph( CoreFont      *thiz,
              unsigned int   index,
              CoreGlyphData *info )
{
     FT_Error       err;
     FT_Face        face;
     FT_Int         load_flags;
     FT_Glyph       ft_glyph;
     FT_Glyph       ft_glyph_tmp;
     FT_BitmapGlyph ft_bitmapglyph, ft_bitmapglyph_tmp;
     FT_Bitmap      ft_bitmapsource;
     u8          *src;
     int          y;
     FT2ImplData *data    = thiz->impl_data;
     CoreSurface *surface = info->surface;
     CoreSurfaceBufferLock  lock;
     char       ft_pixel_mode;


     pthread_mutex_lock ( &library_mutex );

     face = data->face;
     /* NOTE: If this is being rendered to an A8 format, the outline/glyph can be merged into one bitmap ... */
     load_flags = (unsigned long) face->generic.data;
     //load_flags |= FT_LOAD_RENDER; 


     if ((err = FT_Load_Glyph( face, index, load_flags ))) {
          D_DEBUG( "DirectFB/FontFT2: Could not render glyph for character index #%d!\n", index );
          pthread_mutex_unlock ( &library_mutex );
          return DFB_FAILURE;
     }
       
     /* Use stroker object to render border onto glyph ... */
//#if 0

     if (
        (thiz->attributes & DFFA_BOLD) &&
        (thiz->desc_flags & DFDESC_ATTRIBUTES)
        ) 
     {
          New_GlyphSlot_Embolden(face->glyph);
     }

     if (data->outline_render && info->layer == 1)
     {
         
         /* 
         *  This FT_Get_Glyph used to be before the if statement above,
         *  however it was never released if we were not doing outline rendering,
         *  which could cause memory leak issues, so only run Get_Glyph here
         *  if we are to extract a stroked bitmap...
         */
         /* Extract the glyph from the structure ... */
         if ( (err = FT_Get_Glyph(face->glyph, &ft_glyph)) )
         {
              D_DEBUG( "DirectFB/FontFT2: Could not get the glyph!!\n");
              pthread_mutex_unlock ( &library_mutex );
              return DFB_FAILURE;
         }
         
         D_DEBUG("Outline render enter...\n");
         /* Print out the bitmap before and the bitmap after ... */
         ft_glyph_tmp   =   ft_glyph;
         err = FT_Glyph_Stroke( &ft_glyph, data->stroker, 1);
         if (err)
            return DFB_FAILURE; /* Fail this call for now ... */
         
         /* Convert our glyph to a bitmap ... */
         err = FT_Glyph_To_Bitmap(&ft_glyph, (load_flags & FT_LOAD_TARGET_MONO) ? ft_render_mode_mono : ft_render_mode_normal,
                                  NULL, 1);
         
         
                                  
         /* In the case of outlined fonts, the bitmap generated by this subroutine needs to be referenced from now on ... */
         if (err)
         {
            FT_Done_Glyph( ft_glyph);
            /* No need to release if there is no error as Freetype will do it ... */
            D_DEBUG("DirectFB/FontFT2: Could not convert glyph to bitmap!!\n");
            return DFB_FAILURE;
         }
         err = FT_Glyph_To_Bitmap(&ft_glyph, (load_flags & FT_LOAD_TARGET_MONO) ? ft_render_mode_mono : ft_render_mode_normal,
                                  NULL, 1);
                                  
                                  
         ft_bitmapglyph     =   (FT_BitmapGlyph)ft_glyph;   /* Typecast .. */
         //ft_bitmapglyph_tmp =   (FT_BitmapGlyph)ft_glyph_tmp;
         //printf("Stroked...\n");
         //print_bitmap(ft_bitmapglyph);
         //printf("Unstroked...\n");
         //print_bitmap(ft_bitmapglyph_tmp);
         
     }
     else
//#endif
     {
        FT_Render_Glyph(face->glyph, (load_flags & FT_LOAD_TARGET_MONO) ? ft_render_mode_mono : ft_render_mode_normal);

     }
     
     pthread_mutex_unlock ( &library_mutex );

     err = dfb_surface_lock_buffer( surface, CSBR_BACK, CSAID_CPU, CSAF_WRITE, &lock );
     if (err) {
          D_DERROR( err, "DirectFB/FontFT2: Unable to lock surface!\n" );
          return err;
     }
     
     /* Glyph information ... */
     if (!data->outline_render || info->layer == 0)
     {
         info->width = face->glyph->bitmap.width;
         info->height = face->glyph->bitmap.rows;
         info->left = face->glyph->bitmap_left;
         info->top  = thiz->ascender - face->glyph->bitmap_top;
         src = face->glyph->bitmap.buffer;
         D_DEBUG("Glyph from face..\n");
     }
     else
     {
         /* Grab the information from the glyph bitmap... */
         info->width    =   ft_bitmapglyph->bitmap.width;
         info->height   =   ft_bitmapglyph->bitmap.rows;
         
         info->left     =   ft_bitmapglyph->left;
         info->top      =   thiz->ascender - ft_bitmapglyph->top;
         //info->left     = face->glyph->bitmap_left;
         //info->top      = thiz->ascender - face->glyph->bitmap_top;
         src            =   ft_bitmapglyph->bitmap.buffer;
         D_DEBUG("Glyph from stroked bitmap...\n");
     
     }
     if (info->width + info->start > surface->config.size.w)
          info->width = surface->config.size.w - info->start;

     
     if (info->height > surface->config.size.h)
          info->height = surface->config.size.h;
     if (data->fixed_clip) {
          while (info->left + info->width > data->fixed_advance)
               info->left--;

          if (info->left < 0)
               info->left = 0;

          if (info->width > data->fixed_advance)
               info->width = data->fixed_advance;
     }
      
     lock.addr += DFB_BYTES_PER_LINE(surface->config.format, info->start);

     for (y=0; y < info->height; y++) {
          int  i, j, n;
          u8  *dst8  = lock.addr;
          u16 *dst16 = lock.addr;
          u32 *dst32 = lock.addr;
          
          /* Set the correct pixel mode ... */
          if (!data->outline_render || info->layer == 0)    
            ft_pixel_mode   =   face->glyph->bitmap.pixel_mode;
          else                              
            ft_pixel_mode   =   ft_bitmapglyph->bitmap.pixel_mode;
          
          switch (ft_pixel_mode) {
               case ft_pixel_mode_grays:
                    switch (surface->config.format) {
                         case DSPF_ARGB:
                              if (thiz->surface_caps & DSCAPS_PREMULTIPLIED) {
                                   for (i=0; i<info->width; i++)
                                        dst32[i] = src[i] * 0x01010101;
                              }
                              else
                                   for (i=0; i<info->width; i++)
                                        dst32[i] = (src[i] << 24) | 0xFFFFFF;
                              break;
                         case DSPF_AiRGB:
                              for (i=0; i<info->width; i++)
                                   dst32[i] = ((src[i] ^ 0xFF) << 24) | 0xFFFFFF;
                              break;
                         case DSPF_ARGB8565:
                              for (i = 0, j = -1; i < info->width; ++i) {
                                  u32 d;
                                  if (thiz->surface_caps & DSCAPS_PREMULTIPLIED) {
                                       d = src[i] * 0x01010101;
                                       d = ARGB_TO_ARGB8565 (d);
                                  }
                                  else
                                       d = (src[i] << 16) | 0xFFFF;
#ifdef WORDS_BIGENDIAN
                                   dst8[++j] = (d >> 16) & 0xff;
                                   dst8[++j] = (d >>  8) & 0xff;
                                   dst8[++j] = (d >>  0) & 0xff;
#else
                                   dst8[++j] = (d >>  0) & 0xff;
                                   dst8[++j] = (d >>  8) & 0xff;
                                   dst8[++j] = (d >> 16) & 0xff;
#endif
                              }
                              break;
                         case DSPF_ARGB4444:
                         case DSPF_RGBA4444:
                              if (thiz->surface_caps & DSCAPS_PREMULTIPLIED) {
                                   for (i=0; i<info->width; i++)
                                        dst16[i] = (src[i] >> 4) * 0x1111;
                              }
                              else {
                                   if( surface->config.format == DSPF_ARGB4444 ) {
                                        for (i=0; i<info->width; i++)
                                             dst16[i] = (src[i] << 8) | 0x0FFF;
                                   } else {
                                        for (i=0; i<info->width; i++)
                                             dst16[i] = (src[i] >> 4) | 0xFFF0;
                                   }
                              }
                              break;
                         case DSPF_ARGB2554:
                              for (i=0; i<info->width; i++)
                                   dst16[i] = (src[i] << 8) | 0x3FFF;
                              break;
                         case DSPF_ARGB1555:
                              if (thiz->surface_caps & DSCAPS_PREMULTIPLIED) {
                                   for (i=0; i<info->width; i++) {
                                        unsigned short x = src[i] >> 3;
                                        dst16[i] = ((src[i] & 0x80) << 8) |
                                             (x << 10) | (x << 5) | x;
                                   }
                              }
                              else {
                                   for (i=0; i<info->width; i++)
                                        dst16[i] = (src[i] << 8) | 0x7FFF;
                              }
                              break;
                         case DSPF_RGBA5551:
                              if (thiz->surface_caps & DSCAPS_PREMULTIPLIED) {
                                   for (i=0; i<info->width; i++) {
                                        unsigned short x = src[i] >> 3;
                                        dst16[i] =  (x << 11) | (x << 6) | (x << 1) |
                                             (src[i] >> 7);
                                   }
                              }
                              else {
                                   for (i=0; i<info->width; i++)
                                        dst16[i] = 0xFFFE | (src[i] >> 7);
                              }
                              break;
                         case DSPF_A8:
                              direct_memcpy( lock.addr, src, info->width );
                              break;
                         case DSPF_A4:
                              for (i=0, j=0; i<info->width; i+=2, j++)
                                   dst8[j] = (src[i] & 0xF0) | (src[i+1] >> 4);
                              break;
                         case DSPF_A1:
                              for (i=0, j=0; i < info->width; ++j) {
                                   register u8 p = 0;

                                   for (n=0; n<8 && i<info->width; ++i, ++n)
                                        p |= (src[i] & 0x80) >> n;

                                   dst8[j] = p;
                              }
                              break;
                         case DSPF_A1_LSB:
                              for (i=0, j=0; i < info->width; ++j) {
                                   register u8 p = 0;

                                   for (n=0; n<8 && i<info->width; ++i, ++n)
                                        p |= (src[i] & 0x80) >> (7-n);

                                   dst8[j] = p;
                              }
                              break;
                         case DSPF_LUT2:
                              for (i=0, j=0; i < info->width; ++j) {
                                   register u8 p = 0;

                                   for (n=0; n<8 && i<info->width; ++i, n+=2)
                                        p |= (src[i] & 0xC0) >> n;

                                   dst8[j] = p;
                              }
                              break;
                         default:
                              D_UNIMPLEMENTED();
                              break;
                    }
                    break;

               case ft_pixel_mode_mono:
                    switch (surface->config.format) {
                         case DSPF_ARGB:
                              for (i=0; i<info->width; i++)
                                   dst32[i] = (((src[i>>3] & (1<<(7-(i%8)))) ?
                                                0xFF : 0x00) << 24) | 0xFFFFFF;
                              break;
                         case DSPF_AiRGB:
                              for (i=0; i<info->width; i++)
                                   dst32[i] = (((src[i>>3] & (1<<(7-(i%8)))) ?
                                                0x00 : 0xFF) << 24) | 0xFFFFFF;
                              break;
                         case DSPF_ARGB8565:
                              for (i = 0, j = -1; i < info->width; ++i) {
                                   u32 d;
                                   if (thiz->surface_caps & DSCAPS_PREMULTIPLIED) {
                                        d = ((src[i>>3] & (1<<(7-(i%8)))) ?
                                             0xffffff : 0x000000);
                                   }
                                   else
                                       d = (((src[i>>3] & (1<<(7-(i%8)))) ?
                                             0xff : 0x00) << 16) | 0xffff;
#ifdef WORDS_BIGENDIAN
                                   dst8[++j] = (d >> 16) & 0xff;
                                   dst8[++j] = (d >>  8) & 0xff;
                                   dst8[++j] = (d >>  0) & 0xff;
#else
                                   dst8[++j] = (d >>  0) & 0xff;
                                   dst8[++j] = (d >>  8) & 0xff;
                                   dst8[++j] = (d >> 16) & 0xff;
#endif
                              }
                              break;
                         case DSPF_ARGB4444:
                              for (i=0; i<info->width; i++)
                                   dst16[i] = (((src[i>>3] & (1<<(7-(i%8)))) ?
                                                0xF : 0x0) << 12) | 0xFFF;
                              break;
                         case DSPF_RGBA4444:
                              for (i=0; i<info->width; i++)
                                   dst16[i] = (((src[i>>3] & (1<<(7-(i%8)))) ?
                                                0xF : 0x0)      ) | 0xFFF0;
                              break;
                         case DSPF_ARGB2554:
                              for (i=0; i<info->width; i++)
                                   dst16[i] = (((src[i>>3] & (1<<(7-(i%8)))) ?
                                                0x3 : 0x0) << 14) | 0x3FFF;
                              break;
                         case DSPF_ARGB1555:
                              for (i=0; i<info->width; i++)
                                   dst16[i] = (((src[i>>3] & (1<<(7-(i%8)))) ?
                                                0x1 : 0x0) << 15) | 0x7FFF;
                              break;
                         case DSPF_RGBA5551:
                              for (i=0; i<info->width; i++)
                                   dst16[i] = 0xFFFE | (((src[i>>3] & (1<<(7-(i%8)))) ?
                                                0x1 : 0x0));
                              break;
                         case DSPF_A8:
                              for (i=0; i<info->width; i++)
                              {
                                   dst8[i] = (src[i>>3] &
                                              (1<<(7-(i%8)))) ? 0xFF : 0x00;
#if PRINTF_GLYPHDATA
                                   if (dst8[i] == 0xFF) {
                                        printf("%02x ", dst8[i]);
                                   }
                                   else
                                   {
                                        printf("   ");
                                   }
#endif

                              }

                              break;
                         case DSPF_A4:
                              for (i=0, j=0; i<info->width; i+=2, j++)
                                   dst8[j] = ((src[i>>3] &
                                              (1<<(7-(i%8)))) ? 0xF0 : 0x00) |
                                             ((src[(i+1)>>3] &
                                              (1<<(7-((i+1)%8)))) ? 0x0F : 0x00);
                              break;
                         case DSPF_A1:
                              direct_memcpy( lock.addr, src, DFB_BYTES_PER_LINE(DSPF_A1, info->width) );
                              break;
                         case DSPF_A1_LSB:
                              for (i=0, j=0; i < info->width; ++j) {
                                   register u8 p = 0;

                                   for (n=0; n<8 && i<info->width; ++i, ++n)
                                        p |= (((src[i] >> n) & 1) << (7-n));

                                   dst8[j] = p;
                              }
                              break;
                         default:
                              D_UNIMPLEMENTED();
                              break;
                    }
                    break;

               default:
                    break;

          }
          
        if (!data->outline_render || info->layer == 0)
        {
            src += face->glyph->bitmap.pitch;
        }
        else
        {
            src += ft_bitmapglyph->bitmap.pitch;
        }
          lock.addr += lock.pitch;
     }
     
     /* Call FT_DoneGlyph once we are finished to release the resources associated with the glyph ... */
     if (data->outline_render && info->layer == 1)
     {
        /* Adding back locking functions ... */
        pthread_mutex_lock ( &library_mutex );
        FT_Done_Glyph(ft_glyph);
        pthread_mutex_unlock ( &library_mutex );
     }
     dfb_surface_unlock_buffer( surface, &lock );

     return DFB_OK;
}


static DFBResult
get_glyph_info( CoreFont      *thiz,
                unsigned int   index,
                CoreGlyphData *info )
{
     FT_Error       err;
     FT_Face        face;
     FT_Int         load_flags;
     FT_Glyph       ft_glyph;
     FT_BitmapGlyph ft_bitmapglyph;
     
     FT2ImplData *data = (FT2ImplData*) thiz->impl_data;

     pthread_mutex_lock ( &library_mutex );

     face = data->face;

     load_flags = (unsigned long) face->generic.data;

     if ((err = FT_Load_Glyph( face, index, load_flags ))) {
          D_DEBUG( "DirectFB/FontFT2: Could not load glyph for character index #%d!\n", index );

          pthread_mutex_unlock ( &library_mutex );

          return DFB_FAILURE;
     }
     
    
     if (face->glyph->format != ft_glyph_format_bitmap || (data->outline_render && info->layer == 1)) {
          
          if (data->outline_render && info->layer == 1)
          {
              err = FT_Get_Glyph( face->glyph, &ft_glyph);
              if (err) 
              {
                D_ERROR( "DirectFB/FontFT2: Could not get the glyph object.\n");
                pthread_mutex_unlock ( &library_mutex );
                
                return DFB_FAILURE;
              }
              
              /* Change to erase the original glyph */
              FT_Glyph_Stroke( &ft_glyph, data->stroker, 1);
              
              err = FT_Glyph_To_Bitmap(&ft_glyph, (load_flags & FT_LOAD_TARGET_MONO) ? ft_render_mode_mono : ft_render_mode_normal,
                                  NULL, 1);
              
              ft_bitmapglyph    =   (FT_BitmapGlyph)ft_glyph;
          }
          else
          {
          /* It is assumed this subroutine is used to fetch information about the font for planning ... */
            err = FT_Render_Glyph( face->glyph,
                                    (load_flags & FT_LOAD_TARGET_MONO) ? ft_render_mode_mono : ft_render_mode_normal );
          }
          
          
          if (err) {
               D_ERROR( "DirectFB/FontFT2: Could not render glyph for character index #%d!\n", index );

               pthread_mutex_unlock ( &library_mutex );

               return DFB_FAILURE;
          }
     }

     pthread_mutex_unlock ( &library_mutex );
     if (!data->outline_render || info->layer == 0)
     {
         info->width   = face->glyph->bitmap.width;
         info->height  = face->glyph->bitmap.rows;
         info->advance = data->fixed_advance ?
                         data->fixed_advance : (face->glyph->advance.x >> 6);

         if (data->fixed_clip && info->width > data->fixed_advance)
              info->width = data->fixed_advance;

     }
     else
     {
         info->width    =   ft_bitmapglyph->bitmap.width;
         info->height   =   ft_bitmapglyph->bitmap.rows;
         //info->advance  =   data->fixed_advance ?
         //                   data->fixed_advance : (ft_glyph->advance.x >> 6);
         info->advance = data->fixed_advance ?
                         data->fixed_advance : (face->glyph->advance.x >> 6);
                         
         if (data->fixed_clip && info->width > data->fixed_advance)
              info->width = data->fixed_advance;
     }
     
     /* Call FT_DoneGlyph once we are finished to release the resources associated with the glyph ... */
     if (data->outline_render && info->layer == 1)
     {
        /* Adding back locking functions ... */
        pthread_mutex_lock ( &library_mutex );
        FT_Done_Glyph(ft_glyph);
        pthread_mutex_unlock ( &library_mutex );
     }
     
     
     return DFB_OK;
}


static DFBResult
get_kerning( CoreFont     *thiz,
             unsigned int  prev,
             unsigned int  current,
             int          *kern_x,
             int          *kern_y)
{
     FT_Vector vector;

     FT2ImplKerningData *data = thiz->impl_data;
     KerningCacheEntry *cache = NULL;

     D_ASSUME( (kern_x != NULL) || (kern_y != NULL) );

     /*
      * Use cached values if characters are in the
      * cachable range and the cache entry is already filled.
      */
     if (KERNING_DO_CACHE (prev, current)) {
          cache = &KERNING_CACHE_ENTRY (prev, current);

          if (kern_x)
               *kern_x = (int) cache->x;

          if (kern_y)
               *kern_y = (int) cache->y;

          return DFB_OK;
     }

     pthread_mutex_lock ( &library_mutex );

     /* Lookup kerning values for the character pair. */
     FT_Get_Kerning( data->base.face,
                     prev, current, ft_kerning_default, &vector );

     pthread_mutex_unlock ( &library_mutex );

     /* Convert to integer. */
     if (kern_x)
          *kern_x = vector.x >> 6;

     if (kern_y)
          *kern_y = vector.y >> 6;

     return DFB_OK;
}

static void
init_kerning_cache( FT2ImplKerningData *data )
{
     int a, b;

     pthread_mutex_lock ( &library_mutex );

     for (a=KERNING_CACHE_MIN; a<=KERNING_CACHE_MAX; a++) {
          for (b=KERNING_CACHE_MIN; b<=KERNING_CACHE_MAX; b++) {
               FT_Vector          vector;
               KerningCacheEntry *cache = &KERNING_CACHE_ENTRY( a, b );

               /* Lookup kerning values for the character pair. */
               FT_Get_Kerning( data->base.face,
                               a, b, ft_kerning_default, &vector );

               cache->x = (signed char) (vector.x >> 6);
               cache->y = (signed char) (vector.y >> 6);
          }
     }

     pthread_mutex_unlock ( &library_mutex );
}

static DFBResult
init_freetype( void )
{
     FT_Error err;

     pthread_mutex_lock ( &library_mutex );

     if (!library) {
          D_DEBUG( "DirectFB/FontFT2: Initializing the FreeType2 library.\n" );
          err = FT_Init_FreeType( &library );
          if (err) {
               D_ERROR( "DirectFB/FontFT2: "
                         "Initialization of the FreeType2 library failed!\n" );
               library = NULL;
               pthread_mutex_unlock( &library_mutex );
               return DFB_FAILURE;
          }
     }

     library_ref_count++;
     pthread_mutex_unlock( &library_mutex );

     return DFB_OK;
}


static void
release_freetype( void )
{
     pthread_mutex_lock( &library_mutex );

     if (library && --library_ref_count == 0) {
          D_DEBUG( "DirectFB/FontFT2: Releasing the FreeType2 library.\n" );
          FT_Done_FreeType( library );
          library = NULL;
     }

     pthread_mutex_unlock( &library_mutex );
}


static void
IDirectFBFont_FT2_Destruct( IDirectFBFont *thiz )
{
     IDirectFBFont_data *data = (IDirectFBFont_data*)thiz->priv;

     if (data->font->impl_data) {
          FT2ImplData *impl_data = (FT2ImplData*) data->font->impl_data;

          pthread_mutex_lock ( &library_mutex );
          FT_Stroker_Done( impl_data->stroker);
          FT_Done_Face( impl_data->face );
          pthread_mutex_unlock ( &library_mutex );

          D_FREE( impl_data );

          data->font->impl_data = NULL;
     }

     IDirectFBFont_Destruct( thiz );

     release_freetype();
}


static DirectResult
IDirectFBFont_FT2_Release( IDirectFBFont *thiz )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBFont)

     if (--data->ref == 0) {
          IDirectFBFont_FT2_Destruct( thiz );
     }

     return DFB_OK;
}


static DFBResult
Probe( IDirectFBFont_ProbeContext *ctx )
{
     FT_Error err;
     FT_Face  face;

     D_DEBUG( "DirectFB/FontFT2: Probe font `%s'.\n", ctx->filename );

     if ((!ctx->filename) &&
         (!ctx->data_buffer))
          return DFB_UNSUPPORTED;

     if (init_freetype() != DFB_OK) {
          return DFB_FAILURE;
     }

     pthread_mutex_lock ( &library_mutex );
     /*
      * This should be
      * err = FT_New_Face( library, ctx->filename, -1, NULL );
      * but due to freetype bugs it doesn't work.
      */

     if (ctx->filename) 
     {
          err = FT_New_Face( library, ctx->filename, 0, &face );
     }
     else if (ctx->data_buffer)
     {
          err = FT_New_Memory_Face( library, ctx->data_buffer, ctx->data_buffer_size, 0, &face );
     }
     else
     {
          err = 1;
     }

     if (!err)
          FT_Done_Face( face );
     pthread_mutex_unlock ( &library_mutex );

     release_freetype();

     return err ? DFB_UNSUPPORTED : DFB_OK;
}


static DFBResult
Construct( IDirectFBFont *thiz,
	   ... )
{
     int                 i;
     DFBResult           ret;
     CoreFont           *font;
     FT_Face             face;
     FT_Error            err;
     FT_Int              load_flags = FT_LOAD_DEFAULT;
     FT2ImplData        *data;
     bool                disable_charmap = false;
     bool                disable_kerning = false;
     bool                load_mono = false;
     bool                outline_render = false;
     int                 outline_width;
     int                 outline_opacity;
     u32                 mask = 0;
     CoreDFB            *core;
     char               *filename;
     DFBFontDescription *desc;
     IDirectFBDataBuffer * font_buffer;

     va_list tag;
     va_start(tag, thiz);
     core = va_arg(tag, CoreDFB *);
     filename = va_arg(tag, char *);
     desc = va_arg(tag, DFBFontDescription *);
     va_end( tag );

     D_DEBUG( "DirectFB/FontFT2: "
              "Construct font from file `%s' (index %d) at pixel size %d x %d.\n",
              filename,
              (desc->flags & DFDESC_INDEX)  ? desc->index  : 0,
              (desc->flags & DFDESC_WIDTH)  ? desc->width  : 0,
              (desc->flags & DFDESC_HEIGHT) ? desc->height : 0 );

     if (init_freetype() != DFB_OK) {
          DIRECT_DEALLOCATE_INTERFACE( thiz );
          return DFB_FAILURE;
     }

     pthread_mutex_lock ( &library_mutex );

     if (desc->flags & DFDESC_FROM_BUFFER) 
     {

         err = FT_New_Memory_Face( library,
                              desc->data_buffer,    /* first byte in memory */
                              desc->data_buffer_size,      /* size in bytes        */
                              0,         /* face_index           */
                              &face );
     }
     else
     {
         err = FT_New_Face( library, filename,
                            (desc->flags & DFDESC_INDEX) ? desc->index : 0,
                            &face );
     }

     pthread_mutex_unlock ( &library_mutex );
     if (err) {
          switch (err) {
               case FT_Err_Unknown_File_Format:
                    D_ERROR( "DirectFB/FontFT2: "
                              "Unsupported font format in file `%s'!\n", filename );
                    break;
               default:
                    D_ERROR( "DirectFB/FontFT2: "
                              "Failed loading face %d from font file `%s'!\n",
                              (desc->flags & DFDESC_INDEX) ? desc->index : 0,
                              filename );
                    break;
          }
          DIRECT_DEALLOCATE_INTERFACE( thiz );
          return DFB_FAILURE;
     }

     if (dfb_config->font_format == DSPF_A1 ||
               dfb_config->font_format == DSPF_A1_LSB || 
               dfb_config->font_format == DSPF_ARGB1555 ||
               dfb_config->font_format == DSPF_RGBA5551)
          load_mono = true;

     if (desc->flags & DFDESC_ATTRIBUTES) {
          if (desc->attributes & DFFA_NOHINTING)
               load_flags |= FT_LOAD_NO_HINTING;
          if (desc->attributes & DFFA_NOBITMAP)
               load_flags |= FT_LOAD_NO_BITMAP;
          if (desc->attributes & DFFA_NOCHARMAP)
               disable_charmap = true;
          if (desc->attributes & DFFA_NOKERNING)
               disable_kerning = true;
          if (desc->attributes & DFFA_MONOCHROME)
               load_mono = true;
          if (desc->attributes & DFFA_OUTLINED)
               outline_render   = true;
     }

     if (load_mono)
          load_flags |= FT_LOAD_TARGET_MONO;

     if (!disable_charmap) {
          pthread_mutex_lock ( &library_mutex );
          err = FT_Select_Charmap( face, ft_encoding_unicode );
          pthread_mutex_unlock ( &library_mutex );

#if FREETYPE_MINOR > 0

          /* ft_encoding_latin_1 has been introduced in freetype-2.1 */
          if (err) {
               D_DEBUG( "DirectFB/FontFT2: "
                        "Couldn't select Unicode encoding, "
                        "falling back to Latin1.\n");
               pthread_mutex_lock ( &library_mutex );
               err = FT_Select_Charmap( face, ft_encoding_latin_1 );
               pthread_mutex_unlock ( &library_mutex );
          }
#endif
          if (err) {
               D_DEBUG( "DirectFB/FontFT2: "
                        "Couldn't select Unicode/Latin1 encoding, "
                        "trying Symbol.\n");
               pthread_mutex_lock ( &library_mutex );
               err = FT_Select_Charmap( face, ft_encoding_symbol );
               pthread_mutex_unlock ( &library_mutex );

               if (!err)
                    mask = 0xf000;
          }
     }

#if 0
     if (err) {
          D_ERROR( "DirectFB/FontFT2: "
                    "Couldn't select a suitable encoding for face %d from font file `%s'!\n", (desc->flags & DFDESC_INDEX) ? desc->index : 0, filename );
          pthread_mutex_lock ( &library_mutex );
          FT_Done_Face( face );
          pthread_mutex_unlock ( &library_mutex );
          DIRECT_DEALLOCATE_INTERFACE( thiz );
          return DFB_FAILURE;
     }
#endif

     if (desc->flags & (DFDESC_HEIGHT       | DFDESC_WIDTH |
                        DFDESC_FRACT_HEIGHT | DFDESC_FRACT_WIDTH))
     {
          int fw = 0, fh = 0;

          if (desc->flags & DFDESC_FRACT_HEIGHT)
               fh = desc->fract_height;
          else if (desc->flags & DFDESC_HEIGHT)
               fh = desc->height << 6;

          if (desc->flags & DFDESC_FRACT_WIDTH)
               fw = desc->fract_width;
          else if (desc->flags & DFDESC_WIDTH)
               fw = desc->width << 6;

          pthread_mutex_lock ( &library_mutex );
          err = FT_Set_Char_Size( face, fw, fh, 0, 0 );
          pthread_mutex_unlock ( &library_mutex );
          if (err) {
               D_ERROR( "DirectB/FontFT2: "
                         "Could not set pixel size to %d x %d!\n",
                         (desc->flags & DFDESC_WIDTH)  ? desc->width  : 0,
                         (desc->flags & DFDESC_HEIGHT) ? desc->height : 0 );
               pthread_mutex_lock ( &library_mutex );
               FT_Done_Face( face );
               pthread_mutex_unlock ( &library_mutex );
               DIRECT_DEALLOCATE_INTERFACE( thiz );
               return DFB_FAILURE;
          }
     }

     face->generic.data = (void *)(unsigned long) load_flags;
     face->generic.finalizer = NULL;

     ret = dfb_font_create( core, &font );
     if (ret) {
          pthread_mutex_lock ( &library_mutex );
          FT_Done_Face( face );
          pthread_mutex_unlock ( &library_mutex );
          DIRECT_DEALLOCATE_INTERFACE( thiz );
          return ret;
     }

     D_ASSERT( font->pixel_format == DSPF_ARGB ||
               font->pixel_format == DSPF_AiRGB ||
               font->pixel_format == DSPF_ARGB8565 ||
               font->pixel_format == DSPF_ARGB4444 ||
               font->pixel_format == DSPF_RGBA4444 ||
               font->pixel_format == DSPF_ARGB2554 ||
               font->pixel_format == DSPF_ARGB1555 ||
               font->pixel_format == DSPF_RGBA5551 ||
               font->pixel_format == DSPF_A8 ||
               font->pixel_format == DSPF_A4 ||
               font->pixel_format == DSPF_A1_LSB ||
               font->pixel_format == DSPF_A1 );


     font->ascender   = face->size->metrics.ascender >> 6;
     font->descender  = face->size->metrics.descender >> 6;
     //font->height     = font->ascender + ABS(font->descender) + 1;
     font->height     = (face->size->metrics.height >> 6) + 1;
     font->maxadvance = face->size->metrics.max_advance >> 6;
     font->current_encoding = DTEID_UTF8;
     font->disabled = DFXL_NONE;

     D_DEBUG( "DirectFB/FontFT2: height = %d, ascender = %d, descender = %d, maxadvance = %d\n",
              font->height, font->ascender, font->descender, font->maxadvance );

     font->GetGlyphData = get_glyph_info;
     font->RenderGlyph  = render_glyph;

     if (FT_HAS_KERNING(face) && !disable_kerning) {
          font->GetKerning = get_kerning;
          data = D_CALLOC( 1, sizeof(FT2ImplKerningData) );
     }
     else
          data = D_CALLOC( 1, sizeof(FT2ImplData) );

     data->face            = face;
     data->disable_charmap = disable_charmap;
     data->outline_render  = outline_render;
     data->stroker         = NULL;
     /* outline_width and outline_opacity need to be set based on the flags ... */
     if (outline_render)
     {
        if (desc->flags & DFDESC_OUTLINE_WIDTH)
        {
            /* Outline Width ... */
            data->outline_width =   desc->outline_width;
        }
        else
        {
            /* Default to a width of 1 */
            data->outline_width =   FT2_DEFAULT_OUTLINE_WIDTH;
        }
        
        if (desc->flags & DFDESC_OUTLINE_OPACITY)
        {
            /* Opacity ...*/
            data->outline_opacity =   desc->outline_opacity;
           
        }
        else
        {
            /* Default to full opacity ... */
            data->outline_opacity =   FT2_DEFAULT_OUTLINE_OPACITY;
        }
        pthread_mutex_lock ( &library_mutex );
        /* Create FT_Stroker object ... */
        err = FT_Stroker_New( library, &data->stroker);
        if (err)
        {
            D_DEBUG("DirectFB/FontFT2: Failed to create stroker object! \n");
            return DFB_FAILURE;
        }
        /* In addition, set the parameters for the stroker interface ... */
       FT_Stroker_Set( data->stroker,
                        (FT_Fixed)(data->outline_width >> 16),  /* Use upper 8-bits ... */
                        FT_STROKER_LINECAP_ROUND,
                        FT_STROKER_LINEJOIN_ROUND,
                        0
                       );
        //if (err)
        //{
        //    D_DEBUG("DirectFB/FontFT2: Failed to set stroker object! \n");
        //    return DFB_FAILURE;
        //}
        
        pthread_mutex_unlock ( &library_mutex );
        D_DEBUG("DirectFB/FontFT2: Outline Width: 0x%X Outline Opacity: 0x%X \n",
                (desc->outline_width >> 16) , desc->outline_opacity );
        
        
     }

     if (FT_HAS_KERNING(face) && !disable_kerning)
          init_kerning_cache( (FT2ImplKerningData*) data );

     if (desc->flags & DFDESC_FIXEDADVANCE) {
          data->fixed_advance = desc->fixed_advance;
          font->maxadvance    = desc->fixed_advance;

          if ((desc->flags & DFDESC_ATTRIBUTES) && (desc->attributes & DFFA_FIXEDCLIP))
               data->fixed_clip = true;
     }

     for (i=0; i<256; i++)
          data->indices[i] = FT_Get_Char_Index( face, i | mask );

     font->impl_data = data;

     dfb_font_register_encoding( font, "UTF8",   &ft2UTF8Funcs,   DTEID_UTF8 );
     dfb_font_register_encoding( font, "Latin1", &ft2Latin1Funcs, DTEID_OTHER );
     
     /* ViXS: Transfer this over or outlined draw calls will fail ...*/
     //if (desc->flags & DFDESC_ATTRIBUTES)
     font->attributes   =   desc->attributes;
     font->desc_flags   =   desc->flags;
     //else
     //   font->attributes    =   0;

     if ( (font->attributes & DFFA_ITALIC) &&
          (font->desc_flags & DFDESC_ATTRIBUTES)  
        ) 
     {
         float lean = 0.3f;
         FT_Matrix matrix;
         matrix.xx = 0x10000L;
         matrix.xy = lean * 0x10000L;
         matrix.yx = 0;
         matrix.yy = 0x10000L;
         FT_Set_Transform( face, &matrix, 0 );
     }


     IDirectFBFont_Construct( thiz, font );

     thiz->Release = IDirectFBFont_FT2_Release;

     return DFB_OK;
}


static void print_bitmap(FT_BitmapGlyph pglyph)
{
    /* Print this bitmap out ... */
#ifndef  FT_BITMAP_DEBUG
    return;
#endif
    unsigned char *pbuffer  =   pglyph->bitmap.buffer;
    int   rows, width, pitch; 
    int   num_grays;
    int   i, j;
    rows        =   pglyph->bitmap.rows;
    width       =   pglyph->bitmap.width;
    pitch       =   pglyph->bitmap.pitch;
    num_grays   =   pglyph->bitmap.num_grays;
    
    printf("print_bitmap ...\n");
    
    printf("Glyph: Rows %u, Width %u, Pitch %u\n", 
            rows, width, pitch);
            
    printf("Num_grays: %u\n", num_grays);
    
    
    /* Print bitmap ... */
    
    for (i = 0 ; i < rows ; i++)
    {
        for (j = 0 ; j < width ; j++)
        {
            printf("%X ", pbuffer[(i*pitch) + j]);
        }
        printf("\n");
    
    }
    
    return;
}



 // Vertical bolden function
FT_Error Vert_FT_Outline_Embolden( FT_Outline*  outline, FT_Pos strength )
{
    FT_Vector*    points;
    FT_Vector    v_prev, v_first, v_next, v_cur;
    FT_Angle    rotate, angle_in, angle_out;
    FT_Int        c, n, first;
    FT_Int        orientation;

    if ( !outline )
        return FT_Err_Invalid_Argument;

    strength /= 2;
    if ( strength == 0 )
        return FT_Err_Ok;

    orientation = FT_Outline_Get_Orientation( outline );
    if ( orientation == FT_ORIENTATION_NONE )
    {
        if ( outline->n_contours )
            return FT_Err_Invalid_Argument;
        else
            return FT_Err_Ok;
    }

    if ( orientation == FT_ORIENTATION_TRUETYPE )
        rotate = -FT_ANGLE_PI2;
    else
        rotate = FT_ANGLE_PI2;

    points = outline->points;

    first = 0;
    for ( c = 0; c < outline->n_contours; c++ )
    {
        int  last = outline->contours[c];

        v_first = points[first];
        v_prev  = points[last];
        v_cur   = v_first;

        for ( n = first; n <= last; n++ )
        {
            FT_Vector    in, out;
            FT_Angle    angle_diff;
            FT_Pos        d;
            FT_Fixed    scale;

            if ( n < last )
                v_next = points[n + 1];
            else
                v_next = v_first;

             /**//* compute the in and out vectors */
            in.x = v_cur.x - v_prev.x;
            in.y = v_cur.y - v_prev.y;

            out.x = v_next.x - v_cur.x;
            out.y = v_next.y - v_cur.y;

            angle_in   = FT_Atan2( in.x, in.y );
            angle_out  = FT_Atan2( out.x, out.y );
            angle_diff = FT_Angle_Diff( angle_in, angle_out );
            scale      = FT_Cos( angle_diff / 2 );

            if ( scale < 0x4000L && scale > -0x4000L )
                in.x = in.y = 0;
            else
              {
                d = FT_DivFix( strength, scale );

                FT_Vector_From_Polar( &in, d, angle_in + angle_diff / 2 - rotate );
            }

            //outline->points[n].x = v_cur.x + strength + in.x;
            //???????????????
            outline->points[n].y = v_cur.y + strength + in.y;

            v_prev = v_cur;
            v_cur  = v_next;
        }

        first = last + 1;
    }

    return FT_Err_Ok;
}

//this is new bolden function
FT_Error New_FT_Outline_Embolden( FT_Outline*  outline, FT_Pos str_h, FT_Pos str_v )
{
     if ( !outline ) return FT_Err_Invalid_Argument;
     int orientation = FT_Outline_Get_Orientation( outline );
     if ( orientation == FT_ORIENTATION_NONE )
         if ( outline->n_contours ) return FT_Err_Invalid_Argument;
     Vert_FT_Outline_Embolden( outline, str_v );
     FT_Outline_Embolden( outline, str_h );
     return FT_Err_Ok;
}

 // make glyp sloe be embolder, fill the other attributes
 void New_GlyphSlot_Embolden( FT_GlyphSlot  slot)
 {
     FT_Library  library = slot->library;
     FT_Face     face    = slot->face;
     FT_Error    error;
     FT_Pos      xstr, ystr;
 
     if ( slot->format != FT_GLYPH_FORMAT_OUTLINE &&
         slot->format != FT_GLYPH_FORMAT_BITMAP )
         return;

     /* some reasonable strength */
     xstr = FT_MulFix( face->units_per_EM,
                       face->size->metrics.y_scale ) / 24;
     ystr = xstr;

 
     if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
       {
         FT_BBox oldBox;
         FT_Outline_Get_CBox(&slot->outline , &oldBox);
         error = New_FT_Outline_Embolden( &slot->outline, xstr , ystr);
         if ( error )
             return;
 
         FT_BBox newBox;
         FT_Outline_Get_CBox(&slot->outline , &newBox);
         xstr = (newBox.xMax - newBox.xMin) - (oldBox.xMax - oldBox.xMin);
         ystr = (newBox.yMax - newBox.yMin) - (oldBox.yMax - oldBox.yMin);
     }
     else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
       {
         xstr = MY_FT_PIX_FLOOR( xstr );
         if ( xstr == 0 )
             xstr = 1 << 6;
         ystr = MY_FT_PIX_FLOOR( ystr );
 
         error = FT_Bitmap_Embolden( library, &slot->bitmap, xstr, ystr );
         if ( error )
             return;
     }
 
     if ( slot->advance.x )
         slot->advance.x += xstr;
 
     if ( slot->advance.y )
         slot->advance.y += ystr;
 
     slot->metrics.width        += xstr;
     slot->metrics.height       += ystr;
     slot->metrics.horiBearingY += ystr;
     slot->metrics.horiAdvance  += xstr;
     slot->metrics.vertBearingX -= xstr / 2;
     slot->metrics.vertBearingY += ystr;
     slot->metrics.vertAdvance  += ystr;
 
     if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
         slot->bitmap_top += ystr >> 6;
 }








