/*
   (c) Copyright 2001-2009  The world wide DirectFB Open Source Community (directfb.org)
   (c) Copyright 2000-2004  Convergence (integrated media) GmbH

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>,
              Andreas Hundt <andi@fischlustig.de>,
              Sven Neumann <neo@directfb.org>,
              Ville Syrjälä <syrjala@sci.fi> and
              Claudio Ciccani <klan@users.sf.net>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

#include <directfb.h>

#include <display/idirectfbsurface.h>

#include <media/idirectfbimageprovider.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/layers.h>
#include <core/surface.h>

#include <misc/gfx_util.h>
#include <misc/util.h>
#include <direct/interface.h>
#include <direct/mem.h>
#include <direct/memcpy.h>
#include <direct/messages.h>

#include <setjmp.h>
#include <math.h>

#undef HAVE_STDLIB_H
#include "xcode4_img_module/xcode4_img_api.h"
#include "xcode4/xcdfb.h"
#include <core/core.h>
#include <fusion/arena.h>
#include <fusion/fusion.h>

#include "viper.h"
#include "xcode4_img_module/xcode4_img_ycbcr2argb.h"

#include "xcode4_img_module/xcode4_imageprovider_HW.h"

#include "xcode4_img_module/xcode4_img_util.h"


#if XCODE_DISABLE_HW_JPEG_DECODE
#warning "Xcode jpeg hw decoding is disabled!"
#endif



#define PIXELS_NUM_BETWEEN_EACH_SLEEP 8000


#define D_JPEG(...) D_DEBUG_AT( JPEG_T, __VA_ARGS__)

D_DEBUG_DOMAIN( JPEG_T, "jpeg", "JPEG decoding" );


static DFBResult
Probe( IDirectFBImageProvider_ProbeContext *ctx );

static DFBResult
Construct( IDirectFBImageProvider *thiz,
           ... );

#include <direct/interface_implementation.h>

DIRECT_INTERFACE_IMPLEMENTATION( IDirectFBImageProvider, JPEG )

/*
 * private data struct of IDirectFBImageProvider_JPEG
 */
typedef struct {
     int                  ref;      /* reference counter */

     IDirectFBDataBuffer *buffer;

     DIRenderCallback     render_callback;
     void                *render_callback_context;

     int                  width;    /*  width of the JPEG image   */
     int                  height;   /*  height of the JPEG image  */
     
     CoreDFB             *core;
     
     u32                 *image;        /*  decoded image data    */
     int                  image_width;  /*  width of decoded image data   */
     int                  image_height; /*  height of decoded image data  */

     CoreSurface         *surf;
     bool                hw_supported;
     DFBSurfacePixelFormat comp_colorspace;
} IDirectFBImageProvider_JPEG_data;



static DirectResult
IDirectFBImageProvider_JPEG_AddRef  ( IDirectFBImageProvider *thiz );

static DirectResult
IDirectFBImageProvider_JPEG_Release ( IDirectFBImageProvider *thiz );

static DFBResult
IDirectFBImageProvider_JPEG_RenderTo( IDirectFBImageProvider *thiz,
                                      IDirectFBSurface       *destination,
                                      const DFBRectangle     *destination_rect );

static DFBResult
IDirectFBImageProvider_JPEG_NoncacheRenderTo( IDirectFBImageProvider *thiz,
                                      IDirectFBSurface       *destination,
                                      const DFBRectangle     *destination_rect );

static DFBResult
IDirectFBImageProvider_JPEG_SetRenderCallback( IDirectFBImageProvider *thiz,
                                               DIRenderCallback        callback,
                                               void                   *context );

static DFBResult
IDirectFBImageProvider_JPEG_GetSurfaceDescription( IDirectFBImageProvider *thiz,
                                                   DFBSurfaceDescription  *dsc);

static DFBResult
IDirectFBImageProvider_JPEG_GetImageDescription( IDirectFBImageProvider *thiz,
                                                 DFBImageDescription    *dsc );


#define JPEG_PROG_BUF_SIZE    0x10000

typedef struct {
     struct jpeg_source_mgr  pub; /* public fields */

     JOCTET                 *data;       /* start of buffer */

     IDirectFBDataBuffer    *buffer;

     int                     peekonly;
     int                     peekoffset;
} buffer_source_mgr;

typedef buffer_source_mgr * buffer_src_ptr;



struct my_error_mgr {
     struct jpeg_error_mgr pub;     /* "public" fields */
     jmp_buf  setjmp_buffer;          /* for return to caller */
};

static void
jpeglib_panic(j_common_ptr cinfo)
{
     struct my_error_mgr *myerr = (struct my_error_mgr*) cinfo->err;
     longjmp(myerr->setjmp_buffer, 1);
}


static void
buffer_init_source (j_decompress_ptr cinfo)
{
     buffer_src_ptr src          = (buffer_src_ptr) cinfo->src;
     IDirectFBDataBuffer *buffer = src->buffer;

     buffer->SeekTo( buffer, 0 ); /* ignore return value */
}

static boolean
buffer_fill_input_buffer (j_decompress_ptr cinfo)
{
     DFBResult            ret;
     unsigned int         nbytes = 0;
     buffer_src_ptr       src    = (buffer_src_ptr) cinfo->src;
     IDirectFBDataBuffer *buffer = src->buffer;

     buffer->WaitForDataWithTimeout( buffer, JPEG_PROG_BUF_SIZE, 1, 0 );

     if (src->peekonly) {
          ret = buffer->PeekData( buffer, JPEG_PROG_BUF_SIZE,
                                  src->peekoffset, src->data, &nbytes );
          src->peekoffset += MAX( nbytes, 0 );
     }
     else {
          ret = buffer->GetData( buffer, JPEG_PROG_BUF_SIZE, src->data, &nbytes );
     }
     
     if (ret || nbytes <= 0) {
          /* Insert a fake EOI marker */
          src->data[0] = (JOCTET) 0xFF;
          src->data[1] = (JOCTET) JPEG_EOI;
          nbytes = 2;

          if (ret && ret != DFB_EOF) {
               DirectFBError( "(DirectFB/ImageProvider_JPEG) GetData failed", ret );
               
               //struct my_error_mgr *myerr = (struct my_error_mgr*) cinfo->err;
               //longjmp(myerr->setjmp_buffer, 1);
          }
     }

     src->pub.next_input_byte = src->data;
     src->pub.bytes_in_buffer = nbytes;

     return TRUE;
}

static void
buffer_skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
     buffer_src_ptr src = (buffer_src_ptr) cinfo->src;

     if (num_bytes > 0) {
          while (num_bytes > (long) src->pub.bytes_in_buffer) {
               num_bytes -= (long) src->pub.bytes_in_buffer;
               (void)buffer_fill_input_buffer(cinfo);
          }
          src->pub.next_input_byte += (size_t) num_bytes;
          src->pub.bytes_in_buffer -= (size_t) num_bytes;
     }
}

static void
buffer_term_source (j_decompress_ptr cinfo)
{
}

static void
jpeg_buffer_src (j_decompress_ptr cinfo, IDirectFBDataBuffer *buffer, int peekonly)
{
     buffer_src_ptr src;

     cinfo->src = (struct jpeg_source_mgr *)
                  cinfo->mem->alloc_small ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                                           sizeof (buffer_source_mgr));

     src = (buffer_src_ptr) cinfo->src;

     src->data = (JOCTET *)
                  cinfo->mem->alloc_small ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                                           JPEG_PROG_BUF_SIZE * sizeof (JOCTET));

     src->buffer = buffer;
     src->peekonly = peekonly;
     src->peekoffset = 0;

     src->pub.init_source       = buffer_init_source;
     src->pub.fill_input_buffer = buffer_fill_input_buffer;
     src->pub.skip_input_data   = buffer_skip_input_data;
     src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
     src->pub.term_source       = buffer_term_source;
     src->pub.bytes_in_buffer   = 0; /* forces fill_input_buffer on first read */
     src->pub.next_input_byte   = NULL; /* until buffer loaded */
}



static inline void
copy_line32( u32 *argb, const u8 *rgb, int width )
{
     while (width--) {
          *argb++ = 0xFF000000 | (rgb[0] << 16) | (rgb[1] << 8) | rgb[2];

          rgb += 3;
     }
}

static inline void
copy_line_nv16( u16 *yy, u16 *cbcr, const u8 *src_ycbcr, int width )
{
     int x;

     for (x=0; x<width/2; x++) {
#ifdef WORDS_BIGENDIAN
          yy[x] = (src_ycbcr[0] << 8) | src_ycbcr[3];

          cbcr[x] = (((src_ycbcr[1] + src_ycbcr[4]) << 7) & 0xff00) |
                     ((src_ycbcr[2] + src_ycbcr[5]) >> 1);
#else
          yy[x] = (src_ycbcr[3] << 8) | src_ycbcr[0];

          cbcr[x] = (((src_ycbcr[2] + src_ycbcr[5]) << 7) & 0xff00) |
                     ((src_ycbcr[1] + src_ycbcr[4]) >> 1);
#endif

          src_ycbcr += 6;
     }

     if (width & 1) {
          u8 *y = (u8*) yy;

          y[width-1] = src_ycbcr[0];

#ifdef WORDS_BIGENDIAN
          cbcr[x] = (src_ycbcr[1] << 8) | src_ycbcr[2];
#else
          cbcr[x] = (src_ycbcr[2] << 8) | src_ycbcr[1];
#endif
     }
}


static DFBResult
JPEG_SW_cache_cleanup(IDirectFBImageProvider_JPEG_data *data)
{
     if (data->image)
     {
          D_FREE( data->image );
          data->image = NULL;
     }

     return DFB_OK;
}


static void
IDirectFBImageProvider_JPEG_Destruct( IDirectFBImageProvider *thiz )
{
     IDirectFBImageProvider_JPEG_data *data =
                              (IDirectFBImageProvider_JPEG_data*)thiz->priv;

     data->buffer->Release( data->buffer );

     IMG_HW_cache_cleanup(&(data->surf));
     JPEG_SW_cache_cleanup(data);

     DIRECT_DEALLOCATE_INTERFACE( thiz );
}


bool 
check_hw_jpeg_constraints( struct jpeg_decompress_struct *info )
{
     bool is_feature_supported_hw = info->is_baseline;

     if( !is_feature_supported_hw ) {
          D_JPEG("the image is not baseline\n");
          return false;
     }

     int aligned_width = (info->image_width + 0x3f) & ~0x3fUL;   // Decoder required the buffer aligned to 64 pixels
     int aligned_height = (info->image_height + 0x3f) & ~0x3fUL;  // Decoder required the buffer aligned to 64 pixels

     //This limitation from firmware 
     if (aligned_width > 1024 * 6 || (aligned_width * aligned_height) > 1024*1024 * 16)
     {
          D_JPEG("Use SW decoder, exceeding fw limit: width > 6K or total pixels > 16M\n");
          return false;
     }


     float num_megapixels = (info->image_width * info->image_height) / 1048576.0;

     if(num_megapixels < dfb_config->hw_jpeg_min_megapixels 
        || num_megapixels > dfb_config->hw_jpeg_max_megapixels) {
        D_JPEG("the image is either too large or too small\n");
        return false;
     }

     /*add more SOF checks here*/
     
     return true;
}


static DFBResult
Probe( IDirectFBImageProvider_ProbeContext *ctx )
{
     if (ctx->header[0] == 0xff && ctx->header[1] == 0xd8) {
          if (strncmp ((char*) ctx->header + 6, "JFIF", 4) == 0 ||
              strncmp ((char*) ctx->header + 6, "Exif", 4) == 0 ||
              strncmp ((char*) ctx->header + 6, "VVL", 3) == 0 ||
              strncmp ((char*) ctx->header + 6, "WANG", 4) == 0)
               return DFB_OK;

          /* Else look for Quantization table marker or Define Huffman table marker,
           * useful for EXIF thumbnails that have no APPx markers.
           */
          if (ctx->header[2] == 0xff && (ctx->header[3] == 0xdb || ctx->header[3] == 0xc4))
               return DFB_OK;

          if (ctx->filename && strchr (ctx->filename, '.' ) &&
             (strcasecmp ( strchr (ctx->filename, '.' ), ".jpg" ) == 0 ||
              strcasecmp ( strchr (ctx->filename, '.' ), ".jpeg") == 0))
               return DFB_OK;
     }

     return DFB_UNSUPPORTED;
}

static DFBResult
Construct( IDirectFBImageProvider *thiz,
           ... )
{
     struct jpeg_decompress_struct cinfo;
     struct my_error_mgr jerr;
     
     IDirectFBDataBuffer *buffer;
     CoreDFB             *core;
     va_list              tag;
     
     jpeg_component_info *compptr;

     DIRECT_ALLOCATE_INTERFACE_DATA(thiz, IDirectFBImageProvider_JPEG)

     va_start( tag, thiz );
     buffer = va_arg( tag, IDirectFBDataBuffer * );
     core = va_arg( tag, CoreDFB * );
     va_end( tag );

     data->ref    = 1;
     data->buffer = buffer;
     data->core   = core;

     buffer->AddRef( buffer );

     cinfo.err = jpeg_std_error(&jerr.pub);
     jerr.pub.error_exit = jpeglib_panic;

     if (setjmp(jerr.setjmp_buffer)) {
          D_ERROR( "ImageProvider/JPEG: Error while reading headers!\n" );

          jpeg_destroy_decompress(&cinfo);
          buffer->Release( buffer );
          DIRECT_DEALLOCATE_INTERFACE( thiz );
          return DFB_FAILURE;
     }

     jpeg_create_decompress(&cinfo);
     jpeg_buffer_src(&cinfo, buffer, 1);
     jpeg_read_header(&cinfo, TRUE);
     //jpeg_start_decompress(&cinfo);
     
     data->width = cinfo.image_width;
     data->height = cinfo.image_height;
     data->hw_supported = check_hw_jpeg_constraints(&cinfo);

     if(!(data->hw_supported)) {
        D_JPEG("HW decoding is not supported for this image\n");
     }

     //To determine the colorspace of the input image.
     //int ci;
     //for (ci = 0; ci < cinfo.num_components; ci++) {
     if (cinfo.jpeg_color_space == JCS_YCbCr) {
          compptr = cinfo.comp_info;
          //printf("h=%d, v=%d, ci=%d\n", compptr->h_samp_factor, compptr->v_samp_factor, ci);
          if(compptr->v_samp_factor == 2) {
               data->comp_colorspace = DSPF_I420;
          }
          else {
               if (compptr->h_samp_factor == 2) {
                    data->comp_colorspace = DSPF_YV16;
               }
               else {
                    data->comp_colorspace = DSPF_YUV444P;
               }
          }
     }
     else if(cinfo.jpeg_color_space == JCS_RGB) {
          data->comp_colorspace = DSPF_ARGB;
     }
     else {
          data->comp_colorspace = DSPF_I420;
     }
     //}
     
     //jpeg_abort_decompress(&cinfo);
     jpeg_destroy_decompress(&cinfo);

     thiz->AddRef = IDirectFBImageProvider_JPEG_AddRef;
     thiz->Release = IDirectFBImageProvider_JPEG_Release;
     thiz->RenderTo = IDirectFBImageProvider_JPEG_RenderTo;
     thiz->NoncacheRenderTo = IDirectFBImageProvider_JPEG_NoncacheRenderTo;
     thiz->SetRenderCallback = IDirectFBImageProvider_JPEG_SetRenderCallback;
     thiz->GetImageDescription =IDirectFBImageProvider_JPEG_GetImageDescription;
     thiz->GetSurfaceDescription =
     IDirectFBImageProvider_JPEG_GetSurfaceDescription;

     return DFB_OK;
}


static DirectResult
IDirectFBImageProvider_JPEG_AddRef( IDirectFBImageProvider *thiz )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     data->ref++;

     return DFB_OK;
}

static DirectResult
IDirectFBImageProvider_JPEG_Release( IDirectFBImageProvider *thiz )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     if (--data->ref == 0) {
          IDirectFBImageProvider_JPEG_Destruct( thiz );
     }

     return DFB_OK;
}


//useNonCache controls whether to create a cached entire image buffer.
//direct controls whether we need to do further scaling.
//If noncached method is not possible due to scaling, we will change to cache method. 
static DFBResult
SoftwareJPEG_RenderTo( IDirectFBImageProvider *thiz,
                       IDirectFBSurface       *destination,
                       const DFBRectangle     *dest_rect,
                       bool useNonCache)
{
     DFBResult              ret;
     bool                   direct;
     DFBRegion              clip;
     DFBRectangle           rect;
     DFBSurfacePixelFormat  format;
     IDirectFBSurface_data *dst_data;
     CoreSurface           *dst_surface;
     CoreSurfaceBufferLock  lock;
     DIRenderCallbackResult cb_result = DIRCR_OK;

     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     TIME_BEGIN();

     dst_data = (IDirectFBSurface_data*) destination->priv;
     if (!dst_data)
          return DFB_DEAD;

     dst_surface = dst_data->surface;
     if (!dst_surface)
          return DFB_DESTROYED;

     ret = destination->GetPixelFormat( destination, &format );
     if (ret)
          return ret;

     dfb_region_from_rectangle( &clip, &dst_data->area.current );

     if (dest_rect) {
          if (dest_rect->w < 1 || dest_rect->h < 1)
               return DFB_INVARG;
          
          rect = *dest_rect;
          rect.x += dst_data->area.wanted.x;
          rect.y += dst_data->area.wanted.y;

          if (!dfb_rectangle_region_intersects( &rect, &clip ))
               return DFB_OK;
     }
     else {
          rect = dst_data->area.wanted;
     }
       
     ret = dfb_surface_lock_buffer( dst_surface, CSBR_BACK, CSAID_CPU, CSAF_WRITE, &lock );
     if (ret) {
          D_JPEG("Can not lock buffer in system memory\n");
          return ret;
     }


     //clear the cached image, in case that the new size is different from the last one.
     if (data->image &&
         //(rect.x || rect.y || rect.w != data->image_width || rect.h != data->image_height)) {
         (rect.w != data->image_width || rect.h != data->image_height)) {//don't know why check rect.x and rect.y here.
           D_FREE( data->image );
           data->image        = NULL;
           data->image_width  = 0;
           data->image_height = 0;
     }
     
     if (useNonCache) {
          JPEG_SW_cache_cleanup(data);
     }

     /* actual loading and rendering */
     if (!data->image) {
          struct jpeg_decompress_struct cinfo;
          struct my_error_mgr jerr;
          JSAMPARRAY buffer;      /* Output row buffer */
          int row_stride;         /* physical row width in output buffer */
          u32 *row_ptr;
          int y = 0;
          int uv_offset = 0;

          int omit_last_line, omit_last_col;

          cinfo.err = jpeg_std_error(&jerr.pub);
          jerr.pub.error_exit = jpeglib_panic;

          if (setjmp(jerr.setjmp_buffer)) {
               D_ERROR( "ImageProvider/JPEG: Error during decoding!\n" );

               jpeg_destroy_decompress(&cinfo);

               if (data->image) {
                    dfb_scale_linear_32( data->image, data->image_width, data->image_height,
                                         lock.addr, lock.pitch, &rect, dst_surface, &clip );
                    dfb_surface_unlock_buffer( dst_surface, &lock );
                    if (data->render_callback) {
                         DFBRectangle r = { 0, 0, data->image_width, data->image_height };

                         if (data->render_callback( &r, data->render_callback_context ) != DIRCR_OK)
                              return DFB_INTERRUPTED;
                    }

                    return DFB_INCOMPLETE;
               }
               else
                    dfb_surface_unlock_buffer( dst_surface, &lock );

               return DFB_FAILURE;
          }

          jpeg_create_decompress(&cinfo);

          //cinfo.mem->max_memory_to_use = 1 << 20; //1M
          //D_JPEG("libjpeg: max_memory_to_use = %d\n", cinfo.mem->max_memory_to_use);
          
          jpeg_buffer_src(&cinfo, data->buffer, 0);
          jpeg_read_header(&cinfo, TRUE);
#if JPEG_LIB_VERSION >= 70
          cinfo.scale_num = 8;
          cinfo.scale_denom = 8;
#else
          cinfo.scale_num = 1;
          cinfo.scale_denom = 1;
#endif
          jpeg_calc_output_dimensions(&cinfo);

          if (check_direct_mode(format, cinfo.output_width, cinfo.output_height, rect.w, rect.h, 
               &omit_last_line, &omit_last_col)) {
               direct = true;
               D_JPEG("original: direct = %d\n", direct);
          }
          else {
#if JPEG_LIB_VERSION >= 70
             /*  The supported scaling ratios in libjpeg 7 and 8
                          *  are N/8 with all N from 1 to 16.
                          */
               cinfo.scale_num = 1;
               jpeg_calc_output_dimensions( &cinfo );
               while (cinfo.scale_num < 8// replaces 16, since we only want shrinking when decoding, no upscaling.
                      && cinfo.output_width < (unsigned)rect.w
                      && cinfo.output_height < (unsigned)rect.h) {
                    ++cinfo.scale_num;
                    jpeg_calc_output_dimensions( &cinfo );
               }
#else
             /*  The supported scaling ratios in libjpeg 6
                          *  are 1/1, 1/2, 1/4, and 1/8.
                          */
               while (cinfo.scale_denom < 8
                      && ((cinfo.output_width >> 1) >= (unsigned)rect.w)
                      && ((cinfo.output_height >> 1) >= (unsigned)rect.h)) {
                    cinfo.scale_denom <<= 1;
                    jpeg_calc_output_dimensions( &cinfo );
               }
#endif
               D_JPEG("output_width = %d, output_height = %d\n", cinfo.output_width, cinfo.output_height);
               D_JPEG("rect.w = %d, rect.h = %d\n", rect.w, rect.h);
               if (check_direct_mode(format, cinfo.output_width, cinfo.output_height, rect.w, rect.h, 
                    &omit_last_line, &omit_last_col)) {
                    direct = true;
                    D_JPEG("scaling while decoding: ON\n");
               }
               else {
                    if(direct) {
                         D_JPEG("direct mode unavailable, use cache instead.\n");
                    }
                    direct = false;
               }
          }             

          cinfo.output_components = 3;

          switch (format) {
               case DSPF_NV16:
                    uv_offset = dst_surface->config.size.h * lock.pitch;

                    if (direct && !rect.x && !rect.y) {
                         D_JPEG( "Using YCbCr color space directly! (%dx%d)\n",
                                 cinfo.output_width, cinfo.output_height );
                         cinfo.out_color_space = JCS_YCbCr;
                         break;
                    }
                    D_JPEG( "Going through RGB color space! (%dx%d -> %dx%d @%d,%d)\n",
                            cinfo.output_width, cinfo.output_height, rect.w, rect.h, rect.x, rect.y );
               
               default:
                    cinfo.out_color_space = JCS_RGB;
                    break;
          }

          jpeg_start_decompress(&cinfo);

          data->image_width = cinfo.output_width;
          data->image_height = cinfo.output_height;

          row_stride = cinfo.output_width * 3;

          buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo,
                                              JPOOL_IMAGE, row_stride, 1);

          bool use_row_buffer;

          if (useNonCache && direct) {
               use_row_buffer = true;
          }
          else {
               use_row_buffer = false;
          }

          if (use_row_buffer) {
               D_JPEG("only one row allocated.\n");
               row_ptr = D_CALLOC( 1, data->image_width * 4 );
               if (!row_ptr) {
                    dfb_surface_unlock_buffer( dst_surface, &lock );
                    return D_OOM();
               }
          }
          else {
               D_JPEG("the entire image allocated.\n");
               data->image = D_CALLOC( data->image_height, data->image_width * 4 );
               if (!data->image) {
                    dfb_surface_unlock_buffer( dst_surface, &lock );
                    return D_OOM();
               }
               row_ptr = data->image;
          }

#ifdef SLEEP_BETWEEN_LINES
          int lines_num_each_sleep = PIXELS_NUM_BETWEEN_EACH_SLEEP / data->image_width;
          int tmp = lines_num_each_sleep;
#endif
          while (cinfo.output_scanline < (cinfo.output_height - omit_last_line)&& cb_result == DIRCR_OK) {
               jpeg_read_scanlines(&cinfo, buffer, 1);

               switch (format) {
                    case DSPF_NV16:
                         if (direct) {
                              copy_line_nv16( lock.addr, (u16*)lock.addr + uv_offset, *buffer, rect.w );

                              lock.addr = (u8*)lock.addr + lock.pitch;

                              if (data->render_callback) {
                                   DFBRectangle r = { 0, y, data->image_width, 1 };

                                   cb_result = data->render_callback( &r, 
                                                                      data->render_callback_context );
                              }
                              break;
                         }

                    default:
                         copy_line32( row_ptr, *buffer, data->image_width );

                         if (direct) {
                              DFBRectangle r = { rect.x, rect.y+y, rect.w, 1 };
                              dfb_copy_buffer_32( row_ptr, lock.addr, lock.pitch,
                                                  &r, dst_surface, &clip );
                              if (data->render_callback) {
                                   r = (DFBRectangle){ 0, y, data->image_width, 1 };
                                   cb_result = data->render_callback( &r, 
                                                                      data->render_callback_context );
                              }
                         }
                         break;
               }

               //if (!direct){
                    //row_ptr += data->image_width;
               //}
               row_ptr += use_row_buffer ? 0 : data->image_width; //to speed up
               
               y++;
#ifdef SLEEP_BETWEEN_LINES               
               if(!(tmp--)) {
                    tmp = lines_num_each_sleep;
                    usleep(1);//sleep to release cpu.
               }
#endif
          }
          
          TIME("software decoding time");
          
#if 0
          FILE *fp = fopen("/nfs/dump_jpeg.argb", "wb");
          if (fp == NULL) {
               printf("unable to open file\n");
               exit -1;
          }
          fwrite(data->image, 1, data->image_height * data->image_width * 4, fp);

          fclose(fp);
#endif

          if (!direct) {
               dfb_scale_linear_32( data->image, data->image_width, data->image_height,
                                    lock.addr, lock.pitch, &rect, dst_surface, &clip );
               if (data->render_callback) {
                    DFBRectangle r = { 0, 0, data->image_width, data->image_height };
                    cb_result = data->render_callback( &r, data->render_callback_context );
               }
          }
          
          TIME("software scaling time");

          if (cb_result != DIRCR_OK) {
               jpeg_abort_decompress(&cinfo);
               D_FREE( data->image );
               data->image = NULL;
          }
          else {
               if (omit_last_line) {
                    jpeg_abort_decompress(&cinfo);
               } else {
                    jpeg_finish_decompress(&cinfo);
               }
          }
          jpeg_destroy_decompress(&cinfo);

          if (use_row_buffer) {
               D_FREE(row_ptr);
          }

          if (useNonCache) {
               JPEG_SW_cache_cleanup(data);
          }
               
     }
     else {
          dfb_scale_linear_32( data->image, data->image_width, data->image_height,
                               lock.addr, lock.pitch, &rect, dst_surface, &clip );
          if (data->render_callback) {
               DFBRectangle r = { 0, 0, data->image_width, data->image_height };
               data->render_callback( &r, data->render_callback_context );
          }
     }

    //destination->Dump(destination, "/nfs/", "jpeg");
#if 0
     printf("dumping yuv\n");
     printf("pitch = %d\n", lock.pitch);
     FILE *fp = fopen("/nfs/dump_jpeg.yuv", "wb");
     if (fp == NULL) {
          printf("unable to open file\n");
          exit -1;
     }
     fwrite(lock.addr, 1, (data->image_height) * lock.pitch * 3/2, fp);

     fclose(fp);
#endif
     
     dfb_surface_unlock_buffer( dst_surface, &lock );

     if (cb_result != DIRCR_OK)
         return DFB_INTERRUPTED;

     return DFB_OK;
}


static DFBResult
JPEG_RenderTo( IDirectFBImageProvider *thiz,
                      IDirectFBSurface       *destination,
                      const DFBRectangle     *dest_rect ,
                      bool useNonCache)

{
     IDirectFBSurface_data *dst_data;
     CoreSurface           *dst_surface;
     DFBResult              ret = DFB_OK;

     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     dst_data = (IDirectFBSurface_data*) destination->priv;
     if (!dst_data)
          return DFB_DEAD;

     dst_surface = dst_data->surface;
     if (!dst_surface)
          return DFB_DESTROYED;

     // the destination is on system memory we fall back to software implemenation then.
     D_JPEG("Renderto dst surface caps: 0x%08x, format: %s\n", 
                dst_data->caps, dfb_pixelformat_name(dst_surface->config.format) );
     D_JPEG("jpeg image size (%d, %d)\n", data->width, data->height);
#if !XCODE_DISABLE_HW_JPEG_DECODE
     if (!dfb_config->disable_hw_jpeg_decode && data->hw_supported)
     {
          IDirectFBImageProvider_HW_data hw_data;
          hw_data.buffer = data->buffer;
          hw_data.render_callback = data->render_callback;
          hw_data.render_callback_context = data->render_callback_context;
          hw_data.width = data->width;
          hw_data.height = data->height;
          hw_data.core = data->core;
          hw_data.pp_surf = &(data->surf);
          hw_data.comp_colorspace = data->comp_colorspace;
          
          ret = HardwareImg_RenderTo(&hw_data, destination, dest_rect, !useNonCache, DFB_HW_JPEG);
          
          if (ret == DFB_OK) return DFB_OK;
     }
#endif

     return SoftwareJPEG_RenderTo( thiz, destination, dest_rect, useNonCache);
}



DFBResult
IDirectFBImageProvider_JPEG_RenderTo( IDirectFBImageProvider *thiz,
                                      IDirectFBSurface       *destination,
                                      const DFBRectangle     *dest_rect )
{
    return JPEG_RenderTo(thiz, destination, dest_rect, true);

}

static DFBResult
IDirectFBImageProvider_JPEG_NoncacheRenderTo( IDirectFBImageProvider *thiz,
                                              IDirectFBSurface       *destination,
                                              const DFBRectangle     *dest_rect )
{
    return JPEG_RenderTo(thiz, destination, dest_rect, true);
}



static DFBResult
IDirectFBImageProvider_JPEG_SetRenderCallback( IDirectFBImageProvider *thiz,
                                               DIRenderCallback        callback,
                                               void                   *context )
{
     DIRECT_INTERFACE_GET_DATA (IDirectFBImageProvider_JPEG)

     data->render_callback         = callback;
     data->render_callback_context = context;

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_JPEG_GetSurfaceDescription( IDirectFBImageProvider *thiz,
                                                   DFBSurfaceDescription  *dsc )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)
     
     dsc->flags  = DSDESC_WIDTH |  DSDESC_HEIGHT | DSDESC_PIXELFORMAT;
     dsc->height = data->height;
     dsc->width  = data->width;
     dsc->pixelformat = dfb_primary_layer_pixelformat();

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_JPEG_GetImageDescription( IDirectFBImageProvider *thiz,
                                                 DFBImageDescription    *dsc )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_JPEG)

     if (!dsc)
          return DFB_INVARG;

     dsc->caps = DICAPS_NONE;

#if 1
     if (data->hw_supported) {
          dsc->caps |= DICAPS_HW_DECODE;
     }
#else
     if (data->width * data->height <= 2500 * 2000) {
          dsc->caps |= DICAPS_HW_DECODE;
     }
#endif

     return DFB_OK;
}

