/*
   (c) Copyright 2001-2009  The world wide DirectFB Open Source Community (directfb.org)
   (c) Copyright 2000-2004  Convergence (integrated media) GmbH

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>,
              Andreas Hundt <andi@fischlustig.de>,
              Sven Neumann <neo@directfb.org>,
              Ville Syrjälä <syrjala@sci.fi> and
              Claudio Ciccani <klan@users.sf.net>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <config.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <png.h>
#include <string.h>
#include <stdarg.h>

#include <directfb.h>

#include <display/idirectfbsurface.h>

#include <media/idirectfbimageprovider.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/layers.h>
#include <core/palette.h>
#include <core/surface.h>

#include <misc/gfx_util.h>
#include <misc/util.h>

#include <gfx/clip.h>
#include <gfx/convert.h>

#include <direct/interface.h>
#include <direct/mem.h>
#include <direct/memcpy.h>
#include <direct/messages.h>
#include <direct/util.h>

#include "config.h"


#include "xcode4_img_module/xcode4_img_api.h"
#include "xcode4/xcdfb.h"
#include <core/core.h>
#include <fusion/arena.h>
#include <fusion/fusion.h>

#include "viper.h"
#include "xcode4_img_module/xcode4_img_ycbcr2argb.h"

#include "xcode4_img_module/xcode4_imageprovider_HW.h"

#include "xcode4_img_module/xcode4_img_util.h"

#define READ_SIZE 64

//#define PNG_DEBUG

#ifdef PNG_DEBUG
#define D_PRINT(x...) printf(x)
#else
#define D_PRINT(x...) 
#endif



#define D_PNG(...) D_DEBUG_AT( PNG_T, __VA_ARGS__)

D_DEBUG_DOMAIN( PNG_T, "png", "PNG decoding" );



static DFBResult
Probe( IDirectFBImageProvider_ProbeContext *ctx );

static DFBResult
Construct( IDirectFBImageProvider *thiz,
           ... );

#include <direct/interface_implementation.h>

DIRECT_INTERFACE_IMPLEMENTATION( IDirectFBImageProvider, PNG )


enum {
     STAGE_ABORT = -2,
     STAGE_ERROR = -1,
     STAGE_START =  0,
     STAGE_INFO,
     STAGE_IMAGE,
     STAGE_END
};

/*
 * private data struct of IDirectFBImageProvider_PNG
 */
typedef struct {
     int                  ref;      /* reference counter */
     IDirectFBDataBuffer *buffer;

     int                  stage;
     int                  rows;

     png_structp          png_ptr;
     png_infop            info_ptr;

     png_uint_32          width;
     png_uint_32          height;
     int                  bpp;
     int                  color_type;
     png_uint_32          color_key;
     bool                 color_keyed;

     CoreSurface         *surf;

     void                *image;
     int                  pitch;
     u32                  palette[256];
     DFBColor             colors[256];

     DIRenderCallback     render_callback;
     void                *render_callback_context;

     CoreDFB             *core;

     bool                hw_supported;
     bool                use_row_buffer; //special for PNG, because we use png_set_progressive_read_fn callback mechanism
     int                 num_pass;//number of passes of interlace handling for png, used to compute use_row_buffer
     bool                first_renderto;

     //DFBRectangle        cb_clipped;
     DFBRegion           cb_clip;
     DFBRectangle        cb_rect;
     CoreSurface         *cb_dst_surf;
     CoreSurfaceBufferLock cb_lock;
} IDirectFBImageProvider_PNG_data;

static DirectResult
IDirectFBImageProvider_PNG_AddRef  ( IDirectFBImageProvider *thiz );

static DirectResult
IDirectFBImageProvider_PNG_Release ( IDirectFBImageProvider *thiz );

static DFBResult
IDirectFBImageProvider_PNG_RenderTo( IDirectFBImageProvider *thiz,
                                     IDirectFBSurface       *destination,
                                     const DFBRectangle     *destination_rect );

static DFBResult
IDirectFBImageProvider_PNG_NoncacheRenderTo( IDirectFBImageProvider *thiz,
                                     IDirectFBSurface       *destination,
                                     const DFBRectangle     *dest_rect );

static DFBResult
IDirectFBImageProvider_PNG_SetRenderCallback( IDirectFBImageProvider *thiz,
                                              DIRenderCallback        callback,
                                              void                   *context );

static DFBResult
IDirectFBImageProvider_PNG_GetSurfaceDescription( IDirectFBImageProvider *thiz,
                                                  DFBSurfaceDescription  *dsc );

static DFBResult
IDirectFBImageProvider_PNG_GetImageDescription( IDirectFBImageProvider *thiz,
                                                DFBImageDescription    *dsc );

/* Called at the start of the progressive load, once we have image info */
static void
png_info_callback (png_structp png_read_ptr,
                   png_infop   png_info_ptr);

/* Called for each row; note that you will get duplicate row numbers
   for interlaced PNGs */
static void
png_row_callback  (png_structp png_read_ptr,
                   png_bytep   new_row,
                   png_uint_32 row_num,
                   int         pass_num);

/* Called after reading the entire image */
static void
png_end_callback  (png_structp png_read_ptr,
                   png_infop   png_info_ptr);

/* Pipes data into libpng until stage is different from the one specified. */
static DFBResult
push_data_until_stage (IDirectFBImageProvider_PNG_data *data,
                       int                              stage,
                       int                              buffer_size);

/**********************************************************************************************************************/
static DFBResult
write_surf_buffer_png( IDirectFBImageProvider_PNG_data *data,
                         int height, DFBRectangle rect);

/**********************************************************************************************************************/
bool 
check_hw_png_constraints( png_structp png_ptr, png_infop info_ptr )
{
     IDirectFBImageProvider_PNG_data *data;

     data = png_get_progressive_ptr( png_ptr );

     //should not directly read info_ptr->bit_depth, which is always 8.
     //png_get_bit_depth(png_ptr, info_ptr)  also doesn't work.
     int m_bit_depth = data->bpp;

     if(m_bit_depth > 8) {//16 bits
          D_PNG("PNG: the image has bit depth greater than 8.\n");
          return false;
     }

     if(data->width * data->height > 2048 * 1024) {
          D_PNG("PNG: the image exceeds HW decoder limit of 2048x1024 pixels.\n");
          return false;
     }

     if(data->width > 4096 ) {
          D_PNG("PNG: the image exceeds HW decoder maximal width 4096.\n");
          return false;
     }     
     return true;
}


/**********************************************************************************************************************/

static DFBResult
Probe( IDirectFBImageProvider_ProbeContext *ctx )
{
     if (png_check_sig( ctx->header, 8 ))
          return DFB_OK;

     return DFB_UNSUPPORTED;
}

static DFBResult
Construct( IDirectFBImageProvider *thiz,
           ... )
{
     DFBResult ret = DFB_FAILURE;

     IDirectFBDataBuffer *buffer;
     CoreDFB             *core;
     va_list              tag;

     DIRECT_ALLOCATE_INTERFACE_DATA(thiz, IDirectFBImageProvider_PNG)

     va_start( tag, thiz );
     buffer = va_arg( tag, IDirectFBDataBuffer * );
     core = va_arg( tag, CoreDFB * );
     va_end( tag );

     data->ref    = 1;
     data->buffer = buffer;
     data->core   = core;

     /* Increase the data buffer reference counter. */
     buffer->AddRef( buffer );

     /* Create the PNG read handle. */
     data->png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING,
                                             NULL, NULL, NULL );
     if (!data->png_ptr)
          goto error;

     if (setjmp( data->png_ptr->jmpbuf )) {
          D_ERROR( "ImageProvider/PNG: Error reading header!\n" );
          goto error;
     }

     /* Create the PNG info handle. */
     data->info_ptr = png_create_info_struct( data->png_ptr );
     if (!data->info_ptr)
          goto error;


     /* Setup progressive image loading. */
     png_set_progressive_read_fn( data->png_ptr, data,
                                  png_info_callback,
                                  png_row_callback,
                                  png_end_callback );

     data->use_row_buffer = false;
     data->first_renderto = true;


     /* Read until info callback is called. */
     ret = push_data_until_stage( data, STAGE_INFO, READ_SIZE );
     if (ret)
          goto error;

     data->hw_supported = check_hw_png_constraints(data->png_ptr, data->info_ptr);

     thiz->AddRef                = IDirectFBImageProvider_PNG_AddRef;
     thiz->Release               = IDirectFBImageProvider_PNG_Release;
     thiz->RenderTo              = IDirectFBImageProvider_PNG_RenderTo;
     thiz->NoncacheRenderTo      = IDirectFBImageProvider_PNG_NoncacheRenderTo;
     thiz->SetRenderCallback     = IDirectFBImageProvider_PNG_SetRenderCallback;
     thiz->GetImageDescription   = IDirectFBImageProvider_PNG_GetImageDescription;
     thiz->GetSurfaceDescription = IDirectFBImageProvider_PNG_GetSurfaceDescription;

     return DFB_OK;

error:
     if (data->png_ptr)
          png_destroy_read_struct( &data->png_ptr, &data->info_ptr, NULL );

     buffer->Release( buffer );

     if (data->image)
          D_FREE( data->image );

     DIRECT_DEALLOCATE_INTERFACE(thiz);

     return ret;
}

/**********************************************************************************************************************/

static void
IDirectFBImageProvider_PNG_Destruct( IDirectFBImageProvider *thiz )
{
     IDirectFBImageProvider_PNG_data *data =
                              (IDirectFBImageProvider_PNG_data*)thiz->priv;

     png_destroy_read_struct( &data->png_ptr, &data->info_ptr, NULL );

     /* Decrease the data buffer reference counter. */
     data->buffer->Release( data->buffer );

     IMG_HW_cache_cleanup(&(data->surf));
     
     /* Deallocate image data. */
     if (data->image)
          D_FREE( data->image );

     DIRECT_DEALLOCATE_INTERFACE( thiz );
}

static DirectResult
IDirectFBImageProvider_PNG_AddRef( IDirectFBImageProvider *thiz )
{
     DIRECT_INTERFACE_GET_DATA (IDirectFBImageProvider_PNG)

     data->ref++;

     return DFB_OK;
}

static DirectResult
IDirectFBImageProvider_PNG_Release( IDirectFBImageProvider *thiz )
{
     DIRECT_INTERFACE_GET_DATA (IDirectFBImageProvider_PNG)

     if (--data->ref == 0) {
          IDirectFBImageProvider_PNG_Destruct( thiz );
     }

     return DFB_OK;
}

/**********************************************************************************************************************/

static DFBResult
reconstruct_png_read_struct(IDirectFBImageProvider *thiz)
{
     DFBResult ret = DFB_FAILURE;

     DIRECT_INTERFACE_GET_DATA (IDirectFBImageProvider_PNG)
     
     /* Create the PNG read handle. */
     data->png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING,
                                             NULL, NULL, NULL );
     if (!data->png_ptr)
          goto error_reconstruct;

     if (setjmp( data->png_ptr->jmpbuf )) {
          D_ERROR( "ImageProvider/PNG: Error reading header!\n" );
          goto error_reconstruct;
     }


     /* Create the PNG info handle. */
     data->info_ptr = png_create_info_struct( data->png_ptr );
     if (!data->info_ptr)
          goto error_reconstruct;


     /* Setup progressive image loading. */
     png_set_progressive_read_fn( data->png_ptr, data,
                                  png_info_callback,
                                  png_row_callback,
                                  png_end_callback );


     /* Read until info callback is called. */
     ret = push_data_until_stage( data, STAGE_INFO, READ_SIZE );
     if (ret)
          goto error_reconstruct;

     return DFB_OK;

error_reconstruct:
#if 0
     if (data->png_ptr)
          png_destroy_read_struct( &data->png_ptr, &data->info_ptr, NULL );

     data->buffer->Release( data->buffer );

     if (data->image) {
          D_FREE( data->image );
          data->image = NULL;
     }

     DIRECT_DEALLOCATE_INTERFACE(thiz);
#endif
     return ret;
}


static DFBResult
SoftwarePNG_RenderTo( IDirectFBImageProvider *thiz,
                                   IDirectFBSurface       *destination,
                                   const DFBRectangle     *dest_rect,
                                   bool useNonCache)

{
     DFBResult              ret = DFB_OK;
     IDirectFBSurface_data *dst_data;
     CoreSurface           *dst_surface;
     DFBRegion              clip;
     DFBRectangle           rect;
     png_infop              info;
     //int                    x, y;
     DFBRectangle           clipped;

     bool                   direct;
     int omit_last_line, omit_last_col;

     DIRECT_INTERFACE_GET_DATA (IDirectFBImageProvider_PNG)


     if (data->use_row_buffer) { //the last renderto mode determines if we reconstruct the png decoder.
          png_destroy_read_struct( &data->png_ptr, &data->info_ptr, NULL );
          data->rows = 0;
          data->stage = STAGE_START;

          IDirectFBDataBuffer *buffer = data->buffer;

          buffer->SeekTo( buffer, 0 );

          /* Deallocate image data. */
          if (data->image) {
               D_FREE( data->image );
               data->image = NULL;
          }

          ret = reconstruct_png_read_struct(thiz);
          if (ret) {
               D_PRINT("png struct reconstructing failed.\n");
               return ret;
          }
     }
     

     info = data->info_ptr;

     dst_data = (IDirectFBSurface_data*) destination->priv;
     if (!dst_data)
          return DFB_DEAD;

     dst_surface = dst_data->surface;
     if (!dst_surface)
          return DFB_DESTROYED;

     dfb_region_from_rectangle( &clip, &dst_data->area.current );

     if (dest_rect) {
          if (dest_rect->w < 1 || dest_rect->h < 1)
               return DFB_INVARG;
          rect = *dest_rect;
          rect.x += dst_data->area.wanted.x;
          rect.y += dst_data->area.wanted.y;
     }
     else {
          rect = dst_data->area.wanted;
     }

     clipped = rect;

     if (!dfb_rectangle_intersect_by_region( &clipped, &clip ))
          return DFB_INVAREA;
     

     direct = false;

     DFBSurfacePixelFormat format = dst_surface->config.format;

     if( data->rows == 0 && ((data->width * data->height) > 1000*1000) ) { //means the image lines have not been decoded.
          if (data->num_pass == 1) { //make sure interlaced handling is not needed
               if( check_direct_mode(format, data->width, data->height, rect.w, rect.h, 
                    &omit_last_line, &omit_last_col) ) {
                    direct = true;
               }
          }
          D_PRINT("num_pass = %d\n", data->num_pass);
     }

     if (useNonCache && direct) {
          if (data->first_renderto) { //only for the very first time. 
               data->use_row_buffer = true;
          } else {
               //stay the same as last time for data->use_row_buffer
          }
     }
     else {
          data->use_row_buffer = false;
     }

     D_PNG("png sw decoder use_row_buffer = %d\n", data->use_row_buffer);

     data->cb_clip = clip;
     data->cb_rect = rect;
     data->cb_dst_surf = dst_surface;

     
     ret = dfb_surface_lock_buffer( dst_surface, CSBR_BACK, CSAID_CPU, CSAF_WRITE, &data->cb_lock );
     if (ret) {
          return ret;
     }


     if (setjmp( data->png_ptr->jmpbuf )) {
          D_ERROR( "ImageProvider/PNG: Error during decoding!\n" );

          if (data->stage < STAGE_IMAGE)
               return DFB_FAILURE;

          data->stage = STAGE_ERROR;
     }

     /* Read until image is completely decoded. */
     if (data->stage != STAGE_ERROR) {
          ret = push_data_until_stage( data, STAGE_END, 16384 );
          if (ret) {
               dfb_surface_unlock_buffer( dst_surface, &data->cb_lock );
               return ret;
          }
     }

     if (!data->use_row_buffer) {
          ret = write_surf_buffer_png( data, data->height, rect);
     }

     dfb_surface_unlock_buffer( dst_surface, &data->cb_lock );

     if (data->stage != STAGE_END)
          ret = DFB_INCOMPLETE;

     data->first_renderto = false;

     return ret;
}


static DFBResult
write_surf_buffer_png( IDirectFBImageProvider_PNG_data *data,
                         int height, DFBRectangle rect)
{
     DFBResult              ret = DFB_OK;
     int                    x, y;

     CoreSurface *dst_surface = data->cb_dst_surf;
     DFBRegion clip = data->cb_clip;
     
     D_PRINT("000000000\n");
     /* actual rendering */
#if 0
     if (rect.w == data->width && rect.h == height &&
         (data->color_type == PNG_COLOR_TYPE_RGB || data->color_type == PNG_COLOR_TYPE_RGBA) &&
         (dst_surface->config.format == DSPF_RGB32 || dst_surface->config.format == DSPF_ARGB) &&
         !(dst_surface->config.caps & DSCAPS_PREMULTIPLIED))
     {
          D_PRINT("1111111111\n");
          ret = dfb_surface_write_buffer( dst_surface, CSBR_BACK,
                                          data->image +
                                             (clipped.x - rect.x) * 4 +
                                             (clipped.y - rect.y) * data->width * 4,
                                          data->width * 4, &clipped );
     }
#else 
     if (0) {
     }
#endif
     else {
#if 0
          if (is_first_callback) {
               ret = dfb_surface_lock_buffer( dst_surface, CSBR_BACK, CSAID_CPU, CSAF_WRITE, &data->cb_lock );
               if (ret)
                    return ret;
          }
#endif

          CoreSurfaceBufferLock *p_lock = &data->cb_lock;

          D_PRINT("222222222\n");

          switch (data->color_type) {
               case PNG_COLOR_TYPE_PALETTE:
                    D_PRINT("3333333333333333\n");
                    if (dst_surface->config.format == DSPF_LUT8 && data->info_ptr->bit_depth == 8) {
                         /*
                          * Special indexed PNG to LUT8 loading.
                          */

                         /* FIXME: Limitation for LUT8 is to load complete surface only. */
                         dfb_clip_rectangle( &clip, &rect );
                         if (rect.x == 0 && rect.y == 0 &&
                             rect.w == dst_surface->config.size.w  &&
                             rect.h == dst_surface->config.size.h &&
                             rect.w == data->width         &&
                             rect.h == height)
                         {
                              for (y=0; y<height; y++)
                                   direct_memcpy( p_lock->addr + p_lock->pitch * y,
                                                  data->image + data->pitch * y,
                                                  data->width );

                              break;
                         }
                    }
                    /* fall through */

               case PNG_COLOR_TYPE_GRAY: {
                    D_PRINT("444444444444444\n");
                    /*
                                    * Convert to ARGB and use generic loading code.
                                    */

                    // FIXME: allocates four additional bytes because the scaling functions
                    //        in src/misc/gfx_util.c have an off-by-one bug which causes
                    //        segfaults on darwin/osx (not on linux)                
                    int size = data->width * height * 4 + 4;

                    /* allocate image data */
                    void *image_argb = D_MALLOC( size );

                    if (!image_argb) {
                         D_ERROR( "DirectFB/ImageProvider_PNG: Could not "
                                  "allocate %d bytes of system memory!\n", size );
                         ret = DFB_NOSYSTEMMEMORY;
                    }
                    else {
#if 0
                         if (data->color_type == PNG_COLOR_TYPE_GRAY) {
                              int num = 1 << data->info_ptr->bit_depth;
                              if (is_first_callback) {
                                   D_PRINT("*****first here\n");
                                   for (x=0; x<num; x++) {
                                        int value = x * 255 / (num - 1);

                                        data->palette[x] = 0xff000000 | (value << 16) | (value << 8) | value;
                                   }
                              }
                         }
#endif

                         switch (data->info_ptr->bit_depth) {
                              case 8:
                                   for (y=0; y<height; y++) {
                                        u8  *S = data->image + data->pitch * y;
                                        u32 *D = image_argb  + data->width * y * 4;

                                        for (x=0; x<data->width; x++)
                                             D[x] = data->palette[ S[x] ];
                                   }
                                   break;

                              case 4:
                                   for (y=0; y<height; y++) {
                                        u8  *S = data->image + data->pitch * y;
                                        u32 *D = image_argb  + data->width * y * 4;

                                        for (x=0; x<data->width; x++) {
                                             if (x & 1)
                                                  D[x] = data->palette[ S[x>>1] & 0xf ];
                                             else
                                                  D[x] = data->palette[ S[x>>1] >> 4 ];
                                        }
                                   }
                                   break;

                              case 2:
                                   for (y=0; y<height; y++) {
                                        int  n = 6;
                                        u8  *S = data->image + data->pitch * y;
                                        u32 *D = image_argb  + data->width * y * 4;

                                        for (x=0; x<data->width; x++) {
                                             D[x] = data->palette[ (S[x>>2] >> n) & 3 ];

                                             n = (n ? n - 2 : 6);
                                        }
                                   }
                                   break;

                              case 1:
                                   for (y=0; y<height; y++) {
                                        int  n = 7;
                                        u8  *S = data->image + data->pitch * y;
                                        u32 *D = image_argb  + data->width * y * 4;

                                        for (x=0; x<data->width; x++) {
                                             D[x] = data->palette[ (S[x>>3] >> n) & 1 ];

                                             n = (n ? n - 1 : 7);
                                        }
                                   }
                                   break;

                              default:
                                   D_ERROR( "ImageProvider/PNG: Unsupported indexed bit depth %d!\n",
                                            data->info_ptr->bit_depth );
                         }

                         dfb_scale_linear_32( image_argb, data->width, height,
                                              p_lock->addr, p_lock->pitch, &rect, dst_surface, &clip );
 
                         D_FREE( image_argb );
                    }
                    break;
               }
               default:
                    D_PRINT("55555555555555555\n");
                    /*
                                    * Generic loading code.
                                    */
                    dfb_scale_linear_32( data->image, data->width, height,
                                         p_lock->addr, p_lock->pitch, &rect, dst_surface, &clip );
                    break;
          }
          D_PRINT("666666666666666\n");
#if 0
          if (is_last_callback) {
               D_PRINT("*****last here\n");
               dfb_surface_unlock_buffer( dst_surface, &data->cb_lock );
          }
#endif
     }


     return ret;
}


static DFBResult
PNG_RenderTo( IDirectFBImageProvider *thiz,
                     IDirectFBSurface       *destination,
                     const DFBRectangle     *dest_rect,
                     bool useNonCache)

{ 
     IDirectFBSurface_data *dst_data;
     CoreSurface           *dst_surface;
     DFBResult              ret = DFB_OK;

     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_PNG)

     dst_data = (IDirectFBSurface_data*) destination->priv;
     if (!dst_data)
          return DFB_DEAD;

     dst_surface = dst_data->surface;
     if (!dst_surface)
          return DFB_DESTROYED;
     
     // the destination is on system memory we fall back to software implemenation then.
     D_PNG("RenderTo dst surface caps: 0x%08x, format: %s\n", 
                    dst_data->caps, dfb_pixelformat_name(dst_surface->config.format));
     D_PNG("png image size (%d, %d)\n", data->width, data->height);
#if !XCODE_DISABLE_HW_PNG_DECODE
     if (!dfb_config->disable_hw_png_decode && data->hw_supported)
     {
          IDirectFBImageProvider_HW_data hw_data;
          hw_data.buffer = data->buffer;
          hw_data.render_callback = data->render_callback;
          hw_data.render_callback_context = data->render_callback_context;
          hw_data.width = data->width;
          hw_data.height = data->height;
          hw_data.core = data->core;
          hw_data.pp_surf = &(data->surf);
          ret = HardwareImg_RenderTo( &hw_data, destination, dest_rect, !useNonCache, DFB_HW_PNG);
          
          if (ret == DFB_OK) 
               return DFB_OK;
     }
#endif

     return SoftwarePNG_RenderTo( thiz, destination, dest_rect, useNonCache );
    
}


static DFBResult
IDirectFBImageProvider_PNG_RenderTo( IDirectFBImageProvider *thiz,
                                     IDirectFBSurface       *destination,
                                     const DFBRectangle     *dest_rect )
{
    return PNG_RenderTo(thiz, destination, dest_rect, true);
}


/* This version needs to be used when there is no image scaling involved ... */
static DFBResult
IDirectFBImageProvider_PNG_NoncacheRenderTo( IDirectFBImageProvider *thiz,
                                     IDirectFBSurface       *destination,
                                     const DFBRectangle     *dest_rect )
{
     return PNG_RenderTo(thiz, destination, dest_rect, true);

}

static DFBResult
IDirectFBImageProvider_PNG_SetRenderCallback( IDirectFBImageProvider *thiz,
                                              DIRenderCallback        callback,
                                              void                   *context )
{
     DIRECT_INTERFACE_GET_DATA (IDirectFBImageProvider_PNG)

     data->render_callback         = callback;
     data->render_callback_context = context;

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_PNG_GetSurfaceDescription( IDirectFBImageProvider *thiz,
                                                  DFBSurfaceDescription *dsc )
{
     DFBSurfacePixelFormat primary_format = dfb_primary_layer_pixelformat();

     DIRECT_INTERFACE_GET_DATA (IDirectFBImageProvider_PNG)

     dsc->flags  = DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_PIXELFORMAT;
     dsc->width  = data->width;
     dsc->height = data->height;

     if (data->color_type & PNG_COLOR_MASK_ALPHA)
          dsc->pixelformat = DFB_PIXELFORMAT_HAS_ALPHA(primary_format) ? primary_format : DSPF_ARGB;
     else
          dsc->pixelformat = primary_format;

     if (data->color_type == PNG_COLOR_TYPE_PALETTE) {
          dsc->flags  |= DSDESC_PALETTE;

          dsc->palette.entries = data->colors;  /* FIXME */
          dsc->palette.size    = 256;
     }

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_PNG_GetImageDescription( IDirectFBImageProvider *thiz,
                                                DFBImageDescription    *dsc )
{
     DIRECT_INTERFACE_GET_DATA(IDirectFBImageProvider_PNG)

     if (!dsc)
          return DFB_INVARG;

     dsc->caps = DICAPS_NONE;

     if (data->color_type & PNG_COLOR_MASK_ALPHA)
          dsc->caps |= DICAPS_ALPHACHANNEL;

     if (data->color_keyed) {
          dsc->caps |= DICAPS_COLORKEY;

          dsc->colorkey_r = (data->color_key & 0xff0000) >> 16;
          dsc->colorkey_g = (data->color_key & 0x00ff00) >>  8;
          dsc->colorkey_b = (data->color_key & 0x0000ff);
     }

#if 1
     if (data->hw_supported) {
          dsc->caps |= DICAPS_HW_DECODE;
     }
#else
     if (data->width * data->height <= 2000 * 2000) {
          dsc->caps |= DICAPS_HW_DECODE;
     }
#endif

     return DFB_OK;
}

/**********************************************************************************************************************/

#define MAXCOLORMAPSIZE 256

static int SortColors (const void *a, const void *b)
{
     return (*((const u8 *) a) - *((const u8 *) b));
}

/*  looks for a color that is not in the colormap and ideally not
    even close to the colors used in the colormap  */
static u32 FindColorKey( int n_colors, u8 *cmap )
{
     u32   color = 0xFF000000;
     u8    csort[n_colors];
     int   i, j, index, d;

     if (n_colors < 1)
          return color;

     for (i = 0; i < 3; i++) {
          direct_memcpy( csort, cmap + (n_colors * i), n_colors );
          qsort( csort, n_colors, 1, SortColors );

          for (j = 1, index = 0, d = 0; j < n_colors; j++) {
               if (csort[j] - csort[j-1] > d) {
                    d = csort[j] - csort[j-1];
                    index = j;
               }
          }
          if ((csort[0] - 0x0) > d) {
               d = csort[0] - 0x0;
               index = n_colors;
          }
          if (0xFF - (csort[n_colors - 1]) > d) {
               index = n_colors + 1;
          }

          if (index < n_colors)
               csort[0] = csort[index] - (d/2);
          else if (index == n_colors)
               csort[0] = 0x0;
          else
               csort[0] = 0xFF;

          color |= (csort[0] << (8 * (2 - i)));
     }

     return color;
}

/* Called at the start of the progressive load, once we have image info */
static void
png_info_callback( png_structp png_read_ptr,
                   png_infop   png_info_ptr )
{
     int                              i;
     IDirectFBImageProvider_PNG_data *data;

     data = png_get_progressive_ptr( png_read_ptr );

     /* error stage? */
     if (data->stage < 0)
          return;

     /* set info stage */
     data->stage = STAGE_INFO;

     png_get_IHDR( data->png_ptr, data->info_ptr,
                   &data->width, &data->height, &data->bpp, &data->color_type,
                   NULL, NULL, NULL );

     if (png_get_valid( data->png_ptr, data->info_ptr, PNG_INFO_tRNS )) {
          data->color_keyed = true;

          /* generate color key based on palette... */
          if (data->color_type == PNG_COLOR_TYPE_PALETTE) {
               u32        key;
               png_colorp palette    = data->info_ptr->palette;
               png_bytep  trans      = data->info_ptr->trans;
               int        num_colors = MIN( MAXCOLORMAPSIZE,
                                            data->info_ptr->num_palette );
               u8         cmap[3][num_colors];

               for (i=0; i<num_colors; i++) {
                    cmap[0][i] = palette[i].red;
                    cmap[1][i] = palette[i].green;
                    cmap[2][i] = palette[i].blue;
               }

               key = FindColorKey( num_colors, &cmap[0][0] );

               for (i=0; i<data->info_ptr->num_trans; i++) {
                    if (!trans[i]) {
                         palette[i].red   = (key & 0xff0000) >> 16;
                         palette[i].green = (key & 0x00ff00) >>  8;
                         palette[i].blue  = (key & 0x0000ff);
                    }
               }

               data->color_key = key;
          }
          else {
               /* ...or based on trans rgb value */
               png_color_16p trans = &data->info_ptr->trans_values;

               data->color_key = (((trans->red & 0xff00) << 8) |
                                  ((trans->green & 0xff00)) |
                                  ((trans->blue & 0xff00) >> 8));
          }
     }

     switch (data->color_type) {
          case PNG_COLOR_TYPE_PALETTE: {
               png_colorp palette    = data->info_ptr->palette;
               png_bytep  trans      = data->info_ptr->trans;
               int        num_trans  = data->info_ptr->num_trans;
               int        num_colors = MIN( MAXCOLORMAPSIZE, data->info_ptr->num_palette );

               for (i=0; i<num_colors; i++) {
                    data->colors[i].a = (i < num_trans) ? trans[i] : 0xff;
                    data->colors[i].r = palette[i].red;
                    data->colors[i].g = palette[i].green;
                    data->colors[i].b = palette[i].blue;

                    data->palette[i] = PIXEL_ARGB( data->colors[i].a,
                                                   data->colors[i].r,
                                                   data->colors[i].g,
                                                   data->colors[i].b );
               }

               data->pitch = (data->width + 7) & ~7;
               break;
          }

          case PNG_COLOR_TYPE_GRAY:
               data->pitch = data->width;

               if (data->bpp == 16)
                    png_set_strip_16( data->png_ptr );

               break;

          case PNG_COLOR_TYPE_GRAY_ALPHA:
               png_set_gray_to_rgb( data->png_ptr );
               /* fall through */

          default:
               data->pitch = data->width * 4;

               if (data->bpp == 16)
                    png_set_strip_16( data->png_ptr );

#ifdef WORDS_BIGENDIAN
               if (!(data->color_type & PNG_COLOR_MASK_ALPHA))
                    png_set_filler( data->png_ptr, 0xFF, PNG_FILLER_BEFORE );

               png_set_swap_alpha( data->png_ptr );
#else
               if (!(data->color_type & PNG_COLOR_MASK_ALPHA))
                    png_set_filler( data->png_ptr, 0xFF, PNG_FILLER_AFTER );

               png_set_bgr( data->png_ptr );
#endif
               break;
     }

     data->num_pass = png_set_interlace_handling( data->png_ptr );

     /* Update the info to reflect our transformations */
     png_read_update_info( data->png_ptr, data->info_ptr );


     if (data->color_type == PNG_COLOR_TYPE_GRAY) {
          int num = 1 << data->info_ptr->bit_depth;
          D_PRINT("*****first initialize the palette here\n");
          int x;
          for (x=0; x<num; x++) {
               int value = x * 255 / (num - 1);

               data->palette[x] = 0xff000000 | (value << 16) | (value << 8) | value;
          }
     }
}

/* Called for each row; note that you will get duplicate row numbers
   for interlaced PNGs */
static void
png_row_callback( png_structp png_read_ptr,
                  png_bytep   new_row,
                  png_uint_32 row_num,
                  int         pass_num )
{
     IDirectFBImageProvider_PNG_data *data;

     data = png_get_progressive_ptr( png_read_ptr );

     /* error stage? */
     if (data->stage < 0)
          return;

     /* set image decoding stage */
     data->stage = STAGE_IMAGE;

     /* check image data pointer */
     if (!data->image) {
          // FIXME: allocates four additional bytes because the scaling functions
          //        in src/misc/gfx_util.c have an off-by-one bug which causes
          //        segfaults on darwin/osx (not on linux)                
          int size;
          if (data->use_row_buffer) {
               size = data->pitch + 4;
          }
          else {
               size = data->pitch * data->height + 4;
          }
          D_PRINT("png sw decoder buffer size = %d\n", size);

          /* allocate image data */
          data->image = D_CALLOC( 1, size );
          if (!data->image) {
               D_ERROR("DirectFB/ImageProvider_PNG: Could not "
                        "allocate %d bytes of system memory!\n", size);

               /* set error stage */
               data->stage = STAGE_ERROR;

               return;
          }
          data->rows = 0;
     }
     D_PRINT("use_row_buffer=%d\n", data->use_row_buffer);

     /* write to image data */
     if(data->use_row_buffer) {
          png_progressive_combine_row( data->png_ptr, data->image, new_row );
          DFBRectangle rect = { data->cb_rect.x, data->cb_rect.y + row_num, data->cb_rect.w, 1 };
          //DFBRectangle rect = { 0, row_num, data->width, 1 };
          
          //D_PRINT("%d, %d, %d, %d\n", rect.x, rect.y, rect.w, rect.h);
          //DFBRectangle m_sw_clipped = data->cb_clipped;
          
          //if(rect.y >= m_sw_clipped.y) {
               //bool is_first_cb = (rect.y == m_sw_clipped.y);
               //bool is_last_cb = (rect.y == m_sw_clipped.y + m_sw_clipped.h);
               
               //m_sw_clipped.h = 1;
               //m_sw_clipped.y += data->rows - (data->cb_clipped.y - data->cb_rect.y);

               //write_surf_buffer_png( data, 1, rect, m_sw_clipped);
               write_surf_buffer_png( data, 1, rect);
          //}

     }
     else {
          png_progressive_combine_row( data->png_ptr, (png_bytep) (data->image +
                                   row_num * data->pitch), new_row );
     }

     /* increase row counter, FIXME: interlaced? */
     data->rows++;

     if (data->render_callback) {
          DIRenderCallbackResult r;
          DFBRectangle rect = { 0, row_num, data->width, 1 };

          r = data->render_callback( &rect, data->render_callback_context );
          if (r != DIRCR_OK)
              data->stage = STAGE_ABORT;
     }
}

/* Called after reading the entire image */
static void
png_end_callback   (png_structp png_read_ptr,
                    png_infop   png_info_ptr)
{
     IDirectFBImageProvider_PNG_data *data;

     data = png_get_progressive_ptr( png_read_ptr );

     /* error stage? */
     if (data->stage < 0)
          return;

     /* set end stage */
     data->stage = STAGE_END;
}

/* Pipes data into libpng until stage is different from the one specified. */
static DFBResult
push_data_until_stage (IDirectFBImageProvider_PNG_data *data,
                       int                              stage,
                       int                              buffer_size)
{
     DFBResult            ret;
     IDirectFBDataBuffer *buffer = data->buffer;

     while (data->stage < stage) {
          unsigned int  len;
          unsigned char buf[buffer_size];

          if (data->stage < 0)
               return DFB_FAILURE;

          while (buffer->HasData( buffer ) == DFB_OK) {
               D_DEBUG( "ImageProvider/PNG: Retrieving data (up to %d bytes)...\n", buffer_size );

               ret = buffer->GetData( buffer, buffer_size, buf, &len );
               if (ret)
                    return ret;

               D_DEBUG( "ImageProvider/PNG: Got %d bytes...\n", len );

               png_process_data( data->png_ptr, data->info_ptr, buf, len );

               D_DEBUG( "ImageProvider/PNG: ...processed %d bytes.\n", len );

               /* are we there yet? */
               if (data->stage < 0 || data->stage >= stage) {
                   switch (data->stage) {
                        case STAGE_ABORT: return DFB_INTERRUPTED;
                        case STAGE_ERROR: return DFB_FAILURE;
                        default:          return DFB_OK;
                   }
               }
          }

          D_DEBUG( "ImageProvider/PNG: Waiting for data...\n" );

          if (buffer->WaitForData( buffer, 1 ) == DFB_EOF)
               return DFB_FAILURE;
     }

     return DFB_OK;
}
