#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

#include <directfb.h>

#include <display/idirectfbsurface.h>

#include <media/idirectfbimageprovider.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/layers.h>
#include <core/surface.h>

#include <misc/gfx_util.h>
#include <misc/util.h>
#include <direct/interface.h>
#include <direct/mem.h>
#include <direct/memcpy.h>
#include <direct/messages.h>

#include <setjmp.h>
#include <math.h>

#undef HAVE_STDLIB_H
#include "xcode4_img_api.h"
#include "xcode4/xcdfb.h"
#include <core/core.h>
#include <fusion/arena.h>
#include <fusion/fusion.h>

#include "viper.h"
#include "xcode4_img_ycbcr2argb.h"

#include "xcode4_imageprovider_HW.h"


#define FMT_NAME(fmt) dfb_pixelformat_name(fmt)



#define REG_READ(reg) *((volatile u32*)(mmio_base+reg)) 
#define REG_WRITE(reg, value) *((volatile u32*)(mmio_base+reg))=value


#define LOCK_MUTEX_REG_ONCE(reg, id, value) \
     REG_WRITE(reg, value << reg##_##id##_SHIFT); \
     if ((REG_READ(reg) & reg##_##id##_MASK) != (value << reg##_##id##_SHIFT)){\
          ret = -1;\
     }


#define LOCK_MUTEX_REG(reg, id, value) \
    REG_WRITE(reg, value << reg##_##id##_SHIFT); \
    while (1) {\
        if ((REG_READ(reg) & reg##_##id##_MASK) == (value << reg##_##id##_SHIFT)){\
            break; \
        } \
        else{\
            REG_WRITE(reg, value << reg##_##id##_SHIFT);\
        }\
    }

#define UNLOCK_MUTEX_REG(reg, id, value) \
    REG_WRITE(reg, value << reg##_##id##_SHIFT); \
    while (REG_READ(reg) & reg##_##id##_MASK){\
         REG_WRITE(reg, value << reg##_##id##_SHIFT);\
    }

#define LOCK_SIMD_DECODER LOCK_MUTEX_REG_ONCE(MIPS_INTERLOCK4, ID2, 1)
#define UNLOCK_SIMD_DECODER UNLOCK_MUTEX_REG(MIPS_INTERLOCK4, ID2, 1)



#define D_HW_IMG(...) D_DEBUG_AT( HW_IMG_T, __VA_ARGS__)

D_DEBUG_DOMAIN( HW_IMG_T, "hw_img", "H/W decoding" );




DFBResult
IMG_HW_cache_cleanup(CoreSurface **p_surf)
{
     // cleanup the cache
     if (*p_surf)
     {
          dfb_surface_destroy_buffers( *p_surf );
          dfb_surface_unref( *p_surf );
          *p_surf = NULL;
     }

     return DFB_OK;
}


unsigned int 
get_next_sample_color(unsigned int color, int *channel)
{
     int color_channel = (*channel);
     color &=  0xffffff;

     if ((color >> color_channel) >= 255)
     {
          color_channel+=8;
          if (color_channel > 16) {
               color_channel = 0;
          }
     }
     color = (((color >> color_channel) + 1) & 0xff) << color_channel;
     *channel = color_channel;

     color |= 0xff000000;
     return color;
}

void
generate_sample_image(void *addr, DFBRectangle *rect, unsigned int pitch, DFBSurfacePixelFormat fmt)
{
     unsigned char *uaddr = (unsigned char *) addr;
     unsigned int argb_color = 0, start_argb_color = 0;
     unsigned int color_channel = 0xff000000, start_color_channel = 0;
     int i ,j;

     if (fmt == DSPF_ARGB)
     {
          D_HW_IMG("HardwareImgDecode: Generate debug argb test sample!\n");
          for (i=0;i < rect->h;i++)
          {
               unsigned int *p_argb = (unsigned int *)(uaddr+i*pitch);
               argb_color = start_argb_color;
               color_channel = start_color_channel;
               for (j=0;j < rect->w;j++)
               {
                   p_argb[j] = argb_color;
                   argb_color = get_next_sample_color(argb_color, &color_channel); 
               }

               start_argb_color = get_next_sample_color(start_argb_color, &start_color_channel);
          }
     }
     else
     {
          //TODO: waiting for the the YUV to ARGB equations from hardware team.
          D_HW_IMG("HardwareImgDecode: TODO: Generate debug YUV I420 test sample!\n");
     }
}



DFBResult
HardwareImgDecode(IDirectFBImageProvider_HW_data *data
                   , unsigned int width
                   , unsigned int height
                   , unsigned int size
                   , unsigned int address
                   , DFBSurfacePixelFormat format
                   , DFB_HW_img_type img_type)
{
     XC4Shared *shared = NULL;

     void *h_img;
     char *buf = NULL;
     unsigned int buf_len = 0;

     int file_read_size;
     unsigned int curr_buf_pos = 0;
     unsigned int total_size = 0;

     DFBResult ret = DFB_OK;
     int dret = 0;

     static volatile unsigned char * mmio_base = NULL; 


     ret = fusion_arena_get_shared_field( dfb_core_arena( data->core ),
                                          "xcode4", (void **)&shared );
     
     if (ret != DFB_OK)
     {
          D_ERROR( "ImageProvider: Error while retrieving for the lock to povide exclusive access of img decoder!\n" );
          return ret;
     }

     ret = fusion_skirmish_prevail( &shared->img_hw_lock );

     if (ret != 0)
     {
          D_ERROR( "ImageProvider: Error while locking for the exclusive access of img decoder!\n" );
          return ret;
     }

     if(img_type == DFB_HW_JPEG) { //PNG is moved to Proc3, so now do not need to lock mutex.
          if(!mmio_base) {
               mmio_base = xcode_get_mmr_map_addr();
               D_HW_IMG("DFB HW decoder: obtaining mmio_base\n");
               if(!mmio_base) {
                    goto HW_DECODE_EXIT;
               }
                    
          }

          LOCK_SIMD_DECODER;
          if (ret) {
               D_HW_IMG("HW decoder can not be locked!\n");
               goto HW_DECODE_EXIT;
          }
     }

     D_HW_IMG("firmware started decoding....\n");
     h_img = img_decode_init();

     if (!h_img)
     { 
          D_HW_IMG("img_decode_init() failed!\n");
          ret = DFB_FAILURE;
          goto HW_DECODE_EXIT;
     }
     
     ret = img_decoder_open(h_img, width, height, size, address, format, img_type);

     if (ret != 0)
     {
          D_HW_IMG("open img decoder error!\n");
          ret = DFB_FAILURE;
          goto HW_DECODE_DEINT;
     }
     TIME("time taken for opening decoder");
     
     IDirectFBDataBuffer *data_buffer = data->buffer;

     data_buffer->GetPosition(data_buffer, &curr_buf_pos);

     data_buffer->SeekTo(data_buffer, 0);

     buf = (char *)img_get_send_buf_addr(h_img);
     buf_len = img_get_send_buf_size(h_img);

     while (ret != DFB_EOF 
            && dret == 0 
            && ((ret = data_buffer->HasData(data_buffer)) == DFB_OK))
     {
          unsigned int left_read = buf_len;
          char *curr_pos = buf;
          do {
               unsigned int ret_read = 0;
               data_buffer->WaitForDataWithTimeout( data_buffer, left_read, 1, 0 );
               ret = data_buffer->GetData(data_buffer, left_read, curr_pos, &ret_read);
               if (ret != DFB_OK) break;
               left_read -= ret_read;
               curr_pos += ret_read;
          } while (left_read && ret != DFB_EOF);

          if (ret != DFB_OK && ret != DFB_EOF) break;
          total_size += (buf_len - left_read);
          dret = img_decoder_append_data(h_img, buf_len - left_read);
     }

     if (ret == DFB_EOF && dret == 0) {
          TIME("time taken for sending data to decoder");
          
          ret = img_execute_decode(h_img);
          if ( ret != 0 ) {
               ret = DFB_FAILURE;
               D_ERROR( "img_execute_decode() error!\n" );
          }
          else {
               TIME("time taken for executing decoding");
          }
     }
     else
     {
          ret = DFB_FAILURE;
          D_ERROR( "ImageProvider: Error while reading data in hardware decoding!\n" );
     }

     data_buffer->SeekTo(data_buffer, curr_buf_pos);

     img_decoder_close(h_img);

HW_DECODE_DEINT:
     img_decode_deinit(h_img);

     if(img_type == DFB_HW_JPEG) {
          UNLOCK_SIMD_DECODER;
     }
     
HW_DECODE_EXIT:
     D_HW_IMG("firmware finished decoding: %s\n", ret ? "failed":"success");

     fusion_skirmish_dismiss( &shared->img_hw_lock );

     return ret;
}


int calc_surf_size(int pitch, int height, DFBSurfacePixelFormat fmt) 
{
     int size, Y_size, U_size;

     if (fmt == DSPF_ARGB){
          size = pitch * height * 4; //size for the ARGB
     }
     else { 
          Y_size = pitch * height;

          if (fmt == DSPF_YUV444P) {
               U_size = Y_size;                        
          }
          else if (fmt == DSPF_YV16) {
               U_size = Y_size >> 1;
          }
          else {
               U_size = Y_size >> 2;
          }

          size = Y_size + U_size * 2;
     }

     return size;
}

void
surf_blit_internal(CoreSurface *src_surf, CoreSurface *dst_surf, DFBRectangle *p_irect, DFBRegion *p_newclip)
{
     CardState state;
     
     dfb_state_init( &state, NULL );
     dfb_state_set_destination( &state, dst_surf );
     dfb_state_set_source( &state, src_surf );

     dfb_region_from_rectangle( p_newclip, p_irect );
     dfb_state_set_clip( &state, p_newclip );

     dfb_gfxcard_blit( p_irect, 0, 0, &state );
     dfb_state_set_source( &state, NULL );
     dfb_state_set_destination( &state, NULL );
     dfb_state_destroy( &state );
}


void
optimized_videoYUV_to_systemARGB(CoreSurface *src_surf, DFBRectangle irect, 
                    CoreSurface *dst_surf, DFBRectangle drect, DFBRegion clip,
                    CoreDFB *core, CardState *p_dst_state )
{
     CoreSurface           *sys_surface;
     CoreSurfaceBufferLock  lock;
     CoreSurfaceConfig iconfig;
     int ret;
     int Y_size, U_size;

     DFBSurfacePixelFormat src_fmt = src_surf->config.format;

     if (irect.w==drect.w && irect.h==drect.h) {
          unsigned char *input_buf[3];
          unsigned char * output_buf;

          void * source;
          void * dest;
          CoreSurfaceBufferLock  dst_lock;
          
          //dfb_clip_rectangle( &clip, &drect );

          //borrowed from dfb_copy_buffer_32(), simulate the process in _SoftwareJPEG_Renderto
          //clip is the destination surface boundary, drect is the destination rectangle to blit.
          //works the same as dfb_clip_rectangle(). But here, we need variable sx, sy, so give up dfb_clip_rectangle()
          int sx = 0, sy = 0;
          if (drect.x < clip.x1) {
               sx = clip.x1 - drect.x;
               drect.w -= sx;
               drect.x += sx;
          }
          if (drect.y < clip.y1) {
               sy = clip.y1 - drect.y;
               drect.h -= sy;
               drect.y += sy;
          }
          if ((drect.x + drect.w - 1) > clip.x2) {
               drect.w -= drect.x + drect.w - 1 - clip.x2;
          }
          if ((drect.y + drect.h - 1) > clip.y2) {
               drect.h -= drect.y + drect.h - 1 - clip.y2;
          }
          
          DFBRectangle drect_tmp = {sx, sy, drect.w, drect.h};
          if (src_fmt == DSPF_I420 && (drect_tmp.h & 1)) { //avoid the odd line number of I420
              drect_tmp.h++;
          }
          
          //avoid the odd width of 420 and 422
          if ((src_fmt == DSPF_I420 || src_fmt == DSPF_YV16) && (drect_tmp.w & 1)) { 
              drect_tmp.w++;
          }


          /**************copying the video memory to system memory***********/
          iconfig.flags  = CSCONF_SIZE | CSCONF_FORMAT | CSCONF_CAPS;
          iconfig.size.w = drect_tmp.w; 
          iconfig.size.h = drect_tmp.h; 
          iconfig.format = src_fmt;
          iconfig.caps   = DSCAPS_SYSTEMONLY;
          ret = dfb_surface_create( core, &iconfig, CSTF_NONE, 0, NULL, &sys_surface );

          DFBRegion newclip = {0, 0, drect_tmp.w - 1, drect_tmp.h - 1}; //size of sys_surface
          
          surf_blit_internal(src_surf, sys_surface, &drect_tmp, &newclip);

          /**********************doing the conversion*********************/
          D_HW_IMG("optimized: decoded video YUV->dest system ARGB!(%d %d %d %d)->(%d %d %d %d)\n",
                    irect.x, irect.y, irect.w, irect.h, drect.x, drect.y, irect.w, irect.h);

          dfb_surface_lock_buffer( dst_surf, CSBR_FRONT, CSAID_CPU, CSAF_WRITE, &dst_lock );
          dest = dst_lock.addr;
          output_buf = dest;

          dfb_surface_lock_buffer( sys_surface, CSBR_FRONT, CSAID_CPU, CSAF_READ, &lock );
          source = lock.addr;


          Y_size = lock.pitch * drect_tmp.h;
          
          if (src_fmt == DSPF_YUV444P) {
               U_size = Y_size;                        
          }
          else if (src_fmt == DSPF_YV16) {
               U_size = Y_size >> 1;
          }
          else {
               U_size = Y_size >> 2;
          }

          input_buf[0] = source;
          input_buf[1] = input_buf[0] + Y_size;
          input_buf[2] = input_buf[1] + U_size;
          //even if the dest is premultiplied, we are safe here, because JPEGs always have alpha channel 0xFF
          if (src_fmt == DSPF_YUV444P) {
               h1v1_merged_upsample(&drect, lock.pitch, dst_lock.pitch, input_buf, output_buf);
          } 
          else if (src_fmt == DSPF_YV16) {
               unsigned char *tmp = input_buf[1];
               input_buf[1] = input_buf[2];
               input_buf[2] = tmp;
               h2v1_merged_upsample(&drect, lock.pitch, dst_lock.pitch, input_buf, output_buf);
          } 
          else {
               h2v2_merged_upsample(&drect, lock.pitch, dst_lock.pitch, input_buf, output_buf);
          }
          //TIME("time taken system %s -> system ARGB", FMT_NAME(src_fmt));  
          
          dfb_surface_unlock_buffer( sys_surface, &lock);
          dfb_surface_unlock_buffer( dst_surf, &dst_lock);
     }
     else {//if stretch is needed, then use the DFB way to do the conversion
          int surf_width = src_surf->config.size.w;
          int surf_height = src_surf->config.size.h;
          
          iconfig.flags  = CSCONF_SIZE | CSCONF_FORMAT | CSCONF_CAPS;
          iconfig.size.w = surf_width;
          iconfig.size.h = surf_height;
          iconfig.format = src_fmt;
          iconfig.caps   = DSCAPS_SYSTEMONLY;
          ret = dfb_surface_create( core, &iconfig, CSTF_NONE, 0, NULL, &sys_surface );

          if (src_fmt == DSPF_I420 && (irect.h & 1)) { //avoid the odd line number of I420
              irect.h++;
          }
          
          //avoid the odd width of 420 and 422
          if ((src_fmt == DSPF_I420 || src_fmt == DSPF_YV16) && (irect.w & 1)) { 
              irect.w++;
          }

          DFBRegion newclip;
          dfb_region_from_rectangle( &newclip, &irect );//here the irect is the dest clip
          surf_blit_internal(src_surf, sys_surface, &irect, &newclip);


          //stretchblit
          dfb_state_set_source( p_dst_state, sys_surface );
          dfb_gfxcard_stretchblit( &irect, &drect, p_dst_state );
          D_HW_IMG("stretchblit: decoded video %s-> dest system %s (%d %d %d %d)->(%d %d %d %d)\n", 
               FMT_NAME(sys_surface->config.format), FMT_NAME(dst_surf->config.format), 
                      irect.x, irect.y, irect.w, irect.h, drect.x, drect.y, drect.w, drect.h);
     }

     dfb_surface_destroy_buffers( sys_surface );
     dfb_surface_unref( sys_surface );  
}



DFBResult 
HardwareImg_RenderTo( IDirectFBImageProvider_HW_data *data,
                       IDirectFBSurface       *destination,
                       const DFBRectangle     *dest_rect, 
                       bool                    need_cached,
                       DFB_HW_img_type         img_type)
{
     DFBResult              ret = DFB_OK;
     CoreSurfaceBufferLock  lock;
     IDirectFBSurface_data *dst_data = NULL;
     CoreSurface           *dst_surface = NULL;
     CardState             *dst_state = NULL;
     DFBRectangle           irect, drect;
     CoreSurface           *decode_surf = NULL;
     CoreSurface           *cur_surf = NULL;
     unsigned int           phy_addr;
     unsigned int           aligned_width = 0, aligned_height = 0;
     CoreSurfaceConfig      iconfig;
     DFBSurfacePixelFormat  dest_format;

     DFBRegion              clip;


     TIME_BEGIN();
     
     dst_data = (IDirectFBSurface_data*) destination->priv;
     dst_surface = dst_data->surface;
     dst_state = &dst_data->state;

     ret = destination->GetPixelFormat( destination, &dest_format );
     if (ret !=DFB_OK) return ret;

     if (dest_format != DSPF_ARGB && dest_format != DSPF_I420)
     {
          return DFB_UNSUPPORTED;
     }

     // Decoder required the buffer aligned to 64 pixels
     aligned_width = (data->width + 0x3f) & ~0x3fUL;   
     aligned_height = (data->height + 0x3f) & ~0x3fUL;
     
#if 0     
     /*determine if we regress to software for decoding caused by hardware limitation from G2D source pitch.*/
     //if(aligned_width * 4 >= G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK ) {//the condition is not very accurate.
          //D_HW_IMG("width %d (aligned to %d) greater than hardware limitation %d, use software instead.\n", data->width, aligned_width, G2D_SRC_PITCH_FORMAT_SRC_PITCH_MASK>>2);
          //return DFB_UNSUPPORTED;
     //}
#endif

     D_HW_IMG("HW image decoder is used\n");
     D_HW_IMG("image size (%d, %d)\n", data->width, data->height);
     
     
     bool is_align_needed = (aligned_width != data->width) || (aligned_height != data->height);


     dfb_region_from_rectangle( &clip, &dst_data->area.current );

     if (dest_rect) 
     {
          if (dest_rect->w < 1  ||  dest_rect->h < 1)
          {
               D_ERROR( "width or height of destination rectangle is smaller than 1!\n" );
               return DFB_INVARG;
          }
          else
          {
               drect = *dest_rect;

               drect.x += dst_data->area.wanted.x;
               drect.y += dst_data->area.wanted.y;

               if (!dfb_rectangle_region_intersects( &drect, &clip ))
                    return DFB_OK;
          }
     }
     else
     {
          drect = dst_data->area.wanted;
     }

     irect.x = 0;
     irect.y = 0;
     irect.w = data->width;
     irect.h = data->height;


     bool is_dst_sys_mem = false;

     //if is_dst_sys_mem == true, enable the route that firmware output YUV, 
     //then software convert YUV to system memory ARGB.
     //comment this to disable this function. 
     ret = dfb_surface_lock_buffer( dst_surface, CSBR_FRONT, CSAID_GPU, CSAF_WRITE, &lock );
     if (ret) {
          is_dst_sys_mem = true;
     } else {
          is_dst_sys_mem = false;
          dfb_surface_unlock_buffer( dst_surface, &lock );
     }
          
     
     bool is_dst_premultiplied = dst_data->caps & DSCAPS_PREMULTIPLIED;

     D_HW_IMG("dest format = %s\n", FMT_NAME(dest_format));
     D_HW_IMG("dest in system memory = %d\n", is_dst_sys_mem);
     D_HW_IMG("dest premultiplied = %d\n", is_dst_premultiplied);

#if 0
     bool is_dst_preallocated = 0;
     is_dst_preallocated = dst_surface->config.flags & CSCONF_PREALLOCATED;     
     D_HW_IMG("dest preallocated = %d\n", is_dst_preallocated);
     is_dst_sys_mem |= is_dst_preallocated;
#endif


     DFBSurfacePixelFormat format_to_gen;

     if(img_type == DFB_HW_JPEG) {
          DFBSurfacePixelFormat native_format = data->comp_colorspace;

          if((dest_format == DSPF_ARGB) && dfb_config->enable_jpeg_decode_argb_output){
               format_to_gen = DSPF_ARGB;
          } else {
               format_to_gen = native_format;
          }

          D_HW_IMG("native_format = %s, format_to_gen = %s\n", FMT_NAME(native_format),
                 FMT_NAME(format_to_gen));
     }
     else if(img_type == DFB_HW_PNG) {
          format_to_gen = DSPF_ARGB;
     }
     else {
          D_HW_IMG("image type not supported by H/W!\n");
          return DFB_UNSUPPORTED;
     }


     bool need_format_conv = (dest_format != format_to_gen);
     bool need_scale = (irect.w != drect.w) || (irect.h != drect.h);
     

     if (need_cached && *(data->pp_surf))
     {
          decode_surf = *(data->pp_surf);
     }
     else
     {
          IMG_HW_cache_cleanup(data->pp_surf);
          
          bool need_int_dec_surf = need_cached || is_align_needed || need_scale
                               || need_format_conv || is_dst_sys_mem || is_dst_premultiplied;
          
          if (need_int_dec_surf) {
               iconfig.flags  = CSCONF_SIZE | CSCONF_FORMAT | CSCONF_CAPS;
               iconfig.size.w = aligned_width;
               iconfig.size.h = aligned_height;
               iconfig.format = format_to_gen;
               iconfig.caps   = DSCAPS_VIDEOONLY;
               ret = dfb_surface_create( data->core, &iconfig, CSTF_LAYER, 0, NULL, &decode_surf );
               D_HW_IMG("decode surface %s (%d,%d) created in video mem\n", FMT_NAME(format_to_gen),
                              aligned_width, aligned_height); 
               if (ret != DFB_OK)
               {
                    D_HW_IMG("create decode_surf fails: 0x%08x\n", ret);
                    return ret;
               }
          }
          else {
               decode_surf = dst_surface;
          }

          ret = dfb_surface_lock_buffer( decode_surf, CSBR_FRONT, CSAID_GPU, CSAF_WRITE, &lock );

          if (ret != DFB_OK)
          {
               D_HW_IMG("lock decode_surf fails: 0x%08x\n", ret);
               return ret;
          }

          phy_addr = lock.phys;

          int size = calc_surf_size(lock.pitch, aligned_height, format_to_gen);


          TIME("time taken before hardware decoding");
#if DEBUG_WITHOUT_FIRMWARE_DECODE
          generate_sample_image(lock.addr, &irect, lock.pitch, format_to_gen);
#else
          ret = HardwareImgDecode(data, aligned_width, aligned_height,
                              size, phy_addr, format_to_gen, img_type);

          if(ret != DFB_OK) {
               D_HW_IMG("hardware decoding fails: %d\n", ret);
               dfb_surface_unlock_buffer( decode_surf, &lock );

               dfb_surface_destroy_buffers( decode_surf );
               dfb_surface_unref( decode_surf );      

               return ret;
          }
#endif
          TIME("time taken for hardware decoding %s", FMT_NAME(decode_surf->config.format));

          ret = dfb_surface_unlock_buffer( decode_surf, &lock );
          if(ret != DFB_OK) {
               D_HW_IMG("unlock error: %d\n", ret);
          }

          if (need_cached)
          {
               *(data->pp_surf) = decode_surf; 
               D_HW_IMG("decoded image surface cached %s.\n"
                    , FMT_NAME((*(data->pp_surf))->config.format));
          } 
     }

     cur_surf = decode_surf;
     bool need_second_int_surf = false;

     //another intermediate video buffer is needed.
     if (need_format_conv && need_scale && !is_dst_sys_mem)
     {
          CoreSurface           *format_conv_surf = NULL;

          iconfig.flags  = CSCONF_SIZE | CSCONF_FORMAT | CSCONF_CAPS;
          iconfig.size.w = irect.w;
          iconfig.size.h = irect.h;
          iconfig.format = dest_format;
          iconfig.caps   = DSCAPS_VIDEOONLY;
          ret = dfb_surface_create( data->core, &iconfig, CSTF_LAYER, 0, NULL, &format_conv_surf );
          D_HW_IMG("format conversion surface %s (%d,%d) created in video mem\n", FMT_NAME(dest_format), irect.w, irect.h); 
          if (ret != DFB_OK) {
             D_HW_IMG("creating format_conv_surf failed: 0x%08x\n", ret);
             return ret;
          }

          //intermediate surface clip region = irect 
          DFBRegion newclip = {0, 0, irect.w - 1, irect.h - 1}; 

          surf_blit_internal(decode_surf, format_conv_surf, &irect, &newclip);

          //uncomment if we need to time the exact time taken by hardware
          //dfb_gfxcard_sync(); 
          D_HW_IMG("blit: decoded video %s->intermediate video %s\n", FMT_NAME(decode_surf->config.format), FMT_NAME(dest_format));
          TIME("time taken for decoded video %s->intermediate video %s", FMT_NAME(decode_surf->config.format), FMT_NAME(dest_format));

          if(!need_cached){
               dfb_surface_destroy_buffers( decode_surf );
               dfb_surface_unref( decode_surf );             
          }
          
          cur_surf = format_conv_surf;
          need_second_int_surf = true;
     }
     

     if (cur_surf != dst_surface) {
          if(is_ycbcr[DFB_PIXELFORMAT_INDEX(cur_surf->config.format)] 
                         && dst_surface->config.format == DSPF_ARGB
                         && is_dst_sys_mem == true) {
               optimized_videoYUV_to_systemARGB(cur_surf, irect, 
                     dst_surface, drect, clip, data->core, dst_state);

               if (data->render_callback) {
                    DFBRectangle r = { 0, 0, data->width, data->height };
                    if (data->render_callback( &r, data->render_callback_context ) != DIRCR_OK){
                         ret = DFB_INTERRUPTED;
                    }
               }

               TIME("time taken for intermediate video %s->dest system %s", 
                    FMT_NAME(format_to_gen), FMT_NAME(dst_surface->config.format));
          }
          else {//for all other case, leave it to DFB, SW or HW, which depends. 
               if(is_dst_premultiplied){
                    dfb_state_set_blitting_flags(&dst_data->state, DSBLIT_SRC_PREMULTIPLY);
               }
               
               dfb_state_set_source( &dst_data->state, cur_surf );
               if (irect.w==drect.w && irect.h==drect.h) {
                    dfb_gfxcard_blit( &irect, drect.x, drect.y, &dst_data->state );
                    D_HW_IMG("blit: %s video %s->dest %s %s.(%d %d %d %d)->(%d %d %d %d)\n",
                                need_second_int_surf ? "intermediate" : "decoded",
                                FMT_NAME(cur_surf->config.format), 
                                is_dst_sys_mem ? "system" : "video", FMT_NAME(dest_format),
                              irect.x, irect.y, irect.w, irect.h, drect.x, drect.y, irect.w, irect.h);
               } else {
                    dfb_gfxcard_stretchblit( &irect, &drect, &dst_data->state );
                    D_HW_IMG("stretchblit: %s video %s->dest %s %s.(%d %d %d %d)->(%d %d %d %d)\n",
                                need_second_int_surf ? "intermediate" : "decoded",
                                FMT_NAME(cur_surf->config.format),  
                                is_dst_sys_mem ? "system" : "video", FMT_NAME(dest_format),
                                irect.x, irect.y, irect.w, irect.h, drect.x, drect.y, drect.w, drect.h);
               }

               if (data->render_callback) {
                    DFBRectangle r = { 0, 0, data->width, data->height };
                    if (data->render_callback( &r, data->render_callback_context ) != DIRCR_OK){
                         ret = DFB_INTERRUPTED;
                    }
               }

               TIME("time taken for video %s->dest %s %s", FMT_NAME(cur_surf->config.format), 
                    is_dst_sys_mem ? "system" : "video", FMT_NAME(dst_surface->config.format));
          }
          
          /*consider the case when calling renderto repeatly, data->surf == cur_surf*/
          /*in this case the buffer can not be released*/
          if (cur_surf && (*(data->pp_surf) != cur_surf)){ 
               dfb_surface_destroy_buffers( cur_surf );
               dfb_surface_unref( cur_surf );               
          }
     }


     D_HW_IMG("HardwareRenderTo completed\n\n");

     return ret;
}

