#include <config.h>

#include <directfb.h>
#include <directfb_util.h>

#include <direct/debug.h>
#include <direct/mem.h>

#include <fusion/shmalloc.h>

#include <core/core.h>
#include <core/coredefs.h>

#include <core/surface_buffer.h>
#include <core/surface_pool.h>
#include <core/surface_pool_bridge.h>
#include <core/system.h>

#include <gfx/convert.h>
#include <misc/conf.h>

#include "drv_if.h"

#define MAX_DMA_SIZE 0x20000
#define IOCTL_TIMEOUTVALUE  2000

D_DEBUG_DOMAIN( Xcode4_PoolBridge, "Xcode4PoolBridge", "Xcode4PoolBridge" );

#define DPRINTF(...) D_DEBUG_AT( Xcode4_PoolBridge, __VA_ARGS__)

typedef struct {
    int fd;
} xcode4_dma_handle, *pxcode4_dma_handle;

static
int Xcode4PoolBridgeDataSize( void )
{
    return sizeof(xcode4_dma_handle);
}

static
DFBResult Xcode4InitPoolBridge   ( CoreDFB                          *core,
                                CoreSurfacePoolBridge            *bridge,
                                void                             *bridge_data,
                                void                             *bridge_local,
                                void                             *context,
                                CoreSurfacePoolBridgeDescription *ret_desc )
{
    pxcode4_dma_handle handle = (pxcode4_dma_handle)bridge_data;

    ret_desc->caps = CSPBCAPS_NONE;
    strcpy(ret_desc->name, "Xcode4 DMA Pool Bridge");

    handle->fd = open(DEVICE_NAME, O_RDWR);

    if(handle->fd <= 0)
    {
        return DFB_FAILURE;
    }

    DPRINTF("Open XCODE handler for DMA\n");

    return DFB_OK;
}

static
DFBResult Xcode4DestroyPoolBridge   ( CoreSurfacePoolBridge            *bridge,
                                void                             *bridge_data,
                                void                             *bridge_local )
{
    pxcode4_dma_handle handle = (pxcode4_dma_handle)bridge_data;

    if(handle->fd > 0)
    {
        close(handle->fd);
        handle->fd = 0;
    }

    return DFB_OK;
}

static
DFBResult Xcode4CheckTransfer    ( CoreSurfacePoolBridge            *bridge,
                                void                             *bridge_data,
                                void                             *bridge_local,
                                CoreSurfaceBuffer                *buffer,
                                CoreSurfaceAllocation            *from,
                                CoreSurfaceAllocation            *to )
{
     DPRINTF("from-type = %x, to-type = %x\n", from->pool->desc.types, to->pool->desc.types);
     if((from->pool->desc.types & CSTF_EXTERNAL) && (to->pool->desc.types & CSTF_EXTERNAL))
     {
          if(from->buffer->format == to->buffer->format && from->surface->config.size.w == to->surface->config.size.w)
          {
                switch (buffer->format) 
                {
                  case DSPF_YV12:
                  case DSPF_I420:
                  case DSPF_NV12:
                  case DSPF_NV21:
                  case DSPF_NV16:
                  case DSPF_YUV444P:
                      return DFB_UNSUPPORTED;
                  default:
                      break;
                }
                
                return DFB_OK;
          }
     }
     return DFB_UNSUPPORTED;
}

static
DFBResult Xcode4Dma(int XCodeHandle, unsigned int src_addr, unsigned int dst_addr, unsigned int size)
{
    IOCTLDMAFBFB ioctl_dma_fb_fb;

    memset(&ioctl_dma_fb_fb, 0, sizeof(IOCTLDMAFBFB));
    ioctl_dma_fb_fb.ioctl_header.command          = CMD_DMA_FB_TO_FB;
    ioctl_dma_fb_fb.ioctl_header.size_of_buf      = sizeof(IOCTLDMAFBFB);
    ioctl_dma_fb_fb.ioctl_header.time_out_value   = IOCTL_TIMEOUTVALUE;
    ioctl_dma_fb_fb.ioctl_header.return_value     = IOCTL_RET_CMD_FAILED;

    ioctl_dma_fb_fb.fb_dma_address_dst = dst_addr;
    ioctl_dma_fb_fb.fb_dma_address_src = src_addr;
    ioctl_dma_fb_fb.size_in_bytes_of_mem = size;

    ioctl(XCodeHandle, IOCTL_CMD, &ioctl_dma_fb_fb);

    if (ioctl_dma_fb_fb.ioctl_header.return_value == IOCTL_RET_CMD_COMPLETE)
    {
        return DFB_OK;
    }
    else
    {
        return DFB_IO;
    }
}

static
DFBResult Xcode4StartTransfer    ( CoreSurfacePoolBridge            *bridge,
                                void                             *bridge_data,
                                void                             *bridge_local,
                                CoreSurfacePoolTransfer          *transfer,
                                void                             *transfer_data )
{
    pxcode4_dma_handle handle = (pxcode4_dma_handle)bridge_data;
    DFBResult ret;
    CoreSurfaceBufferLock  src;
    CoreSurfaceBufferLock  dst;
    CoreSurfaceAllocation *source = transfer->from;
    CoreSurfaceAllocation *allocation = transfer->to;
    int size, src_addr, dst_addr;
    
    /* Lock the source allocation. */
    dfb_surface_buffer_lock_init( &src, CSAID_CPU, CSAF_READ );
    
    ret = dfb_surface_pool_lock( source->pool, source, &src );
    if (ret) {
         D_DERROR( ret, "Core/SurfBuffer: Could not lock source for transfer!\n" );
         dfb_surface_buffer_lock_deinit( &src );
         return ret;
    }
    
    /* Lock the destination allocation. */
    dfb_surface_buffer_lock_init( &dst, CSAID_CPU, CSAF_WRITE );
    
    ret = dfb_surface_pool_lock( allocation->pool, allocation, &dst );
    if (ret) {
         D_DERROR( ret, "Core/SurfBuffer: Could not lock destination for transfer!\n" );
         dfb_surface_pool_unlock( source->pool, source, &src );
         return ret;
    }

    size = dst.pitch * transfer->buffer->surface->config.size.h;
    src_addr = src.phys;
    dst_addr = dst.phys;

    while(size > MAX_DMA_SIZE && ret == DFB_OK)
    {
        ret = Xcode4Dma(handle->fd, src.phys, dst.phys, MAX_DMA_SIZE);
        size -= MAX_DMA_SIZE;
        src_addr += MAX_DMA_SIZE;
        dst_addr += MAX_DMA_SIZE;
    }

    if(ret == DFB_OK && size)
    {
        ret = Xcode4Dma(handle->fd, src_addr, dst_addr, size);
    }

    DPRINTF("DMA Transfered %d bytes\n", dst.pitch * transfer->buffer->surface->config.size.h);
    
    dfb_surface_pool_unlock( allocation->pool, allocation, &dst );
    dfb_surface_pool_unlock( source->pool, source, &src );
    
    dfb_surface_buffer_lock_deinit( &dst );
    dfb_surface_buffer_lock_deinit( &src );

    return ret;
}

static
DFBResult Xcode4FinishTransfer   ( CoreSurfacePoolBridge            *bridge,
                                void                             *bridge_data,
                                void                             *bridge_local,
                                CoreSurfacePoolTransfer          *transfer,
                                void                             *transfer_data )
{
    return DFB_OK;
}


const SurfacePoolBridgeFuncs xcode4SurfacePoolBridgeFuncs = {

     .PoolBridgeDataSize = Xcode4PoolBridgeDataSize,
     .InitPoolBridge = Xcode4InitPoolBridge,
     .DestroyPoolBridge = Xcode4DestroyPoolBridge,
     .CheckTransfer = Xcode4CheckTransfer,
     .StartTransfer = Xcode4StartTransfer,
     .FinishTransfer = Xcode4FinishTransfer,

};


